import { BABYLON } from "./render3d/babylon";
import { BABYLONAction, CustomScene, ITempFrameBuffer, RenderTargetFactory, Copy, ExtensionTools, getContentWH, RenderTarget, SceneRenderFlags, EFilter, EWrap, GlContext, IPostprocessParam, Constants } from "./index";

export interface IFullScreenScene {
    name: string;
    zIndex: number;
    clearColor: [number, number, number, number] | null;
    cleardepth: number | null;
}

export interface ICustomSizeScene {
    name: string;
    clearColor: [number, number, number, number] | null;
    cleardepth: number | null;
    framebuffer: ITempFrameBuffer;
}
/**
 * 全屏渲染的场景的管理器
 * * 项目可能有多层场景
 * * 多层场景间需要层级控制
 * * 场景可能在UI之下,场景可能在UI之上
 * * UI之下下的场景可能会在渲染前清空屏幕, UI之上的场景一般不会清空屏幕(即会保留UI内容)
 * * 层级设计
 *   * 使用zIndex配置层级
 *   * 按从小到大顺序进行从下到上渲染
 *   * zIndex=0 预留给UI渲染
 *   * zIndex<0 为UI之下场景渲染
 *   * zIndex>0 为UI之上场景渲染
 * * 场景清屏配置
 *   * clearColor 进行赋值即标识需要清屏(UI之下的一般都需要设置,默认为[0,0,0,0],值为 0~1)
 *   * clearDepth 进行赋值即标识需要清屏深度数据(一般都需要设置,默认设置为1)
 * * 更新场景配置
 *   * 调用 FullScreenSceneManager.FullScreenScene 创建/更新配置 场景
 * @example
 * ```ts
 *  const [curScene2D, effScene, uiEffectScene] = FullScreenSceneManager.Initial([
 *      { zIndex: -3, name: SceneControl.mainScene2DName, clearColor: [0, 0, 0, 0], cleardepth: 1, },
 *      { zIndex: -2, name: SceneControl.effectSceneName, clearColor: [0, 0, 0, 0], cleardepth: 1, },
 *      { zIndex:  2, name: SceneControl.uiEffectSceneName, clearColor: null, cleardepth: 1, },
 *  ]);
 *  
 *  const mainScene = FullScreenSceneManager.FullScreenScene({ zIndex: -4, name: SceneControl.mainSceneName, clearColor: null, cleardepth: 1, }, BackGroundScene.scene);
 * ```
 */
export class FullScreenSceneManager {
    private static FullScreenSceneMap: Map<string, CustomScene> = new Map();
    private static FullScreenSceneQueue: Map<string, [number, boolean]> = new Map();
    private static Copy: Copy;
    private static useDepth: boolean;
    private static withoutSafeArea: boolean;
    public static Initial(scenes?: IFullScreenScene[], useDepth: boolean = false, withoutSafeArea: boolean = true): CustomScene[] | undefined {
        ExtensionTools.setPermanent(FullScreenSceneManager.Render);
        FullScreenSceneManager.Copy = new Copy();
        FullScreenSceneManager.useDepth = useDepth;
        FullScreenSceneManager.withoutSafeArea = withoutSafeArea;

        if (scenes) {
            let result: CustomScene[] = [];
            scenes.forEach(function(param) {
                let scene = FullScreenSceneManager.FullScreenScene(param);
                result.push(scene);
            });
            return result;
        }
    }
    public static FullScreenScene(param: IFullScreenScene, scene?: CustomScene): CustomScene {
        if (scene == undefined) {
            scene = FullScreenSceneManager.FullScreenSceneMap.get(param.name);
            if (scene == undefined) {
                scene = new CustomScene(param.name);
                FullScreenSceneManager.FullScreenSceneMap.set(param.name, scene);
            }
        } else {
            let old = FullScreenSceneManager.FullScreenSceneMap.get(param.name);
            if (old !== undefined && old != scene) {
                console.error(`Scene ${param.name} Create Again!!!`);
                return;
            }
            FullScreenSceneManager.FullScreenSceneMap.set(param.name, scene);
        }
        if (param.clearColor != null) {
            scene.renderpipeline.opaqueRenderer.autoClearColor(true);
            scene.renderpipeline.opaqueRenderer.clearColor(...param.clearColor);
        } else {
            scene.renderpipeline.opaqueRenderer.autoClearColor(false);
        }
        if (param.cleardepth != null) {
            scene.renderpipeline.opaqueRenderer.autoClearDepth(true);
            scene.renderpipeline.opaqueRenderer.clearDepth(param.cleardepth);
        } else {
            scene.renderpipeline.opaqueRenderer.autoClearDepth(false);
        }
        FullScreenSceneManager.FullScreenSceneQueue.set(param.name, [param.zIndex, param.clearColor != null]);

        return scene;
    }
    public static Get(name: string): CustomScene | undefined {
        return FullScreenSceneManager.FullScreenSceneMap.get(name);
    }
    public static Remove(name: string) {
        let scene = FullScreenSceneManager.FullScreenSceneMap.get(name);
        if (scene) {
            scene.dispose();
        }
        FullScreenSceneManager.FullScreenSceneMap.delete(name);
        FullScreenSceneManager.FullScreenSceneQueue.delete(name);
    }
    public static RenderStart(resultTarget: RenderTarget) {
        let _context = RenderTargetFactory.getContext();
        let gl = _context.gl;
        gl.bindFramebuffer(gl.FRAMEBUFFER, resultTarget.framebuffer);
        gl.clear(gl.COLOR_BUFFER_BIT | gl.DEPTH_BUFFER_BIT | gl.STENCIL_BUFFER_BIT);
    }
    public static RenderForZIndex0(resultTarget: RenderTarget) {

    }
    public static Render() {
        let vdocument = window["vdocument"];
        let info = getContentWH();
        let renderX         = FullScreenSceneManager.withoutSafeArea ? 0 : info.renderX;
        let renderY         = FullScreenSceneManager.withoutSafeArea ? 0 : info.renderY;
        let renderWidth     = info.renderWidth;
        let renderHeight    = info.renderHeight;
        let frameWidth      = FullScreenSceneManager.withoutSafeArea ? info.renderWidth : info.contextWidth;
        let frameHeight     = FullScreenSceneManager.withoutSafeArea ? info.renderHeight : info.contextHeight;
        let resultTarget = RenderTargetFactory.createRenderTarget(frameWidth, frameHeight, FullScreenSceneManager.useDepth);

        let customrenderer = null;
        {
            CustomRenderTargetScene.List.forEach((item) => {
                item.render(undefined);
            });
        }

        FullScreenSceneManager.RenderStart(resultTarget);

        let queue: [number, boolean, CustomScene][] = [];
        FullScreenSceneManager.FullScreenSceneQueue.forEach((item, key) => {
            queue.push([item[0], item[1], FullScreenSceneManager.FullScreenSceneMap.get(key)]);
        });
        queue.sort((a, b) => a[0] - b[0]);

        let startIdx = Number.MIN_SAFE_INTEGER;

        let currIndex: number = Number.MIN_SAFE_INTEGER;
        let len = queue.length;
        for (let idx = len - 1; idx >= 0; idx--) {
            let val = queue[idx];
            if (val[1] && val[2].renderFlag == SceneRenderFlags.active && val[2].impl.rootNodes.length > 1) {
                startIdx = idx;
                break;
            }
        }
        for (let idx = 0; idx < len; idx++) {
            if (idx < startIdx) continue;

            let [zIndex, _autoclear, scene] = queue[idx];
            if (scene && scene.impl.rootNodes.length > 1) {
                if (currIndex < 0 && 0 < zIndex) {
                    FullScreenSceneManager.RenderForZIndex0(resultTarget);
                }
                currIndex = zIndex;

                let rootframe: ITempFrameBuffer = {
                    rect: [renderX, renderY, renderWidth, renderHeight],
                    ctx: undefined,
                    frameBuffer: resultTarget?.framebuffer
                };
                BABYLONAction.updateEngineDimensions(renderWidth, renderHeight);
                // scene.impl.clientRect.left      = 0;
                // scene.impl.clientRect.top       = 0;
                // scene.impl.clientRect.width     = vdocument.width;
                // scene.impl.clientRect.height    = vdocument.height;
                scene.resize(renderWidth, renderHeight);
                scene.render(undefined, rootframe);
            }
        }

        if (currIndex < 0) {
            FullScreenSceneManager.RenderForZIndex0(resultTarget);
        }

        FullScreenSceneManager.RenderToScreen(resultTarget);

        RenderTargetFactory.unuseRenderTarget(resultTarget);
    }
    public static RenderToScreen(resultTarget: RenderTarget) {
        let _context = RenderTargetFactory.getContext();
        let info =  getContentWH();
        let renderX         = FullScreenSceneManager.withoutSafeArea ? info.renderX : 0;
        let renderY         = FullScreenSceneManager.withoutSafeArea ? info.renderY : 0;
        let frameWidth      = FullScreenSceneManager.withoutSafeArea ? info.renderWidth : info.contextWidth;
        let frameHeight     = FullScreenSceneManager.withoutSafeArea ? info.renderHeight : info.contextHeight;
        let target: ITempFrameBuffer = {
            rect: [renderX, renderY, frameWidth, frameHeight],
            ctx: undefined,
            frameBuffer: undefined
        };

        _context.gl.enable(_context.gl.BLEND);
        _context.gl.blendFunc(_context.gl.ONE, _context.gl.ONE_MINUS_SRC_ALPHA);
        _context.gl.viewport(...target.rect);
        _context.gl.scissor(...target.rect);
        _context.gl.disable(_context.gl.DEPTH_TEST);
        _context.gl.disable(_context.gl.CULL_FACE);
        FullScreenSceneManager.Copy.apply({ etype: "Copy", intensity: 1, alpha: 1 }, [resultTarget], <any>{ x: renderX, y: renderY, width: frameWidth, height: frameHeight, framebuffer: target.frameBuffer }, RenderTargetFactory.getContext());
    }
}

export class CustomRenderTargetScene extends CustomScene {
    public static List: CustomRenderTargetScene[] = [];
    public renderTarget: RenderTarget;
    private _texture: BABYLON.InternalTexture;
    constructor(
        name: string,
        width: number, height: number,
        shadowsEnabled: boolean = false,
        blockMaterialDirtyMechanism: boolean = true,
        needDepth: boolean = false,
        magFilter: EFilter = EFilter.LINEAR, minFilter: EFilter = EFilter.LINEAR, wrapS: EWrap = EWrap.CLAMP_TO_EDGE, wrapT: EWrap = EWrap.CLAMP_TO_EDGE, mipmap: boolean = true
    ) {
        super(name, width, height, shadowsEnabled, blockMaterialDirtyMechanism);

        this.renderTarget = RenderTargetFactory.createRenderTarget(width, height, needDepth, magFilter, minFilter, wrapS, wrapT, mipmap);
        this._texture = new BABYLON.InternalTexture(this.engine, 0);
        this._texture.width = this.renderWidth;
        this._texture.height = this.renderHeight;
        this._texture.baseWidth = this.renderWidth;
        this._texture.baseHeight = this.renderHeight;
        this._texture.isReady = true;
        this._texture._webGLTexture = this.renderTarget._webGLTexture;

        CustomRenderTargetScene.List.push(this);
    }
    getTexture(): BABYLON.Texture {
        let texture = new BABYLON.Texture(null, this.impl, true, true, BABYLON.Texture.LINEAR_LINEAR);
        texture.delayLoadState = BABYLON.Constants.DELAYLOADSTATE_LOADED;
        texture._texture = this._texture;
        // texture.dispose = () => {};
        (<any>texture)._delayedOnLoad = null;
        (<any>texture)._delayedOnError = null;

        texture.uScale = 1 ;
        texture.vScale = 1 ;

        return texture;
    }
    render(ctx: GlContext, target?: ITempFrameBuffer, postprocess?: IPostprocessParam[], blend?: Constants.ALPHA_MODE, forceResetContext?: boolean): boolean {
        super.render(undefined, {
            rect: [0, 0, this.renderTarget.width, this.renderTarget.height],
            frameBuffer: this.renderTarget.framebuffer,
            ctx: undefined,
        });
        return true;
    }
    dispose(): void {
        let idx = CustomRenderTargetScene.List.indexOf(this);
        if (idx >= 0) {
            CustomRenderTargetScene.List.splice(idx, 1);
        }
        super.dispose();
    }
}
