import * as GLTF2 from "./glTF2Interface";
import { LoaderRecord, TypedArrayConstructor, TypedArrayLike } from "./base";
import { BufferViewLoader } from "./bufferView";
import { BABYLON } from "../render3d/babylon";
import { AccessorComponentType, IAccessor } from "./glTF2Interface";
import { ResTab, ExtensionTools } from "../interface/index";

export class AccessorLoader {
    public static loadAsyncUint16(loaders: LoaderRecord, index: number, accessor: GLTF2.IAccessor): Promise<Uint16Array> {
        let loader = loaders.u16_accessor_loaders.get(index);
        if (!loader) {
            let indexBufferView = accessor.bufferView;
            loader = BufferViewLoader.loadAsync(loaders, indexBufferView).then((data) => {
                let temp = new Uint16Array(data.buffer, data.byteOffset + (accessor.byteOffset || 0), accessor.count * AccessorLoader._GetNumComponents(accessor.type));
                loaders.u16_accessors.set(index, temp)
                return temp;
            });
            loaders.u16_accessor_loaders.set(index, loader);
        }
        return loader;
    }
    public static loadAsyncUint32(loaders: LoaderRecord, index: number, accessor: GLTF2.IAccessor): Promise<Uint32Array>  {
        let loader = loaders.u32_accessor_loaders.get(index);
        if (!loader) {
            let indexBufferView = accessor.bufferView;
            loader = BufferViewLoader.loadAsync(loaders, indexBufferView).then((data) => {
                let temp = new Uint32Array(data.buffer, data.byteOffset + (accessor.byteOffset || 0), accessor.count * AccessorLoader._GetNumComponents(accessor.type));
                loaders.u32_accessors.set(index, temp)
                return temp;
            });
            loaders.u32_accessor_loaders.set(index, loader);
        }
        return loader;
    }
    public static loadAsyncFloat32(loaders: LoaderRecord, index: number, accessor: GLTF2.IAccessor): Promise<Float32Array>  {
        let loader = loaders.f32_accessor_loaders.get(index);
        if (!loader) {
            let indexBufferView = accessor.bufferView;
            loader = BufferViewLoader.loadAsync(loaders, indexBufferView).then((data) => {
                let temp = new Float32Array(data.buffer, data.byteOffset + (accessor.byteOffset || 0), accessor.count * AccessorLoader._GetNumComponents(accessor.type));
                loaders.f32_accessors.set(index, temp);
                return temp;
            });
            loaders.f32_accessor_loaders.set(index, loader);
        }
        return loader;
    }
    
    public static _loadAccessorAsync(loaders: LoaderRecord, index: number, accessor: IAccessor, constructor: TypedArrayConstructor): Promise<ArrayBufferView> {
        
        let loader = loaders.arraybufferview_accessor_loaders.get(index);
        if (!loader) {
            const numComponents = AccessorLoader._GetNumComponents(accessor.type);
            const byteStride = numComponents * BABYLON.VertexBuffer.GetTypeByteLength(accessor.componentType);
            const length = numComponents * accessor.count;

            let indexBufferView = accessor.bufferView;
            if (accessor.bufferView == undefined) {
                let result: ArrayBufferView = new constructor(length);
                loader = Promise.resolve(result);
                loaders.arraybufferview_accessors.set(index, result);
            }
            else {
                loader = BufferViewLoader.loadAsync(loaders, indexBufferView).then((bufferviewdata) => {
                    let bufferView = loaders.gltf.bufferViews[accessor.bufferView];
                    let result: ArrayBufferView;
                    if (accessor.componentType === AccessorComponentType.FLOAT && !accessor.normalized && (!bufferView.byteStride || bufferView.byteStride === byteStride)) {
                        result = AccessorLoader._GetTypedArray(``, accessor.componentType, bufferviewdata, accessor.byteOffset, length);
                    }
                    else {
                        const typedArray = new constructor(length);
                        BABYLON.VertexBuffer.ForEach(bufferviewdata, accessor.byteOffset || 0, bufferView.byteStride || byteStride, numComponents, accessor.componentType, typedArray.length, accessor.normalized || false, (value, index) => {
                            typedArray[index] = value;
                        });
                        result = typedArray;
                    }
                    loaders.arraybufferview_accessors.set(index, result);
                    return result;
                });
            }
            loaders.arraybufferview_accessor_loaders.set(index, loader);
        }
        return loader;
    }

    public static _GetTypedArrayConstructor(context: string, componentType: AccessorComponentType): TypedArrayConstructor {
        switch (componentType) {
            case AccessorComponentType.BYTE: return Int8Array;
            case AccessorComponentType.UNSIGNED_BYTE: return Uint8Array;
            case AccessorComponentType.SHORT: return Int16Array;
            case AccessorComponentType.UNSIGNED_SHORT: return Uint16Array;
            case AccessorComponentType.UNSIGNED_INT: return Uint32Array;
            case AccessorComponentType.FLOAT: return Float32Array;
            default: throw new Error(`${context}: Invalid component type ${componentType}`);
        }
    }
    
    public static _GetTypedArray(context: string, componentType: AccessorComponentType, bufferView: ArrayBufferView, byteOffset: number | undefined, length: number): TypedArrayLike {
        const buffer = bufferView.buffer;
        byteOffset = bufferView.byteOffset + (byteOffset || 0);

        const constructor = AccessorLoader._GetTypedArrayConstructor(`${context}/componentType`, componentType);

        try {
            return new constructor(buffer, byteOffset, length);
        }
        catch (e) {
            throw new Error(`${context}: ${e}`);
        }
    }
    
    public static _GetNumComponents(type: string): number {
        switch (type) {
            case "SCALAR": return 1;
            case "VEC2": return 2;
            case "VEC3": return 3;
            case "VEC4": return 4;
            case "MAT2": return 4;
            case "MAT3": return 9;
            case "MAT4": return 16;
        }

        throw new Error(`Invalid type (${type})`);
    }
}

export type GLTFAccessorRes = ArrayBufferView;

export const RES_GLTF_ACCESSOR_BUFFER_VIEW = "RES_GLTF_ACCESSOR_BUFFER_VIEW";

ExtensionTools.RegistStartUp(function() {
    ExtensionTools.ResLoadRegist(
        RES_GLTF_ACCESSOR_BUFFER_VIEW,
        (resTab: ResTab, type: typeof RES_GLTF_ACCESSOR_BUFFER_VIEW, filePath: string, ...args: any[]) => {
            throw new Error(`'RES_GLTF_ACCESSOR_BUFFER_VIEW' Can't Creat By Load().`);
        },
        (link: GLTFAccessorRes) => {
            //
        },
        (link: GLTFAccessorRes) => {
            return link.byteLength;
        }
    );
})