
import { BABYLON } from "../render3d/babylon";
import { IGLTFNew, IMeshNew, IMeshPrimitiveNew, INewIAnimation, INodeNew } from "./interface";

import { ExtensionTools, Res, ResTab } from '../interface/index';
import { OfflineProvider } from "../render3d/offline_provider";

export type GLTFRes = Res<any, IGLTFNew>;

export interface TypedArrayLike extends ArrayBufferView {
    readonly length: number;
    [n: number]: number;
}

export interface TypedArrayConstructor {
    new(length: number): TypedArrayLike;
    new(buffer: ArrayBufferLike, byteOffset: number, length?: number): TypedArrayLike;
}

export interface IRenderState {
    renderQueue: number;
    cullMode: string;
    depthTest: boolean;
    depthWrite: boolean;
    alphaMode: number;
}

export enum ETempNodeAnimationDataKey {
    Parent = "Parent",
    Bone = "Bone",
    Node = "Node",
    Mesh = "Mesh",
    Material = "Material",
    MainTex = "MainTex",
    OpacityTex = "OpacityTex",
    MaskTex = "MaskTex",
    EmissiveTex = "EmissiveTex",
    Camera = "Camera",
    Light = "Light",
    Geometries = "Geometries",
}

export interface ITempNodeAnimationAbout {
    [ETempNodeAnimationDataKey.Node]: BABYLON.TransformNode,
    [ETempNodeAnimationDataKey.Parent]: number,
    [ETempNodeAnimationDataKey.Bone]: { [index: number]: BABYLON.Bone},
    [ETempNodeAnimationDataKey.Mesh]?: BABYLON.AbstractMesh[],
    [ETempNodeAnimationDataKey.Material]?: BABYLON.Material[],
    [ETempNodeAnimationDataKey.MainTex]?: BABYLON.Texture[],
    [ETempNodeAnimationDataKey.OpacityTex]?: BABYLON.Texture[],
    [ETempNodeAnimationDataKey.EmissiveTex]?: BABYLON.Texture[],
    [ETempNodeAnimationDataKey.MaskTex]?: BABYLON.Texture[],
    [ETempNodeAnimationDataKey.Camera]?: BABYLON.Camera,
    [ETempNodeAnimationDataKey.Light]?: BABYLON.Light,
}

export class LoaderRecord {
    gltfres: GLTFRes;
    rootNode: BABYLON.AbstractMesh;
    scene: BABYLON.Scene;
    gltf: IGLTFNew;
    rootDir: string;
    rootUrl: string;
    resTab: ResTab;
    uris: Map<any, Promise<Uint8Array>> = new Map();
    meshLoaders: Promise<any>[] = [];
    bufferLoaders: Map<any, Promise<Uint8Array>> = new Map();
    buffers: Map<any, ArrayBufferView> = new Map();
    bufferViewLoaders: Map<any, Promise<Uint8Array>> = new Map();
    bufferViews: Map<number, ArrayBufferView> = new Map();
    u16_accessor_loaders: Map<any, Promise<Uint16Array>> = new Map();
    u32_accessor_loaders: Map<any, Promise<Uint32Array>> = new Map();
    f32_accessor_loaders: Map<any, Promise<Float32Array>> = new Map();
    u16_accessors: Map<number, Uint16Array> = new Map();
    u32_accessors: Map<number, Uint32Array> = new Map();
    f32_accessors: Map<number, ArrayBufferView> = new Map();
    arraybufferview_accessor_loaders: Map<any, Promise<ArrayBufferView>> = new Map();
    arraybufferview_accessors: Map<number, ArrayBufferView> = new Map();
    nodeAnimationAbout: Map<number, ITempNodeAnimationAbout> = new Map();
    animationFPS: number;
    materials: Map<number, BABYLON.Material> = new Map();
    skeletons: Map<number, BABYLON.Skeleton> = new Map();
    bones: Map<number, BABYLON.Bone[]> = new Map();
    textures: Promise<BABYLON.Texture>[] = [];
    animationLoaderList: Promise<void>[] = [];
    animationGroups: BABYLON.AnimationGroup[] = [];
    nodes: BABYLON.AnimationGroup[] = [];
    particles: any[] = [];
    meshes: BABYLON.Mesh[] = [];
    geometriesLoaders: Promise<any>[] = [];
    geometries: Map<number, BABYLON.Geometry> = new Map();
    public _isDisposed: boolean = false;
    public get babylonScene() {
        return this.scene;
    }
    public promise() {
        let result = [];

        this.uris.forEach((item) => { result.push(item); });

        return Promise.all(result);
    }
    public accessors() {
        let result = [];

        this.meshLoaders.forEach((item) => { result.push(item); });
        this.bufferLoaders.forEach((item) => { result.push(item); });
        this.bufferViewLoaders.forEach((item) => { result.push(item); });
        this.u16_accessor_loaders.forEach((item) => { result.push(item); });
        this.u32_accessor_loaders.forEach((item) => { result.push(item); });
        this.f32_accessor_loaders.forEach((item) => { result.push(item); });
        
        return Promise.all(result);
    }
    public bufferViewsLoadAsync() {
        let result = [];

        this.bufferLoaders.forEach((item) => { result.push(item); });
        this.bufferViewLoaders.forEach((item) => { result.push(item); });
        
        return Promise.all(result);
    }
    public textureLoaders() {
        return Promise.all(this.textures);
    }
    public animationLoaders() {
        return Promise.all(this.animationLoaderList);
    }
    public dispose() {
        if (this.gltfres) {
            this.resTab.delete(this.gltfres);
            console.log("Dispose GLTF");
        }
        this.meshes.forEach((mesh) => {
            if ((<any>mesh).materialData) {
                mesh.material = undefined;
            }
        });
        this._isDisposed = true;
        this.gltfres = undefined;
        this.scene = undefined;
        this.gltf = undefined;
        this.rootUrl = undefined;
        this.resTab = undefined;
        this.uris = undefined;
        this.bufferLoaders = undefined;
        this.bufferViewLoaders = undefined;
        this.u16_accessor_loaders = undefined;
        this.u32_accessor_loaders = undefined;
        this.f32_accessor_loaders = undefined;
    }
}

export interface IBoundingbox {
    boundingbox?: {
        center: [number, number, number],
        extents: [number, number, number],
    },
    extras?: {
        boundingbox?: {
            center: [number, number, number],
            extents: [number, number, number],
        }
    }
}

export interface ICell {
    cell?: [number, number],
    extras?: {
        cell?: [number, number]
    }
}

export class GLTFTool {
    public static meshParticleCfg(node: INodeNew) {
        return node.extras?.meshParticle || (<any>node).meshParticle;
    }
    public static cell(node: INodeNew) {
        return node.extras?.cell || <[number, number]>(<any>node).meshParticle;
    }
    public static boundingbox(node: IBoundingbox): {
        center: [number, number, number];
        extents: [number, number, number];
    } {
        return node.extras?.boundingbox || (<any>node).boundingbox;
    }
    public static subMeshes(node: IMeshPrimitiveNew) {
        return node.extras?.subMeshes || <number[][]>(<any>node).subMeshes;
    }
    public static alphaIndex(mesh: IMeshNew, babylonMesh: BABYLON.Mesh) {
        if (mesh.extras) {
            if (mesh.extras.alphaIndex !== undefined && mesh.extras.alphaIndex !== 0) {
                babylonMesh.alphaIndex = mesh.extras.alphaIndex;
            }
        } else {
            if ((<any>mesh).alphaIndex !== undefined && (<any>mesh).alphaIndex !== 0) {
                babylonMesh.alphaIndex = (<any>mesh).alphaIndex;
            }
        }
    }
    public static layerMask(mesh: IMeshNew, babylonMesh: BABYLON.Mesh) {
        if (babylonMesh._scene?.pi_autoApplyLayerMask) {
            if (mesh.extras) {
                if (mesh.extras.layerMask !== undefined) {
                    babylonMesh.layerMask = mesh.extras.layerMask;
                }
            } else {
                if ((<any>mesh).layerMask !== undefined) {
                    babylonMesh.layerMask = (<any>mesh).layerMask;
                }
            }
        }
    }
    public static receiveShadows(mesh: IMeshNew, babylonMesh: BABYLON.Mesh) {
        // if (babylonMesh._scene?.pi_autoApplyReceiveShadows) {
            if (mesh.extras) {
                if (mesh.extras.receiveShadows) {
                    babylonMesh.receiveShadows = !!(<any>mesh).extras.receiveShadows;
                }
            } else {
                if ((<any>mesh).receiveShadows) {
                    babylonMesh.receiveShadows = !!(<any>mesh).receiveShadows;
                }
            }
        // }
    }
    public static castShadows(mesh: IMeshNew, babylonMesh: BABYLON.Mesh) {
        // if (babylonMesh._scene?.pi_autoApplyReceiveShadows) {
            if (mesh.extras) {
                let flag = (<any>mesh.extras).castShadows || mesh.extras.castShadow;
                if (flag) {
                    babylonMesh.castShadows = !!flag;
                }
            } else {
                let flag = (<any>mesh).castShadows || (<any>mesh).castShadow;
                if (flag) {
                    babylonMesh.castShadows = !!flag;
                }
            }
        // }
    }
    public static isSkybox(mesh: IMeshNew, babylonMesh: BABYLON.Mesh) {
        if (mesh.extras) {
            if (mesh.extras.isSkybox !== undefined) {
                babylonMesh.infiniteDistance = true;
            }
        } else {
            if ((<any>mesh).isSkybox !== undefined) {
                babylonMesh.infiniteDistance = true;
            }
        }
    }
    public static animeEvents(animation: INewIAnimation) {
        return animation.extras?.events || <[number, string][]>(<any>animation).events;
    }
    public static SkinAnim(animation: INewIAnimation) {
        return animation.extras?.SkinAnim || <[number, number, number, number, number, number, number, number, string, string]>(<any>animation).SkinAnim;
    }
    public static light(node: INodeNew, gltf: IGLTFNew) {
        if (node.extras?.light != undefined) {
            return gltf.extensions?.KHR_lights_punctual?.lights[node.extras.light];
        } else if (node.extensions?.KHR_lights_punctual?.light != undefined) {
            return gltf.extensions?.KHR_lights_punctual?.lights[node.extensions.KHR_lights_punctual.light];
        } else {
            return undefined;
        }
    }
    public static applyBoundingbox(node: IBoundingbox, babylonMesh: BABYLON.AbstractMesh) {
        let cfg = node.extras?.boundingbox || (<any>node).boundingbox;
        if (cfg) {
            let x = 0;
            let y = 0;
            let z = 0;
    
            x = cfg.extents[0];
            y = cfg.extents[1];
            z = cfg.extents[2];
            const extend = new BABYLON.Vector3(x, y, z);
    
            x = cfg.center[0];
            y = cfg.center[1];
            z = cfg.center[2];
            const center = new BABYLON.Vector3(x, y, z);
    
            babylonMesh.getBoundingInfo().centerOn(center, extend);
            // babylonMesh.getBoundingInfo().isLocked = true;
            babylonMesh.getBoundingInfo().isImport = true;
            babylonMesh.isPickable = true;
            // babylonMesh.doNotSyncBoundingInfo = true;
        }
        
    }
    public static billboard(node: IBoundingbox, babylonMesh: BABYLON.AbstractMesh) {
        if ((<any>node).billboard != undefined) {
            babylonMesh.billboardMode = 7;
        }
    }
}

export const BABYLON_MATERIAL_NEW = "BABYLON_MATERIAL_NEW";

ExtensionTools.RegistStartUp(() => {
    ExtensionTools.ResLoadRegist(
        BABYLON_MATERIAL_NEW,
        undefined,
        (link: BABYLON.Material) => {
            link.dispose(false, true);
        },
        function(link: BABYLON.Material) {
            if (OfflineProvider.onoffCollectInfo) {
                OfflineProvider.piBABYLONResCollect[BABYLON_MATERIAL_NEW] -= 2048;
            }
            return 2048;
        }
    );
});