import { BABYLON } from "../render3d/babylon";
import { AccessorComponentType, AccessorType, IAccessor, IAnimation, IAnimationChannel, IBuffer, IBufferView, IGLTF, IImage, IMaterial, IMeshPrimitive, IProperty, ISampler, ISkin, ITexture, ITextureInfo, MeshPrimitiveMode, TextureMagFilter, TextureMinFilter, TextureWrapMode } from "./glTF2Interface";
import { BABYLON_MATERIAL_NEW, ETempNodeAnimationDataKey, GLTFTool, ITempNodeAnimationAbout, LoaderRecord, TypedArrayConstructor, TypedArrayLike } from "./base";
import { IGLTFNew, IMeshNew, IMeshPrimitiveNew, INewIAnimation, INodeNew } from "./interface";
import { UriLoader } from "./uri";
import { MeshParticleSystem } from "../mesh_particle/mesh_particle_system";
import { LoadExtraMaterial } from "../render3d/extends/material";
import { PI_material } from "./extensions/material";
import { BufferViewLoader } from "./bufferView";
import { BufferLoader } from "./buffer";
import { AccessorLoader, RES_GLTF_ACCESSOR_BUFFER_VIEW } from "./accessor";
import { KHR_lights } from "./extensions/KHR_lights_punctual";
import { CameraLoader } from "./camera";
import { SkinLoader } from "./skin";
import { IImageSolt, IMaterialReplace, IMaterialSolt } from "../interface/modelobj";
import { ExtensionTools, Res, ResTab } from "../interface/index";
import { StandardMaterialData, StandardMaterialMesh } from "../material/standardmaterial";
import { BoundingInfo } from "../render3d/extends/bounding_info";
import { MathDataPool } from "../render3d/extends/math";


export enum GLTFInterpolationType {
    "LINEAR" = 1,
    "STEP" = 2,
    "CUBICSPLINE" = 3,
    "PICUBICSPLINE" = 4,
}
export enum GLTFInterpolationTypeOld {
    "STEP" = 0,
    "LINEAR" = 1,
    "PICUBICSPLINE" = 2,
    "CUBICSPLINE" = 3,
}
export enum EAnimationPropertyID {
    Position = 0,
    RotationQuaternion = 1,
    "Scaling",
    "MainTexUScale",
    "MainTexVScale",
    "MainTexUOffset",
    "MainTexVOffset",
    "Alpha",
    "MainColor",
    "ort",
    "per",
    "IsActive",
    "Euler",
    "intensity",
    "lightdiffuse",
    "AlphaCutOff",
    "cellId",
    "OpacityTexUScale",
    "OpacityTexVScale",
    "OpacityTexUOffset",
    "OpacityTexVOffset",
    "MaskCutoff",
    "MaskTexUScale",
    "MaskTexVScale",
    "MaskTexUOffset",
    "MaskTexVOffset",
}

const enum AnimationChannelTargetPath {
    /**
     * Translation
     */
    TRANSLATION = "translation",
    /**
     * Rotation
     */
    ROTATION = "rotation",
    // PI_BEGIN
    /**
     * rotation_eular
     */
    ROTATION_EULAR = "rotation_eular",
    // PI_END
    /**
     * Scale
     */
    SCALE = "scale",
    // PI_BEGIN
    // /**
    //  * Weights
    //  */
    WEIGHTS = "weights",
    // PI_END

    // PI_BEGIN
    USCALE = "uScale",

    VSCALE = "vScale",

    UOFFSET = "uOffset",

    VOFFSET = "vOffset",

    ALPHA = "alpha",

    COLOR = "color",

    ORT = "ort",

    PER = "per"
    // PI_END
}

const enum AnimationSamplerInterpolation {
    /**
     * The animated values are linearly interpolated between keyframes
     */
    LINEAR = "LINEAR",
    /**
     * The animated values remain constant to the output of the first keyframe, until the next keyframe
     */
    STEP = "STEP",
    /**
     * The animation's interpolation is computed using a cubic spline with specified tangents
     */
    CUBICSPLINE = "CUBICSPLINE",

    // PI_BEGIN
    PICUBICSPLINE = "PICUBICSPLINE",
    // PI_END
}

export interface GLTFLodResult {
    _root_?: BABYLON.AbstractMesh;
    nodes?: BABYLON.TransformNode[];
    meshes?: BABYLON.AbstractMesh[];
    particles?: any[];
    materials?: any[];
    animationGroups?: BABYLON.AnimationGroup[];
    gltf?: Res<any, IGLTFNew>;
}

interface IMetaData {
    imageSolts?: string[] | IImageSolt[];
    materialSolts?: string[] | IMaterialSolt[];
    materialReplace?: IMaterialReplace[];
};

export class GLTFLoader {
    public static AnimTargetPaths = ["translation", "rotation", "scale", "uScale", "vScale", "uOffset", "vOffset", "alpha", "color", "ort", "per", "piIsActive", "rotation_eular", "intensity", "lightdiffuse", "alphacutoff", "cellId"];
    public static AnimInterpolationType = ["STEP", "LINEAR", "PICUBICSPLINE", "CUBICSPLINE"];
    public static readonly DefaultSampler: ISampler = {};
    public static PiAsyncCall(func: Function, arg: any) {
        ExtensionTools.AddTask(func, [arg], 99999, 0);
    };

    private static _GetTextureWrapMode(context: string, mode: TextureWrapMode | undefined): number {
        // Set defaults if undefined
        mode = mode == undefined ? TextureWrapMode.REPEAT : mode;

        switch (mode) {
            case TextureWrapMode.CLAMP_TO_EDGE: return BABYLON.Texture.CLAMP_ADDRESSMODE;
            case TextureWrapMode.MIRRORED_REPEAT: return BABYLON.Texture.MIRROR_ADDRESSMODE;
            case TextureWrapMode.REPEAT: return BABYLON.Texture.WRAP_ADDRESSMODE;
            default:
                // Logger.Warn(`${context}: Invalid value (${mode})`);
                return BABYLON.Texture.CLAMP_ADDRESSMODE;
        }
    }

    private static _GetTextureSamplingMode(context: string, sampler: ISampler): number {
        // Set defaults if undefined
        const magFilter = sampler.magFilter == undefined ? TextureMagFilter.LINEAR : sampler.magFilter;
        const minFilter = sampler.minFilter == undefined ? TextureMinFilter.LINEAR_MIPMAP_LINEAR : sampler.minFilter;

        if (magFilter === TextureMagFilter.LINEAR) {
            switch (minFilter) {
                case TextureMinFilter.NEAREST: return BABYLON.Texture.LINEAR_NEAREST;
                case TextureMinFilter.LINEAR: return BABYLON.Texture.LINEAR_LINEAR;
                case TextureMinFilter.NEAREST_MIPMAP_NEAREST: return BABYLON.Texture.LINEAR_NEAREST_MIPNEAREST;
                case TextureMinFilter.LINEAR_MIPMAP_NEAREST: return BABYLON.Texture.LINEAR_LINEAR_MIPNEAREST;
                case TextureMinFilter.NEAREST_MIPMAP_LINEAR: return BABYLON.Texture.LINEAR_NEAREST_MIPLINEAR;
                case TextureMinFilter.LINEAR_MIPMAP_LINEAR: return BABYLON.Texture.LINEAR_LINEAR_MIPLINEAR;
                default:
                    // Logger.Warn(`${context}/minFilter: Invalid value (${minFilter})`);
                    return BABYLON.Texture.LINEAR_LINEAR;
            }
        }
        else {
            if (magFilter !== TextureMagFilter.NEAREST) {
                // Logger.Warn(`${context}/magFilter: Invalid value (${magFilter})`);
            }

            switch (minFilter) {
                case TextureMinFilter.NEAREST: return BABYLON.Texture.NEAREST_NEAREST;
                case TextureMinFilter.LINEAR: return BABYLON.Texture.NEAREST_LINEAR;
                case TextureMinFilter.NEAREST_MIPMAP_NEAREST: return BABYLON.Texture.NEAREST_NEAREST_MIPNEAREST;
                case TextureMinFilter.LINEAR_MIPMAP_NEAREST: return BABYLON.Texture.NEAREST_LINEAR_MIPNEAREST;
                case TextureMinFilter.NEAREST_MIPMAP_LINEAR: return BABYLON.Texture.NEAREST_NEAREST_MIPLINEAR;
                case TextureMinFilter.LINEAR_MIPMAP_LINEAR: return BABYLON.Texture.NEAREST_LINEAR_MIPLINEAR;
                default:
                    // Logger.Warn(`${context}/minFilter: Invalid value (${minFilter})`);
                    return BABYLON.Texture.NEAREST_NEAREST;
            }
        }
    }

    public static InstanceSourceMesh(url: string, instanceID: string, sourceMesh: BABYLON.Mesh) {
        //
    }

    public static GLTFFormat(gltfUrl: string, gltf: IGLTFNew): IGLTFNew {
        return gltf;
    }

    public static _GetNumComponents(context: string, type: string): number {
        switch (type) {
            case "SCALAR": return 1;
            case "VEC2": return 2;
            case "VEC3": return 3;
            case "VEC4": return 4;
            case "MAT2": return 4;
            case "MAT3": return 9;
            case "MAT4": return 16;
        }

        throw new Error(`${context}: Invalid type (${type})`);
    }
    /** @hidden */
    public static _GetDrawMode(context: string, mode: number | undefined): number {
        if (mode == undefined) {
            mode = MeshPrimitiveMode.TRIANGLES;
        }

        switch (mode) {
            case MeshPrimitiveMode.POINTS: return BABYLON.Material.PointListDrawMode;
            case MeshPrimitiveMode.LINES: return BABYLON.Material.LineListDrawMode;
            case MeshPrimitiveMode.LINE_LOOP: return BABYLON.Material.LineLoopDrawMode;
            case MeshPrimitiveMode.LINE_STRIP: return BABYLON.Material.LineStripDrawMode;
            case MeshPrimitiveMode.TRIANGLES: return BABYLON.Material.TriangleFillMode;
            case MeshPrimitiveMode.TRIANGLE_STRIP: return BABYLON.Material.TriangleStripDrawMode;
            case MeshPrimitiveMode.TRIANGLE_FAN: return BABYLON.Material.TriangleFanDrawMode;
        }

        throw new Error(`${context}: Invalid mesh primitive mode (${mode})`);
    }

    // PI_BEGIN - 20200309 扩展纹理uri可以为相对路径
    public static formatUrl(rootUrl: string, url: string | undefined) {
        if (url && url.startsWith('../')) {
            const rootList = rootUrl.split('/');
            const urlList = url.split('../');
            if (rootList.length < urlList.length) {
                return rootUrl + url;
            } else {
                let res = '';
                const rootListLen = rootList.length;
                const urlListLen = urlList.length;
                for (let i = 0; i < rootListLen - urlListLen; i++) {
                    res += `${rootList[i]}/`;
                }
                res += urlList[urlListLen - 1];

                return res;
            }
        } else {
            return rootUrl + url;
        }
    }
    // PI_END

    private _babylonScene: BABYLON.Scene;
    private _gltf: Res<any, IGLTFNew>;
    private _gltfJSON: IGLTFNew;
    private _resTab: ResTab;
    private displayName: string;
    private rootDir: string;
    public readonly rootUrl: string;
    private loaders: LoaderRecord;

    private _metadata: IMetaData;
    private _materialCustomModify: boolean = false;

    private includeNodes: number[];
    includeNodeNames: string[];

    private _rootBabylonMesh: BABYLON.AbstractMesh;

    public InstanceMeshMap: Map<string, BABYLON.Mesh>;
    public _defaultBabylonMaterialData: { [drawMode: number]: BABYLON.Material } = {};

    public get _scene() {
        return this._babylonScene;
    }
    public get babylonScene() {
        return this._babylonScene;
    }
    public get resTab() {
        return this._resTab;
    }
    public get gltf() {
        return this._gltfJSON;
    }
    public get loaderRecord() {
        return this.loaders;
    }
    public get fileName() {
        return this.rootUrl;
    }
    public get rootBabylonMesh() {
        return this._rootBabylonMesh;
    }

    constructor(scene: BABYLON.Scene, resTab: ResTab, rootPath: string, rootUrl: string, displayName: string, includeNodes?: number[], customInfo?: IMetaData) {
        this._babylonScene = scene;
        this._resTab = resTab;
        this.includeNodes = includeNodes;
        this.displayName = displayName;
        this.rootDir = rootPath;
        this.rootUrl = rootPath + rootUrl;
        this.loaders = new LoaderRecord();
        this.loaders.scene = this._scene;
        this.loaders.rootUrl = rootPath + rootUrl;
        this.loaders.rootDir = rootPath;
        this.loaders.resTab = this.resTab;
        this._metadata = customInfo;
    }

    dispose() {
        this.loaders.dispose();
        this._rootBabylonMesh?.dispose(false, true);
        this._gltf = undefined;
        // this._babylonScene = undefined;
    }

    public loadAsync(js: boolean = true): Promise<GLTFLodResult> {
        return UriLoader._loadGLTF(this.rootUrl, this.resTab).then((data) => {
            if (this._scene.isDisposed || this.loaders._isDisposed) {
                return {};
            }
            // const str = PISYS.UTF8.utf8Decode(data);
            // let json = JSON.parse(str);
            data.link = GLTFLoader.GLTFFormat(this.rootUrl, data.link);
            this._gltf = data;

            if (this._metadata && (this._metadata.imageSolts || this._metadata.materialReplace || this._metadata.materialSolts)) {
                this._materialCustomModify = true;
                let json = clone(data.link);
                this._applyCustomData(json);
                this._gltfJSON = json;
            } else {
                this._gltfJSON = data.link;
            }

            // if (this.imageSolts) {
            //     _gltfImageSolt(this.gltf, this.imageSolts);
            // }
            // if (this.materialSolts) {
            //     _gltfMaterialSolts(this.gltf, this.materialSolts);
            // }
            // if (this.materialReplace) {
            //     _gltfMaterialReplace(this.gltf, this.materialReplace);
            // }

            this.loaders.gltf = this.gltf;

            return new Promise((resolve, reject) => {
                GLTFLoader.PiAsyncCall(
                    () => {
                        this._analy().then(resolve).catch(reject);
                    },
                    null
                );
            });
        });
    }



    private _analy(): Promise<GLTFLodResult> {
        // if (this.gltf.asset.extras?.VersionNumber != "20230626") {
        //     return Promise.reject("GLTF Version Error: Need 20230626 But Give " + this.gltf.asset.extras?.VersionNumber);
        // }

        let accessorVerticeNeedLoad: number[] = [];
        let accessorIndicesNeedLoad: number[] = [];
        let materialIndexsNeedLoad: number[] = [];
        let skinIndexsNeedLoad: number[] = [];

        let nodes: BABYLON.TransformNode[] = [];
        let meshes: BABYLON.AbstractMesh[] = [];
        let meshParticles: MeshParticleSystem[] = [];
        let animationGroups: BABYLON.AnimationGroup[] = [];
        let materials: BABYLON.Material[] = [];

        //#region Nodes
        let _root_ = this.loaders.rootNode;
        if (!_root_) {
            _root_ = new BABYLON.AbstractMesh(this.displayName, this._scene);
            _root_.metadata = {};
            this.loaders.rootNode = _root_;
        }
        _root_.setEnabled(false);
        this._rootBabylonMesh = _root_;

        if (!this._scene.isDisposed && !this.loaders._isDisposed) {
            this.gltf.nodes?.forEach((nodeInfo, nodeIndex) => {
                let animationAbout: ITempNodeAnimationAbout = {
                    [ETempNodeAnimationDataKey.Parent]: -1,
                    [ETempNodeAnimationDataKey.Bone]: null,
                    [ETempNodeAnimationDataKey.Node]: null,
                    Mesh: [],
                    Material: [],
                    MainTex: [],
                    EmissiveTex: [],
                    OpacityTex: [],
                    MaskTex: [],
                };
                this.loaders.nodeAnimationAbout.set(nodeIndex, animationAbout);
            });
            this.gltf.nodes?.forEach((nodeInfo, nodeIndex) => {
                if (nodeInfo.children) {
                    nodeInfo.children.forEach((indexChild) => {
                        let child = this.loaders.nodeAnimationAbout.get(indexChild);
                        child[ETempNodeAnimationDataKey.Parent] = nodeIndex;
                    });
                }
            });

            this.gltf.nodes?.forEach((nodeInfo, nodeIndex) => {
                let flag = true;
                if (this.includeNodes) {
                    if (!this.includeNodes.includes(nodeIndex)) {
                        flag = false;
                    }
                }
                if (flag && this.includeNodeNames) {
                    if (!this.includeNodeNames.includes(nodeInfo.name)) {
                        flag = false;
                    }
                }
                if (flag == false) {
                    return;
                }

                let animationAbout: ITempNodeAnimationAbout = this.loaders.nodeAnimationAbout.get(nodeIndex);

                this.loadNode(``, nodeIndex, nodeInfo);

                // Transform
                if (nodeInfo.skin == undefined) {
                    let node: BABYLON.TransformNode = <any>animationAbout[ETempNodeAnimationDataKey.Node];
                    if (nodeInfo.translation) {
                        MathDataPool.Recycle(<BABYLON.Vector3>node.position);
                        node.position = MathDataPool.Vector3(...(<[number, number, number]>nodeInfo.translation));
                    }
                    if (nodeInfo.scale) {
                        MathDataPool.Recycle(<BABYLON.Vector3>node.scaling);
                        node.scaling = MathDataPool.Vector3(...(<[number, number, number]>nodeInfo.scale));
                    }
                    if (nodeInfo.rotation) {
                        MathDataPool.Recycle(<BABYLON.Quaternion>node.rotationQuaternion);
                        let rotationQuaternion = MathDataPool.Quaternion(...(<[number, number, number, number]>nodeInfo.rotation));
                        node.rotationQuaternion = rotationQuaternion;
                    }
                    if (nodeInfo.matrix) {
                        let matrix = MathDataPool.Matrix(); BABYLON.Matrix.FromArrayToRef(nodeInfo.matrix, 0, matrix);
                        let position = MathDataPool.Vector3(0, 0, 0);
                        let scaling = MathDataPool.Vector3(1, 1, 1);
                        let quaternion = MathDataPool.Quaternion(0, 0, 0, 1);
                        if (matrix.decompose(scaling, quaternion, position)) {
                            MathDataPool.Recycle(<BABYLON.Vector3>node.position);
                            MathDataPool.Recycle(<BABYLON.Vector3>node.scaling);
                            MathDataPool.Recycle(<BABYLON.Quaternion>node.rotationQuaternion);
                            node.position = position;
                            node.scaling = scaling;
                            node.rotationQuaternion = quaternion;
                        }
                        MathDataPool.Recycle(matrix);
                    }
                }
            });

            this.gltf.nodes?.forEach((nodeInfo, nodeIndex) => {
                let parent = this.loaders.nodeAnimationAbout.get(nodeIndex);
                let babylonNode = parent[ETempNodeAnimationDataKey.Node];
                nodes.push(babylonNode);
                if (nodeInfo.children) {
                    nodeInfo.children.forEach((indexChild) => {
                        let child = this.loaders.nodeAnimationAbout.get(indexChild);
                        if (parent && child) {
                            child[ETempNodeAnimationDataKey.Node].parent = babylonNode;
                        }
                    });
                }
                if (nodeInfo.mesh == undefined) {
                    // PI_BEGIN
                    this._addBoundingBox(nodeInfo, babylonNode);
                    // PI_END
                }
            });

            // this.gltf.nodes?.forEach((nodeInfo, nodeIndex) => {
            //     let parent = this.loaders.nodeAnimationAbout.get(nodeIndex);
            //     let babylonNode = parent[ETempNodeAnimationDataKey.Node];
            //     SkinLoader.loadNodeAsync(nodeIndex, nodeInfo, this.loaders, babylonNode);
            // });
            // this.gltf.nodes?.forEach((nodeInfo, nodeIndex) => {
            //     let parent = this.loaders.nodeAnimationAbout.get(nodeIndex);
            //     let babylonBone = parent[ETempNodeAnimationDataKey.Bone];
            //     nodeInfo.children?.forEach((indexChild) => {
            //         let child = this.loaders.nodeAnimationAbout.get(indexChild);
            //         let bone = child[ETempNodeAnimationDataKey.Bone];
            //         if (babylonBone && bone) {
            //             bone.parent = babylonBone;
            //         }
            //     });
            // });
            //#endregion
            return this.loaders.accessors().then(() => {
                if (this._scene.isDisposed || this.loaders._isDisposed) {
                    return Promise.reject(`Scene Disposed .`);
                }
                this.gltf.nodes?.forEach((nodeInfo, nodeIndex) => {
                    let animationAbout: ITempNodeAnimationAbout = this.loaders.nodeAnimationAbout.get(nodeIndex);
                    animationAbout[ETempNodeAnimationDataKey.Mesh]?.forEach((babylonMesh) => {
                        meshes.push(babylonMesh);
                    });

                    let babylonNode = animationAbout[ETempNodeAnimationDataKey.Node];
                    SkinLoader.loadNodeAsync(nodeIndex, nodeInfo, this.loaders, babylonNode);
                });
                return this.loaders.textureLoaders().then(() => {
                    if (this._scene.isDisposed || this.loaders._isDisposed) {
                        return Promise.reject(`Scene Disposed .`);
                    } else {
                        this._loadAnimationsAsync();
                        return this.loaders.animationLoaders().then(() => {
                            if (this._scene.isDisposed || this.loaders._isDisposed) {
                                return Promise.reject(`Scene Disposed .`);
                            } else {
                                this.loaders.animationGroups.forEach((animation) => {
                                    animation.normalize(0);
                                })
                                this.loaders.materials.forEach((val, index) => {
                                    materials[index] = val;
                                });
                                return <GLTFLodResult>{
                                    _root_: this.rootBabylonMesh,
                                    nodes,
                                    particles: this.loaders.particles,
                                    meshes,
                                    gltf: this._gltf,
                                    materials,
                                    animationGroups: this.loaders.animationGroups,
                                    skeletons: this.loaders.skeletons,
                                };
                            }
                        }).catch((e) => {
                            return Promise.reject(e);
                        });
                    }
                }).catch((e) => {
                    return Promise.reject(e);
                });
            }).catch((e) => {
                return Promise.reject(e);
            });

        } else {
            return Promise.reject("Disposed");
        }
    }

    public loadMeshParticleAsync() {

    }

    public loadNodePure(context: string, nodeIndex: number, nodeInfo: INodeNew) {
        let animationAbout = this.loaders.nodeAnimationAbout.get(nodeIndex);
        if (nodeInfo.mesh != undefined) {
            let node = new BABYLON.TransformNode(nodeInfo.name, this._scene);
            let meshInfo = this.gltf.meshes[nodeInfo.mesh];
            let load = this._loadMeshAsync(``, nodeInfo, nodeIndex, meshInfo, node, () => {});
            this.loaders.meshLoaders.push(<any>load);
            animationAbout[ETempNodeAnimationDataKey.Node] = node;
        } else {
            let node = new BABYLON.TransformNode(nodeInfo.name, this._scene);
            node.metadata = {};
            node.parent = this._rootBabylonMesh;
            animationAbout[ETempNodeAnimationDataKey.Node] = node;
        }
    }
    public loadNode(context: string, nodeIndex: number, nodeInfo: INodeNew) {
        this.loadNodePure(context, nodeIndex, nodeInfo);

        let babylonNode = this.loaders.nodeAnimationAbout.get(nodeIndex)[ETempNodeAnimationDataKey.Node];

        CameraLoader.loadNodeAsync(nodeIndex, nodeInfo, this.loaders, babylonNode);

        KHR_lights.loadNodeAsync(nodeIndex, nodeInfo, this.loaders, babylonNode);
    }

    protected _loadMeshAsync(context: string, node: INodeNew, nodeIndex: number, mesh: IMeshNew, parent: BABYLON.TransformNode, assign: (babylonMesh: BABYLON.AbstractMesh) => void): Promise<BABYLON.AbstractMesh> {
        const primitives = mesh.primitives;
        if (!primitives || !primitives.length) {
            throw new Error(`${context}: Primitives are missing`);
        }

        const name = node.name;

        let babylonNode: BABYLON.AbstractMesh;
        let tempMeshes: BABYLON.AbstractMesh[] = this.loaders.nodeAnimationAbout.get(nodeIndex)[ETempNodeAnimationDataKey.Mesh];

        let newassign = (mesh: BABYLON.AbstractMesh) => {
            tempMeshes.push(<BABYLON.Mesh>mesh);

            if (babylonNode == undefined) {
                babylonNode = mesh;
                babylonNode.parent = parent;
            } else {
                mesh.parent = babylonNode;
            }

            assign(mesh);
        }
        if (primitives.length === 1) {
            const primitive = mesh.primitives[0];
            let promise = this._loadMeshPrimitiveAsync(``, name, nodeIndex, node, mesh, primitive, newassign).then((babylonMesh) => {

                //#region PI_BEGIN - SHADOW ABOUT
                GLTFTool.castShadows(mesh, <BABYLON.Mesh>babylonMesh);
                //#endregion
                return babylonMesh;
            });
            this.loaders.meshLoaders.push(promise);
            return promise;
        }
        else {
            let loaders = [];
            let index = 0;
            for (const primitive of primitives) {
                loaders.push(
                    this._loadMeshPrimitiveAsync(``, `${name}_primitive${index}`, nodeIndex, node, mesh, primitive, newassign).then((babylonMesh) => {

                        //#region PI_BEGIN - SHADOW ABOUT
                        GLTFTool.castShadows(mesh, <BABYLON.Mesh>babylonMesh);
                        //#endregion
                    })
                );


                index += 1;
            }

            let promise = Promise.all(loaders).then(() => { return babylonNode });
            this.loaders.meshLoaders.push(promise);
            return promise;
        }


        // return babylonNode;
    }

    public _loadMeshPrimitiveAsync(context: string, name: string, nodeIndex: number, node: INodeNew, mesh: IMeshNew, primitive: IMeshPrimitive, assign: (babylonMesh: BABYLON.AbstractMesh) => void): Promise<BABYLON.AbstractMesh> {
        let babylonAbstractMesh: BABYLON.AbstractMesh;

        // PI_BEGIN - 加载时处理实例化 - 20201020
        if (!this.babylonScene.metadata) {
            this.babylonScene.metadata = {};
        }
        // if (!this.babylonScene.metadata.InstanceMeshLoadMap) {
        //     this.babylonScene.metadata.InstanceMeshLoadMap = new Map();
        // }
        if (!this.babylonScene.metadata.InstanceMeshMap) {
            this.babylonScene.metadata.InstanceMeshMap = new Map();
        }

        // this.InstanceMeshLoadMap    = this.babylonScene.metadata.InstanceMeshLoadMap;
        this.InstanceMeshMap = this.babylonScene.metadata.InstanceMeshMap;
        // PI_END

        let instanceID: string = null; // (<any>mesh).instanceID;

        if (instanceID) {
            return this._extend_loadMesh(context, name, nodeIndex, mesh, primitive, GLTFTool.meshParticleCfg(node) !== undefined, assign).then((sourceMesh) => {
                babylonAbstractMesh = sourceMesh.createInstance(`${(sourceMesh).name}-instance-${(sourceMesh).instances.length}`);
                // 网格设置包围盒后再设置导出的包围盒
                GLTFTool.applyBoundingbox(mesh, babylonAbstractMesh);
                GLTFTool.billboard(node, babylonAbstractMesh);

                // PI_BEGIN - IMAGE Cell
                let cell = GLTFTool.cell(node);
                if (cell) {
                    (<any>babylonAbstractMesh).cellWidth = cell[0];
                    (<any>babylonAbstractMesh).cellHeight = cell[1];
                }
                // PI_END

                return babylonAbstractMesh;
            });
        } else {
            return this._extend_loadMesh(context, name, nodeIndex, mesh, primitive, GLTFTool.meshParticleCfg(node) !== undefined, assign).then((babylonAbstractMesh) => {
                GLTFTool.billboard(node, babylonAbstractMesh);

                // PI_BEGIN - IMAGE Cell
                let cell = GLTFTool.cell(node);
                if (cell) {
                    (<any>babylonAbstractMesh).cellWidth = cell[0];
                    (<any>babylonAbstractMesh).cellHeight = cell[1];
                }
                // PI_END

                return babylonAbstractMesh;
            })
        }
    }

    public _extend_loadMesh(context: string, name: string, nodeIndex: number, mesh: IMeshNew, primitive: IMeshPrimitiveNew, materialIsCommonUse: boolean, assign: (babylonMesh: BABYLON.AbstractMesh) => void): Promise<BABYLON.Mesh> {
        let babylonMesh: BABYLON.Mesh;
        let instanceID: string = null;

        materialIsCommonUse = materialIsCommonUse || this._materialCustomModify;

        // if (instanceID) {
        //     instanceID = this.rootUrl + "#" + instanceID;
        //     babylonMesh = this.InstanceMeshMap.get(instanceID);
        // }

        if (!babylonMesh) {
            babylonMesh = new BABYLON.Mesh(name, this._babylonScene);
            this.loaders.meshes.push(babylonMesh);
            babylonMesh.metadata = {};
            babylonMesh.subMeshes = [];
            if (primitive.indices == undefined) {
                babylonMesh.isUnIndexed = true;
            }
            if (instanceID) {
                let temp = this.loaders.nodeAnimationAbout.get(nodeIndex);
                let parentIndex = temp[ETempNodeAnimationDataKey.Parent];
                if (parentIndex >= 0) {
                    let parent = this.loaders.nodeAnimationAbout.get(parentIndex)[ETempNodeAnimationDataKey.Node];
                    babylonMesh.parent = parent;
                }
                babylonMesh.setEnabled(false);
                this.InstanceMeshMap.set(instanceID, babylonMesh);
                babylonMesh.onDisposeObservable.addOnce(() => {
                    this.InstanceMeshMap.delete(instanceID);
                });
                GLTFLoader.InstanceSourceMesh(this.rootUrl, instanceID, babylonMesh);
            }

            //#region Geometry
            {
                // PI_BEIGN - 实例化相关处理
                if (babylonMesh.geometry) {
                    babylonMesh.geometry.releaseForMesh(babylonMesh, true);
                }
                // PI_END
                // PI_BEIGN - 实例化相关处理
                if (babylonMesh.instancedBuffers) {
                    for (let kind in babylonMesh.instancedBuffers) {
                        babylonMesh.setVerticesBuffer(babylonMesh._userInstancedBuffersStorage.vertexBuffers[kind]!);
                    }
                }
                // PI_END
                {
                    let subMeshes = GLTFTool.subMeshes(primitive);
                    if (subMeshes) {
                        const count = subMeshes.length;
                        for (let i = 0; i < count; i++) {
                            const [subMaterialIndex, , verticesStart, verticesCount, indexStart, indexCount, alphaIndex] = subMeshes[i];
                            let subMesh = new BABYLON.SubMesh(subMaterialIndex, verticesStart, verticesCount, indexStart, indexCount, babylonMesh);
                            subMesh._alphaIndex = alphaIndex;
                        }
                    }
                }
                babylonMesh.metadata.loadPromise = this._loadVertexDataAsync(context, primitive, babylonMesh).then((babylonGeometry) => {
                    this._loadMorphTargetsAsync(context, primitive, babylonMesh, babylonGeometry);

                    if (instanceID) {
                        GLTFLoader.InstanceSourceMesh(this.rootUrl, instanceID, babylonMesh);
                    }

                    // 网格设置包围盒后再设置导出的包围盒
                    GLTFTool.applyBoundingbox(mesh, babylonMesh);
                    // PI_END

                    // PI_BEGIN - 20200909 导入时自动处理 alphaIndex
                    GLTFTool.alphaIndex(mesh, babylonMesh);
                    // PI_END

                    // PI_BEGIN - 20200909 导入时自动处理 layerMask
                    GLTFTool.layerMask(mesh, babylonMesh);
                    // PI_END

                    // PI_BEGIN - 导入时自动处理 ReceiveShadows
                    GLTFTool.receiveShadows(mesh, babylonMesh);
                    GLTFTool.castShadows(mesh, babylonMesh);
                    // PI_END

                    // PI-BEGIN - 20201106 导入时自动处理 SKYBOX
                    GLTFTool.isSkybox(mesh, babylonMesh);
                    // PI_END

                    //#region Material
                    if (this.loaders._isDisposed == false) {
                        let temp = this.loaders.nodeAnimationAbout.get(nodeIndex);
                        const babylonDrawMode = GLTFLoader._GetDrawMode(context, primitive.mode);
                        if (primitive.material == undefined) {
                            if (primitive.subMaterials) {
                                let multiMaterial = new BABYLON.MultiMaterial('multi' + mesh.name, this.babylonScene);
                                babylonMesh.material = multiMaterial;
                                const count = primitive.subMaterials.length;

                                for (let i = 0; i < count; i++) {

                                    let indexMat = primitive.subMaterials[i];
                                    let babylonMaterial = this.loaders.materials.get(indexMat);
                                    if (babylonMaterial == undefined) {
                                        const material = this.gltf.materials[indexMat];
                                        babylonMaterial = this._loadMaterialAsync(``, indexMat, nodeIndex, material, babylonMesh, babylonDrawMode, false);

                                        this.loaders.materials.set(indexMat, babylonMaterial);
                                    }
                                    multiMaterial.subMaterials[i] = babylonMaterial;
                                    temp[ETempNodeAnimationDataKey.Material].push(babylonMaterial);
                                }
                            }
                            else {
                                let babylonMaterial = this._defaultBabylonMaterialData[babylonDrawMode];
                                if (!babylonMaterial) {
                                    babylonMaterial = this._createDefaultMaterial("__GLTFLoader._default", babylonDrawMode);
                                    // this._parent.onMaterialLoadedObservable.notifyObservers(babylonMaterial);
                                    this._defaultBabylonMaterialData[babylonDrawMode] = babylonMaterial;
                                }
                                babylonMesh.material = babylonMaterial;
                                temp[ETempNodeAnimationDataKey.Material].push(babylonMaterial);
                            }
                        } else {
                            let indexMat = primitive.material;
                            let babylonMaterial = this.loaders.materials.get(indexMat);
                            if (babylonMaterial == undefined) {
                                const material = this.gltf.materials[indexMat];
                                babylonMaterial = this._loadMaterialAsync(``, indexMat, nodeIndex, material, babylonMesh, babylonDrawMode, !materialIsCommonUse);

                                this.loaders.materials.set(indexMat, babylonMaterial);
                            }

                            if (!materialIsCommonUse && ExtensionTools.Option.optimizeMaterialWithResTab) {
                                let mesh = <StandardMaterialMesh>babylonMesh;
                                mesh.materialData = new StandardMaterialData(this._scene);
                                mesh.materialData.fromMaterial(<any>babylonMaterial);
                                (<BABYLON.Mesh>babylonMesh).onDisposeObservable.addOnce(function() {
                                    (<StandardMaterialData>mesh.materialData)?.dispose();
                                    mesh.materialData = undefined;
                                });
                            }

                            babylonMesh.material = babylonMaterial;
                            temp[ETempNodeAnimationDataKey.Material].push(babylonMaterial);
                        }
                    }
                    //#endregion

                    return babylonMesh;
                });

            }
            //#endregion

        }

        assign(babylonMesh);
        return babylonMesh.metadata.loadPromise;
    }

    public _loadVertexDataAsync(context: string, primitive: IMeshPrimitive, babylonMesh: BABYLON.Mesh): Promise<BABYLON.Geometry> {

        const attributes = primitive.attributes;
        if (!attributes) {
            throw new Error(`${context}: Attributes are missing`);
        } else {
            const babylonGeometry = new BABYLON.Geometry(babylonMesh.name, this._babylonScene);

            if (primitive.indices) {
                const accessor = this.gltf.accessors[primitive.indices];
                this._loadIndicesAccessorAsync(``, primitive.indices, accessor).then((data) => {
                    babylonGeometry.setIndices(data);
                });
            }

            let promises = [];
            let accessor = null;
            accessor = this.loadAttribute(attributes, "POSITION", BABYLON.VertexBuffer.PositionKind, babylonGeometry, babylonMesh, promises);
            accessor = this.loadAttribute(attributes, "NORMAL", BABYLON.VertexBuffer.NormalKind, babylonGeometry, babylonMesh, promises);
            accessor = this.loadAttribute(attributes, "TANGENT", BABYLON.VertexBuffer.TangentKind, babylonGeometry, babylonMesh, promises);
            accessor = this.loadAttribute(attributes, "TEXCOORD_0", BABYLON.VertexBuffer.UVKind, babylonGeometry, babylonMesh, promises);
            accessor = this.loadAttribute(attributes, "TEXCOORD_1", BABYLON.VertexBuffer.UV2Kind, babylonGeometry, babylonMesh, promises);
            accessor = this.loadAttribute(attributes, "JOINTS_0", BABYLON.VertexBuffer.MatricesIndicesKind, babylonGeometry, babylonMesh, promises);
            accessor = this.loadAttribute(attributes, "WEIGHTS_0", BABYLON.VertexBuffer.MatricesWeightsKind, babylonGeometry, babylonMesh, promises);
            accessor = this.loadAttribute(attributes, "JOINTS_1", BABYLON.VertexBuffer.MatricesIndicesExtraKind, babylonGeometry, babylonMesh, promises);
            accessor = this.loadAttribute(attributes, "WEIGHTS_1", BABYLON.VertexBuffer.MatricesWeightsExtraKind, babylonGeometry, babylonMesh, promises);
            accessor = this.loadAttribute(attributes, "COLOR_0", BABYLON.VertexBuffer.ColorKind, babylonGeometry, babylonMesh, promises);
            if (accessor) {
                if (accessor.type === AccessorType.VEC4) {
                    babylonMesh.hasVertexAlpha = true;
                }
            }

            (<any>babylonGeometry)._bakeframes = primitive.extras?.bakeframes;

            return Promise.all(promises).then(() => {
                if (babylonMesh.isDisposed()) return;

                let geometry = babylonMesh.geometry;
                babylonGeometry.applyToMesh(babylonMesh);
                if ((babylonMesh as BABYLON.Mesh)._boundingInfo && (babylonMesh as BABYLON.Mesh)._boundingInfo?.isImport) {
                    babylonMesh._updateBoundingInfo();
                }
                else if (geometry && (geometry!.useBoundingInfoFromGeometry)) {
                    // simply apply the world matrices to the bounding info - the extends are already ok
                    babylonMesh._updateBoundingInfo();
                } else {
                    babylonMesh.refreshBoundingInfo(true);
                }
                babylonMesh.instances.forEach(ins => {
                    if (ins._boundingInfo && ins._boundingInfo?.isImport) {
                        ins._updateBoundingInfo();
                    } else {
                        ins.getBoundingInfo().reConstruct(babylonMesh.getBoundingInfo().minimum.clone(), babylonMesh.getBoundingInfo().maximum.clone());
                        ins._updateBoundingInfo();
                    }
                });
                return babylonGeometry;
            });
        }
    }

    public loadAttribute(attributes: { [name: string]: number; }, attribute: string, kind: string, babylonGeometry: BABYLON.Geometry, babylonMesh: BABYLON.Mesh, promises: Promise<void>[]): IAccessor {
        if (attributes[attribute] == undefined) {
            return;
        }

        let indexAccessor = attributes[attribute];
        const accessor = this.gltf.accessors[indexAccessor];
        promises.push(
            this._loadVertexAccessorAsync(``, indexAccessor, accessor, kind).then((babylonVertexBuffer) => {
                if (babylonVertexBuffer.getKind() === BABYLON.VertexBuffer.PositionKind && !babylonMesh.skeleton) {
                    const mmin = accessor.min as [number, number, number], mmax = accessor.max as [number, number, number];
                    if (mmin !== undefined && mmax !== undefined) {
                        const min = BABYLON.TmpVectors.Vector3[0], max = BABYLON.TmpVectors.Vector3[1];
                        min.copyFromFloats(...mmin);
                        max.copyFromFloats(...mmax);
                        if (ExtensionTools.Option.optimizeBoundingInfoOnlySphere) {
                            babylonGeometry._boundingInfo = new BoundingInfo(min, max, null);
                        } else {
                            babylonGeometry._boundingInfo = new BABYLON.BoundingInfo(min, max);
                        }
                        babylonGeometry.useBoundingInfoFromGeometry = true;
                    }
                }
                babylonGeometry.setVerticesBuffer(babylonVertexBuffer, accessor.count);
            })
        );

        if (kind == BABYLON.VertexBuffer.MatricesIndicesExtraKind) {
            babylonMesh.numBoneInfluencers = 8;
        }

        return accessor;
    }

    public _loadMorphTargetsAsync(context: string, primitive: IMeshPrimitive, babylonMesh: BABYLON.Mesh, babylonGeometry: BABYLON.Geometry) {

        if (!primitive.targets) {
            return;
        }

        // const morphTargetManager = babylonMesh.morphTargetManager!;
        // for (let index = 0; index < morphTargetManager.numTargets; index++) {
        //     const babylonMorphTarget = morphTargetManager.getTarget(index);
        //     this._loadMorphTargetVertexDataAsync(`${context}/targets/${index}`, babylonGeometry, primitive.targets[index], babylonMorphTarget);
        // }
    }

    public _loadMorphTargetVertexDataAsync(context: string, babylonGeometry: BABYLON.Geometry, attributes: { [name: string]: number }, babylonMorphTarget: BABYLON.MorphTarget) {
    }

    public _loadIndicesAccessorAsync(context: string, indexAccessor: number, accessor: IAccessor): Promise<BABYLON.IndicesArray> {
        if (accessor.type !== AccessorType.SCALAR) {
            throw new Error(`${context}/type: Invalid value ${accessor.type}`);
        }

        if (accessor.componentType !== AccessorComponentType.UNSIGNED_BYTE &&
            accessor.componentType !== AccessorComponentType.UNSIGNED_SHORT &&
            accessor.componentType !== AccessorComponentType.UNSIGNED_INT) {
            throw new Error(`${context}/componentType: Invalid value ${accessor.componentType}`);
        }

        // if (accessor._data) {
        //     return accessor._data as Promise<IndicesArray>;
        // }

        if (accessor.sparse) {
            let key = `${this.rootUrl}#Indices#${indexAccessor}`;
            let res: Res<any, BABYLON.IndicesArray> = this.resTab.get(RES_GLTF_ACCESSOR_BUFFER_VIEW, key);
            if (res) {
                return Promise.resolve(res.link);
            } else {
                const constructor = AccessorLoader._GetTypedArrayConstructor(`${context}/componentType`, accessor.componentType);
                return AccessorLoader._loadAccessorAsync(this.loaders, indexAccessor, accessor, constructor).then((data) => {
                    let link = new Uint16Array(data.buffer.slice(data.byteOffset, data.byteLength));
                    this.resTab.createRes(key, RES_GLTF_ACCESSOR_BUFFER_VIEW, link, []);
                    return link;
                });
            }
        }
        else {
            let key = `${this.rootUrl}#Indices#${indexAccessor}`;
            let res: Res<any, BABYLON.IndicesArray> = this.resTab.get(RES_GLTF_ACCESSOR_BUFFER_VIEW, key);
            if (res) {
                return Promise.resolve(res.link);
            } else {
                const constructor = AccessorLoader._GetTypedArrayConstructor(`${context}/componentType`, accessor.componentType);
                return AccessorLoader._loadAccessorAsync(this.loaders, indexAccessor, accessor, constructor).then((data) => {
                    this.resTab.createRes(key, RES_GLTF_ACCESSOR_BUFFER_VIEW, data, []);
                    return data as Uint16Array;
                });
            }
            // let bufferView = this.gltf.bufferViews[accessor.bufferView];
            // let data = this.loaders.bufferViews.get(accessor.bufferView);
            // return AccessorLoader._GetTypedArray(context, accessor.componentType, data, accessor.byteOffset, accessor.count) as BABYLON.IndicesArray;
            // const bufferView = ArrayItem.Get(`${context}/bufferView`, this.gltf.bufferViews, accessor.bufferView);
            // accessor._data = this.loadBufferViewAsync(`/bufferViews/${bufferView.index}`, bufferView).then((data) => {
            //     return GLTFLoader._GetTypedArray(context, accessor.componentType, data, accessor.byteOffset, accessor.count);
            // });
        }

        // return accessor._data as Promise<IndicesArray>;
    }
    public _loadVertexAccessorAsync(context: string, indexAccessor: number, accessor: IAccessor, kind: string): Promise<BABYLON.VertexBuffer> {

        if (accessor.sparse) {
            // accessor._babylonVertexBuffer = this._loadFloatAccessorAsync(`/accessors/${accessor.index}`, accessor).then((data) => {
            //     return new VertexBuffer(this._babylonScene.getEngine(), data, kind, false);
            // });
            throw new Error("Not impl sparse Load..");
        }
        // HACK: If byte offset is not a multiple of component type byte length then load as a float array instead of using Babylon buffers.
        else if (accessor.byteOffset && accessor.byteOffset % BABYLON.VertexBuffer.GetTypeByteLength(accessor.componentType) !== 0) {
            // Logger.Warn("Accessor byte offset is not a multiple of component type byte length");
            // accessor._babylonVertexBuffer = this._loadFloatAccessorAsync(`/accessors/${accessor.index}`, accessor).then((data) => {
            //     return new BABYLON.VertexBuffer(this._babylonScene.getEngine(), data, kind, false);
            // });
            return this._loadFloatAccessorAsync(indexAccessor, accessor).then((data) => {
                return new BABYLON.VertexBuffer(this._babylonScene.getEngine(), data, kind, false);
            });
        }
        // Load joint indices as a float array since the shaders expect float data but glTF uses unsigned byte/short.
        // This prevents certain platforms (e.g. D3D) from having to convert the data to float on the fly.
        else if (kind === BABYLON.VertexBuffer.MatricesIndicesKind || kind === BABYLON.VertexBuffer.MatricesIndicesExtraKind) {
            // accessor._babylonVertexBuffer = this._loadFloatAccessorAsync(`/accessors/${accessor.index}`, accessor).then((data) => {
            //     return new VertexBuffer(this._babylonScene.getEngine(), data, kind, false);
            // });
            return this._loadFloatAccessorAsync(indexAccessor, accessor).then((data) => {
                return new BABYLON.VertexBuffer(this._babylonScene.getEngine(), data, kind, false);
            });
        }
        else {
            // const bufferView = ArrayItem.Get(`${context}/bufferView`, this.gltf.bufferViews, accessor.bufferView);
            // accessor._babylonVertexBuffer = this._loadVertexBufferViewAsync(bufferView, kind).then((babylonBuffer) => {
            //     const size = GLTFLoader._GetNumComponents(context, accessor.type);
            //     return new VertexBuffer(this._babylonScene.getEngine(), babylonBuffer, kind, false, false, bufferView.byteStride,
            //         false, accessor.byteOffset, size, accessor.componentType, accessor.normalized, true, 1, true);
            // });
            const bufferView = this.gltf.bufferViews[accessor.bufferView];
            // let babylonBuffer = this.loaders.bufferViews.get(accessor.bufferView);
            const size = GLTFLoader._GetNumComponents(context, accessor.type);
            return this._loadVertexBufferViewAsync(accessor.bufferView, bufferView, accessor.type).then((babylonBuffer) => {
                return new BABYLON.VertexBuffer(this._babylonScene.getEngine(), babylonBuffer.buffer, kind, false, false, bufferView.byteStride,
                    false, accessor.byteOffset + babylonBuffer.byteOffset, size, accessor.componentType, accessor.normalized, true, 1, true)
            });
        }

        // return accessor._babylonVertexBuffer;
    }

    // public loadBufferViewAsync(context: string, bufferView: IBufferView) {
    //     this.loaders.bufferViews.get(bufferView.)
    // }

    public _loadVertexBufferViewAsync(indexBufferView: number, bufferView: IBufferView, _kind: string) {
        // if (bufferView._babylonBuffer) {
        //     return <any>bufferView._babylonBuffer;
        // }

        // bufferView._babylonBuffer = this.loadBufferViewAsync(`#/bufferViews/${bufferView.index}`, bufferView).then((data) => {
        //     return <any>data;
        // });

        // return <any>bufferView._babylonBuffer;
        return BufferViewLoader.loadAsync(this.loaders, indexBufferView);

        // let buffer = this.loaders.bufferViews.get(indexBufferView);
        // return buffer;
    };

    public _loadFloatAccessorAsync(indexAccessor: number, accessor: IAccessor): Promise<Float32Array> {
        return AccessorLoader._loadAccessorAsync(this.loaders, indexAccessor, accessor, Float32Array) as Promise<Float32Array>;
    }

    public _loadMaterialAsync(context: string, indexMat: number, indexNode: number, material: IMaterial, babylonMesh: BABYLON.Nullable<StandardMaterialMesh>, babylonDrawMode: number, notParticleSys?: boolean) {
        let pimaterial = new PI_material();
        pimaterial._loader = this;
        let key = `${this._scene.uid}#${this.rootUrl}${context}#${indexMat}`;
        if (notParticleSys && ExtensionTools.Option.optimizeMaterialWithResTab && !this._materialCustomModify) {
            let pmat = this.resTab.get(BABYLON_MATERIAL_NEW, key);
            if (pmat == undefined) {
                const babylonMaterial = pimaterial.createMaterial(context, indexMat, indexNode, material, babylonDrawMode, notParticleSys);
                pimaterial.loadMaterialPropertiesAsync(context, material, babylonMaterial);
                pmat = this.resTab.createRes(key, BABYLON_MATERIAL_NEW, babylonMaterial, []);
            }
            const babylonMaterial = <BABYLON.Material>pmat.link;
            return babylonMaterial;
        } else {
            const babylonMaterial = pimaterial.createMaterial(context, indexMat, indexNode, material, babylonDrawMode, notParticleSys && !this._materialCustomModify);
            pimaterial.loadMaterialPropertiesAsync(context, material, babylonMaterial);
            return babylonMaterial;
        }
    }

    public _createDefaultMaterial(name: string, babylonDrawMode: number): BABYLON.Material {
        // this._babylonScene._blockEntityCollection = this._forAssetContainer;
        // PI_BEGIN - 默认屏蔽 PBR材质
        // const babylonMaterial = new PBRMaterial(name, this._babylonScene);
        // this._babylonScene._blockEntityCollection = false;
        // // Moved to mesh so user can change materials on gltf meshes: babylonMaterial.sideOrientation = this._babylonScene.useRightHandedSystem ? Material.CounterClockWiseSideOrientation : Material.ClockWiseSideOrientation;
        // babylonMaterial.fillMode = babylonDrawMode;
        // babylonMaterial.enableSpecularAntiAliasing = true;
        // babylonMaterial.useRadianceOverAlpha = !this._parent.transparencyAsCoverage;
        // babylonMaterial.useSpecularOverAlpha = !this._parent.transparencyAsCoverage;
        // babylonMaterial.transparencyMode = PBRMaterial.PBRMATERIAL_OPAQUE;
        // babylonMaterial.metallic = 1;
        // babylonMaterial.roughness = 1;
        // /////////////////////////////////
        const babylonMaterial = new BABYLON.StandardMaterial(name, this._babylonScene);
        // this._babylonScene._blockEntityCollection = false;
        // Moved to mesh so user can change materials on gltf meshes: babylonMaterial.sideOrientation = this._babylonScene.useRightHandedSystem ? Material.CounterClockWiseSideOrientation : Material.ClockWiseSideOrientation;
        babylonMaterial.fillMode = babylonDrawMode;
        // PI_END
        return babylonMaterial;
    }

    // PI_BEGIN - BoundingBox
    public _addBoundingBox(node: INodeNew, babylonNode: BABYLON.TransformNode, boxBone?: BABYLON.Bone): void {
        let boundingbox = GLTFTool.boundingbox(node);
        if (boundingbox) {
            // const box = BABYLON.Mesh.CreateBox((boundingbox.boneName || node.name) + "Box", 1, this.babylonScene);
            const box = BABYLON.Mesh.CreateBox((node.name) + "Box", 1, this.babylonScene);
            box.position = BABYLON.Vector3.FromArray(boundingbox.center);
            box.scaling = BABYLON.Vector3.FromArray(boundingbox.extents);
            box.isVisible = false;
            box.isPickable = true;

            if (!box.metadata) {
                box.metadata = {};
            }
            if (!boxBone) {
                box.metadata.link = babylonNode;
                box.parent = babylonNode;
            } else {
                box.metadata.link = babylonNode.parent;
                box.attachToBone(boxBone, babylonNode);
            }
        }
    }
    // PI_END

    public loadTextureInfoAsync(context: string, textureInfo: ITextureInfo): BABYLON.BaseTexture {
        const texture = this.gltf.textures[textureInfo.index];
        let babylonTexture = this._loadTextureAsync(context, texture);
        babylonTexture.coordinatesIndex = textureInfo.texCoord || 0;
        return babylonTexture;
    }

    /** @hidden */
    public _loadTextureAsync(context: string, texture: ITexture): BABYLON.BaseTexture {
        // const extensionPromise = this._extensionsLoadTextureAsync(context, texture, assign);
        // if (extensionPromise) {
        //     return extensionPromise;
        // }

        // this.logOpen(`${context} ${texture.name || ""}`);

        const invertY = false; // !!texture.invertY;

        if (texture.sampler >= 0) {
            let sampler = this.gltf.samplers[texture.sampler];
            const image = this.gltf.images[texture.source];
            return this._createTextureAsync(context, sampler, image, invertY);
        } else {
            const image = this.gltf.images[texture.source];
            return this._createTextureAsync(context, {}, image, invertY);
        }
    }

    /** @hidden */
    public _createTextureAsync(context: string, sampler: ISampler, image: IImage, invertY: boolean, textureLoaderOptions?: any): BABYLON.BaseTexture {
        const samplerData = this._loadSampler(``, sampler);

        // const deferred = new Deferred<void>();
        // this._babylonScene._blockEntityCollection = this._forAssetContainer;
        // const babylonTexture = new BABYLON.Texture(null, this._babylonScene, samplerData.noMipMaps, invertY, samplerData.samplingMode, () => {
        //     if (!this._disposed) {
        //         deferred.resolve();
        //     }
        // }, (message, exception) => {
        //     if (!this._disposed) {
        //         deferred.reject(new Error(`${context}: ${(exception && exception.message) ? exception.message : message || "Failed to load texture"}`));
        //     }
        // }, undefined, undefined, undefined, image.mimeType, textureLoaderOptions);
        // this._babylonScene._blockEntityCollection = false;
        // promises.push(deferred.promise);

        const babylonTexture = new BABYLON.Texture(null, this._babylonScene, samplerData.noMipMaps, invertY, samplerData.samplingMode, undefined, undefined, undefined, undefined, undefined, image.mimeType, textureLoaderOptions);

        this.loaders.textures.push(new Promise((resolve, reject) => {
            babylonTexture.onLoadObservable.addOnce((data) => {
                // if (!this._disposed) {
                resolve(data);
                // }
            });
            // babylonTexture.loa
        }));

        // promises.push(this.loadImageAsync(`/images/${image.index}`, image).then((data) => {
        //     const name = image.uri || `${this._fileName}#image${image.index}`;
        //     const dataUrl = `data:${this._uniqueRootUrl}${name}`;
        //     babylonTexture.updateURL(dataUrl, data);
        // }));

        babylonTexture.wrapU = samplerData.wrapU;
        babylonTexture.wrapV = samplerData.wrapV;

        let path = GLTFLoader.formatUrl(this.rootDir, image.uri);
        // PI_BEGIN - 20200309 扩展纹理uri可以为相对路径
        babylonTexture.updateURL(path);
        // PI_END
        babylonTexture.hasAlpha = image.uri.endsWith('.png');

        return babylonTexture;
    }

    private _loadSampler(context: string, sampler: ISampler) {
        return {
            noMipMaps: (sampler.minFilter === TextureMinFilter.NEAREST || sampler.minFilter === TextureMinFilter.LINEAR || sampler.minFilter === undefined),
            samplingMode: GLTFLoader._GetTextureSamplingMode(context, sampler),
            wrapU: GLTFLoader._GetTextureWrapMode(`${context}/wrapS`, sampler.wrapS),
            wrapV: GLTFLoader._GetTextureWrapMode(`${context}/wrapT`, sampler.wrapT)
        };
    }

    /**
     * Loads a glTF image.
     * @param context The context when loading the asset
     * @param image The glTF image property
     * @returns A promise that resolves with the loaded data when the load is complete
     */
    public loadImageAsync(context: string, image: IImage): ArrayBufferView {
        // if (!image._data) {
        //     this.logOpen(`${context} ${image.name || ""}`);

        //     if (image.uri) {
        //         image._data = this.loadUriAsync(`${context}/uri`, image, image.uri);
        //     }
        //     else {
        //         const bufferView = ArrayItem.Get(`${context}/bufferView`, this.gltf.bufferViews, image.bufferView);
        //         image._data = this.loadBufferViewAsync(`/bufferViews/${bufferView.index}`, bufferView);
        //     }

        //     this.logClose();
        // }

        throw new Error("Not Impl.")

        return;
    }

    /**
     * Loads a glTF uri.
     * @param context The context when loading the asset
     * @param property The glTF property associated with the uri
     * @param uri The base64 or relative uri
     * @returns A promise that resolves with the loaded data when the load is complete
     */
    public loadUriAsync(context: string, property: IProperty, uri: string): ArrayBufferView {
        // const extensionPromise = this._extensionsLoadUriAsync(context, property, uri);
        // if (extensionPromise) {
        //     return extensionPromise;
        // }

        // if (!GLTFLoader._ValidateUri(uri)) {
        //     throw new Error(`${context}: '${uri}' is invalid`);
        // }

        // if (Tools.IsBase64(uri)) {
        //     const data = new Uint8Array(Tools.DecodeBase64(uri));
        //     this.log(`Decoded ${uri.substr(0, 64)}... (${data.length} bytes)`);
        //     return Promise.resolve(data);
        // }

        // this.log(`Loading ${uri}`);

        // // PI_BEGIN - 20200309 扩展纹理uri可以为相对路径
        // // return this._parent.preprocessUrlAsync(this._rootUrl + uri).then((url) => {
        // // //////////////////////////////
        // return this._parent.preprocessUrlAsync(GLTFLoader.formatUrl(this._rootUrl, uri)).then((url) => {
        // // PI_END
        //     return new Promise((resolve, reject) => {
        //         this._parent._loadFile(url, this._babylonScene, (data) => {
        //             if (!this._disposed) {
        //                 this.log(`Loaded ${uri} (${(data as ArrayBuffer).byteLength} bytes)`);
        //                 resolve(new Uint8Array(data as ArrayBuffer));
        //             }
        //         }, true, (request) => {
        //             reject(new LoadFileError(`${context}: Failed to load '${uri}'${request ? ": " + request.status + " " + request.statusText : ""}`, request));
        //         });
        //     });
        // });
        return;
    }


    private _loadAnimationsAsync() {
        let animations = this.gltf.animations;
        if (animations) {
            for (let index = 0; index < animations.length; index++) {
                const animation = animations[index];
                let animeGroup = this.loadAnimationAsync(index, animation);
                this.loaders.animationGroups.push(animeGroup);
            }
        }
    }

    _loadAnimationKeysAsync(context: string, animationContext: string, animation: IAnimation, channel: number[], targetPath: string): Promise<Res<void, BABYLON.IAnimationKey[]>> {
        throw new Error("Not Impl.");
    }

    public loadAnimationAsync(indexGroup: number, animation: INewIAnimation) {

        const babylonAnimationGroup = new BABYLON.AnimationGroup(animation.name || `animation${indexGroup}`, this.babylonScene);

        // //////////////////////////////////
        let PiChannel: number[][] = (<any>animation).PiChannel;
        if (PiChannel) {
            let tempAnimIDs = [];
            for (const channel of PiChannel) {
                let [propertyID, targetID] = channel;
                let aID = `${propertyID}#${targetID}`;
                if (tempAnimIDs.includes(aID) == false) {
                    tempAnimIDs.push(aID);
                    this._loadAnimationChannelAsync(``, ``, animation, channel, babylonAnimationGroup);
                }
            }
        }
        else if (animation.channels) {
            (<any>animation).buffer = 0;
            let lenChannel = animation.channels.length;
            for (let indexChannel = 0; indexChannel < lenChannel; indexChannel++) {
                let channel = animation.channels[indexChannel];

                let propertyID: EAnimationPropertyID;
                if (channel.target.path) {
                    switch (<any>channel.target.path) {
                        case (AnimationChannelTargetPath.TRANSLATION): {
                            propertyID = EAnimationPropertyID.Position;
                            break;
                        }
                        case (AnimationChannelTargetPath.SCALE): {
                            propertyID = EAnimationPropertyID.Scaling;
                            break;
                        }
                        case (AnimationChannelTargetPath.ROTATION): {
                            propertyID = EAnimationPropertyID.RotationQuaternion;
                            break;
                        }
                    }
                }
                if (channel.extras.property != undefined) {
                    propertyID = channel.extras.property;
                }

                let sampler = animation.samplers[channel.sampler];
                let interpolation = GLTFInterpolationTypeOld.LINEAR;
                if (sampler.interpolation) {
                    switch (<any>sampler.interpolation) {
                        case (AnimationSamplerInterpolation.STEP): {
                            interpolation = GLTFInterpolationTypeOld.STEP;
                            break;
                        }
                        case (AnimationSamplerInterpolation.CUBICSPLINE): {
                            interpolation = GLTFInterpolationTypeOld.CUBICSPLINE;
                            break;
                        }
                    }
                } else if (sampler.extras?.interpolation != undefined) {
                    switch (<any>sampler.extras?.interpolation) {
                        case (GLTFInterpolationType.STEP): {
                            interpolation = GLTFInterpolationTypeOld.STEP;
                            break;
                        }
                        case (GLTFInterpolationType.PICUBICSPLINE): {
                            interpolation = GLTFInterpolationTypeOld.PICUBICSPLINE;
                            break;
                        }
                        case (GLTFInterpolationType.CUBICSPLINE): {
                            interpolation = GLTFInterpolationTypeOld.CUBICSPLINE;
                            break;
                        }
                    }
                }

                let targetID = channel.target.node;
                let input = this.gltf.accessors[sampler.input];
                let output = this.gltf.accessors[sampler.output];
                let inputBufferview = this.gltf.bufferViews[input.bufferView];
                let outputBufferview = this.gltf.bufferViews[output.bufferView];

                // let [propertyID, targetID, inputOffset, inputCount, inputBufferOffset, inputBufferLen, outputOffset, outputCount, outputBufferOffset, outputBufferLen, interpolation]
                let channelOld = [propertyID, targetID, input.byteOffset, input.count, inputBufferview.byteOffset, inputBufferview.byteLength, output.byteOffset, output.count, outputBufferview.byteOffset, outputBufferview.byteLength, interpolation];

                this._loadAnimationChannelAsync(``, ``, animation, channelOld, babylonAnimationGroup);
            }
        }
        // PI_END

        // PI_BEGIN - 20200814 - 动画事件
        let animeEvents = GLTFTool.animeEvents(animation);
        if (animeEvents) {
            var babylonAnimation = new BABYLON.Animation(`event_${animation.name}`, `_pi_meta_event`, 1, BABYLON.Animation.ANIMATIONTYPE_FLOAT);
            const count = animeEvents.length;
            const keys = [];
            for (let i = 0; i < count; i++) {
                const event = animeEvents[i];
                keys.push({
                    frame: event[0],
                    value: 0
                });
                const animEvent = new BABYLON.AnimationEvent(event[0], <any>undefined, false);
                animEvent._pi_eventKey = event[1];
                animEvent.action = function (curframe: number) {
                    const listenList: (() => boolean)[] = (<any>this)._pi_listenList;
                    if (listenList) {
                        listenList.forEach((func: any, index: number) => {
                            if (func) {
                                const onlyOnce = func(curframe);
                                if (onlyOnce) {
                                    listenList[index] = <any>undefined;
                                }
                            }
                        });
                    }
                };

                babylonAnimation.addEvent(animEvent);
            }
            babylonAnimation.setKeys(keys);
            babylonAnimationGroup.addTargetedAnimation(babylonAnimation, babylonAnimation); // 目标不限,
        }
        // PI_END

        // PI_BEGIN - 20200902 - 骨骼动画缓存
        let SkinAnim = GLTFTool.SkinAnim(animation);
        if (SkinAnim) {
            const [, , start, end, totalFrame, , , skinIndex, skinName, animName] = SkinAnim;
            BABYLON.Animation.AddSkinAnimCfg(skinName, animName, SkinAnim);
            var babylonAnimation = new BABYLON.Animation(`skin_cache_${animation.name}`, `pi_actionFrame`, 1, BABYLON.Animation.ANIMATIONTYPE_FLOAT_PRECISION);

            const keys = [
                {
                    frame: 0,
                    value: 0
                },
                {
                    frame: (end - start),
                    value: totalFrame - 1
                }
            ];
            babylonAnimation.setKeys(keys);
            babylonAnimation.pi_floatPrecision = 1;

            let indexSkin = this.gltf.skins.findIndex((v: ISkin) => {
                return v.name === skinName;
            });
            let skin = this.loaders.skeletons.get(indexSkin);
            if (skin) {
                skin.pi_cacheMatrixInfoMap.set(animName, SkinAnim);
                babylonAnimationGroup.addTargetedAnimation(babylonAnimation, skin);
                babylonAnimationGroup.onAnimationGroupPlayObservable.add((e, s) => {
                    skin.pi_actionName = animName;
                });
            }
        }
        // PI_END

        // babylonAnimationGroup.normalize(0);
        return babylonAnimationGroup;
    }

    public _loadAnimationChannelAsync(context: string, animationContext: string, animation: IAnimation, channel: number[], babylonAnimationGroup: BABYLON.AnimationGroup): Promise<void> {
        throw new Error("Not Impl .")
    }

    public _loadPiAnimationAsync(context: string, bufferIndex: number, channel: number[]) {
        return Promise.all([
            this._loadPiFloatAccessorAsync(channel[2], channel[3], channel[4], channel[5], bufferIndex, -1),
            this._loadPiFloatAccessorAsync(channel[6], channel[7], channel[8], channel[9], bufferIndex, channel[0])
        ]).then(([inputData, outputData]) => {
            return {
                input: inputData,
                interpolation: channel[10],
                output: outputData,
            };
        });
    };

    public loadPiBufferAsync = function (this: GLTFLoader, bufferByteOffset: number, bufferLength: number, index: number): Promise<ArrayBufferView> {
        return BufferLoader._loadBufferAsync(this.loaders, index).then((data) => {
            return new Uint8Array(data.buffer, (bufferByteOffset || 0), bufferLength);
        });
    }

    public _loadBufferAsync(index: number, buffer: IBuffer, offset: number, length: number) {
        return BufferLoader._loadBufferAsync(this.loaders, index).then((data) => {
            return new Uint8Array(data.buffer, (offset || 0), length);
        });
    }

    public _loadPiFloatAccessorAsync(byteOffset: number, count: number, bufferByteOffset: number, bufferLength: number, index: number, type: number): Promise<Float32Array> {
        let type1 = "SCALAR";
        /**
         * translation
         * scale
         * color
         * rotation_eular
         * lightdiffuse
         */
        if (type == 0 || type == 2 || type == 8 || type == 12 || type == 14) {
            type1 = "VEC3";
        } else if (type == 1) {
            type1 = "VEC4";
        }
        // TODO: support normalized and stride
        const numComponents = _GetPiComponents('', type1);
        const length = numComponents * count;

        return this.loadPiBufferAsync(bufferByteOffset, bufferLength, index).then(data => {
            return <Float32Array>AccessorLoader._GetTypedArray('', 5126, data, byteOffset, length);
        });
    };


    public _applyCustomData(json: IGLTFNew) {
        // PI_BEGIN - 20200309 - 拓展gltf imageSolts
        this._gltfImageSolt(json);
        this._gltfMaterialSolts(json);
        // PI_END
    }
    // PI_BEGIN - 20200309 扩展可变纹理数据
    private _gltfImageSolt(data: IGLTFNew) {
        if (this._metadata && this._metadata.imageSolts) {
            let solts = data.imageSolts;
            const images = data.images;
            const textures = data.textures;
            if (images) {
                const imagesLen = images.length;
                const imageSolts: string[] | IImageSolt[] = this._metadata.imageSolts;
                if (typeof imageSolts[0] === 'string') {
                    if (solts) {
                        for (let i = 0, len = imageSolts.length; i < len; i++) {
                            const solt = solts.find((s) => {
                                return s[0] === imageSolts[i];
                            });
                            if (solt && solt[1] <= imagesLen && solt[1] >= 0) {
                                images[solt[1]].uri = solt[2];
                                const tex = (<ITexture[]>textures).find((tex) => {
                                    return tex.source === solt[1];
                                });
                                tex && ((<any>tex).invertY = !!solt[3]);
                            }
                        }
                    }
                } else {
                    solts = <IImageSolt[]>imageSolts;
                    for (let i = 0, len = solts.length; i < len; i++) {
                        const solt = solts[i];
                        if (solt && solt[1] <= imagesLen && solt[1] >= 0) {
                            images[solt[1]].uri = solt[2];
                            const tex = (<ITexture[]>textures).find((tex) => {
                                return tex.source === solt[1];
                            });
                            tex && ((<any>tex).invertY = !!solt[3]);
                        }
                    }
                }
            }
        }
    }
    // PI_END

    // PI_BEGIN - 20200310 - 拓展gltf materialSolts
    private _gltfMaterialSolts(data: IGLTFNew) {
        if (this._metadata && this._metadata.materialSolts) {
            let solts = data.materialSolts;
            const materials = data.materials;
            if (materials) {
                const materialLen = materials.length;
                const materialSolts: string[] | IMaterialSolt[] = this._metadata.materialSolts;

                if (typeof materialSolts[0] === 'string') {
                    if (solts) {
                        let material;
                        for (let i = 0, len = materialSolts.length; i < len; i++) {
                            material = undefined;
                            const solt = solts.find((s) => {
                                return s[0] === materialSolts[i];
                            });

                            if (solt && solt[1] <= materialLen && solt[1] >= 0) {
                                material = (<any>materials[solt[1]]).extensions.PI_material;
                                if (material) {
                                    const attrList = solt[2];
                                    if (attrList) {
                                        this._gltfMaterialSolt(attrList, material);
                                    }
                                }
                            }
                        }
                    }
                } else {
                    solts = <IMaterialSolt[]>materialSolts;
                    let material;
                    for (let i = 0, len = solts.length; i < len; i++) {
                        const solt = solts[i];
                        material = undefined;
                        if (solt && solt[1] <= materialLen && solt[1] >= 0 && materials[solt[1]]) {
                            material = (<any>materials[solt[1]]).extensions.PI_material;
                        }
                        if (material) {
                            const attrList = solt[2];
                            if (attrList) {
                                this._gltfMaterialSolt(attrList, material);
                            }
                        }
                    }
                }
            }
        }
    }

    private _gltfMaterialSolt(attrList: [string, any][], material: any) {
        if (attrList) {
            let key = '';
            let value = undefined;
            for (let k = 0, len = attrList.length; k < len; k++) {
                key = attrList[k][0];
                value = attrList[k][1];
                switch (key) {
                    case ('scale'):
                    case ('offset'): {
                        material.diffuseTexture[key] = value;
                        break;
                    }
                    default: {
                        material[key] = value;
                    }
                }
            }
        }
    }
    // PI_END

}

function _GetPiComponents(context: string, type: string): number {
    switch (type) {
        case "SCALAR": return 1;
        case "VEC2": return 2;
        case "VEC3": return 3;
        case "VEC4": return 4;
        case "MAT2": return 4;
        case "MAT3": return 9;
        case "MAT4": return 16;
    }

    throw new Error(`${context}: Invalid type (${type})`);
}
/**
 * 深度克隆
 * @param   obj 待克隆的对象
 * @return      生成的对象
 */
const clone = (obj) => {
    let o;
    if (typeof obj === 'object') {
        if (Array.isArray(obj)) {
            o = [];
            for (const v of obj) {
                o.push(clone(v));
            }
        } else if (obj instanceof Map) {
            o = new Map();
            obj.forEach((v, k) => o.set(k, clone(v)));
        } else {
            o = {};
            Object.getOwnPropertyNames(obj).map((k) => o[k] = clone(obj[k]));
        }
    } else {
        o = obj;
    }

    return o;
};
