import { BABYLON } from "../render3d/babylon";
import { AccessorLoader, RES_GLTF_ACCESSOR_BUFFER_VIEW } from "./accessor";
import { ETempNodeAnimationDataKey, ITempNodeAnimationAbout, LoaderRecord } from "./base";
import { ISkin } from "./glTF2Interface";
import { INodeNew } from "./interface";

export class SkinLoader {
    public static loadNodeAsync(nodeIndex: number, nodeInfo: INodeNew, loaders: LoaderRecord, babylonNode: BABYLON.TransformNode) {
        let indexSkin = nodeInfo.skin;
        if (indexSkin != undefined) {
            if (loaders.skeletons.get(indexSkin) == undefined) {
                let skin = loaders.gltf.skins[indexSkin];
                let babylonSkeleton = SkinLoader._loadSkinAsync(``, nodeInfo, nodeIndex, skin, loaders);
                loaders.skeletons.set(indexSkin, babylonSkeleton);
            }
        }
    }
    
    
    public static _loadSkinAsync(context: string, node: INodeNew, nodeIndex: number, skin: ISkin, loaders: LoaderRecord): BABYLON.Skeleton {
        let babylonSkeleton: BABYLON.Skeleton = new BABYLON.Skeleton(skin.name, skin.name, loaders.babylonScene);
        babylonSkeleton.overrideMesh = loaders.rootNode;
        SkinLoader._loadBones(context, skin, loaders, babylonSkeleton);
        {
            let temp = loaders.nodeAnimationAbout.get(nodeIndex);
            let meshes = temp[ETempNodeAnimationDataKey.Mesh];
            meshes.forEach((babylonMesh) => {
                babylonMesh.skeleton = babylonSkeleton;
                const _scene = loaders.babylonScene;

                babylonMesh.onDispose = () => {
                    for (let bone of babylonSkeleton.bones) {
                        bone.dispose();
                    }
                    for (let i = 0; i < (<any>_scene)._activeSkeletons.data.length; i++) {
                        if ((<any>_scene)._activeSkeletons.data[i] == babylonSkeleton) {
                            (<any>_scene)._activeSkeletons.data.splice(i, 1);
                            break;
                        }
                    }
                    babylonSkeleton.dispose();
                };
    
                // PI_BEGIN - 20200916 - 加载时提取克隆数据
                // SkinLoader._skeletonCloneInfo.push([babylonMesh.metadata.gltf, babylonSkeleton.id]);
                // PI_END
            });
        }

        SkinLoader._loadSkinInverseBindMatricesDataAsync(context, skin, loaders).then((inverseBindMatricesData) => {
            SkinLoader._updateBoneMatrices(babylonSkeleton, inverseBindMatricesData);
        });

        return babylonSkeleton;
    }
    public static _loadSkinInverseBindMatricesDataAsync(context: string, skin: ISkin, loaders: LoaderRecord): Promise<BABYLON.Nullable<Float32Array>> {
        let indexAccessor = skin.inverseBindMatrices;
        const accessor = loaders.gltf.accessors[indexAccessor];
        let key = `${loaders.rootUrl}#inverseBindMatrices#${indexAccessor}`;
        let res = loaders.resTab.get(RES_GLTF_ACCESSOR_BUFFER_VIEW, key);
        if (res) {
            return Promise.resolve(res.link);
        } else {
            const constructor = AccessorLoader._GetTypedArrayConstructor(`${context}/componentType`, accessor.componentType);
            return AccessorLoader._loadAccessorAsync(loaders, indexAccessor, accessor, constructor).then((data) => {
                loaders.resTab.createRes(key, RES_GLTF_ACCESSOR_BUFFER_VIEW, data, []);
                return data as Float32Array;
            });
        }

        // const accessor = SkinLoader.gltf.accessors[skin.inverseBindMatrices];
        // return SkinLoader._loadFloatAccessorAsync(skin.inverseBindMatrices, accessor);
    }
    public static _loadBones(context: string, skin: ISkin, loaders: LoaderRecord, babylonSkeleton: BABYLON.Skeleton): void {
        let bones: { [index: number]: BABYLON.Bone } = {};
        let boneInfos: number[] = [];
        let boneNodeIndexList: number[] = [];
        skin.joints.forEach((nodeIndex, boneIndex) => {
            SkinLoader._collectBoneInfo(boneInfos, nodeIndex, boneIndex, boneNodeIndexList, loaders.nodeAnimationAbout);
        });
        boneNodeIndexList.forEach((nodeIndex) => {
            let boneIndex = boneInfos[nodeIndex];
            if (bones[nodeIndex] == undefined) {
                const node = loaders.gltf.nodes[nodeIndex];
                const temp = loaders.nodeAnimationAbout.get(nodeIndex);

                let parentIndex = temp[ETempNodeAnimationDataKey.Parent];
                let parentBone = bones[parentIndex];

                let babylonBone = SkinLoader._loadBone(parentBone, nodeIndex, boneIndex, node, babylonSkeleton, loaders);
                bones[nodeIndex] = babylonBone;
            }
        });
        // boneInfos.forEach((boneIndex, nodeIndex) => {
        //     let babylonBone = bones[nodeIndex];
        //     const temp = loaders.nodeAnimationAbout.get(nodeIndex);
        //     let parentIndex = temp[ETempNodeAnimationDataKey.Parent];
        //     babylonBone.parent = bones[parentIndex];
        // });
    }

    public static _collectBoneInfo(boneInfos: number[], nodeIndex: number, boneIndex: number, boneNodeIndexList: number[], maps: Map<number, ITempNodeAnimationAbout>) {
        if (boneInfos[nodeIndex] == undefined || boneInfos[nodeIndex] < 0) {
            boneInfos[nodeIndex] = boneIndex;
            let temp = maps.get(nodeIndex);
            let parentIndex = temp[ETempNodeAnimationDataKey.Parent];
            if (parentIndex >= 0) {
                SkinLoader._collectBoneInfo(boneInfos, parentIndex, -1, boneNodeIndexList, maps);
            }
            boneNodeIndexList.push(nodeIndex);
        }
    }

    public static _loadBone(parentBone: BABYLON.Bone, nodeIndex: number, boneIndex: number, node: INodeNew, babylonSkeleton: BABYLON.Skeleton, loaders: LoaderRecord): BABYLON.Bone {
        let babylonBone: BABYLON.Bone;

        let temp        = loaders.nodeAnimationAbout.get(nodeIndex);
        let babylonNode = temp[ETempNodeAnimationDataKey.Node];

        babylonBone = new BABYLON.Bone(node.name || `joint${nodeIndex}`, babylonSkeleton, parentBone, SkinLoader._getNodeMatrix(node), null, null, boneIndex);
        babylonBone.linkTransformNode(babylonNode);

        return babylonBone;
    }
    
    public static _updateBoneMatrices(babylonSkeleton: BABYLON.Skeleton, inverseBindMatricesData: BABYLON.Nullable<Float32Array>): void {
        for (const babylonBone of babylonSkeleton.bones) {
            let baseMatrix = BABYLON.Matrix.Identity();
            const boneIndex = babylonBone._index!;
            if (inverseBindMatricesData && boneIndex !== -1) {
                BABYLON.Matrix.FromArrayToRef(inverseBindMatricesData, boneIndex * 16, baseMatrix);
                baseMatrix.invertToRef(baseMatrix);
            }

            const babylonParentBone = babylonBone.getParent();
            if (babylonParentBone) {
                baseMatrix.multiplyToRef(babylonParentBone.getInvertedAbsoluteTransform(), baseMatrix);
            }

            babylonBone.setBindPose(baseMatrix);

            babylonBone.updateMatrix(baseMatrix, false, false);
            babylonBone._updateDifferenceMatrix(undefined, false);
        }
    }

    public static _getNodeMatrix(node: INodeNew): BABYLON.Matrix {
        return node.matrix ?
            BABYLON.Matrix.FromArray(node.matrix) :
            BABYLON.Matrix.Compose(
                node.scale ? BABYLON.Vector3.FromArray(node.scale) : BABYLON.Vector3.One(),
                node.rotation ? BABYLON.Quaternion.FromArray(node.rotation) : BABYLON.Quaternion.Identity(),
                node.translation ? BABYLON.Vector3.FromArray(node.translation) : BABYLON.Vector3.Zero());
    }
}