import * as GLTF2 from "./glTF2Interface";
import { LoaderLog } from "./log";
import { LoaderRecord } from "./base";
import { IGLTFNew } from "./interface";
import { ExtensionTools, ResTab, Res } from "../interface/index";

export class UriLoader {
    public static loadUriAsync(loaders: LoaderRecord, property: GLTF2.IProperty, uri: string): Promise<Uint8Array> {
        if (!UriLoader._ValidateUri(uri)) {
            throw new Error(`'${uri}' is invalid`);
        }

        if (UriLoader.IsBase64(uri)) {
            const data = new Uint8Array(UriLoader.DecodeBase64(uri));
            LoaderLog.log(`Decoded ${uri.substr(0, 64)}... (${data.length} bytes)`);
            return Promise.resolve(data);
        } else {
            let url = UriLoader.formatUrl(loaders.rootDir, uri);
            return UriLoader._loadFile(url, loaders.resTab);
        }
    }

    private static _ValidateUri(uri: string): boolean {
        return (UriLoader.IsBase64(uri) || uri.indexOf("..") === -1);
    }
    
    public static formatUrl(rootUrl: string, url: string | undefined) {
        if (url && url.startsWith('../')) {
            const rootList = rootUrl.split('/');
            const urlList = url.split('../');
            if (rootList.length < urlList.length) {
                return rootUrl + url;
            } else {
                let res = '';
                const rootListLen = rootList.length;
                const urlListLen = urlList.length;
                for (let i = 0; i < rootListLen - urlListLen; i++) {
                    res += `${rootList[i]}/`;
                }
                res += urlList[urlListLen - 1];

                return res;
            }
        } else {
            return rootUrl + url;
        }
    }

    public static _loadFile(url: string, resTab: ResTab) : Promise<Uint8Array> {
        return resTab.load(ExtensionTools.RES_FILE, url, []).then((data) => {
            let val = <Uint8Array>data.link;
            if (!ArrayBuffer.isView(val)) {
                val = new Uint8Array(val);
            } else if (val.byteLength > 0) {
                val = new Uint8Array(val.slice().buffer);
            }
            return val;
        });
        // throw new Error("Not Implement!");
    }

    public static _loadGLTF(url: string, resTab: ResTab) : Promise<Res<any, IGLTFNew>> {
        return resTab.load(ExtensionTools.RES_FILE, url, []).then((data) => {
            let val = <Uint8Array | ArrayBuffer | IGLTFNew>data.link;
            if ((<Uint8Array>val).buffer) {
                if (!ArrayBuffer.isView(val)) {
                    val = new Uint8Array(<ArrayBuffer>val);
                } else if (val.byteLength > 0) {
                    val = new Uint8Array(val.slice().buffer);
                }
                const str = ExtensionTools.UTF8Decode(val);
                data.link = JSON.parse(str);
                return data;
            } else {
                return data;
            }
        });
        // throw new Error("Not Implement!");
    }

    /**
     * Test if the given uri is a base64 string
     * @param uri The uri to test
     * @return True if the uri is a base64 string or false otherwise
     */
    public static IsBase64(uri: string): boolean {
        return uri.length < 5 ? false : uri.substr(0, 5) === "data:";
    }

    /**
     * Decode the given base64 uri.
     * @param uri The uri to decode
     * @return The decoded base64 data.
     */
    public static DecodeBase64(uri: string): ArrayBuffer {
        const decodedString = atob(uri.split(",")[1]);
        const bufferLength = decodedString.length;
        const bufferView = new Uint8Array(new ArrayBuffer(bufferLength));

        for (let i = 0; i < bufferLength; i++) {
            bufferView[i] = decodedString.charCodeAt(i);
        }

        return bufferView.buffer;
    }
    
    /**
     * Gets the absolute url.
     * @param url the input url
     * @return the absolute url
     */
    public static GetAbsoluteUrl(url: string): string {
        const a = document.createElement("a");
        a.href = url;
        return a.href;
    }
}