import { SceneStruct } from "./scene_struct";
import { BABYLON } from "./render3d/babylon";
import { SceneManager, SceneManagerData } from "./scene";
import { MaterialBase } from "./material/base";
import { OfflineProvider } from "./render3d/offline_provider";
import { MainOpacityOpacityFresnelMaterial } from "./material/main_opacity_opacity_fresnel";
import { MainOpacityEmissiveFresnelMaterial } from "./material/main_opacity_emissive_fresnel";
import { TwoOpacityMixMaterial } from "./material/two_opacity_mix";
import { MainOpacityMaterial } from "./material/main_opacity_material";
import { LoadExtraMaterial } from "./render3d/extends/material";
import { registOpacityClipFactory } from "./material/opacity_clip";
import { initialAlphaModeExtends } from "./alpha_blend/alpha_mode";
import { applyGLTFLoaderDataModifier } from "./modifier/gltf_loader_data";
import { registEffectMaskFactory } from "./material/effect_mask_material";
import { registDistortionUVFactory } from "./material/effect_distortion_uv";
import { registPiStandardFactory } from "./standard/pi_standard";
import { initial } from "./mesh_particle/extend";
import { ModelPrefab, PrefabModelObjPool } from "./model_prefab";
import { Constants } from "./interface/constants";
import { RenderTarget, RenderTargetFactory } from "./postprocess/render_target";
import { ManagedWebGLRenderingContext } from "./postprocess/webgl";
import { PostProcess } from "./postprocess/postprocess";
import { Copy } from "./postprocess/copy";
import { IPostprocessParam, IPostprocessTarget, IPostprocessTexture, KeyPostprocess, resortPostprocessParams } from "./interface/postprocess";
import { IPrefabCreation, IPrefabPool } from "./interface/modelprefab";
import { ModelObj } from "./model_obj";
import { Scene } from "./render3d/scene";
import { IModelObj, IModelOpt, SceneRenderFlags } from "./interface/modelobj";
import { KeyCopy } from "./interface/postprocess";
import { ICustomRenderTarget, IInitialOption, ITempFrameBuffer, TAutoRenderTarget } from "./interface/base";
import { IRenderPipeline, IRenderer } from "./interface/renderer";
import { IShadowGenerator } from "./interface/shadow";
import { ExtensionTools, TLoadCall } from "./interface/index";
import { PISYS } from "pi_sys/index";
import { TaskMgr } from 'pi_ucore/util/task_mgr';
import { register, ResTab } from 'pi_res/res_mgr';
import { Res } from "pi_res/res";
import { getLogger } from "pi_tools/logger";
import { Logger } from 'pi_sys/util/logger';
import { getGlobal } from "pi_ucore/util/frame_mgr";
import { loadFile } from "pi_sys/load/object_load";
import { IBakedKeyframes } from "./interface/spine_baked";

import { StandardMaterialData, TextureTillOffData, initStandardMaterialOpt } from "./material/standardmaterial";
import { optBoundingInfo } from "./render3d/extends/bounding_info";
import { MathDataPool } from "./render3d/extends/math";
import { TimeoutPool } from "./util/timeout_pool";
import { ScenePickManager } from "./scene_pick";
import { IColliderItem } from "./interface/extends";
import { initialTransformNodeExtend } from "./transformnode";

export { CustomGeometry } from "./customgeometry/custom_geometry";
export { Sprite } from "./sprite/sprite";
export { SpriteManager } from "./sprite/sprite_manager";

export * from "./postprocess/render_target";
export * from "./postprocess/copy";
export * from "./postprocess/webgl";
export * from "./model_prefab";
export * from "./model_obj";
export * from "./mesh_particle/extend";
export * from "./mesh_particle/mesh_particle_system";
export * from "./scene_tool";
export * from "./scene_struct";
export * from "./interface/index";

export { MainOpacityEmissiveFresnelMaterial } from "./material/main_opacity_emissive_fresnel";
export { MainOpacityMaterial } from "./material/main_opacity_material";
export { MainOpacityOpacityFresnelMaterial } from "./material/main_opacity_opacity_fresnel";
export { TwoOpacityMixMaterial } from "./material/two_opacity_mix";
export { OpsPass } from "./renderpass";
export * from "./material/shader_material";
export * from "./material/opacity_clip";
export * from "./material/effect_mask_material";
export * from "./material/effect_distortion_uv";

// export function init() {
//     // BABYLON.Engine.Initial();
//     // BABYLON.Engine.registBeforeRun(BABYLON.GLTF2.UriLoader.CheckGLTFLoad);
//     // BABYLON.Engine.registBeforeRun(BABYLON.BaseTexture.Query);
//     // SceneManagerData.engine = BABYLON.Engine.getEngine();
//     SceneManager.init(window["canvas"]);
// }

export function getContentWH() {
    let adaptiveResult = ExtensionTools.AdaptiveResult();
    let contextWidth = adaptiveResult.renderSize.width;
    let contextHeight = adaptiveResult.renderSize.height;

    let renderX = Math.round(adaptiveResult.styleLeft / adaptiveResult.styleSize.width * adaptiveResult.renderSize.width);
    let renderY = Math.floor(adaptiveResult.styleBottom / adaptiveResult.styleSize.height * adaptiveResult.renderSize.height);
    let renderWidth = Math.max(Math.floor((1.0 - (adaptiveResult.styleLeft + adaptiveResult.styleRight) / adaptiveResult.styleSize.width) * adaptiveResult.renderSize.width), 0);
    let renderHeight = Math.max(Math.floor((1.0 - (adaptiveResult.styleTop + adaptiveResult.styleBottom) / adaptiveResult.styleSize.height) * adaptiveResult.renderSize.height), 0);

    return {
        contextWidth,
        contextHeight,
        renderX,
        renderY,
        renderWidth,
        renderHeight,
        layoutSize: adaptiveResult.layoutSize,
        styleTop: adaptiveResult.styleTop,
        styleBottom: adaptiveResult.styleBottom,
        styleLeft: adaptiveResult.styleLeft,
        styleSize: adaptiveResult.styleSize,
        styleRight: adaptiveResult.styleRight
    }
}

declare var window: any;

export class AnimationGroup extends BABYLON.AnimationGroup { }
export class Animation extends BABYLON.Animation { }

export enum TCullMode {
    Off,
    Front,
    Back,
}
export enum TFrontFace {
    Ccw,
    Cw,
}
export enum TAlphaMode {
    ALPHA_COMBINE,
}

export interface GlContext {
    canvas: {
        setBrush(value: number): void;
        width: number;
        height: number;
    }
}

export interface IRefScene {
    camera: BABYLON.Camera;
    scene: SceneStruct;
    prefabPool: IPrefabPool;
    render(uicanvas: any): boolean;
    resize(): void;
    renderWidth: number;
    renderHeight: number;
}

export const REPEAT = 0x2901;
export const CLAMP_TO_EDGE = 0x812F;
export const MIRRORED_REPEAT = 0x8370;
export type AddressMode = typeof REPEAT | typeof CLAMP_TO_EDGE | typeof MIRRORED_REPEAT;

export const MATERIAL_OPAQUE = 0;
export const MATERIAL_ALPHATEST = 1;
export const MATERIAL_ALPHABLEND = 2;
export const MATERIAL_ALPHATESTANDBLEND = 3;
export type TransparentcyMode = typeof MATERIAL_OPAQUE | typeof MATERIAL_ALPHATEST | typeof MATERIAL_ALPHABLEND | typeof MATERIAL_ALPHATESTANDBLEND;

interface IUINode {
    offsetLeft: number;
    offsetTop: number;
    offsetWidth: number;
    offsetHeight: number;
}
interface IUIBoundingClientRect {
    left: number;
    top: number;
    width: number;
    height: number;
}
export interface IActionD3 {
    initial(opt: IInitialOption): void;
    engine(): BABYLON.Engine;
    piSizeForWidth(camera: BABYLON.FreeCamera, val: boolean): void;
    piOrthSize(camera: BABYLON.FreeCamera, val: number): void;
    cullMode(mesh: BABYLON.Mesh, val: TCullMode): void;
    frontFace(mesh: BABYLON.Mesh, val: TFrontFace): void;
    depthWrite(mesh: BABYLON.Mesh, val: boolean): void;
    depthTest(mesh: BABYLON.Mesh, val: boolean): void;
    alphaMode(mesh: BABYLON.Mesh, val: Constants.ALPHA_MODE): void;
    alpha(mesh: { material: BABYLON.Material, visibility: number }, val: number): void;
    autoClearColor(camera: BABYLON.Camera, val: boolean): void;
    autoClearDepth(camera: BABYLON.Camera, val: boolean): void;
    clearColor(camera: BABYLON.Camera, val: BABYLON.Color4): void;
    clearDepth(camera: BABYLON.Camera, val: number): void;
    activeCamera(camera: BABYLON.Camera, val: boolean): void;
    aspect(camera: BABYLON.Camera, val: number): void;
    render(scene: SceneStruct, val: GlContext): void;
    isDisposed(scene: BABYLON.Scene): void;
    alphaCutOff(mesh: { material: BABYLON.Material }, val: number): void;
    mainColor(material: BABYLON.Material, r: number, g: number, b: number): void;
    useMaterial(mesh: BABYLON.Mesh, material: BABYLON.Material, transparency: TransparentcyMode, alphaMode: Constants.ALPHA_MODE): void;
    wrapU(texture: BABYLON.Texture, val: AddressMode): void;
    wrapV(texture: BABYLON.Texture, val: AddressMode): void;
    disableDepthWrite(mesh: BABYLON.Mesh, val: boolean): void;
    useAlphaFromDiffuseTexture(mesh: BABYLON.Mesh, val: boolean): void;
    hasAlpha(texture: BABYLON.Texture, val: boolean): void;
    transparencyMode(mesh: BABYLON.Mesh, val: TransparentcyMode): void;
    updateEngineDimensions(width: number, height: number): void;
    backFaceCulling(mesh: { material: BABYLON.Material }, val: boolean): void;
    /**
     * 
     * @param mesh 
     * @param texturePropertyName 目标纹理属性名: "diffuseTexture" 、 "opacityTexture" ...
     * @param val 
     */
    uScale(mesh: BABYLON.Mesh, texturePropertyName: string, val: number);
    /**
     * 
     * @param mesh 
     * @param texturePropertyName 目标纹理属性名: "diffuseTexture" 、 "opacityTexture" ...
     * @param val 
     */
    vScale(mesh: BABYLON.Mesh, texturePropertyName: string, val: number);
    /**
     * 
     * @param mesh 
     * @param texturePropertyName 目标纹理属性名: "diffuseTexture" 、 "opacityTexture" ...
     * @param val 
     */
    uOffset(mesh: BABYLON.Mesh, texturePropertyName: string, val: number);
    /**
     * 
     * @param mesh 
     * @param texturePropertyName 目标纹理属性名: "diffuseTexture" 、 "opacityTexture" ...
     * @param val 
     */
    vOffset(mesh: BABYLON.Mesh, texturePropertyName: string, val: number);
    /**
     * UI 节点映射到场景中的位置和缩放
     * @param content 与相机渲染范围对应的容器节点, 与UI完整对齐可使用root节点
     * @param target 要计算的目标节点, 与UI完整对齐可使用root节点
     * @param camera 目标相机
     * @param resultPosition 映射的全局位置
     * @param resultSize 映射的全局缩放
     */
    uiNodeTransformToSceneRef(content: IUINode, target: IUINode, camera: BABYLON.Camera, resultPosition: BABYLON.Vector3, resultSize: BABYLON.Vector3);
    /**
     * 场景坐标映射到目标节点下相对坐标
     * @param content 与相机渲染范围对应的容器节点, 与UI完整对齐可使用root节点
     * @param target 定位的目标父节点, 与UI完整对齐可使用root节点
     * @param camera 目标相机
     * @param x 
     * @param y 
     * @param z 
     * @param result 映射的局部UI坐标
     */
    scenePositionToUILeftTopRef(content: IUINode, target: IUINode, camera: BABYLON.Camera, x: number, y: number, z: number, result: [number, number]);
    /**
     * UI 节点映射到场景中的位置和缩放
     * @param content 与相机渲染范围对应的容器节点, 与UI完整对齐可使用root节点 - 获取方法 element.getBoundingClientRect()
     * @param target 要计算的目标节点, 与UI完整对齐可使用root节点 - 获取方法 element.getBoundingClientRect()
     * @param camera 目标相机
     * @param resultPosition 映射的全局位置
     * @param resultSize 映射的全局缩放
     */
    uiBoundingClientRectTransformToSceneRef(content: IUIBoundingClientRect, target: IUIBoundingClientRect, camera: BABYLON.Camera, resultPosition: BABYLON.Vector3, resultSize: BABYLON.Vector3);
    /**
     * 场景坐标映射到目标节点下相对坐标
     * @param content 与相机渲染范围对应的容器节点, 与UI完整对齐可使用root节点 - 获取方法 element.getBoundingClientRect()
     * @param target 定位的目标父节点, 与UI完整对齐可使用root节点 - 获取方法 element.getBoundingClientRect()
     * @param camera 目标相机
     * @param x 
     * @param y 
     * @param z 
     * @param result 映射的局部UI坐标
     */
    scenePositionToUIBoundingClientRectLeftTopRef(content: IUIBoundingClientRect, target: IUIBoundingClientRect, camera: BABYLON.Camera, x: number, y: number, z: number, result: [number, number]);
}

export interface ICustomScene {
    readonly prefabPool: PrefabModelObjPool;
    readonly camera: BABYLON.Camera;
    readonly renderWidth: number;
    readonly renderHeight: number;
    renderQuality: number;
    resize(width: number, height: number): void;
    render(ctx: GlContext, target?: ITempFrameBuffer, postprocess?: IPostprocessParam[]): boolean;
}

export interface IResourceState {
    _0_count_bindbuffer: number,
    _1_count_bindgroup: number,
    _2_count_gltf: number,
    _3_count_imgtexture: number,
    _4_count_shader: number,
    _5_count_pipeline: number,
    _6_size_vertexbuffer: number;
    _7_count_vertexbuffer: number,
    _8_count_shadermeta: number;
    _9_mem_shadermeta: number;
    _10_mem_shader: number;
    _11_mem_bindbuffer: number;
    _12_mem_imgtexture: number;

    _14_count_passset0: number;
    _15_count_passset1: number;
    _16_count_passset2: number;
    _17_count_passbindgroups: number;
    _18_count_passshader: number;
    _19_count_passpipeline: number;
    _20_count_passdraw: number;

    _21_count_passmat: number,
    _22_count_passtexs: number,
    _23_count_vertex: number,
}

export interface IMaterialState {
    _13_count_material: number;

    _31_count_material_tex3: number;
    _32_count_material_tex3_ok: number;
    _33_count_material_tex4: number;
    _34_count_material_tex4_ok: number;
}

export interface ISceneState {
    _0_drawcalls: number,
    _1_count_vertex: number,
    _2_count_particlesystem: number,
    _3_count_particles: number,
    _4_count_animationgroup: number,
    _5_count_trail: number,
    _6_count_trail_points: number,
}

export interface IPerformance {
    _0_animation: number,
    _1_animationgroup: number,
    _2_particlesystem: number,
    _3_worldmatrix: number,
    _4_culling: number,
    _5_gltfanaly: number,
    _6_drawobjs: number,
    _7_uniformupdate: number,
    _8_uniformbufferupdate: number,
    _9_trailrenderer: number,
}

export interface IMeshState {
    _0_abstruct_mesh_count: number,
    _1_abstruct_mesh_count_enable: number,
    _2_mesh_count: number,
    _3_mesh_count_geometry_ready: number,
    _4_instance: number,
}

export interface IMaterialState2 {
    _0_count: number,
    _1_ready: number,
    _2_tex0: number,
    _3_tex0_ready: number,
    _4_tex1: number,
    _5_tex1_ready: number,
    _6_tex2: number,
    _7_tex2_ready: number,
    _8_tex3: number,
    _9_tex3_ready: number,
    _10_tex4: number,
    _11_tex4_ready: number,
    _12_tex5: number,
    _13_tex5_ready: number,
    _14_tex6: number,
    _15_tex6_ready: number,
}

export interface ITransformState {
    _0_count: number,
    _1_enable: number,
    _2_global_eanble: number,
    _3_calc_local_matrix: number,
    _4_calc_world_matrix: number,
    _5_max_level: number,
}

export interface ICameraState {
    _0_includes: number,
    _1_culling: number,
    _2_culling_time: number,
}

export interface ITextureLoaderState {
    _0_image_count: number;
    _1_image_fail: number;
    _2_image_success: number;
    _3_image_waiting: number;
    _4_texview_count: number;
    _5_texview_fail: number;
    _6_texview_success: number;
    _7_texview_waiting: number;
    _8_imgtexture: number;
    _9_imgtexture_mem: number;
    _10_imgtextureview: number;
    _11_imgtextureview_mem: number;
}

export interface IErrorState {
    entity: number,
    error: number,
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

export class _BABYLONAction implements IActionD3 {
    private framebufferDimensionsObject: {
        framebufferWidth: number,
        framebufferHeight: number
    } = {
            framebufferWidth: 1,
            framebufferHeight: 1
        };
    piOrthSize(camera: BABYLON.FreeCamera, val: number): void {
        camera.piOrthSize = val;
    }
    piSizeForWidth(camera: BABYLON.FreeCamera, val: boolean) {
        camera.piSizeForWidth = val;
    }
    initial(opt: IInitialOption) {
        ExtensionTools.AdaptiveResult = function () {
            return PISYS.Env.get(`adaptiveResult`);
        }
        ExtensionTools.AddTask = function (func: Function, args: any[], priority: number, type: 0 | 1 | 2) {
            TaskMgr.set(func, args, priority, type);
        }
        ExtensionTools.RES_FILE = "RES_FILE";
        ExtensionTools.RES_IMAGE = "RES_IMAGE";
        ExtensionTools.LoadFile = function (resTab: ResTab, url: string): Promise<ArrayBuffer | ArrayBufferView> {

            return loadFile(url);
            // return resTab.load(ExtensionTools.RES_FILE, url, []).then((data) => {
            //     let val = <Uint8Array>data.link;
            //     if (!ArrayBuffer.isView(val)) {
            //         val = new Uint8Array(val);
            //     } else if (val.byteLength > 0) {
            //         val = new Uint8Array(val.slice().buffer);
            //     }
            //     return val;
            // });
        }
        ExtensionTools.CreateResTab = function () {
            return new ResTab();
        }
        ExtensionTools.Domains = function () {
            return PISYS.Env.get(`domains`);
        }
        ExtensionTools.FileNameWithoutDirectory = function (file: string) {
            let i = file.lastIndexOf("/");
            return (i < 0) ? file : file.slice(i + 1);
        }
        ExtensionTools.GetLogger = function (module: string, tag: string) {
            return getLogger(module, tag);
        }
        ExtensionTools.GetPureResPath = function (url: string) {
            return PISYS.Depend.getPureResPath(url);
        }
        ExtensionTools.QueryRuntimeFullName = function (url: string) {
            return PISYS.Depend.queryRuntimeFullName(url);
        }
        ExtensionTools.Logger = function (module: string, tag: string) {
            return new Logger(module, tag);
        }
        ExtensionTools.QuerySupportCompressedTextureFormat = function () {
            return PISYS.Env.get(`compressed_tex_format`);
        }
        ExtensionTools.ResLoadRegist = function <T, A, D>(type: string, load: TLoadCall<T, A, D>, destroy: (link: any) => void, sizeOf?: (link: any) => number, replace?: boolean, timeout?: number, min_capacity?: number, max_capacity?: number) {
            return register(type, load, destroy, sizeOf, replace, timeout, min_capacity, max_capacity);
        }
        ExtensionTools.UTF8Decode = function (data: ArrayBuffer) {
            return PISYS.UTF8.utf8Decode(data);
        }
        ExtensionTools.setPermanentBefore = function (func: () => void) {
            getGlobal().setPermanentBefore(func);
        }
        ExtensionTools.setPermanent = function (func: () => void) {
            getGlobal().setPermanent(func);
        }
        ExtensionTools.clearPermanent = function (func: () => void) {
            getGlobal().clearPermanent(func);
        }
        ExtensionTools.getAtom = function(v: number | string | any) {
            return v;
        }
        ExtensionTools.getFileSize = function(file: string) {
            return PISYS.Depend.getFile(file)?.size;
        }
        TimeoutPool.Init();
        ExtensionTools.Ready();
        ExtensionTools.Option = opt;

        if (opt.DefaultWhitePath) {
            MaterialBase.DefaultWhitePath = opt.DefaultWhitePath;
        }
        if (opt.KHR_lights_punctual === false) {
            BABYLON.GLTF2.GLTFLoader.UnregisterExtension("KHR_lights_punctual");
        }
        if (opt.PI_environment === false) {
            BABYLON.GLTF2.GLTFLoader.UnregisterExtension("PI_environment");
        }
        if (opt.PI_particle === false) {
            BABYLON.GLTF2.GLTFLoader.UnregisterExtension("PI_particle");
        }
        if (opt.onoffCollectInfo !== undefined) {
            OfflineProvider.onoffCollectInfo = opt.onoffCollectInfo;
            window.piBABYLONFileRecord = OfflineProvider.piBABYLONFileRecord;
            window.piBABYLONGeometryRecord = OfflineProvider.piBABYLONGeometryRecord;
            window.piBABYLONResCollect = OfflineProvider.piBABYLONResCollect; // babylon文件总大小
            window.piBABYLONTextureRecord = OfflineProvider.piBABYLONTextureRecord; // 贴图信息，文件名和尺寸
        }

        initial();
        initialTransformNodeExtend();
        MathDataPool.Init();
        if (opt.optimizeMaterialWithResTab) {
            initStandardMaterialOpt();
        }
        if (opt.optimizeBoundingInfoOnlySphere) {
            optBoundingInfo();
        }
        registPiStandardFactory();
        registDistortionUVFactory();
        registEffectMaskFactory();
        applyGLTFLoaderDataModifier();
        initialAlphaModeExtends();
        registOpacityClipFactory();

        LoadExtraMaterial.init();
        MainOpacityMaterial.regist();
        TwoOpacityMixMaterial.regist();
        MainOpacityEmissiveFresnelMaterial.regist();
        MainOpacityOpacityFresnelMaterial.regist();

        SceneManager.init(window["canvas"]);
        ScenePickManager.Initial();

        BABYLON.Texture.prototype.getTextureMatrix = TextureTillOffData.prototype.getTextureMatrix;
        BABYLON.Scene.prototype.attachControl = function(attachUp?: boolean, attachDown?: boolean, attachMove?: boolean) {};

        this.framebufferDimensionsObject.framebufferWidth = opt.width;
        this.framebufferDimensionsObject.framebufferHeight = opt.height;
        SceneManagerData.engine.framebufferDimensionsObject = this.framebufferDimensionsObject;

        let ctx = new ManagedWebGLRenderingContext(SceneManagerData.canvas);
        RenderTargetFactory.init(SceneManagerData.canvas, ctx);
    }
    engine() {
        return SceneManagerData.engine;
    }
    useMaterial(mesh: BABYLON.Mesh, material: BABYLON.Material, transparency: TransparentcyMode, alphaMode: Constants.ALPHA_MODE): void {
        mesh.material = material;
        material.transparencyMode = transparency;
        material.alphaMode = alphaMode;
    }
    wrapU(texture: BABYLON.Texture, val: AddressMode): void {
        if (val === CLAMP_TO_EDGE) {
            texture.wrapU = BABYLON.Texture.CLAMP_ADDRESSMODE;
        } else if (val === MIRRORED_REPEAT) {
            texture.wrapU = BABYLON.Texture.MIRROR_ADDRESSMODE;
        } else {
            texture.wrapU = BABYLON.Texture.WRAP_ADDRESSMODE;
        }
    }
    wrapV(texture: BABYLON.Texture, val: AddressMode): void {
        if (val === CLAMP_TO_EDGE) {
            texture.wrapV = BABYLON.Texture.CLAMP_ADDRESSMODE;
        } else if (val === MIRRORED_REPEAT) {
            texture.wrapV = BABYLON.Texture.MIRROR_ADDRESSMODE;
        } else {
            texture.wrapV = BABYLON.Texture.WRAP_ADDRESSMODE;
        }
    }
    disableDepthWrite(mesh: BABYLON.Mesh, val: boolean): void {
        if (mesh.material) {
            mesh.material.disableDepthWrite = val;
        }
    }
    useAlphaFromDiffuseTexture(mesh: BABYLON.Mesh, val: boolean): void {
        if (mesh.material) {
            (<BABYLON.StandardMaterial>mesh.material).useAlphaFromDiffuseTexture = val;
        }
    }
    hasAlpha(texture: BABYLON.Texture, val: boolean): void {
        texture.hasAlpha = val;
    }
    transparencyMode(mesh: BABYLON.Mesh, val: TransparentcyMode): void {
        if (mesh.material) {
            (<BABYLON.StandardMaterial>mesh.material).transparencyMode = val;
        }
    }
    cullMode(mesh: BABYLON.Mesh, val: TCullMode) {
        switch (val) {
            case (TCullMode.Off): {
                mesh.material.backFaceCulling = false;
                break;
            }
            default: {
                mesh.material.backFaceCulling = true;
            }
        }
    }
    frontFace(mesh: BABYLON.Mesh, val: TFrontFace) {

    }
    depthWrite(mesh: BABYLON.Mesh, val: boolean) {
        mesh.material.forceDepthWrite = val;
    }
    depthTest(mesh: BABYLON.Mesh, val: boolean) {

    }
    alphaMode(mesh: BABYLON.Mesh, val: number) {
        mesh.material.alphaMode = val;
    }
    alpha(mesh: { material: BABYLON.Material, materialData?: StandardMaterialData }, val: number) {
        (mesh.materialData || <BABYLON.StandardMaterial>mesh.material).alpha = val;
    }
    autoClearColor(camera: BABYLON.Camera, val: boolean) {
        let scene = camera.getScene();
    }
    autoClearDepth(camera: BABYLON.Camera, val: boolean) {
        let scene = camera.getScene();
    }
    clearColor(camera: BABYLON.Camera, val: BABYLON.Color4) {
        let scene = camera.getScene();
    }
    clearDepth(camera: BABYLON.Camera, val: number) {
        // let scene = camera.getScene();
        // camera.clearDepth = val;
    }
    activeCamera(camera: BABYLON.Camera, val: boolean) {
        // camera.active = val;
    }
    aspect(camera: BABYLON.Camera, val: number) {

    }
    render(scene: SceneStruct, val: GlContext) {
        // scene?.render();
        // return true;
    }
    isDisposed(scene: BABYLON.Scene) {
        return scene.isDisposed;
    }
    alphaCutOff(mesh: { material: BABYLON.Material, materialData?: StandardMaterialData }, val: number) {
        (mesh.materialData || <BABYLON.StandardMaterial>mesh.material).alphaCutOff = val;
    }
    mainColor(material: BABYLON.Material, r: number, g: number, b: number) {
        (<BABYLON.StandardMaterial>material).emissiveColor.copyFromFloats(r, g, b);
    }
    updateEngineDimensions(width: number, height: number) {
        this.framebufferDimensionsObject.framebufferWidth = width;
        this.framebufferDimensionsObject.framebufferHeight = height;
    }
    backFaceCulling(mesh: { material: BABYLON.Material }, val: boolean): void {
        if (mesh.material) {
            mesh.material.backFaceCulling = val;
        }
    }
    uScale(mesh: { material: BABYLON.Material, materialData?: StandardMaterialData }, texturePath: string, val: number) {
        if ((mesh.materialData || mesh.material)[texturePath]) {
            (mesh.materialData || mesh.material)[texturePath].uScale = val;
        }
    }
    vScale(mesh: { material: BABYLON.Material, materialData?: StandardMaterialData }, texturePath: string, val: number) {
        if ((mesh.materialData || mesh.material)[texturePath]) {
            (mesh.materialData || mesh.material)[texturePath].vScale = val;
        }
    }
    uOffset(mesh: { material: BABYLON.Material, materialData?: StandardMaterialData }, texturePath: string, val: number) {
        if ((mesh.materialData || mesh.material)[texturePath]) {
            (mesh.materialData || mesh.material)[texturePath].uOffset = val;
        }
    }
    vOffset(mesh: { material: BABYLON.Material, materialData?: StandardMaterialData }, texturePath: string, val: number) {
        if ((mesh.materialData || mesh.material)[texturePath]) {
            (mesh.materialData || mesh.material)[texturePath].vOffset = val;
        }
    }
    uiNodeTransformToSceneRef(content: IUINode, target: IUINode, camera: BABYLON.Camera, resultPosition: BABYLON.Vector3, resultSize: BABYLON.Vector3) {
        let matrix = MathDataPool.Matrix();
        let source = MathDataPool.Vector3();

        // 渲染范围的矩形信息 - 0~1 - 左上角为原点,Y轴指向下方
        let dx = (target.offsetLeft   - content.offsetLeft  ) / content.offsetWidth ;
        let dy = (target.offsetTop    - content.offsetTop   ) / content.offsetHeight;
        let dw = target.offsetWidth  / content.offsetWidth ;
        let dh = target.offsetHeight / content.offsetHeight;
        // 渲染范围的矩形信息 - -1~1 - 渲染中心为原点,Y轴指向上方
        let rw = dw;
        let rh = dw;
        let rx =      (dx + dw * 0.5)  * 2 - 1;
        let ry = (1 - (dy + dh * 0.5)) * 2 - 1;

        camera.getTransformationMatrix().invertToRef(matrix);
        
        source.copyFromFloats(rx,       ry,         0);
        BABYLON.Vector3._UnprojectFromInvertedMatrixToRef(source, matrix, resultPosition);
        source.copyFromFloats(rx + rw,  ry + rh,    0);
        BABYLON.Vector3._UnprojectFromInvertedMatrixToRef(source, matrix, resultSize);
        resultSize.subtractInPlace(resultPosition);

        MathDataPool.Recycle(matrix);
        MathDataPool.Recycle(source);
    }
    scenePositionToUILeftTopRef(content: IUINode, target: IUINode, camera: BABYLON.Camera, x: number, y: number, z: number, result: [number, number]) {
        let matrix = MathDataPool.Matrix();
        let source = MathDataPool.Vector3();

        matrix.copyFrom(camera.getTransformationMatrix());
        
        // 世界坐标转视口坐标 - -1~1
        BABYLON.Vector3.TransformCoordinatesFromFloatsToRef(x, y, z, matrix, source);
        let ux = (source.x + 1) * 0.5;
        let uy = 1.0 - (source.y + 1) * 0.5;
        let left = ux * content.offsetWidth  - target.offsetLeft;
        let top  = uy * content.offsetHeight - target.offsetTop;

        result[0] = left;
        result[1] = top;

        MathDataPool.Recycle(matrix);
        MathDataPool.Recycle(source);
    }
    /**
     * UI 节点映射到场景中的位置和缩放
     * @param content 与相机渲染范围对应的容器节点, 与UI完整对齐可使用root节点 - 获取方法 element.getBoundingClientRect()
     * @param target 要计算的目标节点, 与UI完整对齐可使用root节点 - 获取方法 element.getBoundingClientRect()
     * @param camera 目标相机
     * @param resultPosition 映射的全局位置
     * @param resultSize 映射的全局缩放
     */
    uiBoundingClientRectTransformToSceneRef(content: IUIBoundingClientRect, target: IUIBoundingClientRect, camera: BABYLON.Camera, resultPosition: BABYLON.Vector3, resultSize: BABYLON.Vector3) {
        BABYLONAction.uiNodeTransformToSceneRef(
            { offsetLeft: content.left, offsetTop: content.top, offsetWidth: content.width, offsetHeight: content.height },
            { offsetLeft: target.left, offsetTop: target.top, offsetWidth: target.width, offsetHeight: target.height },
            camera,
            resultPosition,
            resultSize
        );
    }
    /**
     * 场景坐标映射到目标节点下相对坐标
     * @param content 与相机渲染范围对应的容器节点, 与UI完整对齐可使用root节点 - 获取方法 element.getBoundingClientRect()
     * @param target 定位的目标父节点, 与UI完整对齐可使用root节点 - 获取方法 element.getBoundingClientRect()
     * @param camera 目标相机
     * @param x 
     * @param y 
     * @param z 
     * @param result 映射的局部UI坐标
     */
    scenePositionToUIBoundingClientRectLeftTopRef(content: IUIBoundingClientRect, target: IUIBoundingClientRect, camera: BABYLON.Camera, x: number, y: number, z: number, result: [number, number]) {
        BABYLONAction.scenePositionToUILeftTopRef(
            { offsetLeft: content.left, offsetTop: content.top, offsetWidth: content.width, offsetHeight: content.height },
            { offsetLeft: target.left, offsetTop: target.top, offsetWidth: target.width, offsetHeight: target.height },
            camera,
            x, y, z,
            result
        );
    }
}

export const BABYLONAction = new _BABYLONAction;

export class SpineBakeModel extends ModelObj {
    protected _changeAnim() {
        super._changeAnim();
        if (!this.isLoaded || this.isDisposed) { return }
        if (this.animationMap && this.animOpt) {
            this.meshes.forEach((mesh) => {
                let enable = (<any>mesh) == this.rootImpl || mesh.name == this.animOpt.animName;
                mesh.setEnabled(enable);
            });
        }
    }
    protected extendLoaded(): void {
        this._meshMap.forEach((item) => {
            let data: IBakedKeyframes = (<any>item.geometry)?._bakeframes;
            if (data) {
                item.hasVertexAlpha = true;
                let material = <BABYLON.StandardMaterial>item.material;
                if (material) {
                    material.disableLighting = true;
                    material.emissiveColor = new BABYLON.Color3(1, 1, 1);
                }
                Object.defineProperty(item, "bakeFrame", {
                    set: (val: number) => {
                        let data: IBakedKeyframes = (<any>item.geometry)?._bakeframes;
                        if (data && val != undefined) {
                            let info = data.keyframes[Math.floor(val)];
                            if (item.subMeshes) {
                                item.subMeshes[0].indexStart = info.data[0].start;
                                item.subMeshes[0].indexCount = info.data[0].count;
                            }
                        }
                    }
                });
                material.sideOrientation = BABYLON.Material.CounterClockWiseSideOrientation;

                let keyFrames: BABYLON.IAnimationKey[] = [];
                keyFrames.push({
                    frame: 0,
                    value: 0
                });
                let _lastFrame = data.keyframes[data.keyframes.length - 1].frames;
                keyFrames.push({
                    frame: data.duration,
                    value: _lastFrame + 0.5
                });

                item.subMeshes[0].indexStart = data.keyframes[0].data[0].start;
                item.subMeshes[0].indexCount = data.keyframes[0].data[0].count;

                let animation = new BABYLON.Animation(item.name, "bakeFrame", 1, BABYLON.Animation.ANIMATIONTYPE_FLOAT);
                animation.setKeys(keyFrames);
                let group = new BABYLON.AnimationGroup(item.name, this.scene.impl);
                group.addTargetedAnimation(animation, item);

                this._animationMap.set(item.name, group);
                this._animations?.push(group);
            }
        });
    }
    public static Creation(name: string, scene: CustomScene, opt: IModelOpt = {}) {
        return new SpineBakeModel(name, scene, opt);
    }
}

class TempRenderer implements IRenderer {
    private _autoClearColor: boolean = true;
    private _autoClearDepth: boolean = true;
    private _autoClearStencil: boolean = true;
    private _clearColor: [number, number, number, number] = [0, 0, 0, 0];
    private _clearDepth: number = 1.0;
    private _clearStencil: number = 0;
    private _target: ICustomRenderTarget | TAutoRenderTarget | ITempFrameBuffer;
    private _transparent: boolean = false;
    _isDisposed: boolean = false;
    onDisposeObservable: any;
    name?: string;
    constructor(camera: BABYLON.Camera, name: string, pass_tag: number, transparent: boolean) {
        this._transparent = transparent;
        this.name = name;
    }
    viewport(x: number, y: number, width: number, height: number): void {
        throw new Error("Method not implemented.");
    }
    autoClearColor(flag: boolean): void {
        this._autoClearColor = flag;
    }
    autoClearDepth(flag: boolean): void {
        this._autoClearDepth = flag;
    }
    autoClearStencil(flag: boolean): void {
        this._autoClearStencil = flag;
    }
    clearColor(r: number, g: number, b: number, a: number): void {
        this._clearColor = [r / 255, g / 255, b / 255, a / 255];
    }
    clearDepth(depth: number): void {
        this._clearDepth = depth;
    }
    clearStencil(value: number): void {
        this._clearStencil = value;
    }
    setRenderTarget(target: ICustomRenderTarget | TAutoRenderTarget | ITempFrameBuffer): void {
        this._target = target;
    }
    _render(ctx: WebGLRenderingContext, target: IPostprocessTarget): any {
        let result = null;
        if (target) {
            ctx.bindFramebuffer(ctx.FRAMEBUFFER, target.framebuffer);
            ctx.viewport(target.x || 0, target.y || 0, target.width, target.height);
            ctx.scissor(target.x || 0, target.y || 0, target.width, target.height);
        } else {
            ctx.bindFramebuffer(ctx.FRAMEBUFFER, null);
            let info = getContentWH();
            ctx.viewport(info.renderX, info.renderY, info.renderWidth, info.renderHeight);
            ctx.scissor(info.renderX, info.renderY, info.renderWidth, info.renderHeight);
        }

        let clear = 0;
        if (this._autoClearColor) { clear |= ctx.COLOR_BUFFER_BIT; ctx.clearColor(...this._clearColor); }
        if (this._autoClearDepth) { clear |= ctx.DEPTH_BUFFER_BIT; ctx.depthMask(true); ctx.clearDepth(this._clearDepth); }
        if (this._autoClearStencil) { clear |= ctx.STENCIL_BUFFER_BIT; ctx.clearStencil(this._clearStencil); }
        if (clear != 0) { ctx.clear(clear); }
        if (this._transparent) { ctx.enable(ctx.BLEND) } else { ctx.disable(ctx.BLEND); }
        ctx.blendColor(0, 0, 0, 0);

        return result;
    }
    dispose(): void {
        this._isDisposed = true;
    }

}

export class TempShadowGenerator extends TempRenderer implements IShadowGenerator {
    private _atlasSize: number;
    private _minZ: number;
    private _maxZ: number;
    private _frustumSize: number;
    private _bias: number;
    private _normalBias: number;
    private _depthScale: number;
    public generator: BABYLON.ShadowGenerator;
    constructor(
        public scene: BABYLON.Scene,
        public light: BABYLON.DirectionalLight,
        atlasize: number,
        pass: Constants.PassTag,
        bias: number = 0.005,
        normalBias: number = 0.005,
        depthScale: number = 10,
        minZ: number = 0.1,
        maxZ: number = 100,
        frustumSize: number = 10,
    ) {
        super(scene.activeCamera, `Shadow`, pass, false);

        this.generator = new BABYLON.ShadowGenerator(atlasize, light);

        this.bias = bias;
        this.normalBias = normalBias;
        this.minZ = minZ;
        this.maxZ = maxZ;
        this.frustumSize = frustumSize;
        this.depthScale = depthScale;
    }

    public set bias(val: number) {
        this.generator.bias = val;
    }
    public get bias() {
        return this._bias;
    }

    public set normalBias(val: number) {
        this.generator.normalBias = val;
    }
    public get normalBias() {
        return this._normalBias;
    }

    public set minZ(val: number) {
        this.light.shadowMinZ = val;
    }
    public get minZ() {
        return this._minZ;
    }

    public set maxZ(val: number) {
        this.light.shadowMaxZ = val;
    }
    public get maxZ() {
        return this._maxZ;
    }

    public set frustumSize(val: number) {
        this.light.shadowFrustumSize = val;
    }
    public get frustumSize() {
        return this._frustumSize;
    }

    public set depthScale(val: number) {
        this.generator.depthScale = val;
    }
    public get depthScale() {
        return this._depthScale;
    }
}

export class TempRenderPipeline implements IRenderPipeline {
    public readonly predepthRenderer: TempRenderer;
    public readonly opaqueRenderer: TempRenderer;
    public readonly opaqueTextureRenderer: TempRenderer;
    public readonly skyRenderer: TempRenderer;
    public readonly transparentRenderer: TempRenderer;
    public readonly shadowGenerator: TempShadowGenerator;

    public readonly PASS_SHADOW = Constants.PASS_TAG_01;
    public readonly PASS_PREDEPTH = Constants.PASS_TAG_02;
    public readonly PASS_OPAQUE = Constants.PASS_TAG_03;
    public readonly PASS_SKY = Constants.PASS_TAG_04;
    public readonly PASS_OPAQUE_TEXTURE = Constants.PASS_TAG_05;
    public readonly PASS_TRANSPARENT = Constants.PASS_TAG_06;
    private _renders: IRenderer[] = [];

    constructor(
        public readonly scene: BABYLON.Scene,
        public readonly maincamera: BABYLON.Camera,
        public readonly shadowLight: BABYLON.DirectionalLight,
        public readonly predepthEanble = true,
        public readonly opaqueEanble = true,
        public readonly opaqueTextureEanble = true,
        public readonly skyEanble = true,
        public readonly transparentEanble = true,
    ) {
        if (this.shadowLight) {
            let atlasize = 1024;
            this.shadowGenerator = new TempShadowGenerator(scene, shadowLight, atlasize, this.PASS_SHADOW);
            this._renders.push(this.shadowGenerator);
        }
        // if (this.predepthEanble) {
        //     this.predepthRenderer = new TempRenderer(maincamera, "PreDepth", this.PASS_PREDEPTH, false);
        //     this._renders.push(this.predepthRenderer);
        // }
        // if (this.opaqueEanble) {
        this.opaqueRenderer = new TempRenderer(maincamera, "Opaque", this.PASS_OPAQUE, false);
        this._renders.push(this.opaqueRenderer);
        // }
        // if (this.skyEanble) {
        //     this.skyRenderer = new TempRenderer(maincamera, "Sky", this.PASS_SKY, false);
        //     this.skyRenderer.setRenderTarget(this.rendertarget);
        //     this._renders.push(this.skyRenderer);
        // }
        // if (this.opaqueTextureEanble) {
        //     this.opaqueTextureRenderer = new TempRenderer(maincamera, "OpaqueTexture", this.PASS_OPAQUE_TEXTURE, false);
        //     this.opaqueTextureRenderer.setRenderTarget(this.rendertarget);
        //     this._renders.push(this.opaqueTextureRenderer);
        // }
        // if (this.transparentEanble) {
        this.transparentRenderer = new TempRenderer(maincamera, "Transparent", this.PASS_TRANSPARENT, true);
        this._renders.push(this.transparentRenderer);
        // }

        this.scene.onBeforeDrawPhaseObservable.add(() => {
            this.opaqueRenderer._render(<any>SceneManagerData.engine._gl, this._tempTarget);
        });
        this.scene.onAfterDrawPhaseObservable.add(() => {
            this.transparentRenderer._render(<any>SceneManagerData.engine._gl, this._tempTarget);
        });
    }
    renderers: IRenderer[];
    queryPass(queue: number): Constants.PassTag {
        return queue < 3000 ? this.PASS_OPAQUE : this.PASS_TRANSPARENT;
    }
    get first() {
        return this._renders[0];
    }
    get final() {
        return this._renders[this._renders.length - 1];
    }

    private _tempTarget: IPostprocessTarget;
    render(target?: IPostprocessTarget): any {
        if (target) {
            BABYLONAction.updateEngineDimensions(target.width, target.height);
        } else {
            let info = getContentWH();
            BABYLONAction.updateEngineDimensions(info.renderWidth, info.renderHeight);
        }
        this._tempTarget = target;
        this.scene.render();
        this._tempTarget = undefined;
        return target;
    }
    dispose() {
        this._renders.forEach((renderer) => {
            renderer.dispose();
        });
        this._renders.length = 0;
    }
}

export class CustomScene extends SceneStruct implements ICustomScene {
    public readonly prefabPool: PrefabModelObjPool;
    private _renderpipeline: TempRenderPipeline;
    private _renderQuality: number = 1;
    public readonly postprocesses: Map<KeyPostprocess, PostProcess> = new Map();
    private width: number;
    private height: number;
    get renderpipeline(): TempRenderPipeline {
        return this._renderpipeline;
    }
    get renderWidth(): number {
        return this.width;
    }
    get renderHeight(): number {
        return this.height;
    }
    set renderQuality(val: 1.2 | 1.5 | 2.0 | 1.0 | 0.8 | 0.5) {
        // this._renderQuality = val;
        // this.resize();
    }
    /**
     * @hidden
     */
    protected pickmgr: ScenePickManager;
    resize(width: number, height: number): void {
        width = Math.floor(width * this._renderQuality / 2) * 2;
        height = Math.floor(height * this._renderQuality / 2) * 2;

        this.width = width;
        this.height = height;
    }
    render(ctx?: GlContext, target?: ITempFrameBuffer, postprocess: IPostprocessParam[] = [], blend: Constants.ALPHA_MODE = Constants.ALPHA_PREMULTIPLIED, forceResetContext: boolean = true): boolean {
        postprocess = resortPostprocessParams(postprocess, this.postprocesses);
        if (this.renderFlag === SceneRenderFlags.active) {
            if (postprocess.length > 0) {
                let postprocssTarget: RenderTarget = RenderTargetFactory.createRenderTarget(this.renderWidth, this.renderHeight, true);

                {
                    this.beforeRenderCallList.forEach((f) => f(this));

                    if (this.impl.animatables.length > 0 || this.impl.particleSystems.length > 0) {
                        this.isDirty = true;
                    }
                    this._frameCount++;

                    if (!this.autoOptimize || (this.isDirty && (this._frameCount % this.frameStep == 0))) {
                        this._renderpipeline.opaqueRenderer.autoClearColor(true);
                        this.impl.onBeforeDrawPhaseObservable.addOnce(this.onBeforeDrawPhaseOnceCall);
                        (<any>this.impl.getEngine())._setProgram(null);
                        this.impl.getEngine().wipeCaches(forceResetContext);
                        this.impl.getEngine().setAlphaMode(BABYLON.Engine.ALPHA_DISABLE);
                        this._renderpipeline.render(postprocssTarget);
                        this.isDirty = false;
                        this.impl.getEngine().wipeCaches(forceResetContext);
                    } else {
                        this.onBeforeDrawPhaseOnceCall();
                    }

                    this.afterRenderCallList.forEach((f) => f(this));
                }

                let needMoreCopy = target != undefined;

                RenderTargetFactory.unuseRenderTarget(<RenderTarget>postprocssTarget);
                let postprocssTexture: IPostprocessTexture = <RenderTarget>postprocssTarget;

                postprocess.forEach((item) => {
                    postprocssTarget = RenderTargetFactory.createRenderTarget(this.renderWidth, this.renderHeight, false);
                    let renderer = this.postprocesses.get(item.etype);
                    if (renderer) {
                        renderer.apply(item, [postprocssTexture], postprocssTarget, RenderTargetFactory.getContext());
                        RenderTargetFactory.unuseRenderTarget(<RenderTarget>postprocssTarget);
                        postprocssTexture = <RenderTarget>postprocssTarget;
                    }
                });

                if (needMoreCopy) {
                    let _context = RenderTargetFactory.getContext();
                    let renderer = this.postprocesses.get(KeyCopy);
                    if (renderer == undefined) {
                        renderer = new Copy();
                        this.postprocesses.set(KeyCopy, renderer);
                    }

                    // _context.gl.viewport(...target.rect);
                    // _context.gl.scissor(...target.rect);
                    _context.gl.disable(_context.gl.DEPTH_TEST);
                    _context.gl.disable(_context.gl.CULL_FACE);
                    PostProcess.blend(_context, blend);
                    renderer.apply({ etype: KeyCopy, alpha: 1, intensity: 1. }, [postprocssTexture], { x: target.rect[0], y: target.rect[1], width: target.rect[2], height: target.rect[3], framebuffer: target.frameBuffer }, RenderTargetFactory.getContext());

                    // RenderTargetFactory.unuseRenderTarget(postprocssTarget);
                    postprocssTexture = postprocssTarget;
                }
            } else {
                // let postprocssTarget: RenderTarget = RenderTargetFactory.createRenderTarget(this.renderWidth, this.renderHeight, true);
                {
                    this.beforeRenderCallList.forEach((f) => f(this));

                    if (this.impl.animatables.length > 0 || this.impl.particleSystems.length > 0) {
                        this.isDirty = true;
                    }
                    this._frameCount++;

                    if (!this.autoOptimize || (this.isDirty && (this._frameCount % this.frameStep == 0))) {
                        this.impl.onBeforeDrawPhaseObservable.addOnce(this.onBeforeDrawPhaseOnceCall);
                        (<any>this.impl.getEngine())._setProgram(null);
                        this.impl.getEngine().wipeCaches(forceResetContext);
                        this.impl.getEngine().setAlphaMode(BABYLON.Engine.ALPHA_DISABLE);
                        this._renderpipeline.render({ x: target.rect[0], y: target.rect[1], width: target.rect[2], height: target.rect[3], framebuffer: target.frameBuffer });
                        this.isDirty = false;
                        this.impl.getEngine().wipeCaches(forceResetContext);
                    } else {
                        this.onBeforeDrawPhaseOnceCall();
                    }

                    this.afterRenderCallList.forEach((f) => f(this));
                }
            }
            return true;
        } else {
            return false;
        }
    }
    constructor(
        name: string,
        width?: number, height?: number,
        shadowsEnabled: boolean = false,
        blockMaterialDirtyMechanism: boolean = true,
    ) {
        let stype = name;
        super(stype, BABYLONAction.engine())
        if (!this.impl.metadata) {
            this.impl.metadata = {};
        }
        // 设置动画的旋转方向
        this.impl.metadata.gltfAsLeftHandedSystem = true;
        this.impl.autoClear = false;
        this.impl.autoClearDepthAndStencil = false;
        this.impl.shadowsEnabled = shadowsEnabled;
        this.impl.blockMaterialDirtyMechanism = blockMaterialDirtyMechanism;

        this.camera = new BABYLON.FreeCamera(`${stype}_camera`, new BABYLON.Vector3(0, 0, -50), this.impl);
        // 设置相机模式——正交相机
        // camera.mode = BABYLON.FreeCamera.ORTHOGRAPHIC_CAMERA;
        this.camera.mode = BABYLON.Camera.ORTHOGRAPHIC_CAMERA;
        BABYLONAction.piOrthSize(<BABYLON.FreeCamera>this.camera, 20);
        this.addCamera(this.camera);
        this.setCurrCamera(this.camera.name);

        this.prefabPool = new PrefabModelObjPool(this);

        this._renderpipeline = new TempRenderPipeline(this.impl, this.camera, undefined, false, true, false, false, true);
        this._renderpipeline.transparentRenderer.autoClearColor(false);
        this._renderpipeline.transparentRenderer.autoClearDepth(false);
        this._renderpipeline.transparentRenderer.autoClearStencil(false);

        this.postprocesses.set(KeyCopy, new Copy());

        if (width && height) {
            this.width = width;
            this.height = height;
        }

        this.pickmgr = new ScenePickManager(this.impl);
    }
    dispose() {
        this._renderpipeline.dispose();
        super.dispose();
    }
    // public insertMesh(objName: string, opt?: IModelOpt, single: boolean = false, cacheCount: number = 4): IModelObj {
    //     if (opt && opt.path && opt.fileName) {
    //         if ((opt.imageSolts != undefined && opt.imageSolts.length > 0) || (opt.materialReplace != undefined && opt.materialReplace.length > 0) || (opt.materialSolts != undefined && opt.materialSolts.length > 0)) {
    //             let model = new ModelObj(objName, this, opt);
    //             model.setEnabled(true);
    //             return model;
    //         } else {
    //             let vtype = `${opt.path}${opt.fileName}`;
    //             if (this.prefabPool.cfgmap.get(vtype) == undefined) {
    //                 PrefabModelObjPool.registPrefabFactory(vtype, ModelPrefab.Creation);
    //                 PrefabModelObjPool.registPrefabResFactory(vtype, ModelPrefab.CreationResource);
    //                 (<IPrefabCreation>opt).single = single;
    //                 (<IPrefabCreation>opt).cacheCount = cacheCount;
    //                 this.prefabPool.registPrefabCfg(vtype, opt);
    //             }
    //             return this.prefabPool.createPrefab(objName, vtype);
    //         }
    //     } else {
    //         let model = new ModelObj(objName, this, opt);;
    //         model.setEnabled(true);
    //         return model;
    //     }
    // }
    
    public pick(e: PointerEvent, items?: IColliderItem[]): IColliderItem | undefined {
        let info = getContentWH();
        let pointerX = (e.clientX - info.styleLeft) / (info.styleSize.width - info.styleLeft - info.styleRight);
        let pointerY = (e.clientY - info.styleTop)  / (info.styleSize.height - info.styleBottom - info.styleTop);
        return this.pickmgr.pick(pointerX, pointerY, items);
    }
    public onPointerDown = (e: PointerEvent) => {
        let info = getContentWH();
        let pointerX = (e.clientX - info.styleLeft) / (info.styleSize.width - info.styleLeft - info.styleRight);
        let pointerY = (e.clientY - info.styleTop)  / (info.styleSize.height - info.styleBottom - info.styleTop);
        this.pickmgr.onPointerDown(pointerX, pointerY);
    }
    public onPointerUp = (e: PointerEvent) => {
        let info = getContentWH();
        let pointerX = (e.clientX - info.styleLeft) / (info.styleSize.width - info.styleLeft - info.styleRight);
        let pointerY = (e.clientY - info.styleTop)  / (info.styleSize.height - info.styleBottom - info.styleTop);
        this.pickmgr.onPointerUp(pointerX, pointerY);
    }
    /**
     * @hidden
     * @param target 
     * @param etype 
     * @param func 
     */
    public addEventListener(target: IColliderItem, etype: "pointerup" | "pointerdown", func: Function) {
        this.pickmgr.addEventListener(target, etype, func);
    }
    /**
     * @hidden
     * @param target 
     * @param etype 
     * @param func 
     */
    public addCollider(target: IColliderItem, minimum?: BABYLON.DeepImmutable<BABYLON.Vector3>, maximum?: BABYLON.DeepImmutable<BABYLON.Vector3>) {
        this.pickmgr.addCollider(target, minimum, maximum);
    }
    /**
     * 
     * @param clientX 
     * @param clientY 
     * @param target 用于参考的全局坐标(例如需要移动的目标节点的全局坐标)
     * @returns 
     */
    public clientPositionToScene(clientX: number, clientY: number, target?: BABYLON.Vector3): BABYLON.Vector3 {
        let info = getContentWH();
        let pointerX = (clientX - info.styleLeft) / (info.styleSize.width - info.styleLeft - info.styleRight);
        let pointerY = (clientY - info.styleTop)  / (info.styleSize.height - info.styleBottom - info.styleTop);
        pointerX =  pointerX * 2 - 1;
        pointerY = (1 - pointerY) * 2 - 1;

        let temp = MathDataPool.Vector3();
        let matrix = MathDataPool.Matrix();
        let source = MathDataPool.Vector3();
        if (target == undefined) {
            temp.copyFromFloats(0, 0, 0);
        } else {
            temp.copyFrom(target);
        }

        let transform = this.camera.getTransformationMatrix();
        transform.invertToRef(matrix);
        BABYLON.Vector3.TransformCoordinatesToRef(temp, transform, temp); 
        
        source.copyFromFloats(pointerX,      pointerY,         temp.z);
        BABYLON.Vector3._UnprojectFromInvertedMatrixToRef(source, matrix, temp);

        MathDataPool.Recycle(matrix);
        MathDataPool.Recycle(source);

        return temp;
    }
}

export class EngineState {
    public static DebugState(flag: boolean) {
        if (flag) {
            SceneManagerData.engine.performanceMonitor.enable()
        } else {
            SceneManagerData.engine.performanceMonitor.disable()
        }
    }
    public static ResourceState(state: IResourceState, performanceinfo: IPerformance) {
        let _12_mem_imgtexture = 0;
        OfflineProvider.piBABYLONTextureRecord.forEach((v, k) => {
            _12_mem_imgtexture += v[0] * v[1] * 4;
        });
        let _6_size_vertexbuffer = 0;
        OfflineProvider.piBABYLONGeometryRecord.forEach((v) => {
            _6_size_vertexbuffer += v;
        });
        let _2_count_gltf = 0;
        OfflineProvider.piBABYLONFileRecord.forEach((v, k) => {
            _2_count_gltf += k.endsWith(".gltf") ? 1 : 0;
        });

        let _23_count_vertex = 0;
        SceneStruct.SceneMap.forEach((v) => {
            _23_count_vertex += v.impl.getTotalVertices();
        });

        let _2_particlesystem = 0;
        let _1_animationgroup = 0;
        SceneStruct.SceneMap.forEach((v) => {
            if ((<any>v.impl).__hackAnime == undefined) {
                let old = v.impl.animate;
                v.impl.animate = function () {
                    let time = (performance || Date).now();
                    old.apply(this);
                    (<any>v.impl).__hackAnime = (performance || Date).now() - time;
                };
                (<any>v.impl).__hackAnime = 0;
            }
            _1_animationgroup += v.renderFlag == SceneRenderFlags.active ? (<any>v.impl).__hackAnime : 0;
            _2_particlesystem += (<Scene>v.impl).particleSystemTime;
            // v._modelMap.forEach((v) => {
            //     _2_particlesystem += v.particleSystems ? v.particleSystems.length : 0;
            // });
        });

        state._0_count_bindbuffer = 0;
        state._1_count_bindgroup = 0;
        state._2_count_gltf = _2_count_gltf;
        state._3_count_imgtexture = 0;
        state._4_count_shader = 0;
        state._5_count_pipeline = 0;
        state._6_size_vertexbuffer = _6_size_vertexbuffer;
        state._7_count_vertexbuffer = OfflineProvider.piBABYLONGeometryRecord.size;
        state._8_count_shadermeta = 0;
        state._9_mem_shadermeta = 0;
        state._10_mem_shader = 0;
        state._11_mem_bindbuffer = 0;
        state._12_mem_imgtexture = _12_mem_imgtexture;

        state._14_count_passset0 = 0;
        state._15_count_passset1 = 0;
        state._16_count_passset2 = 0;
        state._17_count_passbindgroups = 0;
        state._18_count_passshader = 0;
        state._19_count_passpipeline = 0;
        state._20_count_passdraw = 0;

        state._21_count_passmat = 0;
        state._22_count_passtexs = 0;
        state._23_count_vertex = _23_count_vertex;

        performanceinfo._0_animation = 0;
        performanceinfo._1_animationgroup = 0;
        performanceinfo._2_particlesystem = _2_particlesystem * 1000;
        performanceinfo._3_worldmatrix = 0;
        performanceinfo._4_culling = 0;
        performanceinfo._5_gltfanaly = 0;
        performanceinfo._6_drawobjs = 0;
        performanceinfo._7_uniformupdate = 0;
        performanceinfo._8_uniformbufferupdate = 0;
        performanceinfo._9_trailrenderer = 0;
    }
    public static SceneState(scene: SceneStruct, state: ISceneState) {
        // p3d.p3d_query_scene_state(Engine.native(), Engine.param(), scene.entity, Engine._stateInfo);
        let _2_count_particlesystem = 0;
        let _3_count_particles = 0;
        state._0_drawcalls = 0;
        state._5_count_trail = 0;
        state._6_count_trail_points = 0;
        state._4_count_animationgroup = 0;
        if (scene.renderFlag == SceneRenderFlags.active) {
            scene._modelMap.forEach((v) => {
                _2_count_particlesystem += v.particleSystems ? v.particleSystems.length : 0;
                v.particleSystems?.forEach(v => {
                    _3_count_particles += v.psTool.activeParticleList.length;
                });
            });
            state._0_drawcalls = scene.impl.getActiveMeshes().length;
            state._4_count_animationgroup = scene.impl.animationGroups.length;
        }
        state._1_count_vertex = scene.impl.getTotalVertices();
        state._2_count_particlesystem = _2_count_particlesystem;
        state._3_count_particles = _3_count_particles;
    }
    public static MaterialState(state: IMaterialState2) {
        // p3d.p3d_material_state(Engine.native(), Engine.param(), Engine._stateInfo);
        state._0_count = 0;
        state._1_ready = 0;
        state._2_tex0 = 0;
        state._3_tex0_ready = 0;
        state._4_tex1 = 0;
        state._5_tex1_ready = 0;
        state._6_tex2 = 0;
        state._7_tex2_ready = 0;
        state._8_tex3 = 0;
        state._9_tex3_ready = 0;
        state._10_tex4 = 0;
        state._11_tex4_ready = 0;
        state._12_tex5 = 0;
        state._13_tex5_ready = 0;
        state._14_tex6 = 0;
        state._15_tex6_ready = 0;
    }
    public static MeshState(state: IMeshState, scene: SceneStruct) {
        // p3d.p3d_mesh_state(Engine.native(), Engine.param(), scene?.entity, Engine._stateInfo);
        state._0_abstruct_mesh_count = 0;
        state._1_abstruct_mesh_count_enable = 0;
        state._2_mesh_count = 0;
        state._3_mesh_count_geometry_ready = 0;
        state._4_instance = 0;
    }
    public static TransformState(state: ITransformState, scene: SceneStruct) {
        // p3d.p3d_transform_state(Engine.native(), Engine.param(), scene?.entity, Engine._stateInfo);

        // scene.
        state._0_count = 0;
        state._1_enable = 0;
        state._2_global_eanble = 0;
        state._3_calc_local_matrix = 0;
        state._4_calc_world_matrix = 0;
        state._5_max_level = 0;
    }
    public static CameraState(state: ICameraState, camera: BABYLON.Camera) {
        let scene = camera._scene;
        if ((<any>scene).__hackculling === undefined) {
            (<any>scene).__hackculling = 0;
            let old: Function = (<any>scene)._evaluateActiveMeshes;
            (<any>camera._scene)._evaluateActiveMeshes = function () {
                let time = (performance || Date).now();
                old.apply(scene);
                (<any>scene).__hackculling = (performance || Date).now() - time;
            }
        }
        // p3d.p3d_camera_state(Engine.native(), Engine.param(), camera?.entity, Engine._stateInfo);
        state._0_includes = scene.meshes.length;
        state._1_culling = scene.getActiveMeshes().length;
        state._2_culling_time = (<any>scene).__hackculling * 1000;
    }
    public static TextureLoaderState(state: ITextureLoaderState) {
        let _9_imgtexture_mem = 0;
        OfflineProvider.piBABYLONTextureRecord.forEach((v, k) => {
            _9_imgtexture_mem += v[0] * v[1] * 4;
        });
        state._0_image_count = 0;
        state._1_image_fail = 0;
        state._2_image_success = 0;
        state._3_image_waiting = 0;
        state._4_texview_count = 0;
        state._5_texview_fail = 0;
        state._6_texview_success = 0;
        state._7_texview_waiting = 0;
        state._8_imgtexture = 0;
        state._9_imgtexture_mem = _9_imgtexture_mem;
        state._10_imgtextureview = 0;
        state._11_imgtextureview_mem = 0;
    }
    public static ErrorState(func: (entity: number, error: number) => void, flag) {
        // Engine._errorFlag = flag;
        // let count = p3d.p3d_errors(Engine.native(), Engine.param(), Engine._errorInfo, Engine._errorFlag);
        // for (let i = 0; i < count / 2; i++) {
        //     func(Engine._errorInfo[i * 2 + 0], Engine._errorInfo[i * 2 + 1]);
        // }
    }
}
