import { IModel, IModelObj } from "./modelobj";
import { ITransformNode } from "./transform";

export type IPrefabType = string | number;
export type IPrefabID = number;

/**
 * 预制内容创建指令
 */
export interface IPrefabCreation {
    /**
     * * 是否为单例内容
     * * 非单例内容将进行多实例缓存
     */
    single?: boolean;
    /**
     * * 非单例内容, 实例的最大缓存数目
     */
    cacheCount?: number;
    /**
     * 文件名称 - 不带 .gltf 等文件后缀
     */
    fileName?: string;
    /**
     * 文件所属路径
     * * gltf 文件会配合 LoaderTool.getResPath() LoaderTool.getModelPath()  使用
     */
    path?: string;
}

export interface IPrefab extends IModel {
    /**
     * * 获取指定序号的子内容 - 如果有加载流程 loadPromise 需在加载成功后使用
     * * 未指定 index 则获取预制件根内容
     * @param index 
     */
    getChild(index?: number): ITransformNode;
    /**
     * * 同步操作
     * * 设置预制件父级
     * @param parent 
     * @param indexOfParent 
     */
    setParent(parent: IPrefab, indexOfParent?: number): void;
    id(): number;
    loadPromise?: Promise<IPrefab>;
    /**
     * @hidden
     */
    resource?: IModel;
    /**
     * @hidden
     */
    pool: IPrefabPool;
    /**
     * * 同步操作
     * @param x 
     * @param y 
     * @param z 
     */
    setPosition(x: number, y: number, z: number): void;
    /**
     * * 同步操作
     * @param x 
     * @param y 
     * @param z 
     */
    setScale(x: number, y: number, z: number): void;
    /**
     * * 同步操作
     * @param x 
     * @param y 
     * @param z 
     */
    setRotate(x: number, y: number, z: number): void;
    setLookAt(x: number, y?: number, z?: number): void;
    resourceUnuse(): void;
}

export interface IPrefabTypeFactory {
    /**
     * 注册预制件资源的创建方法
     * @param type 
     * @param constructor 
     */
    registPrefabResFactory<C extends Object, IC extends IPrefabCreation>(type: IPrefabType, constructor: (scene: IPrefabTypeFactory, param: IC) => C): void;
    /**
     * 生成预制件的唯一ID 
     */
    prefabID(): number;
}

export interface IPrefabPool {
    /**
     * 注册预制件的创建方法
     * @param type 
     * @param param 
     */
    registPrefabCfg<IC extends IPrefabCreation>(type: IPrefabType, param: IC): void;
    /**
     * 创建一个内容
     * @param id 唯一ID由 IPrefabTypeFactory 分配 - 实际实现为 SceneStruct
     * @param type 
     * @param param 
     */
    createPrefab<T extends IPrefab, IC extends IPrefabCreation>(id: any, type: IPrefabType, param?: IC): T;
    /**
     * 销毁清空所有内容
     * @returns 
     */
    clearPrefabs: () => void;
}
