import { IPITextureInfo } from "../material_info";
import { BABYLON } from "../render3d/babylon";
import { ShaderMaterial } from "./shader_material";
import { LoadExtraMaterial } from "../render3d/extends/material";
import { StandardMaterialMesh } from "./standardmaterial";
import { MathDataPool } from "../render3d/extends/math";
const KEY_DISTORTION_UV = "DistortionUV";
//laserEffect
BABYLON.Effect.ShadersStore[`${KEY_DISTORTION_UV}VertexShader`] = `
    precision highp float;

    attribute vec3 position;
    attribute vec3 color;
    attribute vec2 uv;

    attribute vec4 a_diffuseTilloff;
    attribute vec4 a_maskTilloff;
    attribute vec4 a_opacityTilloff;
    attribute vec4 a_distortionInfo;
    attribute vec4 a_control;

    uniform mat4 worldViewProjection;

    varying vec2 vDiffuseUV;
    varying vec2 vMaskUV;
    varying vec2 vMaskDiff;
    varying vec4 vColor;
    varying vec2 vOpacityUV;
    
    varying vec4 diffuseMatrix;
    varying vec4 maskMatrix;
    varying vec4 opacityMatrix;
    varying vec4 distortionInfo;
    varying vec4 control;

    void main(void) {
        vec3 positionUpdated = position;
        vec2 uvUpdated = uv;

        gl_Position = worldViewProjection*vec4(positionUpdated, 1.0);

        control = a_control;
        distortionInfo = a_distortionInfo;
        control = a_control;

        float progress  = control.x;
        float maskValueFlow = control.y;
        float maskUVFlow = 1.0 - maskValueFlow;
        vec2 diffValue = progress * distortionInfo.zw;

        vDiffuseUV = vec2(uvUpdated * diffuseMatrix.xy + diffuseMatrix.zw);

        vMaskUV = vec2(uvUpdated * maskMatrix.xy + maskMatrix.zw);
        vMaskUV = vMaskUV + maskUVFlow * diffValue;

        vMaskDiff = maskValueFlow * diffValue;

        vColor = color;
        vOpacityUV = vec2(uvUpdated * opacityMatrix.xy + opacityMatrix.zw);
    }
`;

BABYLON.Effect.ShadersStore[`${KEY_DISTORTION_UV}FragmentShader`] =
    `
    precision highp float;

    varying vec2 vDiffuseUV;
    varying vec2 vMaskUV;
    varying vec2 vMaskDiff;
    varying vec4 vColor;
    varying vec2 vOpacityUV;

    varying vec4 distortionInfo;
    varying vec4 control;

    uniform sampler2D diffuseTexture;
    uniform sampler2D maskTexture;
    uniform sampler2D mask2Texture;
    uniform sampler2D opacityTexture;
    unfiorm float visibility;

    uniform vec4 tintColor;

    vec2 uvAddress(vec2 uv, mat4 matrix) {
        vec2 mode  = vec2(matrix*vec4(0., 0., 0., 1.0));
        vec4 tilloff    = vec4(matrix[0][3], matrix[1][3], matrix[2][3], matrix[3][3]);

        vec2 h = 0.5 * uv;
        vec2 n = floor(uv);
        vec2 a = 1.0 - floor(h - floor(h) + 0.5);

        vec2 repeat = abs(uv - n);
        vec2 mirrorRepeat = abs(1.0 - a - repeat + 2.0 * a * repeat);
        vec2 clamp2edge = abs(abs(h) - abs(h - 0.5) + 0.5);

        vec2 temp = mix(
                clamp2edge,
                mix(
                    repeat,
                    mirrorRepeat,
                    vec2(step(2.0, mode.x), step(2.0, mode.y))
                ),
                vec2(step(1.0, mode.x), step(1.0, mode.y))
            );

        return temp * tilloff.xy + tilloff.zw;
        // return uv;
    }

    void main(void) {
        vec4 textureMask = texture2D(maskTexture, vMaskUV);
        vec4 textureMask2 = texture2D(mask2Texture, vMaskUV);
        vec2 uvDiff = vec2(textureMask.r, textureMask2.g) * 2. - 1. + vMaskDiff;

        vec4 textureColor = texture2D(diffuseTexture, vDiffuseUV + uvDiff * distortionInfo.xy);
        vec4 result = textureColor * tintColor;

        if (control.w > 0.0) {
            result *= texture2D(opacityTexture, vOpacityUV).b;
        }
        result *= mix(vec4(1.0), vColor, control.z);

        result.a *= visibility;

        gl_FragColor = result;
    }
`;

interface IPIMaterialStandard {
    /**
     * 材质类型
     */
    type: string;

    /**
     * 共用材质的ID
     */
    instanceID?: string;

    alphaMode: number;
    needAlphaBlend?: 1;
    zWrite?: 1;

    disableLighting: boolean;
    diffuseColor?: [number, number, number];
    specularColor?: [number, number, number];
    specularPower?: number;
    ambientColor?: [number, number, number];
    reflectivityColor: [number, number, number];

    alpha: number;
    cull: string;

    useAlphaFromDiffuseTexture: boolean;
    diffuseTexture?: IPITextureInfo;

    opacityTexture?: IPITextureInfo;

    emissionColor: [number, number, number];
    emissionTexture?: IPITextureInfo;

    bumpTexture?: IPITextureInfo;
    normalScale?: number;

    useLightmapAsShadowmap?: 1;
    lightmapTexture?: IPITextureInfo;

    // Fresnel
    EFBias?: number;
    EFPower?: number;
    EFLeft?: [number, number, number];
    EFRight?: [number, number, number];
    emissiveMapLevel?: number;

    // alphaCutOff
    alphaCutOff?: number;
    /**
     * cutoffEdgeThickness
     */
    coET?: number;
    /**
     * cutoffEdgeColor
     */
    coEColor?: [number, number, number];
    /**
     * cutoffEdgeLevel
     */
    coEL?: number;
    /**
     * cutoffEdgeAlphaDiff
     */
    coEAD?: number;

    fog?: boolean;

    alphaClip?: boolean;
    renderQueue?: number;
    diffuseOU?: number;
    diffuseOV?: number;
    emissionOU?: number;
    emissionOV?: number;
    opacityOU?: number;
    opacityOV?: number;
    diffuseLevel?: number;
    opacityLevel?: number;

    ambientTexture?: IPITextureInfo;
    ambientMapLevel?: number;

    detailTexture?: IPITextureInfo;
    detailMapLevel?: number;
    detailDiffuseBlend?: number;
    detailBumpScale?: number;
    detailRoughnessBlend?: number;

    // Mask
    maskTexture?: IPITextureInfo;
    mask2Texture?: IPITextureInfo;
    maskFlowMode?: 0 | 1;
    distortionX?: number;
    distortionY?: number;
    distortionSX?: number;
    distortionSY?: number;
    opacityFromMaskBlue?: 0 | 1;
    vertexColorFactor?: 0 | 1;
}

export class DistortionUVMaterial extends ShaderMaterial {
    public static readonly KEY_DISTORTION_UV = "DistortionUV";
    public static readonly KEY_OPACITY = "Opacity";

    private _time: number = Date.now();
    private _distortionInfo: BABYLON.Vector4 = MathDataPool.Vector4(0, 0, 0, 0);
    private _tintColor: BABYLON.Color4 = new BABYLON.Color4(0, 0, 0, 0);
    private _control: BABYLON.Vector4 = MathDataPool.Vector4(0, 0, 0, 0);
    public _emissiveColor: BABYLON.Color3 = MathDataPool.Color3(1, 1, 1);
    public get emissiveColor() {
        return this._emissiveColor;
    }
    public set emissiveColor(value: BABYLON.Color3) {
        this._emissiveColor = value;
    }
    public get distortionX() {
        return this._distortionInfo.x;
    }
    public set distortionX(value: number) {
        this._distortionInfo.x = value;
    }
    public get distortionY() {
        return this._distortionInfo.y;
    }
    public set distortionY(value: number) {
        this._distortionInfo.y = value;
    }
    public get distortionSpeedX() {
        return this._distortionInfo.z;
    }
    public set distortionSpeedX(value: number) {
        this._distortionInfo.z = value;
    }
    public get distortionSpeedY() {
        return this._distortionInfo.w;
    }
    public set distortionSpeedY(value: number) {
        this._distortionInfo.w = value;
    }
    public get alpha() {
        return this._tintColor.a;
    }
    public set alpha(value: number) {
        this._tintColor.a = value;
    }
    public get maskFlowMode() {
        return this._control.y;
    }
    public set maskFlowMode(value: number) {
        this._control.y = value;
    }
    public get vertexColorFactor() {
        return this._control.z;
    }
    public set vertexColorFactor(value: number) {
        this._control.z = value;
    }
    public get opacityFromMaskBlue() {
        return this._control.w > 0;
    }
    public set opacityFromMaskBlue(value: boolean) {
        this._control.w = value ? 1 : 0;
    }
    private _diffuseMatrices: BABYLON.Matrix = MathDataPool.Matrix();
    private _maskMatrices: BABYLON.Matrix = MathDataPool.Matrix();
    private _mask2Matrices: BABYLON.Matrix = MathDataPool.Matrix();
    private _opacityMatrices: BABYLON.Matrix = MathDataPool.Matrix();
    constructor(babylonScene: any, materialName: string) {
        super(
            materialName,
            babylonScene,
            `${DistortionUVMaterial.KEY_DISTORTION_UV}`,
            {
                attributes: ["position", "uv", "color"],
                uniforms: ["worldViewProjection", "tintColor", "distortionInfo", "diffuseMatrix", "maskMatrix", "mask2Matrix", "opacityMatrix", "control", "visibility"],
                samplers: ["diffuseTexture", "maskTexture", "mask2Texture", "opacityTexture"],
                needAlphaBlending: true
            }
        );
        this._tintColor.r = this._emissiveColor.r;
        this._tintColor.g = this._emissiveColor.g;
        this._tintColor.b = this._emissiveColor.b;
        this.setColor4('tintColor', this._tintColor);
        this.setVector4('distortionInfo', this._distortionInfo);
        this.setMatrix('diffuseMatrix', this._diffuseMatrices);
        this.setMatrix('maskMatrix', this._maskMatrices);
        this.setMatrix('mask2Matrix', this._mask2Matrices);
        this.setMatrix('opacityMatrix', this._opacityMatrices);
        this.setVector4("control", this._control);
    }
    private _diffuseTexture: BABYLON.Texture;
    private _maskTexture: BABYLON.Texture;
    private _mask2Texture: BABYLON.Texture;
    private _opacityTexture: BABYLON.Texture;

    public get diffuseTexture(): BABYLON.Texture {
        return this._diffuseTexture;
    }
    public get maskTexture(): BABYLON.Texture {
        return this._maskTexture;
    }
    public get mask2Texture(): BABYLON.Texture {
        return this._mask2Texture;
    }
    public get opacityTexture(): BABYLON.Texture {
        return this._opacityTexture;
    }
    public set diffuseTexture(val: BABYLON.Texture) {
        this._diffuseTexture = val;
        this.setTexture("diffuseTexture", val);
    }
    public set maskTexture(val: BABYLON.Texture) {
        this._maskTexture = val;
        this.setTexture("maskTexture", val);
    }
    public set mask2Texture(val: BABYLON.Texture) {
        this._mask2Texture = val;
        this.setTexture("mask2Texture", val);
    }
    public set opacityTexture(val: BABYLON.Texture) {
        this._opacityTexture = val;
        this.setTexture("opacityTexture", val);
    }

    isReady(mesh?: BABYLON.AbstractMesh, useInstances?: boolean): boolean {
        let result = !!this.diffuseTexture && !!this.maskTexture && !!this.mask2Texture;
        return result && super.isReady(mesh, useInstances);
    }
    bind(world: BABYLON.Matrix, mesh?: StandardMaterialMesh, effectOverride?: BABYLON.Effect): void {
        let matData = mesh?.materialData || this;
        this._tintColor.r = matData.emissiveColor.r;
        this._tintColor.g = matData.emissiveColor.g;
        this._tintColor.b = matData.emissiveColor.b;
        this._tintColor.a = matData.alpha;
        this._diffuseMatrices.copyFrom(matData.diffuseTexture.getTextureMatrix());
        this._maskMatrices.copyFrom(matData.maskTexture.getTextureMatrix());
        this._mask2Matrices.copyFrom(this.mask2Texture.getTextureMatrix());
        if (this.opacityTexture) {
            this._opacityMatrices.copyFrom(matData.opacityTexture.getTextureMatrix());
        }
        this._control.x = (Date.now() - this._time) / 1000;

        this.setColor4('tintColor', this._tintColor.clone());
        this.setVector4('distortionInfo', this._distortionInfo.clone());
        this.setMatrix('diffuseMatrix', this._diffuseMatrices.clone());
        this.setMatrix('maskMatrix', this._maskMatrices.clone());
        this.setMatrix('mask2Matrix', this._mask2Matrices.clone());
        this.setMatrix('opacityMatrix', this._opacityMatrices.clone());
        this.setVector4("control", this._control.clone());

        this.setFloat("visibility", mesh.visibility);

        this.getScene().resetCachedMaterial();
        super.bind(world, mesh, effectOverride);
    }
    dispose(forceDisposeEffect?: boolean, forceDisposeTextures?: boolean, notBoundToMesh?: boolean): void {
        this.maskTexture?.dispose();
        this.mask2Texture?.dispose();
        this.opacityTexture?.dispose();
        this.diffuseTexture?.dispose();
        MathDataPool.Recycle(this._maskMatrices);
        MathDataPool.Recycle(this._mask2Matrices);
        MathDataPool.Recycle(this._diffuseMatrices);
        MathDataPool.Recycle(this._opacityMatrices);
        MathDataPool.Recycle(this._emissiveColor);
        MathDataPool.Recycle(this._control);
        MathDataPool.Recycle(this._distortionInfo);
        this._maskMatrices  = undefined;
        this._mask2Matrices     = undefined;
        this._diffuseMatrices   = undefined;
        this._opacityMatrices   = undefined;
        this._emissiveColor   = undefined;
        this._control   = undefined;
        this._distortionInfo   = undefined;
        super.dispose(forceDisposeEffect, forceDisposeTextures, notBoundToMesh);
    }
}

//#region PiStandard
function create(context: string, babylonScene: any, materialName: string, index: number, matInfo: any, babylonDrawMode: number): any {
    let mat = new DistortionUVMaterial(
        babylonScene,
        materialName
    );
    let materialInfo: IPIMaterialStandard = matInfo;

    mat.distortionX = materialInfo.distortionX || 0;
    mat.distortionY = materialInfo.distortionY || 0;
    mat.distortionSpeedX = materialInfo.distortionSX || 0;
    mat.distortionSpeedY = materialInfo.distortionSY || 0;
    mat.emissiveColor.copyFromFloats(materialInfo.diffuseColor[0] || 1, materialInfo.diffuseColor[1] || 1, materialInfo.diffuseColor[2] || 1);
    mat.alpha = materialInfo.alpha || 0;
    mat.maskFlowMode = materialInfo.maskFlowMode || 0;
    mat.vertexColorFactor = materialInfo.vertexColorFactor || 0;
    mat.opacityFromMaskBlue = !!materialInfo.opacityFromMaskBlue;

    return mat;
}

function loadAsync(context: string, materialInfo: IPIMaterialStandard, mat: DistortionUVMaterial, _loader: any): Promise<void> {
    let promises: Promise<any>[] = [];

    if (materialInfo.diffuseTexture) {
        promises.push(_loader.loadTextureInfoAsync(context + "/PI_material/diffuseTexture", materialInfo.diffuseTexture, function (texture: BABYLON.Texture) {
            mat.diffuseTexture = texture;
            // mat.setTexture('diffuseTexture', texture);
            if (materialInfo.diffuseTexture.offset) {
                (<any>texture).uOffset = materialInfo.diffuseTexture.offset[0];
                (<any>texture).vOffset = materialInfo.diffuseTexture.offset[1];
            }
            if (materialInfo.diffuseTexture.scale) {
                (<any>texture).uScale = materialInfo.diffuseTexture.scale[0];
                (<any>texture).vScale = materialInfo.diffuseTexture.scale[1];
            }
            (<any>texture).level = materialInfo.diffuseTexture.level || materialInfo.diffuseLevel || 1.0;
            if (materialInfo.diffuseTexture.hasAlpha) {
                texture.hasAlpha = true;
            }
            if (materialInfo.diffuseOU) {
                texture.offsetSpeedU = materialInfo.diffuseOU;
            }
            if (materialInfo.diffuseOV) {
                texture.offsetSpeedV = materialInfo.diffuseOV;
            }
        }));
    }

    if (materialInfo.maskTexture) {
        promises.push(_loader.loadTextureInfoAsync(context + "/PI_material/maskTexture", materialInfo.maskTexture, function (texture: BABYLON.Texture) {
            mat.maskTexture = texture;
            // mat.setTexture('maskTexture', texture);
            if (materialInfo.maskTexture.offset) {
                (<any>texture).uOffset = materialInfo.maskTexture.offset[0];
                (<any>texture).vOffset = materialInfo.maskTexture.offset[1];
            }
            if (materialInfo.maskTexture.scale) {
                (<any>texture).uScale = materialInfo.maskTexture.scale[0];
                (<any>texture).vScale = materialInfo.maskTexture.scale[1];
            }
        }));
        if (materialInfo.mask2Texture == undefined) {
            promises.push(_loader.loadTextureInfoAsync(context + "/PI_material/maskTexture", materialInfo.maskTexture, function (texture: BABYLON.Texture) {
                mat.mask2Texture = texture;
                // mat.setTexture('mask2Texture', texture);
            }));
        }
    }
    if (materialInfo.mask2Texture) {
        promises.push(_loader.loadTextureInfoAsync(context + "/PI_material/mask2Texture", materialInfo.mask2Texture, function (texture: BABYLON.Texture) {
            mat.mask2Texture = texture;
            // mat.setTexture('mask2Texture', texture);
            if (materialInfo.mask2Texture.offset) {
                (<any>texture).uOffset = materialInfo.mask2Texture.offset[0];
                (<any>texture).vOffset = materialInfo.mask2Texture.offset[1];
            }
            if (materialInfo.mask2Texture.scale) {
                (<any>texture).uScale = materialInfo.mask2Texture.scale[0];
                (<any>texture).vScale = materialInfo.mask2Texture.scale[1];
            }
        }));
    }

    if (materialInfo.opacityTexture) {
        promises.push(_loader.loadTextureInfoAsync(context + "/PI_material/opacityTexture", materialInfo.opacityTexture, function (texture: BABYLON.Texture) {
            mat.opacityTexture = texture;
            // mat.setTexture('opacityTexture', texture);
            if (materialInfo.opacityTexture.offset) {
                (<any>texture).uOffset = materialInfo.opacityTexture.offset[0];
                (<any>texture).vOffset = materialInfo.opacityTexture.offset[1];
            }
            if (materialInfo.opacityTexture.scale) {
                (<any>texture).uScale = materialInfo.opacityTexture.scale[0];
                (<any>texture).vScale = materialInfo.opacityTexture.scale[1];
            }
            (<any>texture).level = materialInfo.opacityTexture.level || materialInfo.opacityLevel || 1.0;
            if (materialInfo.opacityOU) {
                texture.offsetSpeedU = materialInfo.opacityOU;
            }
            if (materialInfo.opacityOV) {
                texture.offsetSpeedV = materialInfo.opacityOV;
            }
        }));
    }

    return Promise.all(promises).then(function () {
        if (materialInfo.zWrite !== undefined) {
            mat.forceDepthWrite = true;
        }
        if (materialInfo.fog !== undefined) {
            mat.fogEnabled = materialInfo.fog;
        }
        if (materialInfo.alpha !== undefined) {
            mat.alpha = materialInfo.alpha;
        }
        if (materialInfo.alphaMode !== undefined) {
            mat.alphaMode = materialInfo.alphaMode;
        }
        if (mat) {
            mat.sideOrientation = _loader.babylonScene.useRightHandedSystem ? BABYLON.Material.CounterClockWiseSideOrientation : BABYLON.Material.ClockWiseSideOrientation;
        }
        if (materialInfo.cull === "off") {
            mat.backFaceCulling = false;
        }
        if (materialInfo.cull === "front") {
            mat.sideOrientation = !_loader.babylonScene.useRightHandedSystem ? BABYLON.Material.CounterClockWiseSideOrientation : BABYLON.Material.ClockWiseSideOrientation;
        }
        if (materialInfo.renderQueue) {
            if (materialInfo.renderQueue < 2450) {
                mat.transparencyMode = BABYLON.Material.MATERIAL_OPAQUE;
            }
            else if (materialInfo.renderQueue < 3000) {
                mat.transparencyMode = BABYLON.Material.MATERIAL_ALPHATEST;
            }
            else {
                if (materialInfo.alphaClip) {
                    mat.transparencyMode = BABYLON.Material.MATERIAL_ALPHATESTANDBLEND;
                } else {
                    mat.transparencyMode = BABYLON.Material.MATERIAL_ALPHABLEND;
                }
                mat.needAlphaBlending = () => true;
            }
        }

        (<any>mat)._isReadyForSubMesh = function (this: BABYLON.StandardMaterial, subMesh: BABYLON.SubMesh) {
            const defines = subMesh._materialDefines;
            if (!this.checkReadyOnEveryCall && subMesh.effect && defines) {
                // 如果不受灯光影响 - 则灯光脏标识不受外部影响
                if (this.disableLighting) {
                    defines._areLightsDirty = false;
                }
                if (defines._renderId === this.getScene().getRenderId()) {
                    return true;
                }
            }

            return false;
        };
    });
}

function loadAsyncNew(context: string, matInfo: any, mat: DistortionUVMaterial, _loader: any): void {
    let promises: Promise<any>[] = [];
    let materialInfo: IPIMaterialStandard = matInfo;

    if (materialInfo.diffuseTexture) {
        LoadExtraMaterial.diffuseTexture(mat, _loader, context, matInfo);
    }

    if (materialInfo.maskTexture) {
        LoadExtraMaterial.maskTexture(mat, _loader, context, matInfo);
        if (materialInfo.mask2Texture == undefined) {
            let info = matInfo;
            let texture = <BABYLON.Texture>_loader.loadTextureInfoAsync(context + "/PI_material/maskTexture", info.maskTexture);
            if (mat.mask2Texture) {
                mat.mask2Texture.dispose();
            }
            mat.mask2Texture = texture;
            // let numbers = info.maskTexture.scale;
            // if (numbers) {
            //     texture.uScale = numbers[0];
            //     texture.vScale = numbers[1];
            // }
            // numbers = info.maskTexture.offset;
            // if (numbers) {
            //     texture.uOffset = numbers[0];
            //     texture.vOffset = numbers[1];
            // }
            // let tilloff = info.maskTexture.extras?.atlasTilloff;
            // if (tilloff) {
            //     texture.atlasTilloff = tilloff;
            // }
        }
    }
    if (materialInfo.mask2Texture) {
        LoadExtraMaterial.mask2Texture(mat, _loader, context, matInfo);
    }

    if (materialInfo.opacityTexture) {
        LoadExtraMaterial.opacityTexture(mat, _loader, context, matInfo);
    }

    // return Promise.all(promises).then(function() {
    if (materialInfo.zWrite !== undefined) {
        mat.forceDepthWrite = true;
    }
    if (materialInfo.fog !== undefined) {
        mat.fogEnabled = materialInfo.fog;
    }
    if (materialInfo.alpha !== undefined) {
        mat.alpha = materialInfo.alpha;
    }
    if (materialInfo.alphaMode !== undefined) {
        mat.alphaMode = materialInfo.alphaMode;
    }
    if (mat) {
        mat.sideOrientation = _loader.babylonScene.useRightHandedSystem ? BABYLON.Material.CounterClockWiseSideOrientation : BABYLON.Material.ClockWiseSideOrientation;
    }
    if (materialInfo.cull === "off") {
        mat.backFaceCulling = false;
    }
    if (materialInfo.cull === "front") {
        mat.sideOrientation = !_loader.babylonScene.useRightHandedSystem ? BABYLON.Material.CounterClockWiseSideOrientation : BABYLON.Material.ClockWiseSideOrientation;
    }
    if (materialInfo.renderQueue) {
        if (materialInfo.renderQueue < 2450) {
            mat.transparencyMode = BABYLON.Material.MATERIAL_OPAQUE;
        }
        else if (materialInfo.renderQueue < 3000) {
            mat.transparencyMode = BABYLON.Material.MATERIAL_ALPHATEST;
        }
        else {
            if (materialInfo.alphaClip) {
                mat.transparencyMode = BABYLON.Material.MATERIAL_ALPHATESTANDBLEND;
            } else {
                mat.transparencyMode = BABYLON.Material.MATERIAL_ALPHABLEND;
            }
            mat.needAlphaBlending = () => true;
        }
    }

    (<any>mat)._isReadyForSubMesh = function (this: BABYLON.StandardMaterial, subMesh: BABYLON.SubMesh) {
        const defines = subMesh._materialDefines;
        if (!this.checkReadyOnEveryCall && subMesh.effect && defines) {
            // 如果不受灯光影响 - 则灯光脏标识不受外部影响
            if (this.disableLighting) {
                defines._areLightsDirty = false;
            }
            if (defines._renderId === this.getScene().getRenderId()) {
                return true;
            }
        }

        return false;
    };
    // });
}
//#endregion

export function registDistortionUVFactory() {
    BABYLON.GLTF2.Loader.Extensions.PI_material.regist(
        "DistortionUV",
        <any>create,
        <any>loadAsync
    );
    LoadExtraMaterial.loadCalls.set("DistortionUV", create);
    LoadExtraMaterial.loadPropertyCalls.set("DistortionUV", loadAsyncNew);
    BABYLON.GLTF2.Loader.Extensions.PI_material.regist(
        "distortionUV",
        <any>create,
        <any>loadAsync
    );
    LoadExtraMaterial.loadCalls.set("distortionUV", create);
    LoadExtraMaterial.loadPropertyCalls.set("distortionUV", loadAsyncNew);
}

// Effect
BABYLON.Effect.ShadersStore[`${DistortionUVMaterial.KEY_DISTORTION_UV}VertexShader`] = `
    precision highp float;

    attribute vec3 position;
    attribute vec2 uv;
    attribute vec4 color;

    uniform mat4 worldViewProjection;
    uniform mat4 diffuseMatrix;
    uniform mat4 maskMatrix;
    uniform mat4 opacityMatrix;
    uniform vec4 distortionInfo;
    uniform vec4 control;

    varying vec2 vDiffuseUV;
    varying vec2 vMaskUV;
    varying vec2 vMaskDiff;
    varying vec4 vColor;
    varying vec2 vOpacityUV;

    void main(void) {
        vec3 positionUpdated = position;
        vec2 uvUpdated = uv;

        gl_Position = worldViewProjection*vec4(positionUpdated, 1.0);

        float progress  = control.x;
        float maskValueFlow = control.y;
        float maskUVFlow = 1.0 - maskValueFlow;
        vec2 diffValue = progress * distortionInfo.zw;

        vDiffuseUV = vec2(diffuseMatrix*vec4(uvUpdated, 1.0, 0.0));
        vMaskUV = uvUpdated;
        vMaskUV = vMaskUV + maskUVFlow * diffValue;
        vMaskUV = vec2(maskMatrix*vec4(vMaskUV, 1.0, 0.0));
        vMaskDiff = maskValueFlow * diffValue;
        vColor = color;
        vOpacityUV = vec2(opacityMatrix*vec4(uvUpdated, 1.0, 0.0));
    }
`;

BABYLON.Effect.ShadersStore[`${DistortionUVMaterial.KEY_DISTORTION_UV}FragmentShader`] =
    `
    precision highp float;

    varying vec2 vDiffuseUV;
    varying vec2 vMaskUV;
    varying vec2 vMaskDiff;
    varying vec4 vColor;
    varying vec2 vOpacityUV;

    uniform sampler2D diffuseTexture;
    uniform sampler2D maskTexture;
    uniform sampler2D mask2Texture;
    uniform vec4 tintColor;
    uniform vec4 distortionInfo;
    uniform vec4 control;

    void main(void) {
        vec4 textureMask = texture2D(maskTexture, vMaskUV);
        vec4 textureMask2 = texture2D(mask2Texture, vMaskUV);
        vec2 uvDiff = vec2(textureMask.r, textureMask2.g) * 2. - 1. + vMaskDiff;

        vec4 textureColor = texture2D(diffuseTexture, vDiffuseUV + uvDiff * distortionInfo.xy);

        vec4 result = textureColor * tintColor;

        if (control.w > 0.0) {
            result *= texture2D(maskTexture, vOpacityUV).b;
        }

        result *= mix(vec4(1.0), vColor, control.z);

        if (result.a < 0.001) {
            discard;
        }

        gl_FragColor = result;
    }
`;