import { BABYLON } from "../render3d/babylon";
import { IPITextureInfo } from "../material_info";
import { ShaderMaterial } from "./shader_material";
import { LoadExtraMaterial } from "../render3d/extends/material";
import { StandardMaterialMesh } from "./standardmaterial";
import { MathDataPool } from "../render3d/extends/math";

const KEY_EFFECT_MASK = "EffectMask";
//laserEffect
BABYLON.Effect.ShadersStore[`${KEY_EFFECT_MASK}VertexShader`] = `
    precision highp float;

    attribute vec3 position;
    attribute vec2 uv;

    uniform mat4 worldViewProjection;
    uniform mat4 diffuseMatrix;
    uniform mat4 maskMatrix;

    varying vec2 vDiffuseUV;
    varying vec2 vMaskUV;

    void main(void) {
        vec3 positionUpdated = position;
        vec2 uvUpdated = uv;

        gl_Position = worldViewProjection*vec4(positionUpdated, 1.0);

        vDiffuseUV = vec2(diffuseMatrix*vec4(uvUpdated, 1.0, 0.0));
        vMaskUV = vec2(maskMatrix*vec4(uvUpdated, 1.0, 0.0));
    }
`;

BABYLON.Effect.ShadersStore[`${KEY_EFFECT_MASK}FragmentShader`] =
    `
    precision highp float;

    varying vec2 vDiffuseUV;
    varying vec2 vMaskUV;

    uniform sampler2D diffuseTexture;
    uniform sampler2D maskTexture;
    uniform vec4 tintColor;
    uniform vec2 maskInfo;
    uniform float visibility;

    void main(void) {
        float maskCutoff = maskInfo.y;
        float diff = fract(maskCutoff * 10.0) * 3.0;

        vec4 textureColor = texture2D(diffuseTexture, vDiffuseUV);
        vec4 textureMask = texture2D(maskTexture, vMaskUV);

        float noise = textureMask.r;
        if (diff > 2.0) {
            noise = textureMask.b;
        } else if (diff > 1.0) {
            noise = textureMask.g;
        }
        float displaceNoise = step(maskCutoff, min(noise, 0.99));

        vec4 replaceColor = tintColor;
        replaceColor.a *= textureColor.a;

        vec4 result = mix(textureColor, replaceColor, displaceNoise);
        result.a *= visibility;
        gl_FragColor = result;
    }
`;

interface IPIMaterialStandard {
    /**
     * 材质类型
     */
    type: string;

    /**
     * 共用材质的ID
     */
    instanceID?: string;

    alphaMode: number;
    needAlphaBlend?: 1;
    zWrite?: 1;

    disableLighting: boolean;
    diffuseColor?: [number, number, number];
    specularColor?: [number, number, number];
    specularPower?: number;
    ambientColor?: [number, number, number];
    reflectivityColor: [number, number, number];

    alpha: number;
    cull: string;

    useAlphaFromDiffuseTexture: boolean;
    diffuseTexture?: IPITextureInfo;

    opacityTexture?: IPITextureInfo;

    emissionColor: [number, number, number];
    emissionTexture?: IPITextureInfo;

    bumpTexture?: IPITextureInfo;
    normalScale?: number;

    useLightmapAsShadowmap?: 1;
    lightmapTexture?: IPITextureInfo;

    // Fresnel
    EFBias?: number;
    EFPower?: number;
    EFLeft?: [number, number, number];
    EFRight?: [number, number, number];
    emissiveMapLevel?: number;

    // alphaCutOff
    alphaCutOff?: number;
    /**
     * cutoffEdgeThickness
     */
    coET?: number;
    /**
     * cutoffEdgeColor
     */
    coEColor?: [number, number, number];
    /**
     * cutoffEdgeLevel
     */
    coEL?: number;
    /**
     * cutoffEdgeAlphaDiff
     */
    coEAD?: number;

    fog?: boolean;

    alphaClip?: boolean;
    renderQueue?: number;
    diffuseOU?: number;
    diffuseOV?: number;
    emissionOU?: number;
    emissionOV?: number;
    opacityOU?: number;
    opacityOV?: number;
    diffuseLevel?: number;
    opacityLevel?: number;

    ambientTexture?: IPITextureInfo;
    ambientMapLevel?: number;

    detailTexture?: IPITextureInfo;
    detailMapLevel?: number;
    detailDiffuseBlend?: number;
    detailBumpScale?: number;
    detailRoughnessBlend?: number;

    // Mask
    maskTexture?: IPITextureInfo;
    maskCutoff?: number;
}

export class EffectMaskMaterial extends ShaderMaterial {
    private _maskInfo: BABYLON.Vector2 = new BABYLON.Vector2(0, 0);
    private _tintColor: BABYLON.Color4 = new BABYLON.Color4(0, 0, 0, 0);
    public get maskCutoff() {
        return this._maskInfo.y;
    }
    public set maskCutoff(value: number) {
        this._maskInfo.y = value;
    }
    public get alpha() {
        return this._tintColor.a;
    }
    public set alpha(value: number) {
        this._tintColor.a = value;
    }
    public get r() {
        return this._tintColor.r;
    }
    public set r(value: number) {
        this._tintColor.r = value;
    }
    public get g() {
        return this._tintColor.g;
    }
    public set g(value: number) {
        this._tintColor.g = value;
    }
    public get b() {
        return this._tintColor.b;
    }
    public set b(value: number) {
        this._tintColor.b = value;
    }
    private _diffuseMatrices: BABYLON.Matrix = MathDataPool.Matrix();
    private _maskMatrices: BABYLON.Matrix = MathDataPool.Matrix();
    constructor(babylonScene: any, materialName: string) {
        super(
            materialName,
            babylonScene,
            `${KEY_EFFECT_MASK}`,
            {
                attributes: ["position", "uv"],
                uniforms: ["worldViewProjection", "tintColor", "maskInfo", "diffuseMatrix", "maskMatrix", "visibility"],
                samplers: ["diffuseTexture", "maskTexture"],
                needAlphaBlending: true
            }
        );

        this.setColor4('tintColor', this._tintColor);
        this.setVector2('maskInfo', this._maskInfo);
        this.setMatrix('diffuseMatrix', this._diffuseMatrices);
        this.setMatrix('maskMatrix', this._maskMatrices);
    }
    public diffuseTexture: BABYLON.Texture;
    public maskTexture: BABYLON.Texture;
    isReady(mesh?: BABYLON.AbstractMesh, useInstances?: boolean): boolean {
        let result = !!this.diffuseTexture && !!this.maskTexture;
        return result && super.isReady(mesh, useInstances);
    }
    bind(world: BABYLON.Matrix, mesh?: StandardMaterialMesh, effectOverride?: BABYLON.Effect): void {
        let matData = mesh?.materialData || this;

        this._diffuseMatrices.copyFrom(matData.diffuseTexture.getTextureMatrix());
        this._maskMatrices.copyFrom(matData.maskTexture.getTextureMatrix());

        this.maskCutoff = matData.maskCutoff;
        this.alpha = matData.alpha;

        this.setMatrix('diffuseMatrix', this._diffuseMatrices.clone());
        this.setMatrix('maskMatrix', this._maskMatrices.clone());
        this.setColor4('tintColor', this._tintColor.clone());
        this.setVector2('maskInfo', this._maskInfo.clone());

        this.setFloat("visibility", mesh.visibility);

        this.getScene().resetCachedMaterial();
        super.bind(world, mesh, effectOverride);
    }
    dispose(forceDisposeEffect?: boolean, forceDisposeTextures?: boolean, notBoundToMesh?: boolean): void {
        this.maskTexture.dispose();
        this.diffuseTexture.dispose();

        MathDataPool.Recycle(this._maskMatrices);
        MathDataPool.Recycle(this._diffuseMatrices);
        this._maskMatrices = undefined;
        this._diffuseMatrices = undefined;
        super.dispose(forceDisposeEffect, forceDisposeTextures, notBoundToMesh);
    }
}

//#region PiStandard
function create(context: string, babylonScene: any, materialName: string, index: number, matInfo: any, babylonDrawMode: number): any {
    let mat = new EffectMaskMaterial(
        babylonScene,
        materialName
    );
    let materialInfo: IPIMaterialStandard = matInfo;

    mat.maskCutoff = materialInfo.maskCutoff || 0;
    mat.r = materialInfo.diffuseColor[0] || 0;
    mat.g = materialInfo.diffuseColor[1] || 0;
    mat.b = materialInfo.diffuseColor[2] || 0;
    mat.alpha = materialInfo.alpha || 0;

    return mat;
}

function loadAsync(context: string, matInfo: any, mat: EffectMaskMaterial, _loader: any): Promise<void> {
    let promises: Promise<any>[] = [];
    let materialInfo: IPIMaterialStandard = matInfo;

    if (materialInfo.diffuseTexture) {
        promises.push(_loader.loadTextureInfoAsync(context + "/PI_material/diffuseTexture", materialInfo.diffuseTexture, function (texture: BABYLON.Texture) {
            mat.diffuseTexture = texture;
            mat.setTexture('diffuseTexture', texture);
            if (materialInfo.diffuseTexture.offset) {
                (<any>texture).uOffset = materialInfo.diffuseTexture.offset[0];
                (<any>texture).vOffset = materialInfo.diffuseTexture.offset[1];
            }
            if (materialInfo.diffuseTexture.scale) {
                (<any>texture).uScale = materialInfo.diffuseTexture.scale[0];
                (<any>texture).vScale = materialInfo.diffuseTexture.scale[1];
            }
            (<any>texture).level = materialInfo.diffuseTexture.level || materialInfo.diffuseLevel || 1.0;
            if (materialInfo.diffuseTexture.hasAlpha) {
                texture.hasAlpha = true;
            }
            if (materialInfo.diffuseOU) {
                texture.offsetSpeedU = materialInfo.diffuseOU;
            }
            if (materialInfo.diffuseOV) {
                texture.offsetSpeedV = materialInfo.diffuseOV;
            }
        }));
    }

    if (materialInfo.maskTexture) {
        promises.push(_loader.loadTextureInfoAsync(context + "/PI_material/maskTexture", materialInfo.maskTexture, function (texture: BABYLON.Texture) {
            mat.maskTexture = texture;
            mat.setTexture('maskTexture', texture);
            if (materialInfo.maskTexture.offset) {
                (<any>texture).uOffset = materialInfo.maskTexture.offset[0];
                (<any>texture).vOffset = materialInfo.maskTexture.offset[1];
            }
            if (materialInfo.maskTexture.scale) {
                (<any>texture).uScale = materialInfo.maskTexture.scale[0];
                (<any>texture).vScale = materialInfo.maskTexture.scale[1];
            }
        }));
    }

    return Promise.all(promises).then(function () {
        if (materialInfo.zWrite !== undefined) {
            mat.forceDepthWrite = true;
        }
        if (materialInfo.fog !== undefined) {
            mat.fogEnabled = materialInfo.fog;
        }
        if (materialInfo.alpha !== undefined) {
            mat.alpha = materialInfo.alpha;
        }
        if (materialInfo.alphaMode !== undefined) {
            mat.alphaMode = materialInfo.alphaMode;
        }
        if (mat) {
            mat.sideOrientation = _loader.babylonScene.useRightHandedSystem ? BABYLON.Material.CounterClockWiseSideOrientation : BABYLON.Material.ClockWiseSideOrientation;
        }
        if (materialInfo.cull === "off") {
            mat.backFaceCulling = false;
        }
        if (materialInfo.cull === "front") {
            mat.sideOrientation = !_loader.babylonScene.useRightHandedSystem ? BABYLON.Material.CounterClockWiseSideOrientation : BABYLON.Material.ClockWiseSideOrientation;
        }
        if (materialInfo.renderQueue) {
            if (materialInfo.renderQueue < 2450) {
                mat.transparencyMode = BABYLON.Material.MATERIAL_OPAQUE;
            }
            else if (materialInfo.renderQueue < 3000) {
                mat.transparencyMode = BABYLON.Material.MATERIAL_ALPHATEST;
            }
            else {
                if (materialInfo.alphaClip) {
                    mat.transparencyMode = BABYLON.Material.MATERIAL_ALPHATESTANDBLEND;
                } else {
                    mat.transparencyMode = BABYLON.Material.MATERIAL_ALPHABLEND;
                }
                mat.needAlphaBlending = () => true;
            }
        }

        (<any>mat)._isReadyForSubMesh = function (this: BABYLON.StandardMaterial, subMesh: BABYLON.SubMesh) {
            const defines = subMesh._materialDefines;
            if (!this.checkReadyOnEveryCall && subMesh.effect && defines) {
                // 如果不受灯光影响 - 则灯光脏标识不受外部影响
                if (this.disableLighting) {
                    defines._areLightsDirty = false;
                }
                if (defines._renderId === this.getScene().getRenderId()) {
                    return true;
                }
            }

            return false;
        };
    });
}

function loadAsyncNew(context: string, matInfo: any, mat: EffectMaskMaterial, _loader: any): void {
    let promises: Promise<any>[] = [];
    let materialInfo: IPIMaterialStandard = matInfo;

    if (materialInfo.diffuseTexture) {
        LoadExtraMaterial.diffuseTexture(mat, _loader, context, matInfo);
    }

    if (materialInfo.maskTexture) {
        LoadExtraMaterial.maskTexture(mat, _loader, context, matInfo);
    }

        if (materialInfo.zWrite !== undefined) {
            mat.forceDepthWrite = true;
        }
        if (materialInfo.fog !== undefined) {
            mat.fogEnabled = materialInfo.fog;
        }
        if (materialInfo.alpha !== undefined) {
            mat.alpha = materialInfo.alpha;
        }
        if (materialInfo.alphaMode !== undefined) {
            mat.alphaMode = materialInfo.alphaMode;
        }
        if (mat) {
            mat.sideOrientation = _loader.babylonScene.useRightHandedSystem ? BABYLON.Material.CounterClockWiseSideOrientation : BABYLON.Material.ClockWiseSideOrientation;
        }
        if (materialInfo.cull === "off") {
            mat.backFaceCulling = false;
        }
        if (materialInfo.cull === "front") {
            mat.sideOrientation = !_loader.babylonScene.useRightHandedSystem ? BABYLON.Material.CounterClockWiseSideOrientation : BABYLON.Material.ClockWiseSideOrientation;
        }
        if (materialInfo.renderQueue) {
            if (materialInfo.renderQueue < 2450) {
                mat.transparencyMode = BABYLON.Material.MATERIAL_OPAQUE;
            }
            else if (materialInfo.renderQueue < 3000) {
                mat.transparencyMode = BABYLON.Material.MATERIAL_ALPHATEST;
            }
            else {
                if (materialInfo.alphaClip) {
                    mat.transparencyMode = BABYLON.Material.MATERIAL_ALPHATESTANDBLEND;
                } else {
                    mat.transparencyMode = BABYLON.Material.MATERIAL_ALPHABLEND;
                }
                mat.needAlphaBlending = () => true;
            }
        }

        (<any>mat)._isReadyForSubMesh = function (this: BABYLON.StandardMaterial, subMesh: BABYLON.SubMesh) {
            const defines = subMesh._materialDefines;
            if (!this.checkReadyOnEveryCall && subMesh.effect && defines) {
                // 如果不受灯光影响 - 则灯光脏标识不受外部影响
                if (this.disableLighting) {
                    defines._areLightsDirty = false;
                }
                if (defines._renderId === this.getScene().getRenderId()) {
                    return true;
                }
            }

            return false;
        };
}

//#endregion

export function registEffectMaskFactory() {
    BABYLON.GLTF2.Loader.Extensions.PI_material.regist(
        "EffectMask",
        <any>create,
        <any>loadAsync
    );
    LoadExtraMaterial.loadCalls.set("EffectMask", create);
    LoadExtraMaterial.loadPropertyCalls.set("EffectMask", loadAsyncNew);
}