import { IPITextureInfo } from "../material_info";
import { BABYLON } from "../render3d/babylon";
import { ShaderMaterial } from "./shader_material";
import { LoadExtraMaterial } from "../render3d/extends/material";
const KEY_HSV = "HSVTEX";
//laserEffect
BABYLON.Effect.ShadersStore[`${KEY_HSV}VertexShader`] = `
    precision mediump float;

    attribute vec3 position;
    attribute vec4 color;
    attribute vec2 uv;

    uniform mat4 worldViewProjection;

    varying vec3 vHSV;
    varying float vAlpha;
    varying vec2 vUV;

    void main(void) {
        vec3 positionUpdated = position;
        vec2 uvUpdated = uv;

        gl_Position = worldViewProjection*vec4(positionUpdated, 1.0);

        vUV = uv;

        vHSV = color.rgb;
        vAlpha = color.a;
    }
`;

BABYLON.Effect.ShadersStore[`${KEY_HSV}FragmentShader`] =
    `
    precision mediump float;

    varying vec3 vHSV;
    varying float vAlpha;
    varying vec2 vUV;

    uniform sampler2D diffuseTexture;
    
    vec3 rgb2hsv(vec3 c)
    {
        vec4 K = vec4(0.0, -1.0 / 3.0, 2.0 / 3.0, -1.0);
        vec4 p = mix(vec4(c.bg, K.wz), vec4(c.gb, K.xy), step(c.b, c.g));
        vec4 q = mix(vec4(p.xyw, c.r), vec4(c.r, p.yzx), step(p.x, c.r));

        float d = q.x - min(q.w, q.y);
        float e = 1.0e-10;
        return vec3(abs(q.z + (q.w - q.y) / (6.0 * d + e)), d / (q.x + e), q.x);
    }

    vec3 hsv2rgb(vec3 c)
    {
        vec4 K = vec4(1.0, 2.0 / 3.0, 1.0 / 3.0, 3.0);
        vec3 p = abs(fract(c.xxx + K.xyz) * 6.0 - K.www);
        return c.z * mix(K.xxx, clamp(p - K.xxx, vec3(0.0), vec3(1.0)), c.y);
    }

    vec3 applyHSV(vec3 c, vec3 pHSV) {
        vec3 hsv = rgb2hsv(c.rgb);
        hsv.r += pHSV.r;
        c.rgb = hsv2rgb(hsv);

        // Note: When saturate is greater than 0, the formula is different from PS
        float gray = max(c.r, max(c.g, c.b)) + min(c.r, min(c.g, c.b));
        c.rgb = mix(c.rgb, vec3(0.5 * gray), -pHSV.g);

        return mix(c.rgb * (1.0 + pHSV.b), mix(c.rgb, vec3(1.0), pHSV.b), step(0., pHSV.b));
        // if (pHSV.b >= 0.0) {
        //     c.rgb = mix(c.rgb, vec3(1.0), pHSV.b);
        // } else {
        //     c.rgb *= 1.0 + pHSV.b;
        // }
        // return c;
    }

    void main(void) {
        vec4 textureColor = texture2D(diffuseTexture, vUV);
        textureColor.rgb = applyHSV(textureColor.rgb, vHSV);
        textureColor.a *= vAlpha;
        gl_FragColor = textureColor;
    }
`;

export class TextureHSVMaterial extends ShaderMaterial {
    public static readonly KEY = KEY_HSV;

    private _time: number = Date.now();
    private _distortionInfo: BABYLON.Vector4 = new BABYLON.Vector4(0, 0, 0, 0);
    private _tintColor: BABYLON.Color4 = new BABYLON.Color4(0, 0, 0, 0);
    private _control: BABYLON.Vector4 = new BABYLON.Vector4(0, 0, 0, 0);
    public _emissiveColor: BABYLON.Color3 = new BABYLON.Color3(1, 1, 1);
    public get emissiveColor() {
        return this._emissiveColor;
    }
    public set emissiveColor(value: BABYLON.Color3) {
        this._emissiveColor = value;
    }
    public get alpha() {
        return this._tintColor.a;
    }
    public set alpha(value: number) {
        this._tintColor.a = value;
    }
    constructor(materialName: string, babylonScene: any) {
        super(
            materialName,
            babylonScene,
            `${TextureHSVMaterial.KEY}`,
            {
                attributes: ["position", "uv", "color"],
                uniforms: ["worldViewProjection"],
                samplers: ["diffuseTexture"],
                needAlphaBlending: true
            }
        );
    }
    protected _diffuseTexture: BABYLON.Texture;
    public set diffuseTexture(tex: BABYLON.Texture) {
        this.setTexture('diffuseTexture', tex);
        this._diffuseTexture = tex;
    }
    public get diffuseTexture() {
        return this._diffuseTexture;
    }
    isReady(mesh?: BABYLON.AbstractMesh, useInstances?: boolean): boolean {
        let result = !!this._diffuseTexture;
        return result && super.isReady(mesh, useInstances);
    }
    bind(world: BABYLON.Matrix, mesh?: BABYLON.Mesh, effectOverride?: BABYLON.Effect): void {
        super.bind(world, mesh, effectOverride);
    }
    dispose(forceDisposeEffect?: boolean, forceDisposeTextures?: boolean, notBoundToMesh?: boolean): void {
        this._diffuseTexture?.dispose();
        super.dispose(forceDisposeEffect, forceDisposeTextures, notBoundToMesh);
    }
}
