import { IPITextureInfo } from "../material_info";
import { BABYLON } from "../render3d/babylon";
import { ShaderMaterial } from "./shader_material";
import { LoadExtraMaterial } from "../render3d/extends/material";
const KEY_HUE_SATURATE = "HueSaturate";
//laserEffect
BABYLON.Effect.ShadersStore[`${KEY_HUE_SATURATE}VertexShader`] = `
    precision mediump float;

    attribute vec3 position;
    attribute vec4 color;
    attribute vec2 uv;

    uniform mat4 worldViewProjection;

    varying vec3 vHSV;
    varying float vAlpha;
    varying vec2 vUV;

    void main(void) {
        vec3 positionUpdated = position;
        vec2 uvUpdated = uv;

        gl_Position = worldViewProjection*vec4(positionUpdated, 1.0);

        vUV = uv;

        vHSV = color.rgb;
        vAlpha = color.a;
    }
`;

BABYLON.Effect.ShadersStore[`${KEY_HUE_SATURATE}FragmentShader`] =
    `
    precision mediump float;

    varying vec3 vHSV;
    varying float vAlpha;
    varying vec2 vUV;

    uniform sampler2D diffuseTexture;
    
    const float A = 1.0/6.0;
    const float B = 1.0/2.0;
    const float C = 2.0/3.0;
    float hue2rgb(float p, float q, float t) {

        float a = step(0., t);
        float e = step(1., t);
        t = mix(t + 1.0, 
                mix(t, t - 1.0, e),
                a
            );

        float b = step(A, t);
        float c = step(B, t);
        float d = step(C, t);

        return mix(
            p + (q - p) * 6.0 * t,
            mix(
                q,
                mix(
                    p + (q - p) * (2.0/3.0 - t) * 6.0,
                    p,
                    d
                ),
                c
            ),
            b
        );
        
        // if(t < 0.0) t += 1.0;
        // if(t > 1.0) t -= 1.0;
        // if(t < 1.0/6.0) return p + (q - p) * 6.0 * t;
        // if(t < 1.0/2.0) return q;
        // if(t < 2.0/3.0) return p + (q - p) * (2.0/3.0 - t) * 6.0;
        // return p;
    }
    vec3 hslToRgb (float h, float s, float l) {
        float q = mix( l * (1.0 + s), l + s - l * s, step(0.5, l));
        float p = 2.0 * l - q;

        return mix(
            vec3(l, l, l),
            vec3(hue2rgb(p, q, h + 1.0/3.0), hue2rgb(p, q, h), hue2rgb(p, q, h - 1.0/3.0)),
            step(0.000001, abs(s))
        );
    }

    void main(void) {
        vec4 src = texture2D(diffuseTexture, vUV);
        float gray = max(src.r, max(src.g, src.b)) + min(src.r, min(src.g, src.b));
        vec3 rgb = hslToRgb(vHSV.x, vHSV.y, gray * 0.5);
        rgb = mix(rgb.rgb * (1.0 + vHSV.b), mix(rgb.rgb, vec3(1.0), vHSV.b), step(0., vHSV.b));
        gl_FragColor = vec4(rgb, src.w * vAlpha);
    }
`;

export class TextureHueSaturateMaterial extends ShaderMaterial {
    public static readonly KEY = KEY_HUE_SATURATE;

    private _time: number = Date.now();
    private _distortionInfo: BABYLON.Vector4 = new BABYLON.Vector4(0, 0, 0, 0);
    private _tintColor: BABYLON.Color4 = new BABYLON.Color4(0, 0, 0, 0);
    private _control: BABYLON.Vector4 = new BABYLON.Vector4(0, 0, 0, 0);
    public _emissiveColor: BABYLON.Color3 = new BABYLON.Color3(1, 1, 1);
    public get emissiveColor() {
        return this._emissiveColor;
    }
    public set emissiveColor(value: BABYLON.Color3) {
        this._emissiveColor = value;
    }
    public get alpha() {
        return this._tintColor.a;
    }
    public set alpha(value: number) {
        this._tintColor.a = value;
    }
    constructor(materialName: string, babylonScene: any) {
        super(
            materialName,
            babylonScene,
            `${TextureHueSaturateMaterial.KEY}`,
            {
                attributes: ["position", "uv", "color"],
                uniforms: ["worldViewProjection"],
                samplers: ["diffuseTexture"],
                needAlphaBlending: true
            }
        );
    }
    protected _diffuseTexture: BABYLON.Texture;
    public set diffuseTexture(tex: BABYLON.Texture) {
        this.setTexture('diffuseTexture', tex);
        this._diffuseTexture = tex;
    }
    public get diffuseTexture() {
        return this._diffuseTexture;
    }
    isReady(mesh?: BABYLON.AbstractMesh, useInstances?: boolean): boolean {
        let result = !!this._diffuseTexture;
        return result && super.isReady(mesh, useInstances);
    }
    bind(world: BABYLON.Matrix, mesh?: BABYLON.Mesh, effectOverride?: BABYLON.Effect): void {
        super.bind(world, mesh, effectOverride);
    }
    dispose(forceDisposeEffect?: boolean, forceDisposeTextures?: boolean, notBoundToMesh?: boolean): void {
        this._diffuseTexture?.dispose();
        super.dispose(forceDisposeEffect, forceDisposeTextures, notBoundToMesh);
    }
}
