import { BABYLON } from "../render3d/babylon";
import { IPITextureInfo } from "../material_info";
import { ShaderMaterial } from "./shader_material";
import { GLTFLoader } from "../gltf/glTFLoader";
import { LoadExtraMaterial } from "../render3d/extends/material";
import { StandardMaterialMesh } from "./standardmaterial";
import { MathDataPool } from "../render3d/extends/math";

export interface IPIMaterialStandard {
    /**
     * 材质类型
     */
    type: string;

    /**
     * 共用材质的ID
     */
    instanceID?: string;

    alphaMode: number;
    needAlphaBlend?: 1;
    zWrite?: 1;

    disableLighting: boolean;
    diffuseColor?: [number, number, number, number];
    emissionColor?: [number, number, number, number];

    alpha: number;
    cull: string;

    useAlphaFromDiffuseTexture: boolean;
    diffuseTexture?: IPITextureInfo;

    opacityTexture?: IPITextureInfo;


    // Fresnel
    EFBias?: number;
    EFPower?: number;
    EFLeft?: [number, number, number];
    EFRight?: [number, number, number];
    emissiveMapLevel?: number;

    // alphaCutOff
    alphaCutOff?: number;
    /**
     * cutoffEdgeThickness
     */
    coET?: number;
    /**
     * cutoffEdgeColor
     */
    coEColor?: [number, number, number];
    /**
     * cutoffEdgeLevel
     */
    coEL?: number;
    /**
     * cutoffEdgeAlphaDiff
     */
    coEAD?: number;

    fog?: boolean;

    alphaClip?: boolean;
    renderQueue?: number;
    diffuseOU?: number;
    diffuseOV?: number;
    emissionOU?: number;
    emissionOV?: number;
    opacityOU?: number;
    opacityOV?: number;
    diffuseLevel?: number;
    opacityLevel?: number;
}

export class OpacityClipMaterial extends ShaderMaterial {
    static KEY_OPACITY_CLIP = "OpacityClip";

    private _time = Date.now();
    private _tintColor = new BABYLON.Color4(0, 0, 0, 0);
    private _control = MathDataPool.Vector4(0, 0, 0, 0);
    private _emissiveColor = MathDataPool.Color3(1, 1, 1);
    private _diffuseMatrices = MathDataPool.Matrix();
    private _maskMatrices = MathDataPool.Matrix();

    constructor(babylonScene, materialName) {
        super(materialName, babylonScene, `${OpacityClipMaterial.KEY_OPACITY_CLIP}`, {
            attributes: ["position", "uv", "color"],
            uniforms: ["worldViewProjection", "tintColor", "diffuseMatrix", "maskMatrix", "control", "visibility"],
            samplers: ["diffuseTexture", "maskTexture"],
            needAlphaBlending: true
        });
        this._time = Date.now();
        this._tintColor.r = this._emissiveColor.r;
        this._tintColor.g = this._emissiveColor.g;
        this._tintColor.b = this._emissiveColor.b;

        this.setColor4('tintColor', this._tintColor);
        //this.setVector4('distortionInfo', this._distortionInfo);
        this.setMatrix('diffuseMatrix', this._diffuseMatrices);
        this.setMatrix('maskMatrix', this._maskMatrices);
        this.setVector4("control", this._control);
    }
    get emissiveColor() {
        return this._emissiveColor;
    }
    set emissiveColor(value) {
        this._emissiveColor = value;
    }
    get alpha() {
        return this._tintColor.a;
    }
    set alpha(value) {
        this._tintColor.a = value;
    }
    get alphaCutOff() {
        return this._control.y;
    }
    set alphaCutOff(value) {
        this._control.y = value;
    }
    public diffuseTexture: BABYLON.Texture;
    public maskTexture: BABYLON.Texture;

    isReady(mesh?: BABYLON.AbstractMesh, useInstances?: boolean): boolean {
        let result = !!this.diffuseTexture && !!this.maskTexture;
        return result && super.isReady(mesh, useInstances);
    }
    bind(world: BABYLON.Matrix, mesh?: StandardMaterialMesh, effectOverride?: BABYLON.Effect): void {
        let matData = mesh?.materialData || this;
        this._tintColor.r = matData.emissiveColor.r;
        this._tintColor.g = matData.emissiveColor.g;
        this._tintColor.b = matData.emissiveColor.b;
        this._tintColor.a = matData.alpha;
        this._diffuseMatrices.copyFrom(matData.diffuseTexture.getTextureMatrix());
        this._maskMatrices.copyFrom(matData.maskTexture.getTextureMatrix());
        this._control.x = (Date.now() - this._time) / 1000;
        this.alphaCutOff = matData.alphaCutOff;

        this.setColor4('tintColor', this._tintColor.clone());
        this.setMatrix('diffuseMatrix', matData.diffuseTexture.getTextureMatrix());
        this.setMatrix('maskMatrix', matData.maskTexture.getTextureMatrix());
        this.setVector4("control", this._control.clone());
        this.setFloat("visibility", mesh.visibility);

        this.getScene().resetCachedMaterial();
        super.bind(world, mesh, effectOverride);
    }
    dispose(forceDisposeEffect?: boolean, forceDisposeTextures?: boolean, notBoundToMesh?: boolean): void {
        this.maskTexture.dispose();
        this.diffuseTexture.dispose();
        MathDataPool.Recycle(this._maskMatrices);
        MathDataPool.Recycle(this._diffuseMatrices);
        MathDataPool.Recycle(this._control);
        MathDataPool.Recycle(this._emissiveColor);
        this._maskMatrices  = undefined;
        this._diffuseMatrices   = undefined;
        this._control   = undefined;
        this._emissiveColor     = undefined;
        super.dispose(forceDisposeEffect, forceDisposeTextures, notBoundToMesh);
    }
}
//#region PiStandard
function create(context: string, babylonScene: any, materialName: string, index: number, materialInfo: IPIMaterialStandard, babylonDrawMode: number): any {
    let mat = new OpacityClipMaterial(babylonScene, materialName);
    mat.alphaCutOff = materialInfo.alphaCutOff || 0;
    mat.emissiveColor.copyFromFloats(materialInfo.emissionColor[0] || 1, materialInfo.emissionColor[1] || 1, materialInfo.emissionColor[2] || 1);
    mat.alpha = materialInfo.emissionColor[3] || 0;
    return mat;
}
function loadAsync(context: string, materialInfo: IPIMaterialStandard, mat: OpacityClipMaterial, _loader: any): Promise<void> {
    let promises = [];
    if (materialInfo.diffuseTexture) {
        promises.push(_loader.loadTextureInfoAsync(context + "/PI_material/diffuseTexture", materialInfo.diffuseTexture, function (texture) {
            mat.diffuseTexture = texture;
            mat.setTexture('diffuseTexture', texture);
            if (materialInfo.diffuseTexture.offset) {
                texture.uOffset = materialInfo.diffuseTexture.offset[0];
                texture.vOffset = materialInfo.diffuseTexture.offset[1];
            }
            if (materialInfo.diffuseTexture.scale) {
                texture.uScale = materialInfo.diffuseTexture.scale[0];
                texture.vScale = materialInfo.diffuseTexture.scale[1];
            }
            texture.level = materialInfo.diffuseTexture.level || materialInfo.diffuseLevel || 1.0;
            if (materialInfo.diffuseTexture.hasAlpha) {
                texture.hasAlpha = true;
            }
            if (materialInfo.diffuseOU) {
                texture.offsetSpeedU = materialInfo.diffuseOU;
            }
            if (materialInfo.diffuseOV) {
                texture.offsetSpeedV = materialInfo.diffuseOV;
            }
        }));
    }
    if (materialInfo.opacityTexture) {
        promises.push(_loader.loadTextureInfoAsync(context + "/PI_material/opacityTexture", materialInfo.opacityTexture, function (texture) {
            mat.maskTexture = texture;
            mat.setTexture('maskTexture', texture);
            if (materialInfo.opacityTexture.offset) {
                texture.uOffset = materialInfo.opacityTexture.offset[0];
                texture.vOffset = materialInfo.opacityTexture.offset[1];
            }
            if (materialInfo.opacityTexture.scale) {
                texture.uScale = materialInfo.opacityTexture.scale[0];
                texture.vScale = materialInfo.opacityTexture.scale[1];
            }
        }));
    }
    return Promise.all(promises).then(function () {
        if (materialInfo.zWrite != undefined) {
            mat.forceDepthWrite = true;
        }
        if (materialInfo.fog != undefined) {
            mat.fogEnabled = materialInfo.fog;
        }
        if (materialInfo.alpha != undefined) {
            mat.alpha = materialInfo.alpha;
        }
        if (materialInfo.alphaMode != undefined) {
            mat.alphaMode = materialInfo.alphaMode;
        }
        if (mat) {
            mat.sideOrientation = _loader.babylonScene.useRightHandedSystem ? BABYLON.Material.CounterClockWiseSideOrientation : BABYLON.Material.ClockWiseSideOrientation;
        }
        if (materialInfo.cull == "off") {
            mat.backFaceCulling = false;
        }
        if (materialInfo.cull == "front") {
            mat.sideOrientation = !_loader.babylonScene.useRightHandedSystem ? BABYLON.Material.CounterClockWiseSideOrientation : BABYLON.Material.ClockWiseSideOrientation;
        }
        if (materialInfo.renderQueue) {
            if (materialInfo.renderQueue < 2450) {
                mat.transparencyMode = BABYLON.Material.MATERIAL_OPAQUE;
            }
            else if (materialInfo.renderQueue < 3000) {
                mat.transparencyMode = BABYLON.Material.MATERIAL_ALPHATEST;
            }
            else {
                if (materialInfo.alphaClip) {
                    mat.transparencyMode = BABYLON.Material.MATERIAL_ALPHATESTANDBLEND;
                }
                else {
                    mat.transparencyMode = BABYLON.Material.MATERIAL_ALPHABLEND;
                }
                mat.needAlphaBlending = () => true;
            }
        }

        (<any>mat)._isReadyForSubMesh = function (this: BABYLON.StandardMaterial, subMesh: BABYLON.SubMesh) {
            const defines = subMesh._materialDefines;
            if (!this.checkReadyOnEveryCall && subMesh.effect && defines) {
                // 如果不受灯光影响 - 则灯光脏标识不受外部影响
                if (this.disableLighting) {
                    defines._areLightsDirty = false;
                }
                if (defines._renderId === this.getScene().getRenderId()) {
                    return true;
                }
            }

            return false;
        };
    });
}
//#endregion
//#region PiStandard
function createNew(context: string, babylonScene: any, materialName: string, index: number, matInfo: any, babylonDrawMode: number): any {
    let materialInfo: IPIMaterialStandard = matInfo;
    let mat = new OpacityClipMaterial(babylonScene, materialName);
    mat.alphaCutOff = materialInfo.alphaCutOff || 0;
    mat.emissiveColor.copyFromFloats(materialInfo.emissionColor[0] || 1, materialInfo.emissionColor[1] || 1, materialInfo.emissionColor[2] || 1);
    mat.alpha = materialInfo.emissionColor[3] || 0;
    return mat;
}
function loadAsyncNew(context: string, matInfo: any, mat: OpacityClipMaterial, _loader: GLTFLoader): void {
    let promises = [];
    let materialInfo: IPIMaterialStandard = matInfo;
    if (materialInfo.diffuseTexture) {
        let texture = <BABYLON.Texture>_loader.loadTextureInfoAsync(context + "/PI_material/diffuseTexture", materialInfo.diffuseTexture);
        mat.diffuseTexture = texture;
        mat.setTexture('diffuseTexture', texture);
        if (materialInfo.diffuseTexture.offset) {
            texture.uOffset = materialInfo.diffuseTexture.offset[0];
            texture.vOffset = materialInfo.diffuseTexture.offset[1];
        }
        if (materialInfo.diffuseTexture.scale) {
            texture.uScale = materialInfo.diffuseTexture.scale[0];
            texture.vScale = materialInfo.diffuseTexture.scale[1];
        }
        texture.level = materialInfo.diffuseTexture.level || materialInfo.diffuseLevel || 1.0;
        if (materialInfo.diffuseTexture.hasAlpha) {
            texture.hasAlpha = true;
        }
        if (materialInfo.diffuseOU) {
            texture.offsetSpeedU = materialInfo.diffuseOU;
        }
        if (materialInfo.diffuseOV) {
            texture.offsetSpeedV = materialInfo.diffuseOV;
        }
        let tilloff = materialInfo.diffuseTexture.extras?.atlasTilloff;
        if (tilloff) {
            texture.atlasTilloff = tilloff;
        }
    }
    if (materialInfo.opacityTexture) {
        let texture = <BABYLON.Texture>_loader.loadTextureInfoAsync(context + "/PI_material/opacityTexture", materialInfo.opacityTexture);
        mat.maskTexture = texture;
        mat.setTexture('maskTexture', texture);
        if (materialInfo.opacityTexture.offset) {
            texture.uOffset = materialInfo.opacityTexture.offset[0];
            texture.vOffset = materialInfo.opacityTexture.offset[1];
        }
        if (materialInfo.opacityTexture.scale) {
            texture.uScale = materialInfo.opacityTexture.scale[0];
            texture.vScale = materialInfo.opacityTexture.scale[1];
        }
        let tilloff = materialInfo.opacityTexture.extras?.atlasTilloff;
        if (tilloff) {
            texture.atlasTilloff = tilloff;
        }
    }
    // return Promise.all(promises).then(function () {
    if (materialInfo.zWrite != undefined) {
        mat.forceDepthWrite = true;
    }
    if (materialInfo.fog != undefined) {
        mat.fogEnabled = materialInfo.fog;
    }
    if (materialInfo.alpha != undefined) {
        mat.alpha = materialInfo.alpha;
    }
    if (materialInfo.alphaMode != undefined) {
        mat.alphaMode = materialInfo.alphaMode;
    }
    if (mat) {
        mat.sideOrientation = _loader.babylonScene.useRightHandedSystem ? BABYLON.Material.CounterClockWiseSideOrientation : BABYLON.Material.ClockWiseSideOrientation;
    }
    if (materialInfo.cull == "off") {
        mat.backFaceCulling = false;
    }
    if (materialInfo.cull == "front") {
        mat.sideOrientation = !_loader.babylonScene.useRightHandedSystem ? BABYLON.Material.CounterClockWiseSideOrientation : BABYLON.Material.ClockWiseSideOrientation;
    }
    if (materialInfo.renderQueue) {
        if (materialInfo.renderQueue < 2450) {
            mat.transparencyMode = BABYLON.Material.MATERIAL_OPAQUE;
        }
        else if (materialInfo.renderQueue < 3000) {
            mat.transparencyMode = BABYLON.Material.MATERIAL_ALPHATEST;
        }
        else {
            if (materialInfo.alphaClip) {
                mat.transparencyMode = BABYLON.Material.MATERIAL_ALPHATESTANDBLEND;
            }
            else {
                mat.transparencyMode = BABYLON.Material.MATERIAL_ALPHABLEND;
            }
            mat.needAlphaBlending = () => true;
        }
    }

    (<any>mat)._isReadyForSubMesh = function (this: BABYLON.StandardMaterial, subMesh: BABYLON.SubMesh) {
        const defines = subMesh._materialDefines;
        if (!this.checkReadyOnEveryCall && subMesh.effect && defines) {
            // 如果不受灯光影响 - 则灯光脏标识不受外部影响
            if (this.disableLighting) {
                defines._areLightsDirty = false;
            }
            if (defines._renderId === this.getScene().getRenderId()) {
                return true;
            }
        }

        return false;
    };
    // });
}
//#endregion
export function registOpacityClipFactory() {
    BABYLON.GLTF2.Loader.Extensions.PI_material.regist(
        OpacityClipMaterial.KEY_OPACITY_CLIP,
        <any>create,
        <any>loadAsync
    );

    LoadExtraMaterial.loadCalls.set(OpacityClipMaterial.KEY_OPACITY_CLIP, createNew);
    LoadExtraMaterial.loadPropertyCalls.set(OpacityClipMaterial.KEY_OPACITY_CLIP, loadAsyncNew);
}
// Effect
BABYLON.Effect.ShadersStore[`${OpacityClipMaterial.KEY_OPACITY_CLIP}VertexShader`] = `
    precision highp float;

    attribute vec3 position;
    attribute vec2 uv;
    attribute vec4 color;

    uniform mat4 worldViewProjection;
    uniform mat4 diffuseMatrix;
    uniform mat4 maskMatrix;
    uniform vec4 control;

    varying vec2 vDiffuseUV;
    varying vec2 vMaskUV;
    varying vec4 vColor;

    void main(void) {
        vec3 positionUpdated = position;
        vec2 uvUpdated = uv;

        gl_Position = worldViewProjection*vec4(positionUpdated, 1.0);

        vDiffuseUV = vec2(diffuseMatrix*vec4(uvUpdated, 1.0, 0.0));
        vMaskUV = vec2(maskMatrix*vec4(uvUpdated, 1.0, 0.0));
        vColor = color;
    }
`;
BABYLON.Effect.ShadersStore[`${OpacityClipMaterial.KEY_OPACITY_CLIP}FragmentShader`] =
    `
    precision highp float;

    varying vec2 vDiffuseUV;
    varying vec2 vMaskUV;
    varying vec4 vColor;

    uniform sampler2D diffuseTexture;
    uniform sampler2D maskTexture;
    uniform vec4 tintColor;
    uniform vec4 control;
    uniform mat4 diffuseMatrix;
    uniform mat4 maskMatrix;
    uniform float visibility;

    vec2 uvAddress(vec2 uv, mat4 matrix) {
        vec2 mode  = vec2(matrix*vec4(0., 0., 0., 1.0));
        vec4 tilloff    = vec4(matrix[0][3], matrix[1][3], matrix[2][3], matrix[3][3]);

        vec2 h = 0.5 * uv;
        vec2 n = floor(uv);
        vec2 a = 1.0 - floor(h - floor(h) + 0.5);

        vec2 repeat = uv - n;
        vec2 mirrorRepeat = 1.0 - a - repeat + 2.0 * a * repeat;
        vec2 clamp2edge = abs(h) - abs(h - 0.5) + 0.5;

        vec2 temp = mix(
                clamp2edge,
                mix(
                    repeat,
                    mirrorRepeat,
                    vec2(step(2.0, mode.x), step(2.0, mode.y))
                ),
                vec2(step(1.0, mode.x), step(1.0, mode.y))
            );

        return temp * tilloff.xy + tilloff.zw;
        // return temp;
    }

    void main(void) {
        vec4 textureMask = texture2D(maskTexture, vMaskUV);

        if (textureMask.a < control.y) {
            discard;
        }

        vec4 textureColor = texture2D(diffuseTexture, vDiffuseUV);

        vec4 result = textureColor * tintColor;
        result.a *= visibility;

        gl_FragColor = result;
    }
`;
