// import * as BABYLON from "../pi_babylon/render3d/babylon";
// import { IPITextureInfo } from "./material_info";
import { BABYLON } from "../render3d/babylon";

export class SpineColoredTextureMaterial extends BABYLON.ShaderMaterial {
    public static KEY_SHADER = "SpineColoredTexture";
    public static U_PARAM0 = "U_PARAM0";
    public static U_PARAM1 = "U_PARAM1";

    maskFlag: number;

    private _parma0: BABYLON.Vector4 = new BABYLON.Vector4(0.5, 0.5, 0.5, 0);
    
    private _parma1: BABYLON.Vector4 = new BABYLON.Vector4(1, 0, 1, 0);

    public set visibility(val: number) {
        this._parma1.x = val;
    }
    
    public set alpha(val: number) {
        this._parma1.z = val;
    }
    
    /**
     * 蒙版颜色/数据
     */
    public set maskData(val: [number, number, number]) {
        this._parma0.x = val[0];
        this._parma0.y = val[1];
        this._parma0.z = val[2];
    }

    /**
     * 蒙版类型
     * * 0 - 无效果
     * * 1 - 使用 maskeColor 替换rgb
     * * 2 - HSV 颜色变换
     */
    public set maskType(val: 0 | 1 | 2) {
        this._parma0.z = val;
    }

    public set resultPremultiplyAlpha(val: boolean) {
        this._parma1.y = val ? 1 : 0;
    }

    private _diffuseTexture: BABYLON.Texture;
    public set diffuseTexture(val: BABYLON.Texture) {
        this._diffuseTexture = val;
        this.setTexture('diffuseTexture', this._diffuseTexture);
    }
    private _diffuseMatrices: BABYLON.Matrix = BABYLON.Matrix.Identity();

    constructor(babylonScene: any, materialName: string) {
        super(
            materialName,
            babylonScene,
            `${SpineColoredTextureMaterial.KEY_SHADER}`,
            {
                attributes: ["position", "uv", "color"],
                uniforms: ["worldViewProjection", "diffuseMatrix", SpineColoredTextureMaterial.U_PARAM0, SpineColoredTextureMaterial.U_PARAM1 ],
                samplers: ["diffuseTexture"],
                needAlphaBlending: true
            }
        );

        this.setMatrix('diffuseMatrix', this._diffuseMatrices);
        this.setVector4(SpineColoredTextureMaterial.U_PARAM0, this._parma0);
        this.setVector4(SpineColoredTextureMaterial.U_PARAM1, this._parma1);

        this.backFaceCulling = false;
    }

    isReady(mesh?: BABYLON.AbstractMesh, useInstances?: boolean): boolean {
        let result = !!this._diffuseTexture && !!this._diffuseTexture;
        return result && super.isReady(mesh, useInstances);
    }

    bind(world: BABYLON.Matrix, mesh?: BABYLON.Mesh, effectOverride?: BABYLON.Effect): void {
        this._diffuseMatrices.copyFrom(this._diffuseTexture.getTextureMatrix());
        super.bind(world, mesh, effectOverride);
    }

    dispose(forceDisposeEffect?: boolean, forceDisposeTextures?: boolean, notBoundToMesh?: boolean): void {
        this._diffuseTexture.dispose();
        super.dispose(forceDisposeEffect, forceDisposeTextures, notBoundToMesh);
    }
}

// Effect
BABYLON.Effect.ShadersStore[`${SpineColoredTextureMaterial.KEY_SHADER}VertexShader`] = `
    precision highp float;

    attribute vec3 position;
    attribute vec2 uv;
    attribute vec4 color;

    uniform mat4 worldViewProjection;
    uniform mat4 diffuseMatrix;

    varying vec2 vDiffuseUV;
    varying vec4 vColor;

    void main(void) {
        vec3 positionUpdated = position;
        vec2 uvUpdated = uv;

        gl_Position = worldViewProjection*vec4(positionUpdated, 1.0);

        vDiffuseUV = vec2(diffuseMatrix*vec4(uvUpdated, 1.0, 0.0));
        vColor = color;
    }
`;

BABYLON.Effect.ShadersStore[`${SpineColoredTextureMaterial.KEY_SHADER}FragmentShader`] =
    `
    precision highp float;

    uniform vec4 ${SpineColoredTextureMaterial.U_PARAM0};
    uniform vec4 ${SpineColoredTextureMaterial.U_PARAM1};
    
    uniform sampler2D diffuseTexture;

    varying vec2 vDiffuseUV;
    varying vec4 vColor;
    
    vec3 rgb2hsv(vec3 c)
    {
        vec4 K = vec4(0.0, -1.0 / 3.0, 2.0 / 3.0, -1.0);
        vec4 p = mix(vec4(c.bg, K.wz), vec4(c.gb, K.xy), step(c.b, c.g));
        vec4 q = mix(vec4(p.xyw, c.r), vec4(c.r, p.yzx), step(p.x, c.r));

        float d = q.x - min(q.w, q.y);
        float e = 1.0e-10;
        return vec3(abs(q.z + (q.w - q.y) / (6.0 * d + e)), d / (q.x + e), q.x);
    }

    vec3 hsv2rgb(vec3 c)
    {
        vec4 K = vec4(1.0, 2.0 / 3.0, 1.0 / 3.0, 3.0);
        vec3 p = abs(fract(c.xxx + K.xyz) * 6.0 - K.www);
        return c.z * mix(K.xxx, clamp(p - K.xxx, vec3(0.0), vec3(1.0)), c.y);
    }

    void main(void) {
        gl_FragColor = texture2D(diffuseTexture, vDiffuseUV);

        if (${SpineColoredTextureMaterial.U_PARAM0}.w == 1.0) {
            gl_FragColor.rgb = ${SpineColoredTextureMaterial.U_PARAM0}.rgb * gl_FragColor.a;
        }
        if (${SpineColoredTextureMaterial.U_PARAM0}.w == 2.0) {
            vec4 c = gl_FragColor;
            vec3 hsvValue = ${SpineColoredTextureMaterial.U_PARAM0}.rgb;

            vec3 hsv = rgb2hsv(c.rgb);
            hsv.r += hsvValue.r;
            c.rgb = hsv2rgb(hsv);

            // ע��saturate����0ʱ����ʽ��PS����һ��
            float gray = max(c.r, max(c.g, c.b)) + min(c.r, min(c.g, c.b));
            c.rgb = mix(c.rgb, vec3(0.5 * gray), -hsvValue.g);

            if (hsvValue.b >= 0.0) {
                c.rgb = mix(c.rgb, vec3(1.0), hsvValue.b);
            } else {
                c.rgb *= 1.0 + hsvValue.b;
            }
            gl_FragColor = c;
        }

        gl_FragColor.rgb *= ${SpineColoredTextureMaterial.U_PARAM1}.x;
        gl_FragColor.a 	*= ${SpineColoredTextureMaterial.U_PARAM1}.z;
        gl_FragColor.rgb *= mix(1.0, gl_FragColor.a, ${SpineColoredTextureMaterial.U_PARAM1}.y);
    }
`;