import { Scene } from "../render3d/scene";
import { BABYLON } from "../render3d/babylon";
import { EffectFallbacks } from "./effect_fallbacks";
import { MathDataPool } from "../render3d/extends/math";

export interface TextureTillOffDataScene extends Scene {
    __TextureTillOffDatas: Map<number, TextureTillOffData>;
    __TextureTillOffPool: TextureTillOffData[];
}

export class TextureTillOffData {
    private static Identity = BABYLON.Matrix.FromValues(1, 0, 0, 0,  0, 1, 0, 0,  0, 0, 0, 0,  0, 0, 0, 1);
    private static isIdentityAs3x2 = function() { return false; }
    private static Counter: number = Number.MIN_SAFE_INTEGER;

    private _uScale: number = 1;
    private _vScale: number = 1;
    private _uOffset: number = 0;
    private _vOffset: number = 0;
    get uScale() { return this._uScale };
    get vScale() { return this._vScale };
    get uOffset() { return this._uOffset };
    get vOffset() { return this._vOffset };
    set uScale(val: number) { this._uScale = val; this._dirty = true; };
    set vScale(val: number) { this._vScale = val; this._dirty = true; };
    set uOffset(val: number) { this._uOffset = val; this._dirty = true; };
    set vOffset(val: number) { this._vOffset = val; this._dirty = true; };
    animations: Array<BABYLON.Animation> = [];
    _cachedTextureMatrix: BABYLON.Matrix;

    private _dirty: boolean = true;
    public get dirty() {
        return this._dirty;
    }
    public readonly uid: number;

    public static Create(scene: TextureTillOffDataScene) {
        if (scene.__TextureTillOffDatas == undefined) {
            scene.__TextureTillOffDatas = new Map();
            scene.__TextureTillOffPool = [];
            scene.onDisposeObservable.addOnce(function(scene: TextureTillOffDataScene) {
                scene.__TextureTillOffDatas.forEach(function(item) {
                    item.dispose();
                });
                scene.__TextureTillOffPool = [];
                scene.__TextureTillOffDatas.clear();
            });
            scene.onAfterRenderObservable.add(function(scene: TextureTillOffDataScene) {
                scene.__TextureTillOffDatas.forEach(function(item) {
                    item._dirty = false;
                });
            });
        }

        let result = scene.__TextureTillOffPool.pop();
        if (result == undefined) {
            result = new TextureTillOffData(scene);
        }
        scene.__TextureTillOffDatas.set(result.uid, result);

        return result;
    }

    constructor(private _scene: TextureTillOffDataScene) {
        this.uid = TextureTillOffData.Counter++;
    }

    // PI_BEGIN
    /**
     * 纹理offset自动变化周期 - 每 offsetSpeedU 秒循环一次
     */
    public offsetSpeedU: number = 0;
    /**
     * 纹理offset自动变化周期 - 每 offsetSpeedV 秒循环一次
     */
    public offsetSpeedV: number = 0;
    // PI_END

    // PI_BEGIN
    public computeTextureOffsetWithSpeedUV() {
        if (this._scene?.animationsEnabled) {
            if (Math.abs(this.offsetSpeedU) > 0.0001) {
                this.uOffset = Math.round(this._scene._animationTime) / this.offsetSpeedU;
            }
            if (Math.abs(this.offsetSpeedV) > 0.0001) {
                this.vOffset = Math.round(this._scene._animationTime) / this.offsetSpeedV;
            }
        }
    }
    // PI_END

    getTextureMatrix(): BABYLON.Matrix {
        this.computeTextureOffsetWithSpeedUV();

        if (this.uScale == 1 && this.vScale == 1 && this.uOffset == 0 && this.vOffset == 0) {
            TextureTillOffData.Identity.isIdentityAs3x2 = TextureTillOffData.isIdentityAs3x2;
            return TextureTillOffData.Identity;
        } else {
            if (!this._cachedTextureMatrix) {
                this._cachedTextureMatrix = MathDataPool.Matrix();
                BABYLON.Matrix.FromValuesToRef(1, 0, 0, 0,  0, 1, 0, 0,  0, 0, 0, 0,  0, 0, 0, 1, this._cachedTextureMatrix);
                this._cachedTextureMatrix.isIdentityAs3x2 = TextureTillOffData.isIdentityAs3x2;
            }
            const m = (<any>this._cachedTextureMatrix)._m;
    
            m[0 * 4 + 0] = this.uScale;
            m[1 * 4 + 1] = this.vScale;
            m[2 * 4 + 2] = 0;
            
            m[3 * 4 + 0] = this.uOffset;
            m[3 * 4 + 1] = this.vOffset;
            m[3 * 4 + 2] = 0;
            this._cachedTextureMatrix.setRowFromFloats(2, this.uOffset, this.vOffset, 0, 1);
            this._cachedTextureMatrix._markAsUpdated();

            return this._cachedTextureMatrix;
        }

    }

    dispose() {
        if (this._scene?.__TextureTillOffPool) {
            if (this._scene.__TextureTillOffDatas.delete(this.uid)) {
                this._scene.__TextureTillOffPool.push(this);
                this.animations.length = 0;
                this.uScale = 1;
                this.vScale = 1;
                this.uOffset = 0;
                this.vOffset = 0;
                this.offsetSpeedU = 0;
                this.offsetSpeedV = 0;
                this._dirty = true;
                MathDataPool.Recycle(this._cachedTextureMatrix);
                this._cachedTextureMatrix = undefined;
            }
        }
    }

    copyFrom(tex: TextureTillOffData) {
        this.uScale = tex.uScale;
        this.vScale = tex.vScale;
        this.uOffset = tex.uOffset;
        this.vOffset = tex.vOffset;
        this.offsetSpeedU = tex.offsetSpeedU;
        this.offsetSpeedV = tex.offsetSpeedV;
    }
}

interface StandardMaterialDataScene extends Scene {
    __StandardMaterialData: Map<number, StandardMaterialData>;
}
export class StandardMaterialData {
    private static Counter: number = Number.MIN_SAFE_INTEGER;
    public static Create(_scene: Scene) {
        let scene = <StandardMaterialDataScene>_scene;
        if (scene.__StandardMaterialData == undefined) {
            scene.__StandardMaterialData = new Map();
            scene.onDisposeObservable.addOnce(() => {
                scene.__StandardMaterialData.forEach((item) => {
                    item.dispose();
                });
                scene.__StandardMaterialData.clear();
            });
            scene.onAfterRenderObservable.add(function(scene: StandardMaterialDataScene) {
                scene.__StandardMaterialData.forEach((item) => {
                    item._dirty = false;
                });
            });
        }

        let result = new StandardMaterialData(scene);

        scene.__StandardMaterialData.set(result.uid, result);

        return result;

    }

    emissiveColor: BABYLON.Color3 = BABYLON.Color3.White();
    diffuseTexture: TextureTillOffData;
    emissiveTexture: TextureTillOffData;
    opacityTexture: TextureTillOffData;
    maskTexture: TextureTillOffData;
    animations: Array<BABYLON.Animation> = [];

    public readonly uid: number;

    private _dirty: boolean = true;
    public get dirty() {
        return this._dirty || this.diffuseTexture.dirty || this.emissiveTexture.dirty || this.opacityTexture.dirty || this.maskTexture.dirty;
    }

    constructor(protected scene: BABYLON.Scene) {
        this.uid = StandardMaterialData.Counter++;

        this.diffuseTexture     = TextureTillOffData.Create(<TextureTillOffDataScene>scene);
        this.emissiveTexture    = TextureTillOffData.Create(<TextureTillOffDataScene>scene);
        this.opacityTexture     = TextureTillOffData.Create(<TextureTillOffDataScene>scene);
        this.maskTexture        = TextureTillOffData.Create(<TextureTillOffDataScene>scene);
    }
    fromMaterial(material: StandardMaterialData) {
        if (material.emissiveColor) {
            this.emissiveColor.copyFrom(material.emissiveColor);
        }
        if (material.diffuseTexture) {
            this.diffuseTexture.copyFrom(material.diffuseTexture);
        }
        if (material.opacityTexture) {
            this.opacityTexture.copyFrom(material.opacityTexture);
        }
        if (material.emissiveTexture) {
            this.emissiveTexture.copyFrom(material.emissiveTexture);
        }
        if (material.maskTexture) {
            this.maskTexture.copyFrom(material.maskTexture);
        }
        this.alpha = material.alpha;
        this.alphaCutOff = material.alphaCutOff;
        this.maskCutoff = material.maskCutoff;
    }
    // alpha: number = 1;
    private _alpha: number = 1;
    get alpha() {
        return this._alpha;
    }
    set alpha(val: number) {
        this._alpha = val;
        this._dirty = true;
    }

    // alphaCutOff: number = 0;
    private _alphaCutOff: number = 0;
    get alphaCutOff() {
        return this._alphaCutOff;
    }
    set alphaCutOff(val: number) {
        this._alphaCutOff = val;
        this._dirty = true;
    }

    maskCutoff: number = 0;

    dispose = () => {
        this.diffuseTexture     ?.dispose();
        this.emissiveTexture    ?.dispose();
        this.opacityTexture     ?.dispose();
        this.maskTexture        ?.dispose();
        this.diffuseTexture     = undefined;
        this.emissiveTexture    = undefined;
        this.opacityTexture     = undefined;
        this.maskTexture        = undefined;
    }
}

export interface StandardMaterialMesh extends BABYLON.Mesh {
    materialData: StandardMaterialData;
}

export function initStandardMaterialOpt() {
    /**
     * Binds the submesh to this material by preparing the effect and shader to draw
     * @param world defines the world transformation matrix
     * @param mesh defines the mesh containing the submesh
     * @param subMesh defines the submesh to bind the material to
     */
    BABYLON.StandardMaterial.prototype.bindForSubMesh = function (this: BABYLON.StandardMaterial, world: BABYLON.Matrix, mesh: StandardMaterialMesh, subMesh: BABYLON.SubMesh): void {
        var scene = this.getScene();

        var defines = <BABYLON.StandardMaterialDefines>subMesh._materialDefines;
        if (!defines) {
            return;
        }

        var effect = subMesh.effect;
        if (!effect) {
            return;
        }
        this._activeEffect = effect;

        // Matrices
        if (!defines.INSTANCES || defines.THIN_INSTANCES) {
            this.bindOnlyWorldMatrix(world);
        }

        // PrePass
        this.prePassConfiguration.bindForSubMesh(this._activeEffect, scene, mesh, world, this.isFrozen);

        // Normal Matrix
        if (defines.OBJECTSPACE_NORMALMAP) {
            world.toNormalMatrix(this._normalMatrix);
            this.bindOnlyNormalMatrix(this._normalMatrix);
        }

        let materialData = mesh.materialData || this;
        let mustRebind = this._mustRebind(scene, effect, mesh.visibility) || (this._materialData != materialData);

        // Bones
        BABYLON.MaterialHelper.BindBonesParameters(mesh, effect);
        let ubo = this._uniformBuffer;
        if (mustRebind) {
            this._materialData == materialData;

            ubo.bindToEffect(effect, "Material");
            

            this.bindViewProjection(effect);
            if (!ubo.useUbo || !this.isFrozen || !ubo.isSync) {

                if (BABYLON.StandardMaterial.FresnelEnabled && defines.FRESNEL) {
                    // Fresnel
                    if (this.diffuseFresnelParameters && this.diffuseFresnelParameters.isEnabled) {
                        ubo.updateColor4("diffuseLeftColor", this.diffuseFresnelParameters.leftColor, this.diffuseFresnelParameters.power);
                        ubo.updateColor4("diffuseRightColor", this.diffuseFresnelParameters.rightColor, this.diffuseFresnelParameters.bias);
                    }

                    if (this.opacityFresnelParameters && this.opacityFresnelParameters.isEnabled) {
                        ubo.updateColor4("opacityParts", new BABYLON.Color3(this.opacityFresnelParameters.leftColor.toLuminance(), this.opacityFresnelParameters.rightColor.toLuminance(), this.opacityFresnelParameters.bias), this.opacityFresnelParameters.power);
                    }

                    if (this.reflectionFresnelParameters && this.reflectionFresnelParameters.isEnabled) {
                        ubo.updateColor4("reflectionLeftColor", this.reflectionFresnelParameters.leftColor, this.reflectionFresnelParameters.power);
                        ubo.updateColor4("reflectionRightColor", this.reflectionFresnelParameters.rightColor, this.reflectionFresnelParameters.bias);
                    }

                    if (this.refractionFresnelParameters && this.refractionFresnelParameters.isEnabled) {
                        ubo.updateColor4("refractionLeftColor", this.refractionFresnelParameters.leftColor, this.refractionFresnelParameters.power);
                        ubo.updateColor4("refractionRightColor", this.refractionFresnelParameters.rightColor, this.refractionFresnelParameters.bias);
                    }

                    if (this.emissiveFresnelParameters && this.emissiveFresnelParameters.isEnabled) {
                        ubo.updateColor4("emissiveLeftColor", this.emissiveFresnelParameters.leftColor, this.emissiveFresnelParameters.power);
                        ubo.updateColor4("emissiveRightColor", this.emissiveFresnelParameters.rightColor, this.emissiveFresnelParameters.bias);
                    }
                }

                // Textures
                if (scene.texturesEnabled) {
                    if (this._diffuseTexture && BABYLON.StandardMaterial.DiffuseTextureEnabled) {
                        ubo.updateFloat3("vDiffuseInfos", this._diffuseTexture.coordinatesIndex, this._diffuseTexture.level, this.diffuseTexturePremultiplyAlpha ? 1 : 0);
                        MaterialHelper.BindTextureMatrix(materialData.diffuseTexture, ubo, "diffuse");

                    }

                    if (this._ambientTexture && BABYLON.StandardMaterial.AmbientTextureEnabled) {
                        ubo.updateFloat2("vAmbientInfos", this._ambientTexture.coordinatesIndex, this._ambientTexture.level);
                        MaterialHelper.BindTextureMatrix(this._ambientTexture, ubo, "ambient");
                    }

                    if (this._opacityTexture && BABYLON.StandardMaterial.OpacityTextureEnabled) {
                        ubo.updateFloat3("vOpacityInfos", this._opacityTexture.coordinatesIndex, this._opacityTexture.level, this.opacityTextureApplyRGB ? 1 : 0);
                        MaterialHelper.BindTextureMatrix(materialData.opacityTexture, ubo, "opacity");
                        //#region PI_BEGIN - opacity 拓展 两层opacity混合
                        if (this.mixOpacity) {
                            this._mixOpacityUVUpdate();
                            ubo.updateFloat4("mixOpacityInfo", 0, this.mixOpacityIntensity, this.mixOpacityControlValue, this.mixOpacityControlIntensity);
                            ubo.updateFloat4("mixOpacity" + "Matrix", this.mixOpacityUScale, this.mixOpacityVScale, this.mixOpacityUOffset, this.mixOpacityVOffset);
                            ubo.updateFloat4("mixOpacityControl" + "Matrix", this.mixOpacityControlUScale, this.mixOpacityControlVScale, this.mixOpacityControlUOffset, this.mixOpacityControlVOffset);
                        }
                        //#endregion
                    }

                    if (this._hasAlphaChannel()) {
                        effect.setFloat("alphaCutOff", materialData.alphaCutOff);
                        //#region PI_BEGIN CUTOFF 拓展
                        ubo.updateColor3("cutoffEdgeColor", this.cutoffEdgeColor);
                        ubo.updateFloat3("cutoffEdgeInfo", this.cutoffEdgeThickness, this.cutoffEdgeLevel, this.cutoffEdgeAlphaDiff);
                        //#endregion PI_END
                    }

                    // if (this._reflectionTexture && BABYLON.StandardMaterial.ReflectionTextureEnabled) {
                    //     ubo.updateFloat2("vReflectionInfos", this._reflectionTexture.level, this.roughness);
                    //     ubo.updateMatrix("reflectionMatrix", this._reflectionTexture.getReflectionTextureMatrix());

                    //     if ((<any>this._reflectionTexture).boundingBoxSize) {
                    //         let cubeTexture = <CubeTexture>this._reflectionTexture;

                    //         ubo.updateVector3("vReflectionPosition", cubeTexture.boundingBoxPosition);
                    //         ubo.updateVector3("vReflectionSize", cubeTexture.boundingBoxSize);
                    //     }
                    // }

                    if (this._emissiveTexture && BABYLON.StandardMaterial.EmissiveTextureEnabled) {
                        ubo.updateFloat2("vEmissiveInfos", this._emissiveTexture.coordinatesIndex, this._emissiveTexture.level);
                        MaterialHelper.BindTextureMatrix(materialData.emissiveTexture, ubo, "emissive");
                    }

                    // PI_BEGIN
                    if (this._maskTexture && BABYLON.StandardMaterial.MaskTextureEnabled) {
                        ubo.updateFloat2("vMaskInfos", this._maskTexture.coordinatesIndex, this._maskTexture.level);
                        MaterialHelper.BindTextureMatrix(materialData.maskTexture, ubo, "mask");
                        ubo.updateColor3("vMaskColor", this.maskColor);
                    }
                    // PI_END

                    // PI_BEGIN
                    let positionStateTexture = scene.positionStateTexture;
                    if (this._applyPositionStateTexture && positionStateTexture && BABYLON.StandardMaterial.PositionStateTextureEnabled) {
                        ubo.updateFloat4("positionStateRange", scene.positionStateRange.x, scene.positionStateRange.y, 1 / scene.positionStateRange.z, 1 / scene.positionStateRange.w);
                    }
                    // PI_END

                    // if (this._lightmapTexture && BABYLON.StandardMaterial.LightmapTextureEnabled) {
                    //     //#region PI_BEGIN - 光照图 适配 Unity RGBM
                    //     // ubo.updateFloat2("vLightmapInfos", this._lightmapTexture.coordinatesIndex, this._lightmapTexture.level);
                    //     ubo.updateFloat4("vLightmapInfos", this._lightmapTexture.coordinatesIndex, this._lightmapTexture.level, this.getScene().lightmapHDRMultiplier, 1.0);
                    //     //#endregion
                    //     MaterialHelper.BindTextureMatrix(this._lightmapTexture, ubo, "lightmap");
                    // }

                    // if (this._specularTexture && BABYLON.StandardMaterial.SpecularTextureEnabled) {
                    //     ubo.updateFloat2("vSpecularInfos", this._specularTexture.coordinatesIndex, this._specularTexture.level);
                    //     MaterialHelper.BindTextureMatrix(this._specularTexture, ubo, "specular");
                    // }

                    // if (this._bumpTexture && scene.getEngine().getCaps().standardDerivatives && BABYLON.StandardMaterial.BumpTextureEnabled) {
                    //     ubo.updateFloat3("vBumpInfos", this._bumpTexture.coordinatesIndex, 1.0 / this._bumpTexture.level, this.parallaxScaleBias);
                    //     MaterialHelper.BindTextureMatrix(this._bumpTexture, ubo, "bump");

                    //     if (scene._mirroredCameraPosition) {
                    //         ubo.updateFloat2("vTangentSpaceParams", this._invertNormalMapX ? 1.0 : -1.0, this._invertNormalMapY ? 1.0 : -1.0);
                    //     } else {
                    //         ubo.updateFloat2("vTangentSpaceParams", this._invertNormalMapX ? -1.0 : 1.0, this._invertNormalMapY ? -1.0 : 1.0);
                    //     }
                    // }

                    // if (this._refractionTexture && StandardMaterial.RefractionTextureEnabled) {
                    //     var depth = 1.0;
                    //     if (!this._refractionTexture.isCube) {
                    //         ubo.updateMatrix("refractionMatrix", this._refractionTexture.getReflectionTextureMatrix());

                    //         if ((<any>this._refractionTexture).depth) {
                    //             depth = (<any>this._refractionTexture).depth;
                    //         }
                    //     }
                    //     ubo.updateFloat4("vRefractionInfos", this._refractionTexture.level, this.indexOfRefraction, depth, this.invertRefractionY ? -1 : 1);
                    // }
                }

                // Point size
                if (this.pointsCloud) {
                    ubo.updateFloat("pointSize", this.pointSize);
                }

                if (defines.SPECULARTERM) {
                    ubo.updateColor4("vSpecularColor", this.specularColor, this.specularPower);
                }
                ubo.updateColor3("vEmissiveColor", BABYLON.StandardMaterial.EmissiveTextureEnabled ? materialData.emissiveColor : BABYLON.Color3.BlackReadOnly);

                // Diffuse
                ubo.updateColor4("vDiffuseColor", this.diffuseColor, materialData.alpha);

            }

            // Visibility
            ubo.updateFloat("visibility", mesh.visibility);

            // Textures
            if (scene.texturesEnabled) {
                if (this._diffuseTexture && BABYLON.StandardMaterial.DiffuseTextureEnabled) {
                    effect.setTexture("diffuseSampler", this._diffuseTexture);
                }

                if (this._ambientTexture && BABYLON.StandardMaterial.AmbientTextureEnabled) {
                    effect.setTexture("ambientSampler", this._ambientTexture);
                }

                if (this._opacityTexture && BABYLON.StandardMaterial.OpacityTextureEnabled) {
                    effect.setTexture("opacitySampler", this._opacityTexture);
                }

                // if (this._reflectionTexture && BABYLON.StandardMaterial.ReflectionTextureEnabled) {
                //     if (this._reflectionTexture.isCube) {
                //         effect.setTexture("reflectionCubeSampler", this._reflectionTexture);
                //     } else {
                //         effect.setTexture("reflection2DSampler", this._reflectionTexture);
                //     }
                // }

                if (this._emissiveTexture && BABYLON.StandardMaterial.EmissiveTextureEnabled) {
                    effect.setTexture("emissiveSampler", this._emissiveTexture);
                }

                // PI_BEGIN
                if (this._maskTexture && BABYLON.StandardMaterial.MaskTextureEnabled) {
                    effect.setTexture("maskSampler", this._maskTexture);
                }
                // PI_END

                // PI_BEGIN
                let positionStateTexture = scene.positionStateTexture;
                if (this._applyPositionStateTexture && positionStateTexture && BABYLON.StandardMaterial.PositionStateTextureEnabled) {
                    effect.setTexture("positionStateSampler", positionStateTexture);
                }
                // PI_END

                // if (this._lightmapTexture && BABYLON.StandardMaterial.LightmapTextureEnabled) {
                //     effect.setTexture("lightmapSampler", this._lightmapTexture);
                // }

                // if (this._specularTexture && BABYLON.StandardMaterial.SpecularTextureEnabled) {
                //     effect.setTexture("specularSampler", this._specularTexture);
                // }

                // if (this._bumpTexture && scene.getEngine().getCaps().standardDerivatives && BABYLON.StandardMaterial.BumpTextureEnabled) {
                //     effect.setTexture("bumpSampler", this._bumpTexture);
                // }

                // if (this._refractionTexture && BABYLON.StandardMaterial.RefractionTextureEnabled) {
                //     var depth = 1.0;
                //     if (this._refractionTexture.isCube) {
                //         effect.setTexture("refractionCubeSampler", this._refractionTexture);
                //     } else {
                //         effect.setTexture("refraction2DSampler", this._refractionTexture);
                //     }
                // }
            }

            this.detailMap.bindForSubMesh(ubo, scene, this.isFrozen);

            // Clip plane
            BABYLON.MaterialHelper.BindClipPlane(effect, scene);

            // Colors
            scene.ambientColor.multiplyToRef(this.ambientColor, this._globalAmbientColor);

            BABYLON.MaterialHelper.BindEyePosition(effect, scene);
            effect.setColor3("vAmbientColor", this._globalAmbientColor);
        }

        if (mustRebind || !this.isFrozen) {
            // // Lights
            // if (scene.lightsEnabled && !this._disableLighting) {
            //     BABYLON.MaterialHelper.BindLights(scene, mesh, effect, defines, this._maxSimultaneousLights, this._rebuildInParallel);
            // }

            // // View
            // if (scene.fogEnabled && mesh.applyFog && scene.fogMode !== BABYLON.Scene.FOGMODE_NONE || this._reflectionTexture || this._refractionTexture) {
            //     this.bindView(effect);
            // }

            // Fog
            BABYLON.MaterialHelper.BindFogParameters(scene, mesh, effect);

            // Morph targets
            if (defines.NUM_MORPH_INFLUENCERS) {
                BABYLON.MaterialHelper.BindMorphTargetParameters(mesh, effect);
            }

            // Log. depth
            if (this.useLogarithmicDepth) {
                BABYLON.MaterialHelper.BindLogDepth(defines, effect, scene);
            }

            // image processing
            if (this._imageProcessingConfiguration && !this._imageProcessingConfiguration.applyByPostProcess) {
                this._imageProcessingConfiguration.bind(this._activeEffect);
            }
        }

        ubo.update();
        this._afterBind(mesh, this._activeEffect);
    }

    
    // /**
    //  * Get if the submesh is ready to be used and all its information available.
    //  * Child classes can use it to update shaders
    //  * @param mesh defines the mesh to check
    //  * @param subMesh defines which submesh to check
    //  * @param useInstances specifies that instances should be used
    //  * @returns a boolean indicating that the submesh is ready or not
    //  */
    // BABYLON.StandardMaterial.prototype.isReadyForSubMesh = function (this: BABYLON.StandardMaterial, mesh: StandardMaterialMesh, subMesh: BABYLON.SubMesh, useInstances: boolean = false): boolean {
    //     if (subMesh.effect && this.isFrozen) {
    //         if (subMesh.effect._wasPreviouslyReady) {
    //             return true;
    //         }
    //     }

    //     if (!subMesh._materialDefines) {
    //         subMesh._materialDefines = new BABYLON.StandardMaterialDefines();
    //     }

    //     var scene = this.getScene();
    //     var defines = <BABYLON.StandardMaterialDefines>subMesh._materialDefines;
    //     if (this._isReadyForSubMesh(subMesh)) {
    //         return true;
    //     }

    //     var engine = scene.getEngine();

    //     // PI_BEGIN - 避免无光照有Fresnel时无法线的的错误
    //     // // Lights
    //     // defines._needNormals = MaterialHelper.PrepareDefinesForLights(scene, mesh, defines, true, this._maxSimultaneousLights, this._disableLighting);
    //     // ////////////////////////////////////
    //     defines._needNormals = defines.FRESNEL || BABYLON.MaterialHelper.PrepareDefinesForLights(scene, mesh, defines, true, this._maxSimultaneousLights, this._disableLighting);
    //     // PI_END

    //     // Multiview
    //     BABYLON.MaterialHelper.PrepareDefinesForMultiview(scene, defines);

    //     // PrePass
    //     BABYLON.MaterialHelper.PrepareDefinesForPrePass(scene, defines, this.canRenderToMRT);

    //     // Textures
    //     if (defines._areTexturesDirty) {
    //         defines._needUVs = false;
    //         defines.MAINUV1 = false;
    //         defines.MAINUV2 = false;
    //         if (scene.texturesEnabled) {
    //             if (this._diffuseTexture && BABYLON.StandardMaterial.DiffuseTextureEnabled) {
    //                 if (!this._diffuseTexture.isReadyOrNotBlocking()) {
    //                     return false;
    //                 } else {
    //                     BABYLON.MaterialHelper.PrepareDefinesForMergedUV(this._diffuseTexture, defines, "DIFFUSE");
    //                 }
    //             } else {
    //                 defines.DIFFUSE = false;
    //             }

    //             if (this._ambientTexture && BABYLON.StandardMaterial.AmbientTextureEnabled) {
    //                 if (!this._ambientTexture.isReadyOrNotBlocking()) {
    //                     return false;
    //                 } else {
    //                     BABYLON.MaterialHelper.PrepareDefinesForMergedUV(this._ambientTexture, defines, "AMBIENT");
    //                 }
    //             } else {
    //                 defines.AMBIENT = false;
    //             }

    //             if (this._opacityTexture && BABYLON.StandardMaterial.OpacityTextureEnabled) {
    //                 if (!this._opacityTexture.isReadyOrNotBlocking()) {
    //                     return false;
    //                 } else {
    //                     BABYLON.MaterialHelper.PrepareDefinesForMergedUV(this._opacityTexture, defines, "OPACITY");
    //                     defines.OPACITYRGB = this._opacityTexture.getAlphaFromRGB;
    //                     //#region PI_BEGIN - opacity 拓展 两层opacity混合
    //                     if (this.mixOpacity) {
    //                         defines.MIX_OPACITY = true;
    //                     }
    //                     //#endregion
    //                 }
    //             } else {
    //                 defines.OPACITY = false;
    //             }

    //             if (this._reflectionTexture && BABYLON.StandardMaterial.ReflectionTextureEnabled) {
    //                 if (!this._reflectionTexture.isReadyOrNotBlocking()) {
    //                     return false;
    //                 } else {
    //                     defines._needNormals = true;
    //                     defines.REFLECTION = true;

    //                     defines.ROUGHNESS = (this._roughness > 0);
    //                     defines.REFLECTIONOVERALPHA = this._useReflectionOverAlpha;
    //                     defines.INVERTCUBICMAP = (this._reflectionTexture.coordinatesMode === BABYLON.Texture.INVCUBIC_MODE);
    //                     defines.REFLECTIONMAP_3D = this._reflectionTexture.isCube;
    //                     defines.RGBDREFLECTION = this._reflectionTexture.isRGBD;

    //                     switch (this._reflectionTexture.coordinatesMode) {
    //                         case BABYLON.Texture.EXPLICIT_MODE:
    //                             defines.setReflectionMode("REFLECTIONMAP_EXPLICIT");
    //                             break;
    //                         case BABYLON.Texture.PLANAR_MODE:
    //                             defines.setReflectionMode("REFLECTIONMAP_PLANAR");
    //                             break;
    //                         case BABYLON.Texture.PROJECTION_MODE:
    //                             defines.setReflectionMode("REFLECTIONMAP_PROJECTION");
    //                             break;
    //                         case BABYLON.Texture.SKYBOX_MODE:
    //                             defines.setReflectionMode("REFLECTIONMAP_SKYBOX");
    //                             break;
    //                         case BABYLON.Texture.SPHERICAL_MODE:
    //                             defines.setReflectionMode("REFLECTIONMAP_SPHERICAL");
    //                             break;
    //                         case BABYLON.Texture.EQUIRECTANGULAR_MODE:
    //                             defines.setReflectionMode("REFLECTIONMAP_EQUIRECTANGULAR");
    //                             break;
    //                         case BABYLON.Texture.FIXED_EQUIRECTANGULAR_MODE:
    //                             defines.setReflectionMode("REFLECTIONMAP_EQUIRECTANGULAR_FIXED");
    //                             break;
    //                         case BABYLON.Texture.FIXED_EQUIRECTANGULAR_MIRRORED_MODE:
    //                             defines.setReflectionMode("REFLECTIONMAP_MIRROREDEQUIRECTANGULAR_FIXED");
    //                             break;
    //                         case BABYLON.Texture.CUBIC_MODE:
    //                         case BABYLON.Texture.INVCUBIC_MODE:
    //                         default:
    //                             defines.setReflectionMode("REFLECTIONMAP_CUBIC");
    //                             break;
    //                     }

    //                     defines.USE_LOCAL_REFLECTIONMAP_CUBIC = (<any>this._reflectionTexture).boundingBoxSize ? true : false;
    //                 }
    //             } else {
    //                 defines.REFLECTION = false;
    //             }

    //             if (this._emissiveTexture && BABYLON.StandardMaterial.EmissiveTextureEnabled) {
    //                 if (!this._emissiveTexture.isReadyOrNotBlocking()) {
    //                     return false;
    //                 } else {
    //                     BABYLON.MaterialHelper.PrepareDefinesForMergedUV(this._emissiveTexture, defines, "EMISSIVE");
    //                 }
    //             } else {
    //                 defines.EMISSIVE = false;
    //             }

    //             // PI_BEGIN
    //             if (this._maskTexture && BABYLON.StandardMaterial.MaskTextureEnabled) {
    //                 if (!this._maskTexture.isReadyOrNotBlocking()) {
    //                     return false;
    //                 } else {
    //                     BABYLON.MaterialHelper.PrepareDefinesForMergedUV(this._maskTexture, defines, "PI_MASK");
    //                 }
    //             } else {
    //                 defines.PI_MASK = false;
    //             }
    //             // PI_END

    //             // PI_BEGIN
    //             let positionStateTexture = scene.positionStateTexture;
    //             if (this._applyPositionStateTexture && positionStateTexture && BABYLON.StandardMaterial.PositionStateTextureEnabled) {
    //                 if (!positionStateTexture.isReadyOrNotBlocking()) {
    //                     return false;
    //                 } else {
    //                     BABYLON.MaterialHelper.PrepareDefinesForMergedUV(positionStateTexture, defines, "PI_POSITION_STATE");
    //                 }
    //             } else {
    //                 defines.PI_POSITION_STATE = false;
    //             }
    //             // PI_END

    //             if (this._lightmapTexture && BABYLON.StandardMaterial.LightmapTextureEnabled) {
    //                 if (!this._lightmapTexture.isReadyOrNotBlocking()) {
    //                     return false;
    //                 } else {
    //                     BABYLON.MaterialHelper.PrepareDefinesForMergedUV(this._lightmapTexture, defines, "LIGHTMAP");
    //                     defines.USELIGHTMAPASSHADOWMAP = this._useLightmapAsShadowmap;
    //                     defines.RGBDLIGHTMAP = this._lightmapTexture.isRGBD;
    //                 }
    //             } else {
    //                 defines.LIGHTMAP = false;
    //             }

    //             if (this._specularTexture && BABYLON.StandardMaterial.SpecularTextureEnabled) {
    //                 if (!this._specularTexture.isReadyOrNotBlocking()) {
    //                     return false;
    //                 } else {
    //                     BABYLON.MaterialHelper.PrepareDefinesForMergedUV(this._specularTexture, defines, "SPECULAR");
    //                     defines.GLOSSINESS = this._useGlossinessFromSpecularMapAlpha;
    //                 }
    //             } else {
    //                 defines.SPECULAR = false;
    //             }

    //             if (scene.getEngine().getCaps().standardDerivatives && this._bumpTexture && BABYLON.StandardMaterial.BumpTextureEnabled) {
    //                 // Bump texure can not be not blocking.
    //                 if (!this._bumpTexture.isReady()) {
    //                     return false;
    //                 } else {
    //                     BABYLON.MaterialHelper.PrepareDefinesForMergedUV(this._bumpTexture, defines, "BUMP");

    //                     defines.PARALLAX = this._useParallax;
    //                     defines.PARALLAXOCCLUSION = this._useParallaxOcclusion;
    //                 }

    //                 defines.OBJECTSPACE_NORMALMAP = this._useObjectSpaceNormalMap;
    //             } else {
    //                 defines.BUMP = false;
    //             }

    //             if (this._refractionTexture && BABYLON.StandardMaterial.RefractionTextureEnabled) {
    //                 if (!this._refractionTexture.isReadyOrNotBlocking()) {
    //                     return false;
    //                 } else {
    //                     defines._needUVs = true;
    //                     defines.REFRACTION = true;

    //                     defines.REFRACTIONMAP_3D = this._refractionTexture.isCube;
    //                     defines.RGBDREFRACTION = this._refractionTexture.isRGBD;
    //                 }
    //             } else {
    //                 defines.REFRACTION = false;
    //             }

    //             defines.TWOSIDEDLIGHTING = !this._backFaceCulling && this._twoSidedLighting;
    //         } else {
    //             defines.DIFFUSE = false;
    //             defines.AMBIENT = false;
    //             defines.OPACITY = false;
    //             //#region PI_BEGIN - opacity 拓展 两层opacity混合
    //             defines.MIX_OPACITY = false;
    //             //#endregion
    //             defines.REFLECTION = false;
    //             defines.EMISSIVE = false;
    //             defines.LIGHTMAP = false;
    //             defines.BUMP = false;
    //             defines.REFRACTION = false;
    //         }

    //         defines.ALPHAFROMDIFFUSE = this._shouldUseAlphaFromDiffuseTexture();

    //         defines.EMISSIVEASILLUMINATION = this._useEmissiveAsIllumination;

    //         defines.LINKEMISSIVEWITHDIFFUSE = this._linkEmissiveWithDiffuse;

    //         defines.SPECULAROVERALPHA = this._useSpecularOverAlpha;

    //         defines.PREMULTIPLYALPHA = (this.alphaMode === BABYLON.Constants.ALPHA_PREMULTIPLIED || this.alphaMode === BABYLON.Constants.ALPHA_PREMULTIPLIED_PORTERDUFF);

    //         defines.ALPHATEST_AFTERALLALPHACOMPUTATIONS = this.transparencyMode !== null;

    //         defines.ALPHABLEND = this.transparencyMode === null || this.needAlphaBlendingForMesh(mesh); // check on null for backward compatibility
    //     }

    //     if (!this.detailMap.isReadyForSubMesh(defines, scene)) {
    //         return false;
    //     }

    //     if (defines._areImageProcessingDirty && this._imageProcessingConfiguration) {
    //         if (!this._imageProcessingConfiguration.isReady()) {
    //             return false;
    //         }

    //         this._imageProcessingConfiguration.prepareDefines(defines);

    //         defines.IS_REFLECTION_LINEAR = (this.reflectionTexture != null && !this.reflectionTexture.gammaSpace);
    //         defines.IS_REFRACTION_LINEAR = (this.refractionTexture != null && !this.refractionTexture.gammaSpace);
    //     }

    //     if (defines._areFresnelDirty) {
    //         if (BABYLON.StandardMaterial.FresnelEnabled) {
    //             // Fresnel
    //             if (this._diffuseFresnelParameters || this._opacityFresnelParameters ||
    //                 this._emissiveFresnelParameters || this._refractionFresnelParameters ||
    //                 this._reflectionFresnelParameters) {

    //                 defines.DIFFUSEFRESNEL = (this._diffuseFresnelParameters && this._diffuseFresnelParameters.isEnabled);

    //                 defines.OPACITYFRESNEL = (this._opacityFresnelParameters && this._opacityFresnelParameters.isEnabled);

    //                 defines.REFLECTIONFRESNEL = (this._reflectionFresnelParameters && this._reflectionFresnelParameters.isEnabled);

    //                 defines.REFLECTIONFRESNELFROMSPECULAR = this._useReflectionFresnelFromSpecular;

    //                 defines.REFRACTIONFRESNEL = (this._refractionFresnelParameters && this._refractionFresnelParameters.isEnabled);

    //                 defines.EMISSIVEFRESNEL = (this._emissiveFresnelParameters && this._emissiveFresnelParameters.isEnabled);

    //                 defines._needNormals = true;
    //                 defines.FRESNEL = true;
    //             }
    //         } else {
    //             defines.FRESNEL = false;
    //         }
    //     }

    //     // Misc.
    //     BABYLON.MaterialHelper.PrepareDefinesForMisc(mesh, scene, this._useLogarithmicDepth, this.pointsCloud, this.fogEnabled, this._shouldTurnAlphaTestOn(mesh) || this._forceAlphaTest, defines);

    //     // Attribs
    //     BABYLON.MaterialHelper.PrepareDefinesForAttributes(mesh, defines, true, true, true);

    //     // Values that need to be evaluated on every frame
    //     BABYLON.MaterialHelper.PrepareDefinesForFrameBoundValues(scene, engine, defines, useInstances, null, subMesh.getRenderingMesh().hasThinInstances);

    //     // External config
    //     this.detailMap.prepareDefines(defines, scene);

    //     // Get correct effect
    //     if (defines.isDirty) {
    //         const lightDisposed = defines._areLightsDisposed;
    //         defines.markAsProcessed();

    //         // Fallbacks
    //         var fallbacks = new EffectFallbacks();
    //         // if (defines.REFLECTION) {
    //         //     fallbacks.addFallback(0, "REFLECTION");
    //         // }

    //         // if (defines.SPECULAR) {
    //         //     fallbacks.addFallback(0, "SPECULAR");
    //         // }

    //         // if (defines.BUMP) {
    //         //     fallbacks.addFallback(0, "BUMP");
    //         // }

    //         // if (defines.PARALLAX) {
    //         //     fallbacks.addFallback(1, "PARALLAX");
    //         // }

    //         // if (defines.PARALLAXOCCLUSION) {
    //         //     fallbacks.addFallback(0, "PARALLAXOCCLUSION");
    //         // }

    //         // if (defines.SPECULAROVERALPHA) {
    //         //     fallbacks.addFallback(0, "SPECULAROVERALPHA");
    //         // }

    //         // if (defines.FOG) {
    //         //     fallbacks.addFallback(1, "FOG");
    //         // }

    //         if (defines.POINTSIZE) {
    //             fallbacks.addFallback(0, "POINTSIZE");
    //         }

    //         // if (defines.LOGARITHMICDEPTH) {
    //         //     fallbacks.addFallback(0, "LOGARITHMICDEPTH");
    //         // }

    //         // MaterialHelper.HandleFallbacksForShadows(defines, fallbacks, this._maxSimultaneousLights);

    //         // if (defines.SPECULARTERM) {
    //         //     fallbacks.addFallback(0, "SPECULARTERM");
    //         // }

    //         if (defines.DIFFUSEFRESNEL) {
    //             fallbacks.addFallback(1, "DIFFUSEFRESNEL");
    //         }

    //         if (defines.OPACITYFRESNEL) {
    //             fallbacks.addFallback(2, "OPACITYFRESNEL");
    //         }

    //         // if (defines.REFLECTIONFRESNEL) {
    //         //     fallbacks.addFallback(3, "REFLECTIONFRESNEL");
    //         // }

    //         if (defines.EMISSIVEFRESNEL) {
    //             fallbacks.addFallback(4, "EMISSIVEFRESNEL");
    //         }

    //         if (defines.FRESNEL) {
    //             fallbacks.addFallback(4, "FRESNEL");
    //         }

    //         // if (defines.MULTIVIEW) {
    //         //     fallbacks.addFallback(0, "MULTIVIEW");
    //         // }

    //         //Attributes
    //         var attribs = [BABYLON.VertexBuffer.PositionKind];

    //         if (defines.NORMAL) {
    //             attribs.push(BABYLON.VertexBuffer.NormalKind);
    //         }

    //         if (defines.UV1) {
    //             attribs.push(BABYLON.VertexBuffer.UVKind);
    //         }

    //         if (defines.UV2) {
    //             attribs.push(BABYLON.VertexBuffer.UV2Kind);
    //         }

    //         if (defines.VERTEXCOLOR) {
    //             attribs.push(BABYLON.VertexBuffer.ColorKind);
    //         }

    //         // MaterialHelper.PrepareAttributesForBones(attribs, mesh, defines, fallbacks);
    //         // MaterialHelper.PrepareAttributesForInstances(attribs, defines);
    //         // MaterialHelper.PrepareAttributesForMorphTargets(attribs, mesh, defines);

    //         var shaderName = "default";

    //         var uniforms = ["world", "view", "viewProjection", "vEyePosition", "vLightsType", "vAmbientColor", "vDiffuseColor", "vSpecularColor", "vEmissiveColor", "visibility",
    //             "vFogInfos", "vFogColor", "pointSize",
    //             "vDiffuseInfos", "vAmbientInfos", "vOpacityInfos", "vReflectionInfos", "vEmissiveInfos", "vSpecularInfos", "vBumpInfos", "vLightmapInfos", "vRefractionInfos",
    //             "mBones",
    //             "vClipPlane", "vClipPlane2", "vClipPlane3", "vClipPlane4", "vClipPlane5", "vClipPlane6", "diffuseMatrix", "ambientMatrix", "opacityMatrix", "reflectionMatrix", "emissiveMatrix", "specularMatrix", "bumpMatrix", "normalMatrix", "lightmapMatrix", "refractionMatrix",
    //             "diffuseLeftColor", "diffuseRightColor", "opacityParts", "reflectionLeftColor", "reflectionRightColor", "emissiveLeftColor", "emissiveRightColor", "refractionLeftColor", "refractionRightColor",
    //             "vReflectionPosition", "vReflectionSize",
    //             "logarithmicDepthConstant", "vTangentSpaceParams", "alphaCutOff", "boneTextureWidth"
    //             //#region PI_BEGIN alphaCutOff 拓展
    //             , "cutoffEdgeColor", "cutoffEdgeInfo"
    //             //#endregion
    //             // PI_BEGIN
    //             , "maskMatrix", "vMaskInfos", "vMaskColor"
    //             // PI_END
    //             // PI_BEGIN
    //             , "positionStateRange"
    //             // PI_END
    //             // PI_BEGIN - opacity 拓展 两层opacity混合
    //             , "mixOpacityInfo", "mixOpacityMatrix", "mixOpacityControlMatrix"
    //             // PI_END
    //         ];

    //         var samplers = ["diffuseSampler", "ambientSampler", "opacitySampler", "reflectionCubeSampler",
    //             "reflection2DSampler", "emissiveSampler", "specularSampler", "bumpSampler", "lightmapSampler",
    //             "refractionCubeSampler", "refraction2DSampler", "boneSampler"
    //             // PI_BEGIN
    //             , "maskSampler"
    //             // PI_END
    //             // PI_BEGIN
    //             , "positionStateSampler"
    //             // PI_END
    //         ];

    //         var uniformBuffers = ["Material", "Scene"];

    //         // BABYLON.DetailMapConfiguration.AddUniforms(uniforms);
    //         // BABYLON.DetailMapConfiguration.AddSamplers(samplers);

    //         // BABYLON.PrePassConfiguration.AddUniforms(uniforms);
    //         // BABYLON.PrePassConfiguration.AddSamplers(uniforms);

    //         // if (BABYLON.ImageProcessingConfiguration) {
    //         //     BABYLON.ImageProcessingConfiguration.PrepareUniforms(uniforms, defines);
    //         //     BABYLON.ImageProcessingConfiguration.PrepareSamplers(samplers, defines);
    //         // }

    //         BABYLON.MaterialHelper.PrepareUniformsAndSamplersList(<BABYLON.IEffectCreationOptions>{
    //             uniformsNames: uniforms,
    //             uniformBuffersNames: uniformBuffers,
    //             samplers: samplers,
    //             defines: defines,
    //             maxSimultaneousLights: this._maxSimultaneousLights
    //         });

    //         const csnrOptions: BABYLON.ICustomShaderNameResolveOptions = {};

    //         if (this.customShaderNameResolve) {
    //             shaderName = this.customShaderNameResolve(shaderName, uniforms, uniformBuffers, samplers, defines, attribs, csnrOptions);
    //         }

    //         var join = defines.toString();

    //         let previousEffect = subMesh.effect;
    //         let effect = scene.getEngine().createEffect(shaderName, <IEffectCreationOptions>{
    //             attributes: attribs,
    //             uniformsNames: uniforms,
    //             uniformBuffersNames: uniformBuffers,
    //             samplers: samplers,
    //             defines: join,
    //             fallbacks: fallbacks,
    //             onCompiled: this.onCompiled,
    //             onError: this.onError,
    //             indexParameters: { maxSimultaneousLights: this._maxSimultaneousLights, maxSimultaneousMorphTargets: defines.NUM_MORPH_INFLUENCERS },
    //             processFinalCode: csnrOptions.processFinalCode,
    //             multiTarget: defines.PREPASS
    //         }, engine);

    //         if (effect) {
    //             if (this._onEffectCreatedObservable) {
    //                 onCreatedEffectParameters.effect = effect;
    //                 onCreatedEffectParameters.subMesh = subMesh;
    //                 this._onEffectCreatedObservable.notifyObservers(onCreatedEffectParameters);
    //             }

    //             // Use previous effect while new one is compiling
    //             if (this.allowShaderHotSwapping && previousEffect && !effect.isReady()) {
    //                 effect = previousEffect;
    //                 this._rebuildInParallel = true;
    //                 defines.markAsUnprocessed();

    //                 if (lightDisposed) {
    //                     // re register in case it takes more than one frame.
    //                     defines._areLightsDisposed = true;
    //                     return false;
    //                 }
    //             } else {
    //                 this._rebuildInParallel = false;
    //                 scene.resetCachedMaterial();
    //                 subMesh.setEffect(effect, defines);
    //                 this.buildUniformLayout();
    //             }
    //         }
    //     }

    //     if (!subMesh.effect || !subMesh.effect.isReady()) {
    //         return false;
    //     }

    //     defines._renderId = scene.getRenderId();
    //     subMesh.effect._wasPreviouslyReady = true;

    //     return true;
    // }
}

    export class MaterialHelper {
        /**
         * Binds a texture matrix value to its corrsponding uniform
         * @param texture The texture to bind the matrix for
         * @param uniformBuffer The uniform buffer receivin the data
         * @param key The channel key "diffuse", "specular"... used in the shader
         */
        public static BindTextureMatrix(texture: { getTextureMatrix: () => BABYLON.Matrix }, uniformBuffer: BABYLON.UniformBuffer, key: string): void {
            var matrix = texture.getTextureMatrix();

            uniformBuffer.updateMatrix(key + "Matrix", matrix);
        }
    }