import { SpriteManager } from "../sprite/sprite_manager";
import { BABYLON } from "../render3d/babylon";
import { ShaderMaterial } from "./shader_material";
import { GlobalAtlasManager } from "../sprite/sprite_atlas";
import { Constants } from "../interface/constants";
const KEY_TEXTUREPROGRESS_CIRCLE = "KEY_TEXTUREPROGRESS_CIRCLE";
//laserEffect
BABYLON.Effect.ShadersStore[`${KEY_TEXTUREPROGRESS_CIRCLE}VertexShader`] = `
    precision mediump float;

    attribute vec3 position;
    attribute vec4 color;
    attribute vec2 uv;

    uniform mat4 worldViewProjection;

    varying vec3 vColor;
    varying float vAlpha;
    varying vec2 vUV;
    varying vec2 vCenter;

    void main(void) {
        vec3 positionUpdated = position;
        vec2 uvUpdated = uv;

        gl_Position = worldViewProjection*vec4(positionUpdated, 1.0);

        vUV = uv;

        vColor = color.rgb;
        vAlpha = (1.0 - color.a) * 1.002 - 0.001;
        vCenter = uv;
    }
`;

BABYLON.Effect.ShadersStore[`${KEY_TEXTUREPROGRESS_CIRCLE}FragmentShader`] =
    `
    precision mediump float;

    varying vec3 vColor;
    varying float vAlpha;
    varying vec2 vUV;
    varying vec2 vCenter;

    uniform sampler2D diffuseTexture;
 
    void main(void) {
        vec4 src = texture2D(diffuseTexture, vUV);
        
        vec2 base_v = vec2(0.0, 1.0);
        vec2 diff   = vCenter - vec2(0.5, 0.5);
        vec2 dist   = normalize(diff);
        float d     = dot(vec2(0.0, 1.0), dist);
        
        float angle = acos(d) * mix(-1., 1., step(0., diff.x));
        angle       = angle + mix(6.283185307179586, 0., step(0., diff.x));
        angle       /= 6.283185307179586;

        float intensity = smoothstep(vAlpha, vAlpha + 0.001, 1.0 - angle);
        gl_FragColor = vec4(src.rgb, src.a * intensity);
    }
`;

export class TextureProgressCircleMaterial extends ShaderMaterial {
    public static readonly KEY = KEY_TEXTUREPROGRESS_CIRCLE;
    private _tintColor: BABYLON.Color4 = new BABYLON.Color4(0, 0, 0, 0);
    public _emissiveColor: BABYLON.Color3 = new BABYLON.Color3(1, 1, 1);
    public get emissiveColor() {
        return this._emissiveColor;
    }
    public set emissiveColor(value: BABYLON.Color3) {
        this._emissiveColor = value;
    }
    public get alpha() {
        return this._tintColor.a;
    }
    public set alpha(value: number) {
        this._tintColor.a = value;
    }
    constructor(materialName: string, babylonScene: any) {
        super(
            materialName,
            babylonScene,
            `${TextureProgressCircleMaterial.KEY}`,
            {
                attributes: ["position", "uv", "color"],
                uniforms: ["worldViewProjection"],
                samplers: ["diffuseTexture"],
                needAlphaBlending: true
            }
        );
    }
    protected _diffuseTexture: BABYLON.Texture;
    public set diffuseTexture(tex: BABYLON.Texture) {
        this.setTexture('diffuseTexture', tex);
        this._diffuseTexture = tex;
    }
    public get diffuseTexture() {
        return this._diffuseTexture;
    }
    isReady(mesh?: BABYLON.AbstractMesh, useInstances?: boolean): boolean {
        let result = !!this._diffuseTexture;
        return result && super.isReady(mesh, useInstances);
    }
    bind(world: BABYLON.Matrix, mesh?: BABYLON.Mesh, effectOverride?: BABYLON.Effect): void {
        super.bind(world, mesh, effectOverride);
    }
    dispose(forceDisposeEffect?: boolean, forceDisposeTextures?: boolean, notBoundToMesh?: boolean): void {
        this._diffuseTexture?.dispose();
        super.dispose(forceDisposeEffect, forceDisposeTextures, notBoundToMesh);
    }
    public static RegistForSprite() {
        SpriteManager.MaterialFactory.set(5, function(scene: BABYLON.Scene, atlasImage: string, noMipmap: boolean, samplerMode: number, alphaMode: number, pass?: Constants.PassTag) {
            let _material = new TextureProgressCircleMaterial(`${atlasImage}`, scene);
            let texture = GlobalAtlasManager.getTexture(atlasImage, scene);
            if (texture == undefined) {
                texture = new BABYLON.Texture(null, scene, noMipmap, false, samplerMode || BABYLON.Texture.LINEAR_LINEAR);
                texture.updateURL(atlasImage);
                texture.hasAlpha = true;
                texture.wrapU = BABYLON.Texture.CLAMP_ADDRESSMODE;
                texture.wrapV = BABYLON.Texture.CLAMP_ADDRESSMODE;
            }
            _material.diffuseTexture = texture;
            _material.emissiveColor = new BABYLON.Color3(1, 1, 1);
            _material.disableDepthWrite = true;
            _material.backFaceCulling = false;
            _material.alphaMode = alphaMode;
            _material.transparencyMode = BABYLON.Material.MATERIAL_ALPHATESTANDBLEND;
            return _material;
        });
    }
}
