import { MathDataPool } from "../../render3d/extends/math";
import { BABYLON } from "../../render3d/babylon";
// import * as BABYLON from "../../pi_babylon/render3d/babylon";
import { computeRadians, EShapeEmitterArcMode, EShapeEmitterDirectionMode, IShapeEmitterType } from "./ishape_emitter_type";

/**
 * 圆形发射器
 */
export class CircleShapeEmitter implements IShapeEmitterType {
    public static MAX_Z = 999999999;
    public directionMode: EShapeEmitterDirectionMode = EShapeEmitterDirectionMode.Unity;

    /**
     * 半径
     */
    public radius: number;
    /**
     * 半径域
     */
    public radiusRange: number;
    /**
     * 弧形范围
     */
    public arcValue: number = Math.PI * 2;
    /**
     * 弧形范围发射模式
     */
    public arcMode: EShapeEmitterArcMode = EShapeEmitterArcMode.Random;
    /**
     * 弧形周围可产生粒子的离散间隔 - 小于0.01 时, 不做间隔计算
     */
    public arcSpread: number = 0;
    /**
     * 弧形范围发射速度
     */
    public arcSpeed: number = 1;
    /**
     * 弧形范围精度
     */
    public arcSpreadLimit: number = 0.001;

    public rotation     : BABYLON.Vector3    = MathDataPool.Vector3(0, 0, 0);
    public position     : BABYLON.Vector3    = MathDataPool.Vector3(0, 0, 0);
    public scaling      : BABYLON.Vector3     = MathDataPool.Vector3(1, 1, 1);
    public localMatrix  : BABYLON.Matrix  = MathDataPool.Matrix();

    alignDirection: boolean = false;
    randomizeDirection: number = 0;
    spherizeDirection: number = 0;
    randomizePosition: number = 0;
    /**
     * 
     */
    constructor(radius = 1, radiusRange = 1,
        /** defines how much to randomize the particle direction [0-1] (default is 0) */
        public directionRandomizer = 0) {
        this.radius = radius;
        this.radiusRange = radiusRange;
    }
    public dispose() {
        MathDataPool.Recycle(this.rotation   );
        MathDataPool.Recycle(this.position   );
        MathDataPool.Recycle(this.scaling    );
        MathDataPool.Recycle(this.localMatrix);
        this.rotation     = undefined;
        this.position     = undefined;
        this.scaling      = undefined;
        this.localMatrix  = undefined;
    }

    public rotateByQuaternionAroundPoint = (position: BABYLON.Vector3, rotateQuat: BABYLON.Quaternion, centerOffset: BABYLON.Vector3, orbtialDiff: BABYLON.Vector3) => {
        position.z = 0;
        position.rotateByQuaternionAroundPointToRef(rotateQuat, centerOffset, orbtialDiff);
    }
    /**
     * 
     */
    public startDirectionFunction(worldMatrix: BABYLON.Matrix, directionToUpdate: BABYLON.Vector3, position: BABYLON.Vector3, localPosition: BABYLON.Vector3, isLocal: boolean): void {
        var direction = BABYLON.Vector3.Zero();

        if (isLocal) {
            localPosition.normalizeToRef(direction);
        }
        else {
            position.subtract(worldMatrix.getTranslation()).normalizeToRef(direction);
        }

        if (this.randomizeDirection != 0) {
            direction.x += BABYLON.Scalar.RandomRange(0, this.randomizeDirection);
            direction.y += BABYLON.Scalar.RandomRange(0, this.randomizeDirection);
            direction.z += BABYLON.Scalar.RandomRange(0, this.randomizeDirection);
            direction.normalize();
        }


        if (isLocal) {
            directionToUpdate.copyFrom(direction);
        }
        else {
            BABYLON.Vector3.TransformNormalFromFloatsToRef(direction.x, direction.y, direction.z, worldMatrix, directionToUpdate);
        }
    }

    /**
     * Called by the particle System when the position is computed for the created particle.
     * @param worldMatrix is the world matrix of the particle system
     * @param positionToUpdate is the position vector to update with the result
     * @param particle is the particle we are computed the position for
     * @param isLocal defines if the position should be set in local space
     */
    startPositionFunction(worldMatrix: BABYLON.Matrix, positionToUpdate: BABYLON.Vector3, emissionLoop: number, emissionProgress: number, emissionIndex: number, emissionTotal: number, isLocal: boolean): void {
        let s = computeRadians(emissionLoop, emissionProgress, emissionIndex, emissionTotal, Math.PI * 2, this.arcValue, this.arcSpread, this.arcSpeed, this.arcMode);

        var randRadius = this.radius - BABYLON.Scalar.RandomRange(0, this.radius * this.radiusRange);
        var randX = randRadius * Math.cos(s);
        var randY = randRadius * Math.sin(s);
        var randZ = 0;

        if (this.randomizePosition != 0) {
            randX += BABYLON.Scalar.RandomRange(-this.randomizePosition, this.randomizePosition);
            randZ += BABYLON.Scalar.RandomRange(-this.randomizePosition, this.randomizePosition);
            randY += BABYLON.Scalar.RandomRange(-this.randomizePosition, this.randomizePosition);
        }

        if (isLocal) {
            positionToUpdate.x = randX;
            positionToUpdate.y = randY;
            positionToUpdate.z = randZ;
        }
        else {
            BABYLON.Vector3.TransformCoordinatesFromFloatsToRef(randX, randY, randZ, worldMatrix, positionToUpdate);
        }
    }

    /**
     * Clones the current emitter and returns a copy of it
     * @returns the new emitter
     */
    public clone(): CircleShapeEmitter {
        let newOne = new CircleShapeEmitter(this.radius, this.radiusRange, this.directionRandomizer);

        BABYLON.DeepCopier.DeepCopy(this, newOne);

        return newOne;
    }

    /**
     * @returns a string containing the class name
     */
    public getClassName(): string {
        return "CircleParticleEmitter";
    }

    /**
     * Serializes the particle system to a JSON object.
     * @returns the JSON object
     */
    public serialize(): any {
        var serializationObject: any = {};

        serializationObject.type = this.getClassName();
        serializationObject.radius = this.radius;
        serializationObject.directionRandomizer = this.directionRandomizer;
        serializationObject.radiusRange = this.radiusRange;

        return serializationObject;
    }

    /**
     * Parse properties from a JSON object
     * @param serializationObject defines the JSON object
     */
    public parse(serializationObject: any): void {
        this.radius = serializationObject.radius;
        this.radiusRange = serializationObject.radiusRange !== undefined ? serializationObject.radiusRange : 1;
        this.directionRandomizer = serializationObject.directionRandomizer;
    }
}
