import { BABYLON } from "../render3d/babylon";
import { IMeshParticleSystem, IMeshParticleSystemBatch } from "../interface/particlesystem";
import { SceneStruct } from "../scene_base";
import { MeshParticleSystem } from "./mesh_particle_system";
import { IMesh } from "../interface/mesh";
import { MathDataPool } from "../render3d/extends/math";

export interface IMeshParticleSytemBatchMesh extends BABYLON.Mesh {
    _matrixDATA: Float32Array;
    _colorDATA: Float32Array;
}

export interface IMeshParticleSytemBatchMeshNoIns extends BABYLON.Mesh {
    _positionDATA: Float32Array;
    _colorDATA: Float32Array;
    _uvDATA: Float32Array;
    _indicesDATA: Uint16Array;
}

export class MeshParticleSystemBatch implements IMeshParticleSystemBatch {
    /**
     * 每个粒子相同同时最多多少个粒子
     */
    public static MAX_COUNT: number = 100;
    /**
     * 最多缓存多少个粒子系统用的Buffer容器
     */
    public static MAX_RECYCLE: number = 40;
    private static poolMatrix: Float32Array[] = [];
    private static poolColor: Float32Array[] = [];
    protected maxBatchParticleCountForNoIns: number;
    protected _actives: Map<number, MeshParticleSystem> = new Map();
    _dirty: boolean;
    protected readonly useInstance: boolean;
    protected _meshes: BABYLON.Mesh[] = [];
    protected _buildedMeshes: IMeshParticleSytemBatchMesh[] = [];
    protected _buildedMeshesNoIns: IMeshParticleSytemBatchMeshNoIns[] = [];
    constructor(private scene: BABYLON.Scene, public readonly key: string, public readonly maxBatchParticleCount: number = MeshParticleSystemBatch.MAX_COUNT) {
        this.useInstance = this.scene.getEngine()._caps.instancedArrays;
        // scene.onBeforeRenderObservable.add(this._update);
    }
    setSourceMesh(mesh: IMesh) {
        if (mesh) {
            this._meshes.push(<any>mesh);
        }
    }
    protected getBuildedMesh(idx: number) {
        let result = this._buildedMeshes[idx];
        if (result == undefined) {
            let mesh = <IMeshParticleSytemBatchMesh>this._meshes.pop();
            if (mesh) {
                if (this.useInstance) {
                    let _matrixDATA = MeshParticleSystemBatch.poolMatrix.pop();
                    if (_matrixDATA == undefined) {
                        _matrixDATA = new Float32Array(this.maxBatchParticleCount * 16);
                    }
                    mesh._matrixDATA = _matrixDATA;
                    let _colorDATA = MeshParticleSystemBatch.poolColor.pop();
                    if (_colorDATA == undefined) {
                        _colorDATA = new Float32Array(this.maxBatchParticleCount * 4);
                    }
                    mesh._colorDATA = _colorDATA;
                    mesh.thinInstanceSetBuffer("matrix", mesh._matrixDATA, 16, false);
                    mesh.thinInstanceSetBuffer("color", mesh._colorDATA, 4, false);
                    this._buildedMeshes.push(mesh);
                    result = mesh;
                }
            }
        }

        return result;
    }
    protected getBuildedMeshNoIns(idx: number) {
        let result = this._buildedMeshesNoIns[idx];
        if (result == undefined) {
            let _sourceMesh = this._meshes[0];
            if (_sourceMesh) {
                let positions = _sourceMesh.geometry.getVerticesData("position");
                let vertexCount = positions.length / 3;
                this.maxBatchParticleCountForNoIns = Math.min(this.maxBatchParticleCount, Math.floor(65536 / Math.ceil(vertexCount)))
                let count = this.maxBatchParticleCountForNoIns;

                let mesh = <IMeshParticleSytemBatchMeshNoIns>(new BABYLON.Mesh(this.key, this.scene));


                mesh._positionDATA = new Float32Array(vertexCount * 3 * count);
                mesh._colorDATA = new Float32Array(vertexCount * 4 * count);
                mesh.setVerticesData("position", mesh._positionDATA, true);
                mesh.setVerticesData(BABYLON.VertexBuffer.ColorKind, mesh._colorDATA, true);

                let uvs = _sourceMesh.geometry.getVerticesData(BABYLON.VertexBuffer.UVKind);
                mesh._uvDATA = new Float32Array(vertexCount * 2 * count);
                for (let i = 0; i < count; i++) {
                    for (let j = 0; j < vertexCount; j++) {
                        mesh._uvDATA[(i * vertexCount + j) * 2 + 0] = uvs[j * 2 + 0];
                        mesh._uvDATA[(i * vertexCount + j) * 2 + 1] = uvs[j * 2 + 1];
                    }
                }
                mesh.setVerticesData(BABYLON.VertexBuffer.UVKind, mesh._uvDATA, true);

                let indices = _sourceMesh.geometry.getIndices();
                let indicesCount = indices.length;
                let instanceIndices = new Uint16Array(indicesCount * count);
                let indicesTotal = indicesCount * count;
                for (let i = 0; i < count; i++) {
                    for (let j = 0; j < indicesCount; j++) {
                        instanceIndices[i * indicesCount + j] = indices[j] + i * vertexCount;
                    }
                }
                mesh.setIndices(instanceIndices, undefined, false);
                result = mesh;
            }
        }

        return result;
    }
    active(particlesys: MeshParticleSystem): void {
        this._actives.set(particlesys.particleSysId, particlesys);
    }
    remove(particlesys: MeshParticleSystem): void {
        this._actives.delete(particlesys.particleSysId);
    }
    disposeParticleSystem(_particlesys: MeshParticleSystem): void {
        let mesh = _particlesys._sourceMesh;
        if (mesh) {
            if ((<IMeshParticleSytemBatchMesh>mesh)._matrixDATA) {
                let _matrixDATA = (<IMeshParticleSytemBatchMesh>mesh)._matrixDATA;
                if (MeshParticleSystemBatch.poolMatrix.length < MeshParticleSystemBatch.MAX_RECYCLE) {
                    MeshParticleSystemBatch.poolMatrix.push(_matrixDATA);
                }
                let _colorDATA = (<IMeshParticleSytemBatchMesh>mesh)._colorDATA;
                if (MeshParticleSystemBatch.poolColor.length < MeshParticleSystemBatch.MAX_RECYCLE) {
                    MeshParticleSystemBatch.poolColor.push(_colorDATA);
                }
            }
            (<IMeshParticleSytemBatchMeshNoIns>mesh)._positionDATA = undefined;
            (<IMeshParticleSytemBatchMeshNoIns>mesh)._colorDATA = undefined;
            (<IMeshParticleSytemBatchMesh>mesh)._matrixDATA = undefined;
            (<IMeshParticleSytemBatchMesh>mesh)._colorDATA = undefined;

            let idx = this._meshes.findIndex(v => v == mesh);
            if (idx >= 0) {
                this._meshes.splice(idx, 1);
                return;
            } else {
                idx = this._buildedMeshes.findIndex(v => v == mesh);
                if (idx >= 0) {
                    this._buildedMeshes.splice(idx, 1);
                    return;
                }
            }
        }
        // let item = this._meshes.pop();
        // if (item == undefined) {
        //     if (this.useInstance) {
        //         item = this._buildedMeshes.pop();
        //     } else {
        //         item = this._buildedMeshesNoIns.pop();
        //     }
        // }
        // item?.dispose(false, true);
    }
    _update = (): void => {
        this._meshes.forEach((mesh) => {
            mesh.setEnabled(false);
        });
        this._buildedMeshes.forEach((mesh) => {
            mesh.setEnabled(false);
        });
        this._buildedMeshesNoIns.forEach((mesh) => {
            mesh.setEnabled(false);
        });
        let queues: Map<number, Map<number, number[]>> = new Map();

        let tempCountMap = new Map();
        this._actives.forEach((item, uniqueId) => {
            let renderingGroupId = item.renderingGroupId;
            let alphaIndex = item.alphaIndex;
            let sameGroup = queues.get(renderingGroupId);
            if (sameGroup == undefined) {
                sameGroup = new Map();
                queues.set(renderingGroupId, sameGroup);
            }
            let sameAlphaindex = sameGroup.get(alphaIndex);
            if (sameAlphaindex == undefined) {
                sameAlphaindex = [];
                sameGroup.set(alphaIndex, sameAlphaindex);
            }
            sameAlphaindex.push(uniqueId);
            item._computeCall();
            let tempCount = item._updateCall();
            tempCountMap.set(uniqueId, tempCount);
        });

        if (this.useInstance) {
            let batchIndex = 0;
            queues.forEach((sameGroup, groupId) => {
                sameGroup.forEach((sameAlphaindex, alphaIndex) => {
                    let count = 0;
                    let mesh = this.getBuildedMesh(batchIndex);
                    sameAlphaindex.forEach((particleSysId) => {
                        let particlesys = this._actives.get(particleSysId);
                        if (particlesys) {
                            let tempCount = tempCountMap.get(particleSysId);

                            if (count + tempCount > this.maxBatchParticleCount) {
                                return;
                            }

                            if (tempCount > 0) {
                                let mpMatrixList = particlesys.psTool.mpMatrixList;
                                let offset = count * 16;
                                for (let i = tempCount * 16; i > 0; i--) {
                                    mesh._matrixDATA[offset + i - 1] = mpMatrixList[i - 1];
                                }
                                let mpColorData = particlesys.psTool.mpColorData;
                                offset = count * 4;
                                for (let i = tempCount * 4; i > 0; i--) {
                                    mesh._colorDATA[offset + i - 1] = mpColorData[i - 1];
                                }
                            }

                            count += tempCount;
                        }
                    });
                    let enable = count > 0;
                    if (enable) {
                        mesh.renderingGroupId = groupId;
                        mesh.alphaIndex = alphaIndex;
                        mesh.thinInstanceCount = count;
                        mesh.thinInstanceBufferUpdated("matrix");
                        mesh.thinInstanceBufferUpdated("color");
                        mesh.setEnabled(enable);
                        batchIndex += 1;
                    }
                });
            });
        } else {
            let batchIndex = 0;
            queues.forEach((sameGroup, groupId) => {
                sameGroup.forEach((sameAlphaindex, alphaIndex) => {
                    let _sourceMesh = this._meshes[0];
                    let count = 0;
                    let mesh = this.getBuildedMeshNoIns(batchIndex);
                    sameAlphaindex.forEach((particleSysId) => {
                        let particlesys = this._actives.get(particleSysId);
                        if (particlesys) {
                            let tempCount = tempCountMap.get(particleSysId);

                            if (count + tempCount > this.maxBatchParticleCountForNoIns) {
                                return;
                            }
                            let mpMatrixList = particlesys.psTool.mpMatrixList;
                            let mpColorData = particlesys.psTool.mpColorData;

                            let matrix = MathDataPool.Matrix();
                            let position = MathDataPool.Vector3(0, 0, 0);
                            let sourcePosition = _sourceMesh.geometry.getVerticesData("position");
                            let vertexCount = Math.floor(sourcePosition.length / 3);
                            for (let i = 0; i < tempCount; i++) {
                                let vertexOff = (count + i) * vertexCount;
                                BABYLON.Matrix.FromArrayToRef(mpMatrixList, i * 16, matrix);
                                for (let j = 0; j < vertexCount; j++) {
                                    BABYLON.Vector3.FromArrayToRef(sourcePosition, j * 3, position);
                                    BABYLON.Vector3.TransformCoordinatesToRef(position, matrix, position);
                                    let offset = (vertexOff + j) * 3;
                                    mesh._positionDATA[offset + 0] = position.x;
                                    mesh._positionDATA[offset + 1] = position.y;
                                    mesh._positionDATA[offset + 2] = position.z;
                                }
                            }
                            for (let i = 0; i < tempCount; i++) {
                                let r = mpColorData[i * 4 + 0];
                                let g = mpColorData[i * 4 + 1];
                                let b = mpColorData[i * 4 + 2];
                                let a = mpColorData[i * 4 + 3];
                                let vertexOff = (count + i) * vertexCount;
                                for (let j = 0; j < vertexCount; j++) {
                                    let offset = (vertexOff + j) * 4;
                                    mesh._colorDATA[offset + 0] = r;
                                    mesh._colorDATA[offset + 1] = g;
                                    mesh._colorDATA[offset + 2] = b;
                                    mesh._colorDATA[offset + 3] = a;
                                }
                            }
                            count += tempCount;
                            MathDataPool.Recycle(matrix);
                            MathDataPool.Recycle(position);
                        }
                    });
                    let enable = count > 0;
                    if (enable) {
                        mesh.renderingGroupId = groupId;
                        mesh.alphaIndex = alphaIndex;
                        mesh.thinInstanceCount = count;
                        mesh.updateVerticesData("position", mesh._positionDATA);
                        mesh.updateVerticesData("color", mesh._colorDATA);
                        // mesh.updateVerticesData("uv", mesh._uvDATA);
                        mesh.setEnabled(true);
                        if (mesh.subMeshes[0] != undefined) {
                            mesh.subMeshes[0].indexCount = count * _sourceMesh.geometry.getIndices().length;
                        }

                        mesh.setEnabled(enable);
                        batchIndex += 1;
                    }
                });
            });
        }
    }
}