
import { PostProcessDataManager } from "./data_manager";
import { PostProcess } from "./postprocess";
import { RenderTexture, RenderTarget } from "./render_target";
import { ShaderCompiler } from "./shader_compiler";
import { ManagedWebGLRenderingContext } from "./webgl";

export class AnalogGlitch extends PostProcess {
    public static shaderName = `AnalogGlitch`;
    public static getShader(context: ManagedWebGLRenderingContext) {
        let manager = PostProcessDataManager.get(context);
        let result = manager.getShaderCompiler(AnalogGlitch.shaderName);
        if (!result) {
            result = new ShaderCompiler(context, PostProcess.getVertexString(), AnalogGlitch.getFragmentString());
            manager.setShaderCompiler(AnalogGlitch.shaderName, result);
        }

        return result;
    }
    private static Fractor = 'fractor';
    private static Time = 'time';
    private static Shake = 'shake';
    private static ColorDrift = 'colorDrift';

    scanLineJitterX: number = 0.1;
    scanLineJitterY: number = 0.1;
    verticalJumpX: number = 0.1;
    verticalJumpY: number = 0.1;
    horizontalShake: number = 0.1;
    colorDriftX: number = 0.1;
    colorDriftY: number = 0.1;
    time: number = 0.1;

    constructor(canvasOrContext: ManagedWebGLRenderingContext) {

        super(canvasOrContext);
    }

    public apply(source: RenderTexture[], target: RenderTarget): ShaderCompiler {
        super.apply(source, target);

        const gl = this.canvasOrContext.gl;
        let shader = AnalogGlitch.getShader(this.canvasOrContext);
        shader.bind();

        gl.activeTexture(gl.TEXTURE0);
        gl.bindTexture(gl.TEXTURE_2D, source[0]._webGLTexture);
        shader.setUniformi(ShaderCompiler.TEXTURE_SAMPLER, 0);
        shader.setUniform4f(AnalogGlitch.Fractor, this.scanLineJitterX, this.scanLineJitterY, this.verticalJumpX, this.verticalJumpY);
        shader.setUniform2f(AnalogGlitch.ColorDrift, this.colorDriftX, this.colorDriftY);
        shader.setUniformf(AnalogGlitch.Time, this.time);
        shader.setUniformf(AnalogGlitch.Shake, this.horizontalShake);

        PostProcess.bindData(this.canvasOrContext, this.manager.getDataBufferCfg(PostProcessDataManager.baseDataBufferCfgKey), shader);

        return shader;
    }

    public afterDraw(): void {
        const gl = this.canvasOrContext.gl;
        gl.bindTexture(gl.TEXTURE_2D, null);
    }

    protected static getFragmentString() {
        return `
        #define SHADER_NAME fragment:AnalogGlitch
        precision highp float;

        #define SAMPLES 10

        varying vec2 sampleCenter;

        uniform sampler2D textureSampler;
        uniform vec4 fractor;
        uniform vec2 colorDrift;
        uniform float time;
        uniform float shake;

        float nrand(float x, float y)
        {
            return fract(sin(dot(vec2(x, y), vec2(12.9898, 78.233))) * 43758.5453);
        }

        vec4 Block(vec2 source, vec4 fractor, float time, float h_shake, vec2 colorDrift) {
            // Scan line jitter
            float jitter = nrand(source.x, time) * 2.0 - 1.0;
            jitter *= step(fractor.y, abs(jitter)) * fractor.x;

            // Vertical jump
            float jump = mix(source.y, fract(source.y + fractor.w), fractor.z);

            // Horizontal shake
            float shake = (nrand(time, 2.0) - 0.5) * h_shake;

            // Color drift
            float drift = sin(jump + colorDrift.y) * colorDrift.x;

            vec4 src1 = texture2D(textureSampler, fract(vec2(source.x + jitter + shake, jump)));
            vec4 src2 = texture2D(textureSampler, fract(vec2(source.x + jitter + shake + drift, jump)));

            return vec4(src1.r, src2.g, src1.b, 1.0);
        }

        void main(void)
        {
            vec2 uv = sampleCenter;

            gl_FragColor = Block(uv, fractor, time, shake, colorDrift);
        }
        `;
    }
}