import { Copy } from "./copy";
import { DataBufferCfg } from "./data_buffer";
import { PostProcessDataManager } from "./data_manager";
import { DualBlurDown, DualBlurUp } from "./dual_blur";
import { IBloom, IBloomCombine, IBloomThreshold, KeyBloom, KeyBloomCombine, KeyBloomThreshold } from "../interface/postprocess";
import { ICopy, KeyCopy } from "../interface/postprocess";
import { IDualBlurDown, IDualBlurUp, KeyDualBlurDown, KeyDualBlurUp } from "../interface/postprocess";
import { PostProcess } from "./postprocess";
import { EFilter, RenderTarget, RenderTargetFactory, RenderTexture } from "./render_target";
import { ShaderCompiler } from "./shader_compiler";
import { ManagedWebGLRenderingContext } from "./webgl";

export class BloomThreshold extends PostProcess {
    public static shaderName = KeyBloomThreshold;
    public static getShader(context: ManagedWebGLRenderingContext) {
        let manager = PostProcessDataManager.get(context);
        let result = manager.getShaderCompiler(BloomThreshold.shaderName);
        if (!result) {
            result = new ShaderCompiler(context, PostProcess.getVertexString(), getThresholdFragmentString());
            manager.setShaderCompiler(BloomThreshold.shaderName, result);
        }

        return result;
    }

    public apply(param: IBloomThreshold, sources: RenderTexture[], target: RenderTarget, ctx: ManagedWebGLRenderingContext) {
        super.apply(param, sources, target, ctx);

        let threshold_x = Math.pow(param.threshold, 2.2);
        let threshold_y = threshold_x * param.thresholdKnee;
        let threshold_z = 2 * threshold_y;
        let threshold_w = 0.25 / (threshold_y + 0.00001);
        threshold_y -= threshold_x;

        const gl: WebGLRenderingContext = this.canvasOrContext.gl;
        // gl.clear(gl.DEPTH_BUFFER_BIT | gl.STENCIL_BUFFER_BIT);

        let shader = BloomThreshold.getShader(this.canvasOrContext);
        shader.bind();

        gl.activeTexture(gl.TEXTURE0);
        gl.bindTexture(gl.TEXTURE_2D, sources[0]);
        shader.setUniformi(ShaderCompiler.TEXTURE_SAMPLER, 0);
        shader.setUniform4f("_BloomThreshold", threshold_x, threshold_y, threshold_z, threshold_w);

        let dataBuffer = this.manager.getDataBufferCfg(PostProcessDataManager.baseDataBufferCfgKey);
        PostProcess.bindData(this.canvasOrContext, dataBuffer, shader);
        PostProcess.drawData(this.canvasOrContext, dataBuffer);
        this.afterDraw();
    }
}

export class BloomCombine extends PostProcess {
    public static shaderName = KeyBloomCombine;
    public static getShader(context: ManagedWebGLRenderingContext) {
        let manager = PostProcessDataManager.get(context);
        let result = manager.getShaderCompiler(BloomCombine.shaderName);
        if (!result) {
            result = new ShaderCompiler(context, PostProcess.getVertexString(), getFragmentCombineString());
            manager.setShaderCompiler(BloomCombine.shaderName, result);
        }

        return result;
    }

    public apply(param: IBloomCombine, sources: RenderTexture[], target: RenderTarget, ctx: ManagedWebGLRenderingContext) {
        super.apply(param, sources, target, ctx);

        const gl: WebGLRenderingContext = this.canvasOrContext.gl;

        let shader = BloomCombine.getShader(this.canvasOrContext);
        shader.bind();

        shader.setUniformf("_BloomIntensity", param.intensity);
        gl.activeTexture(gl.TEXTURE0);
        gl.bindTexture(gl.TEXTURE_2D, sources[0]);
        shader.setUniformi(ShaderCompiler.TEXTURE_SAMPLER, 0);

        // if (!this.changeSource) {
            gl.activeTexture(gl[`TEXTURE1`]);
            gl.bindTexture(gl.TEXTURE_2D, sources[1]._webGLTexture);
            shader.setUniformi(`textureSampler${0 + 1}`, 0 + 1);
        // } else {
        //     gl.enable(gl.BLEND);
        //     gl.blendFuncSeparate(gl.SRC_COLOR, gl.ONE, gl.ONE, gl.ONE_MINUS_SRC_ALPHA);
        // }

        let dataBuffer = this.manager.getDataBufferCfg(PostProcessDataManager.baseDataBufferCfgKey);
        PostProcess.bindData(this.canvasOrContext, dataBuffer, shader);
        PostProcess.drawData(this.canvasOrContext, dataBuffer);
        this.afterDraw();
    }
}


export class Bloom extends PostProcess {
    public static readonly maxBlurRenderIteration = 4;

    public static shaderNameBlur = KeyBloom;

    public readonly blurProcessIter: number;
    
    private _threshold: BloomThreshold;
    private _down: DualBlurDown;
    private _up: DualBlurUp;
    private _copy: Copy;

    constructor() {

        super();

        this._threshold = new BloomThreshold();
        this._down = new DualBlurDown();
        this._up = new DualBlurUp();
        this._copy = new Copy();

        this.blurProcessIter = Bloom.maxBlurRenderIteration;
    }

    /**
     * 应用
     * @param sources [0] 为源纹理, [1] 为 renderLowRT() 获得的RT
     * @param target 目标RT
     */
    public apply(param: IBloom, sources: RenderTexture[], target: RenderTarget, ctx: ManagedWebGLRenderingContext) {
        this.canvasOrContext = ctx;
        const source = sources[0];

        let w = RenderTargetFactory.nearPower2(source.width);
        let h = RenderTargetFactory.nearPower2(source.height);

        let _lowRTs = [];
        let temp = 2;
        let thresholdProcessTarget = RenderTargetFactory.createRenderTarget(w / temp, h / temp, false, EFilter.LINEAR, EFilter.LINEAR, undefined, undefined, false);
        _lowRTs.push(thresholdProcessTarget);

        let iteration = Math.min(Bloom.maxBlurRenderIteration, param.useBlurIteration);
        this._threshold.apply(<IBloomThreshold>{ etype: KeyBloomThreshold, threshold: param.threshold, thresholdKnee: param.thresholdKnee}, [source], thresholdProcessTarget, ctx);

        let itered = 0;
        let downSource = thresholdProcessTarget;
        while (w > temp && itered < iteration) {
            itered ++;
            temp *= 2;

            let rt = RenderTargetFactory.createRenderTarget(w / temp, h / temp, false, EFilter.LINEAR, EFilter.LINEAR, undefined, undefined, false);
            this._down.apply(<IDualBlurDown>{ etype: KeyDualBlurDown, blurRadius: param.blurRadius, intensity: 1. }, [downSource], rt, ctx);

            downSource = rt;
            _lowRTs.push(rt);
        }

        for (; itered > 0; itered--) {
            let source = _lowRTs[itered];
            let target = _lowRTs[itered - 1];

            this._up.apply(<IDualBlurUp>{ etype: KeyDualBlurUp, blurRadius: param.blurRadius, intensity: param.intensity }, [source], target, ctx);
        }

        this._copy.apply(<ICopy>{ etype: KeyCopy, intensity: 1. }, [source], target, ctx);
        this._up.apply(<IDualBlurUp>{ etype: KeyDualBlurUp, blurRadius: param.blurRadius, intensity: param.intensity }, [thresholdProcessTarget], target, ctx);
    }

    public performance(param: IBloom, width: number, height: number) {
        return performance(width, height, Math.min(this.blurProcessIter, param.useBlurIteration), param.useBlurIteration);
    }
}

// https://www.shadertoy.com/view/4stSRM
function getFragmentCombineString(changeSource: boolean = false) {
    if (!changeSource) {
        return `
        #define SHADER_NAME fragment:bloom
        precision highp float;

        varying vec2 sampleCenter;

        uniform sampler2D textureSampler;
        uniform sampler2D textureSampler1;

        uniform float _BloomIntensity;

        vec4 getTexel( vec2 p , sampler2D myTex )
        {
            return texture2D( myTex, p );
        }

        void main(void)
        {
            vec3 highRes = texture2D(textureSampler, sampleCenter).rgb;
            vec3 lowRes = getTexel(sampleCenter, textureSampler1).rgb;
            gl_FragColor = vec4(lowRes * _BloomIntensity + highRes, 1.0);
        }
        `;
    } else {
        return `
        #define SHADER_NAME fragment:bloom
        precision highp float;

        varying vec2 sampleCenter;

        uniform sampler2D textureSampler;

        uniform float _BloomIntensity;

        vec4 getTexel( vec2 p , sampler2D myTex )
        {
            return texture2D( myTex, p );
        }

        void main(void)
        {
            vec3 lowRes = getTexel(sampleCenter, textureSampler).rgb;
            gl_FragColor = vec4(lowRes * _BloomIntensity, 1.0);
        }
        `;
    }
}

function getBlurFragmentString(iteration: number = 16) {
    return `
    #define SHADER_NAME fragment:bloomBokehblur
    precision highp float;

    #define GLODEN_COS -0.7373688782616119
    #define GLODEN_SIN 0.675490294061441

    varying vec2 sampleCenter;

    uniform sampler2D textureSampler;

    uniform vec2 offset;

    const int ITERATION = ${iteration};

    vec4 BokehBlur(sampler2D tex, vec2 uv) {
        mat2 rot = mat2(GLODEN_COS, GLODEN_SIN, -GLODEN_SIN, GLODEN_COS);

        vec4 accumulator = vec4(0.0);
        vec4 divisor = vec4(0.0);

        float r = 1.0;
        vec2 angle = offset;

        for (int j = 0; j < ITERATION; j++)
        {
            r += 1.0 / r;
            angle = rot * angle;

            uv = sampleCenter + vec2(r - 1.0) * angle;
            vec4 bokeh = texture2D(tex, uv);

            accumulator += bokeh * bokeh;
            divisor += bokeh;
        }

        return accumulator / divisor;
    }

    void main(void)
    {
        gl_FragColor = BokehBlur(textureSampler, sampleCenter);
    }
    `;
}

function getThresholdFragmentString() {
    return `
    #define SHADER_NAME fragment:bloomThreshold
    precision highp float;

    varying vec2 sampleCenter;

    uniform sampler2D textureSampler;

    uniform vec4 _BloomThreshold;

    vec4 getTexel( vec2 p , sampler2D myTex )
    {
        return texture2D( myTex, p );
    }

    vec3 ApplyBloomThreshold (vec3 color) {
        float brightness = max(color.r, max(color.g, color.b));
        float soft = brightness + _BloomThreshold.y;
        soft = clamp(soft, 0.0, _BloomThreshold.z);
        soft = soft * soft * _BloomThreshold.w;
        float contribution = max(soft, brightness - _BloomThreshold.x);
        contribution /= max(brightness, 0.00001);
        return color * contribution;
    }

    void main(void)
    {
        vec3 color = ApplyBloomThreshold(getTexel(sampleCenter, textureSampler).rgb);
        gl_FragColor = vec4(color, 1.0);
    }
    `;
}

function performance(width: number, height: number, iteration: number, samplerIteration: number) {

    let tWidth = RenderTargetFactory.nearPower2(width);
    let tHeight = RenderTargetFactory.nearPower2(height);

    let temp = 2;
    let pixelSamplerTimes = 0;
    pixelSamplerTimes += tWidth / temp * tHeight / temp;

    temp *= 2;
    for (let i = 0; i < iteration; i++) {
        pixelSamplerTimes += tWidth / temp * tHeight / temp * samplerIteration;

        temp *= 2;
    }

    pixelSamplerTimes += width * height * 2;

    return pixelSamplerTimes;
}