import { DataBufferCfg } from "./data_buffer";
import { PostProcessDataManager } from "./data_manager";
import { DualBlur } from "./dual_blur";
import { PostProcess } from "./postprocess";
import { EFilter, RenderTarget, RenderTargetFactory, RenderTexture } from "./render_target";
import { ShaderCompiler } from "./shader_compiler";
import { ManagedWebGLRenderingContext } from "./webgl";

export class BloomDual extends PostProcess {
    public static readonly maxBlurRenderIteration = 4;

    public static shaderNameThreshold = `BloomDualThreshold`;
    public static getShaderThreshold(context: ManagedWebGLRenderingContext) {
        let manager = PostProcessDataManager.get(context);
        let result = manager.getShaderCompiler(BloomDual.shaderNameThreshold);
        if (!result) {
            result = new ShaderCompiler(context, PostProcess.getVertexString(), getThresholdFragmentString());
            manager.setShaderCompiler(BloomDual.shaderNameThreshold, result);
        }

        return result;
    }

    public static shaderName = `BloomDualCombine`;
    public static getShader(context: ManagedWebGLRenderingContext) {
        let manager = PostProcessDataManager.get(context);
        let name = `${BloomDual.shaderName}`;
        let result = manager.getShaderCompiler(name);
        if (!result) {
            result = new ShaderCompiler(context, PostProcess.getVertexString(), BloomDual.getFragmentString());
            manager.setShaderCompiler(name, result);
        }

        return result;
    }

    public thresholdKnee: number = 0.0;
    public threshold: number = 0.0;
    public intensity: number = 0.0;
    public blurRadius: number = 2;
    public useBlurIteration: number = 1;

    private iteration: number = 12;

    public isEnabled: boolean = true;

    public readonly blurProcessIter: number;

    private _lowRT0: RenderTarget;
    private _lowRT1: RenderTarget;

    private dual: DualBlur;

    constructor(
        canvasOrContext: ManagedWebGLRenderingContext,
        private changeSource: boolean = false
    ) {

        super(canvasOrContext);

        this.blurProcessIter = BloomDual.maxBlurRenderIteration;

        this.dual = new DualBlur(this.canvasOrContext);
    }

    /**
     * 应用
     * @param sources [0] 为源纹理, [1] 为 renderLowRT() 获得的RT
     * @param target 目标RT
     */
    public apply(sources: RenderTexture[], target: RenderTarget): ShaderCompiler {

        const gl: WebGLRenderingContext = this.canvasOrContext.gl;
        let shader = BloomDual.getShader(this.canvasOrContext);
        shader.bind();

        gl.enable(gl.BLEND);
        gl.blendFuncSeparate(gl.ONE, gl.ONE, gl.ONE, gl.ONE_MINUS_SRC_ALPHA);

        gl.activeTexture(gl.TEXTURE0);
        gl.bindTexture(gl.TEXTURE_2D, sources[0]._webGLTexture);
        shader.setUniformi(ShaderCompiler.TEXTURE_SAMPLER, 0);

        shader.setUniformf("_BloomIntensity", this.intensity);

        PostProcess.bindData(this.canvasOrContext, this.manager.getDataBufferCfg(PostProcessDataManager.baseDataBufferCfgKey), shader);

        return shader;
    }

    public renderLowRT(source: RenderTexture) {

        let w = RenderTargetFactory.nearPower2(source.width);
        let h = RenderTargetFactory.nearPower2(source.height);

        let manager = PostProcessDataManager.get(this.canvasOrContext);
        let dataBuffer = manager.getDataBufferCfg(PostProcessDataManager.baseDataBufferCfgKey);

        let temp = 2;
        let thresholdProcessTarget = RenderTargetFactory.createRenderTarget(
            w / temp,
            h / temp,
            false,
            EFilter.LINEAR,
            EFilter.LINEAR,
            undefined, undefined,
            false
        );

        temp *= 2;
        this._lowRT0 = RenderTargetFactory.createRenderTarget(
            w / temp,
            h / temp,
            false,
            EFilter.LINEAR,
            EFilter.LINEAR,
            undefined, undefined,
            false
        );
        temp *= 2;
        this._lowRT1 = RenderTargetFactory.createRenderTarget(
            w / temp,
            h / temp,
            false,
            EFilter.LINEAR,
            EFilter.LINEAR,
            undefined, undefined,
            false
        );

        this._renderThreshold(source._webGLTexture, thresholdProcessTarget, dataBuffer);

        const gl: WebGLRenderingContext = this.canvasOrContext.gl;

        this.dual.downIntensity = 1;
        this.dual.upIntensity = this.intensity;
        this._renderBlurs(dataBuffer, thresholdProcessTarget, [this._lowRT0, this._lowRT1]);

        gl.enable(gl.BLEND);
        gl.blendFuncSeparate(gl.ONE, gl.ONE, gl.ONE, gl.ONE_MINUS_SRC_ALPHA);

        gl.bindFramebuffer(gl.FRAMEBUFFER, this._lowRT0.framebuffer);
        gl.viewport(0, 0, this._lowRT0.width, this._lowRT0.height);
        this.dual.apply([this._lowRT1], this._lowRT0, true);

        PostProcess.drawData(this.canvasOrContext, dataBuffer);
        RenderTargetFactory.unuseRenderTarget(thresholdProcessTarget);
        RenderTargetFactory.unuseRenderTarget(this._lowRT1);
        this._lowRT1 = undefined;

        return this._lowRT0;
    }

    public afterDraw(): void {
        RenderTargetFactory.unuseRenderTarget(this._lowRT0);
        this._lowRT0 = undefined;

        const gl: WebGLRenderingContext = this.canvasOrContext.gl;
        gl.bindTexture(gl.TEXTURE_2D, null);
    }

    private _renderThreshold(sourceTx: WebGLTexture, target: RenderTarget, dataBuffer: DataBufferCfg) {
        let threshold_x = Math.pow(this.threshold, 2.2);
        let threshold_y = threshold_x * this.thresholdKnee;
        let threshold_z = 2 * threshold_y;
        let threshold_w = 0.25 / (threshold_y + 0.00001);
        threshold_y -= threshold_x;

        const gl: WebGLRenderingContext = this.canvasOrContext.gl;
        gl.bindFramebuffer(gl.FRAMEBUFFER, target.framebuffer);
        gl.viewport(0, 0, target.width, target.height);
        // gl.clear(gl.DEPTH_BUFFER_BIT | gl.STENCIL_BUFFER_BIT);

        let shader = BloomDual.getShaderThreshold(this.canvasOrContext);
        shader.bind();

        gl.activeTexture(gl.TEXTURE0);
        gl.bindTexture(gl.TEXTURE_2D, sourceTx);
        shader.setUniformi(ShaderCompiler.TEXTURE_SAMPLER, 0);
        shader.setUniform4f("_BloomThreshold", threshold_x, threshold_y, threshold_z, threshold_w);

        PostProcess.bindData(this.canvasOrContext, this.manager.getDataBufferCfg(PostProcessDataManager.baseDataBufferCfgKey), shader);

        PostProcess.drawData(this.canvasOrContext, dataBuffer);
    }

    private _renderBlurs(dataBuffer: DataBufferCfg, source: RenderTarget, targets: RenderTarget[]) {

        let sourceRT = source;
        let targetRT = targets[0];
        // let useBlurIteration = Math.min(this.useBlurIteration, this.blurProcessIter);
        let useBlurIteration = Math.min(2, this.blurProcessIter);
        let lastBlurProcessTarget: RenderTarget;
        for (let i = 0; i < useBlurIteration; i++) {
            if (i > 0) {
                sourceRT = targets[i - 1];
                targetRT = targets[i];
            }
            this._renderBlur(sourceRT, targetRT, dataBuffer);
            lastBlurProcessTarget = targetRT;
        }
    }
    private _renderBlur(source: RenderTexture, target: RenderTarget, dataBuffer: DataBufferCfg) {
        const gl: WebGLRenderingContext = this.canvasOrContext.gl;
        gl.bindFramebuffer(gl.FRAMEBUFFER, target.framebuffer);
        gl.viewport(0, 0, target.width, target.height);
        // gl.clear(gl.DEPTH_BUFFER_BIT | gl.STENCIL_BUFFER_BIT);
        this.dual.blurRadius = this.blurRadius;
        this.dual.blurFactor = 1;
        this.dual.apply([source], target, false);
        PostProcess.drawData(this.canvasOrContext, dataBuffer);
    }

    // https://www.shadertoy.com/view/4stSRM
    protected static getFragmentString() {
        return `
            #define SHADER_NAME fragment:bloom
            precision highp float;

            varying vec2 sampleCenter;

            uniform sampler2D textureSampler;

            uniform float _BloomIntensity;

            vec4 getTexel( vec2 p , sampler2D myTex )
            {
                return texture2D( myTex, p );
            }

            void main(void)
            {
                vec3 lowRes = getTexel(sampleCenter, textureSampler).rgb;
                gl_FragColor = vec4(lowRes * _BloomIntensity, 1.0);
            }
        `;
    }

    public performance(width: number, height: number) {
        return performance(width, height, Math.min(this.blurProcessIter, this.useBlurIteration), this.iteration);
    }
}

function getThresholdFragmentString() {
    return `
    #define SHADER_NAME fragment:bloomThreshold
    precision highp float;

    varying vec2 sampleCenter;

    uniform sampler2D textureSampler;

    uniform vec4 _BloomThreshold;

    vec4 getTexel( vec2 p , sampler2D myTex )
    {
        return texture2D( myTex, p );
    }

    vec3 ApplyBloomThreshold (vec3 color) {
        float brightness = max(color.r, max(color.g, color.b));
        float soft = brightness + _BloomThreshold.y;
        soft = clamp(soft, 0.0, _BloomThreshold.z);
        soft = soft * soft * _BloomThreshold.w;
        float contribution = max(soft, brightness - _BloomThreshold.x);
        contribution /= max(brightness, 0.00001);
        return color * contribution;
    }

    void main(void)
    {
        vec3 color = ApplyBloomThreshold(getTexel(sampleCenter, textureSampler).rgb);
        gl_FragColor = vec4(color, 1.0);
    }
    `;
}

function performance(width: number, height: number, iteration: number, samplerIteration: number) {

    let tWidth = RenderTargetFactory.nearPower2(width);
    let tHeight = RenderTargetFactory.nearPower2(height);

    let temp = 2;
    let pixelSamplerTimes = 0;
    pixelSamplerTimes += tWidth / temp * tHeight / temp;

    temp *= 2;
    for (let i = 0; i < iteration; i++) {
        pixelSamplerTimes += tWidth / temp * tHeight / temp * samplerIteration;

        temp *= 2;
    }

    pixelSamplerTimes += width * height * 2;

    return pixelSamplerTimes;
}