import { Blur, EIrisAreaMode } from "./blur";
import { PostProcessDataManager } from "./data_manager";
import { PostProcess } from "./postprocess";
import { RenderTarget, RenderTexture } from "./render_target";
import { ShaderCompiler } from "./shader_compiler";
import { ManagedWebGLRenderingContext } from "./webgl";

export class BokehBlur extends Blur {

    public static shaderName = `BokehBlur`;
    public static getShader(context: ManagedWebGLRenderingContext, iteration: number, irisAreaMode: EIrisAreaMode) {
        let manager = PostProcessDataManager.get(context);
        let shaderKey = `${BokehBlur.shaderName}_${iteration}_${irisAreaMode}`;
        let result = manager.getShaderCompiler(shaderKey);
        if (!result) {
            result = new ShaderCompiler(context, PostProcess.getVertexString(), BokehBlur.getFragmentString(iteration, irisAreaMode));
            manager.setShaderCompiler(shaderKey, result);
        }

        return result;
    }

    private static BlurRadius = 'blurRadius';
    private static Iteration = 'iteration';
    private static TextureSize = 'textureSize';

    blurRadius: number = 1;
    public iteration: number = 6;

    constructor(canvasOrContext: ManagedWebGLRenderingContext) {

        super(canvasOrContext);
    }

    public apply(source: RenderTexture[], target: RenderTarget): ShaderCompiler {
        super.apply(source, target);

        const gl: WebGLRenderingContext = this.canvasOrContext.gl;
        let shader = BokehBlur.getShader(this.canvasOrContext, this.iteration, this.irisAreaMode);
        shader.bind();

        gl.activeTexture(gl.TEXTURE0);
        gl.bindTexture(gl.TEXTURE_2D, source[0]._webGLTexture);
        shader.setUniformi(ShaderCompiler.TEXTURE_SAMPLER, 0);
        shader.setUniformf("blurRadius", this.blurRadius / Math.max(source[0].width, source[0].height));

        this.applyIris(shader);

        PostProcess.bindData(this.canvasOrContext, this.manager.getDataBufferCfg(PostProcessDataManager.baseDataBufferCfgKey), shader);

        return shader;
    }

    public afterDraw() {
        const gl: WebGLRenderingContext = this.canvasOrContext.gl;
        gl.bindTexture(gl.TEXTURE_2D, null);
    }

    // https://www.shadertoy.com/view/4stSRM
    protected static getFragmentString(iteration: number = 6, irisAreaMode: EIrisAreaMode = EIrisAreaMode.Circle) {
        return `
        #define SHADER_NAME fragment:bokehblur
        precision highp float;

        #define GLODEN_COS -0.7373688782616119
        #define GLODEN_SIN 0.675490294061441

        #define SAMPLES 10

        varying vec2 sampleCenter;

        uniform sampler2D textureSampler;
        uniform float blurRadius;
        ${Blur.IrisFragmentDefined()}

        const int ITERATION = ${iteration};

        ${Blur.GetIrisMaskFragmentDefined(irisAreaMode)}

        vec4 BokehBlur(sampler2D tex, vec2 uv) {
            float iris = ${Blur.GetIrisMaskFragmentCall(irisAreaMode, 'sampleCenter')};
            if (iris < 0.0001) return texture2D(tex, uv.xy);

            mat2 rot = mat2(GLODEN_COS, GLODEN_SIN, -GLODEN_SIN, GLODEN_COS);

            vec4 accumulator = vec4(0.0);
            vec4 divisor = vec4(0.0);

            float r = 1.0;
            vec2 angle = vec2(0.0, blurRadius) * vec2(iris);

            for (int j = 0; j < ITERATION; j++)
            {
                r += 1.0 / r;
                angle = rot * angle;

                uv = sampleCenter + vec2(r - 1.0) * angle;
                vec4 bokeh = texture2D(tex, uv);

                accumulator += bokeh * bokeh;
                divisor += bokeh;
            }

            return accumulator / divisor;
        }

        void main(void)
        {
            gl_FragColor = BokehBlur(textureSampler, sampleCenter);
        }
        `;
    }
}