
import { PostProcessDataManager } from "./data_manager";
import { IPostprocessTarget } from "../interface/postprocess";
import { ICopy } from "../interface/postprocess";
import { PostProcess } from "./postprocess";
import { RenderTarget, RenderTexture } from "./render_target";
import { ShaderCompiler } from "./shader_compiler";
import { ManagedWebGLRenderingContext } from "./webgl";

export interface CopyRenderParam {
    context: ManagedWebGLRenderingContext;
    source: RenderTarget;
    traget?: RenderTarget;
    viewPort: [number, number, number, number];
    clearColor?: [number, number, number, number];
    scissor?: boolean;
    depth?: boolean;
    stencil?: boolean;
    alphaMode?: 0 | 1 | 2 | 7;
}

export interface CopyOutParam {
    renderTarget?: RenderTarget;
}

export class Copy extends PostProcess {
    public static shaderName = `Copy`;
    public static getShader(context: ManagedWebGLRenderingContext) {
        let manager = PostProcessDataManager.get(context);
        let shaderKey = `${Copy.shaderName}`;
        let result = manager.getShaderCompiler(shaderKey);
        if (!result) {
            result = new ShaderCompiler(context, PostProcess.getVertexString(), Copy.getFragmentString());
            manager.setShaderCompiler(shaderKey, result);
        }

        return result;
    }

    public static apply(canvasOrContext: ManagedWebGLRenderingContext, source: RenderTexture[], target: RenderTarget, alpha: number = 1) {
        let gl = canvasOrContext.gl;
        let shader = Copy.getShader(canvasOrContext);
        shader.bind();

        gl.activeTexture(gl.TEXTURE0);
        gl.bindTexture(gl.TEXTURE_2D, source[0]._webGLTexture);
        shader.setUniformi(ShaderCompiler.TEXTURE_SAMPLER, 0);
        shader.setUniformf(`uAlpha`, alpha);

        return shader;
    }

    public apply(param: ICopy, source: RenderTexture[], target: IPostprocessTarget, ctx: ManagedWebGLRenderingContext) {
        super.apply(param, source, target, ctx);

        let gl = this.canvasOrContext.gl;
        let shader = Copy.getShader(this.canvasOrContext);
        shader.bind();

        gl.activeTexture(gl.TEXTURE0);
        gl.bindTexture(gl.TEXTURE_2D, source[0]._webGLTexture);
        shader.setUniformi(ShaderCompiler.TEXTURE_SAMPLER, 0);
        shader.setUniformf(`uAlpha`, param.alpha);

        let dataBuffer = this.manager.getDataBufferCfg(PostProcessDataManager.baseDataBufferCfgKey);
        PostProcess.bindData(this.canvasOrContext, dataBuffer, shader);
        PostProcess.drawData(this.canvasOrContext, dataBuffer);
        this.afterDraw();

        return shader;
    }

    public static getFragmentString() {
        return `
        #define SHADER_NAME fragment:copy
        precision highp float;

        uniform sampler2D textureSampler;
        uniform float uAlpha;
        varying vec2 sampleCenter;

        void main(void){
            vec4 color = texture2D(textureSampler, sampleCenter);
            color.a *= uAlpha;
            gl_FragColor = color;
        }
        `;
    }
}

function getVertexString() {
    return `
    #define SHADER_NAME vertex:copy
    precision highp float;

    attribute vec3 position;

    varying vec2 sampleCenter;
    const vec2 madd = vec2(0.5, 0.5);
    void main(void) {
        sampleCenter = (position.xy*madd+madd);
        gl_Position = vec4(position.xy, 0.0, 1.0);
    }
    `;
}