import { PostProcessDataManager } from "./data_manager";
import { PostProcess } from "./postprocess";
import { RenderTexture, RenderTarget } from "./render_target";
import { ShaderCompiler } from "./shader_compiler";
import { ManagedWebGLRenderingContext } from "./webgl";

/**
 * 均匀图块错位
 */
export class DigitalGlitch extends PostProcess {
    public static shaderName = `DigitalGlitch`;
    public static getShader(context: ManagedWebGLRenderingContext) {
        let manager = PostProcessDataManager.get(context);
        let shaderKey = `${DigitalGlitch.shaderName}`;
        let result = manager.getShaderCompiler(shaderKey);
        if (!result) {
            result = new ShaderCompiler(context, PostProcess.getVertexString(), DigitalGlitch.getFragmentString());
            manager.setShaderCompiler(shaderKey, result);
        }

        return result;
    }

    private static Fractor = 'fractor';
    private static Time = 'time';
    private static Shake = 'shake';
    private static ColorDrift = 'colorDrift';

    blockSizeX: number = 0.1;
    blockSizeY: number = 0.1;
    time: number = 0.1;
    speed: number = 0.1;

    public apply(sources: RenderTexture[], target: RenderTarget): ShaderCompiler {
        super.apply(sources, target);

        const gl = this.canvasOrContext.gl;
        let shader = DigitalGlitch.getShader(this.canvasOrContext);
        shader.bind();

        gl.activeTexture(gl.TEXTURE0);
        gl.bindTexture(gl.TEXTURE_2D, sources[0]._webGLTexture);

        shader.setUniformi(ShaderCompiler.TEXTURE_SAMPLER, 0);
        shader.setUniform4f(DigitalGlitch.Fractor, this.blockSizeX, this.blockSizeY, this.time, this.speed);

        PostProcess.bindData(this.canvasOrContext, this.manager.getDataBufferCfg(PostProcessDataManager.baseDataBufferCfgKey), shader);

        return shader;
    }

    protected static getFragmentString() {
        return `
        #define SHADER_NAME fragment:DigitalGlitch
        precision highp float;

        #define GLODEN_COS -0.7373688782616119
        #define GLODEN_SIN 0.675490294061441

        #define SAMPLES 10

        varying vec2 sampleCenter;

        uniform sampler2D textureSampler;
        uniform vec4 fractor;

        float randomNoise(vec2 seed, float time, float speed)
        {
            return fract(sin(dot(seed * floor(time * speed), vec2(17.13, 3.71))) * 43758.5453123);
        }

        float nrand(float x, float y)
        {
            return fract(sin(dot(vec2(x, y), vec2(12.9898, 78.233))) * 43758.5453123);
        }

        float nrand2(vec2 st)
        {
            return fract(sin(dot(st, vec2(12.9898, 78.233))) * 43758.5453);
        }

        float randomNoise1(float seed)
        {
            return nrand(seed, 1.0);
        }

        vec4 Block(vec2 source, vec4 fractor) {
            vec2 blockSize = fractor.xy;
            float time = fractor.z;
            float cutoff = fractor.w;
            float block = nrand2(floor(source * blockSize) + vec2(time, .0)); // randomNoise(floor(source * blockSize), time, cutoff);
		    float displaceNoise = block; //pow(block, 8.0) * pow(block, 3.0);

            displaceNoise = displaceNoise > cutoff ? displaceNoise : 0.0;

            float colordirty = randomNoise1(7.0);

            vec4 src1 = texture2D(textureSampler, fract(source * (1.0 + displaceNoise)));
            vec4 src2 = texture2D(textureSampler, fract(source + vec2(displaceNoise * 0.05 * colordirty, 0.0)));
            vec4 src3 = texture2D(textureSampler, fract(source - vec2(displaceNoise * 0.05 * colordirty, 0.0)));

            return vec4(src1.r, src2.g, src3.b, 1.0);
        }

        void main(void)
        {
            vec2 uv = sampleCenter;

            gl_FragColor = Block(uv, fractor);
        }
        `;
    }
}