
import { Blur, EIrisAreaMode } from "./blur";
import { PostProcessDataManager } from "./data_manager";
import { PostProcess } from "./postprocess";
import { RenderTexture, RenderTarget } from "./render_target";
import { ShaderCompiler } from "./shader_compiler";
import { ManagedWebGLRenderingContext } from "./webgl";

export class DirectionalBlur extends Blur {
    public static shaderName = `DirectionalBlur`;
    public static getShader(context: ManagedWebGLRenderingContext, iteration: number, irisAreaMode: EIrisAreaMode) {
        let manager = PostProcessDataManager.get(context);
        let shaderKey = `${DirectionalBlur.shaderName}_${iteration}_${irisAreaMode}`;
        let result = manager.getShaderCompiler(shaderKey);
        if (!result) {
            result = new ShaderCompiler(context, PostProcess.getVertexString(), DirectionalBlur.getFragmentString(iteration, irisAreaMode));
            manager.setShaderCompiler(shaderKey, result);
        }

        return result;
    }

    private static BlurRadius = 'blurRadius';
    private static RadialCenter = 'radialCenter';
    private static Direction = 'direction';
    /**
     * 模糊半径 - 像素值
     */
    blurRadius: number = 10;
    /**
     * 径向角度
     */
    private _angle: number = 0;
    public get angle() {
        return this._angle;
    }
    public set angle(v: number) {
        if (v == undefined || v === null) {
            v = 0;
        }

        v = v / 180 * Math.PI;

        if (this._angle !== v) {
            this._angle = v;
        }
    }
    private direction: [number, number] = [0, 0];

    public iteration = 6;

    public apply(sources: RenderTexture[], target: RenderTarget): ShaderCompiler {
        super.apply(sources, target);

        const gl: WebGLRenderingContext = this.canvasOrContext.gl;
        let shader = DirectionalBlur.getShader(this.canvasOrContext, this.iteration, this.irisAreaMode);
        shader.bind();

        gl.activeTexture(gl.TEXTURE0);
        gl.bindTexture(gl.TEXTURE_2D, sources[0]._webGLTexture);

        shader.setUniformi(ShaderCompiler.TEXTURE_SAMPLER, 0);

        const radius = this.blurRadius / sources[0].width / this.iteration;
        this.direction[0] = Math.sin(this._angle) * radius ;
        this.direction[1] = Math.cos(this._angle) * radius ;
        shader.setUniform2f(DirectionalBlur.Direction, this.direction[0], this.direction[1]);
        this.applyIris(shader);

        PostProcess.bindData(this.canvasOrContext, this.manager.getDataBufferCfg(PostProcessDataManager.baseDataBufferCfgKey), shader);

        return shader;
    }

    protected static getFragmentString(iteration: number = 6, irisAreaMode: EIrisAreaMode = EIrisAreaMode.Circle) {
        return `
        #define SHADER_NAME fragment:directinalblur
        precision highp float;

        varying vec2 sampleCenter;

        uniform sampler2D textureSampler;
        uniform vec2 direction;
        ${Blur.IrisFragmentDefined()}

        ${Blur.GetIrisMaskFragmentDefined(irisAreaMode)}

        vec4 RadialBlur(sampler2D tex, vec2 uv) {

            float iris = ${Blur.GetIrisMaskFragmentCall(irisAreaMode, 'uv')};
            if (iris < 0.0001) return texture2D(tex, uv.xy);

            vec4 acumulateColor = vec4(0.0);

            for (int j = -${iteration}; j < ${iteration}; j++)
            {
                acumulateColor += texture2D(tex, uv.xy - iris * direction * float(j));
            }

            return acumulateColor / (${iteration}.0 * 2.0);
        }

        void main(void) {
            gl_FragColor = RadialBlur(textureSampler, sampleCenter);
        }
        `;
    }
}