
import { PostProcessDataManager } from "./data_manager";
import { IDualBlur, IDualBlurDown, IDualBlurUp, KeyDualBlurDown, KeyDualBlurUp } from "../interface/postprocess";
import { PostProcess } from "./postprocess";
import { EFilter, RenderTarget, RenderTargetFactory, RenderTexture } from "./render_target";
import { ShaderCompiler } from "./shader_compiler";
import { ManagedWebGLRenderingContext } from "./webgl";

export class DualBlurDown extends PostProcess {
    public static getShader(context: ManagedWebGLRenderingContext) {
        let manager = PostProcessDataManager.get(context);
        let shaderKey = KeyDualBlurDown;
        let result = manager.getShaderCompiler(shaderKey);
        if (!result) {
            result = new ShaderCompiler(context, getVertexString(), getFragmentDownString());
            manager.setShaderCompiler(shaderKey, result);
        }

        return result;
    }
    
    public apply(param: IDualBlurDown, sources: RenderTexture[], target: RenderTarget, ctx: ManagedWebGLRenderingContext) {
        super.apply(param, sources, target, ctx);

        const gl: WebGLRenderingContext = this.canvasOrContext.gl;
        let shader = DualBlurDown.getShader(this.canvasOrContext);
        shader.bind();

        gl.activeTexture(gl.TEXTURE0);
        gl.bindTexture(gl.TEXTURE_2D, sources[0]._webGLTexture);
        shader.setUniformi(ShaderCompiler.TEXTURE_SAMPLER, 0);
        shader.setUniformf("intensity", param.intensity);
        shader.setUniform2f("offset", param.blurRadius / sources[0].width, param.blurRadius / sources[0].height);

        let dataBuffer = this.manager.getDataBufferCfg(PostProcessDataManager.baseDataBufferCfgKey);
        PostProcess.bindData(this.canvasOrContext, dataBuffer, shader);
        PostProcess.drawData(this.canvasOrContext, dataBuffer);
        this.afterDraw();
    }
}

export class DualBlurUp extends PostProcess {
    public static getShader(context: ManagedWebGLRenderingContext) {
        let manager = PostProcessDataManager.get(context);
        let shaderKey = KeyDualBlurUp;
        let result = manager.getShaderCompiler(shaderKey);
        if (!result) {
            result = new ShaderCompiler(context, getVertexString(), getFragmentUpString());
            manager.setShaderCompiler(shaderKey, result);
        }

        return result;
    }
    
    public apply(param: IDualBlurUp, sources: RenderTexture[], target: RenderTarget, ctx: ManagedWebGLRenderingContext) {
        super.apply(param, sources, target, ctx);

        const gl: WebGLRenderingContext = this.canvasOrContext.gl;
        let shader = DualBlurUp.getShader(this.canvasOrContext);
        shader.bind();

        gl.activeTexture(gl.TEXTURE0);
        gl.bindTexture(gl.TEXTURE_2D, sources[0]._webGLTexture);
        shader.setUniformi(ShaderCompiler.TEXTURE_SAMPLER, 0);
        shader.setUniformf("intensity", param.intensity);
        shader.setUniform2f("offset", param.blurRadius / sources[0].width, param.blurRadius / sources[0].height);

        let dataBuffer = this.manager.getDataBufferCfg(PostProcessDataManager.baseDataBufferCfgKey);
        PostProcess.bindData(this.canvasOrContext, dataBuffer, shader);
        PostProcess.drawData(this.canvasOrContext, dataBuffer);
        this.afterDraw();
    }
}

export class DualBlur extends PostProcess {
    public static readonly maxBlurRenderIteration = 4;

    private _down: DualBlurDown;
    private _up: DualBlurUp;

    constructor() {
        super();
        this._down = new DualBlurDown();
        this._up = new DualBlurUp();
    }

    public apply(param: IDualBlur, sources: RenderTexture[], target: RenderTarget, ctx: ManagedWebGLRenderingContext) {
        let source = sources[0];
        let w = RenderTargetFactory.nearPower2(source.width);
        let h = RenderTargetFactory.nearPower2(source.height);

        let temp = 2;
        let _lowRT0 = RenderTargetFactory.createRenderTarget(w / temp, h / temp, false, EFilter.LINEAR, EFilter.LINEAR, undefined, undefined, false);

        let downparam = <IDualBlurDown>{ etype: KeyDualBlurDown, blurRadius: param.blurRadius, intensity: 1 };
        this._down.apply(downparam, [source], _lowRT0, ctx);
        
        let upparam = <IDualBlurUp>{ etype: KeyDualBlurUp, blurRadius: param.blurRadius, intensity: param.intensity };
        this._up.apply(upparam, [_lowRT0], target, ctx);

        RenderTargetFactory.unuseRenderTarget(_lowRT0);
    }
}

function getVertexString() {
    return `
    #define SHADER_NAME vertex:DualBlur
    precision highp float;

    attribute vec3 position;

    const vec2 madd = vec2(0.5, 0.5);

    uniform vec2 offset;

    varying vec4 uv01;
    varying vec4 uv23;
    varying vec4 uv45;
    varying vec4 uv67;

    void main(void) {
        vec2 uv = (position.xy*madd+madd);

        uv01 = vec4(uv + vec2( offset.x,  offset.y), uv + vec2(-offset.x,  offset.y));
        uv23 = vec4(uv + vec2( offset.x, -offset.y), uv + vec2(-offset.x, -offset.y));
        uv45 = vec4(uv + vec2( offset.x * 2.,  0.),  uv + vec2(-offset.x * 2.,  0.));
        uv67 = vec4(uv + vec2( 0.,  offset.y * 2.),  uv + vec2(0.,  offset.y * 2.));

        gl_Position = vec4(position.xy, 0.0, 1.0);
    }
    `;
}

// https://zhuanlan.zhihu.com/p/125744132
function getFragmentDownString() {
    return `
    #define SHADER_NAME fragment:DualBlurDown
    precision highp float;

    uniform sampler2D textureSampler;
    uniform float intensity;

    varying vec4 uv01;
    varying vec4 uv23;
    varying vec4 uv45;
    varying vec4 uv67;

    void main(void)
    {
        vec4 color =  texture2D( textureSampler, uv01.xy )
                    + texture2D( textureSampler, uv01.zw )
                    + texture2D( textureSampler, uv23.xy )
                    + texture2D( textureSampler, uv23.zw );
        color.rgb *= intensity;
        gl_FragColor = color * 0.25;
    }
    `;
}
// https://zhuanlan.zhihu.com/p/125744132
function getFragmentUpString() {
    return `
    #define SHADER_NAME fragment:DualBlurUp
    precision highp float;

    uniform sampler2D textureSampler;
    uniform float intensity;

    varying vec4 uv01;
    varying vec4 uv23;
    varying vec4 uv45;
    varying vec4 uv67;

    void main(void)
    {
        vec4 color =  texture2D( textureSampler, uv01.xy ) * 2.
                    + texture2D( textureSampler, uv01.zw ) * 2.
                    + texture2D( textureSampler, uv23.xy ) * 2.
                    + texture2D( textureSampler, uv23.zw ) * 2.
                    + texture2D( textureSampler, uv45.xy )
                    + texture2D( textureSampler, uv45.zw )
                    + texture2D( textureSampler, uv67.xy )
                    + texture2D( textureSampler, uv67.zw );
        color.rgb *= intensity;
        gl_FragColor = color * 0.0833333; // 1/12
    }
    `;
}