
import { Blur, EIrisAreaMode } from "./blur";
import { PostProcessDataManager } from "./data_manager";
import { PostProcess } from "./postprocess";
import { RenderTexture, RenderTarget } from "./render_target";
import { ShaderCompiler } from "./shader_compiler";
import { ManagedWebGLRenderingContext } from "./webgl";

export class GaussianBlur extends Blur {
    public static shaderName = `GaussianBlur`;
    public static getShader(context: ManagedWebGLRenderingContext, irisAreaMode: EIrisAreaMode) {
        let manager = PostProcessDataManager.get(context);
        let shaderKey = `${GaussianBlur.shaderName}_${irisAreaMode}`;
        let result = manager.getShaderCompiler(shaderKey);
        if (!result) {
            result = new ShaderCompiler(context, GaussianBlur.getVertexString(), GaussianBlur.getFragmentString(irisAreaMode));
            manager.setShaderCompiler(shaderKey, result);
        }

        return result;
    }

    direction: [number, number] = [0, 0];
    kernel: number;

    public apply(sources: RenderTexture[], target: RenderTarget): ShaderCompiler {
        super.apply(sources, target);

        const gl = this.canvasOrContext.gl;
        let shader = GaussianBlur.getShader(this.canvasOrContext, this.irisAreaMode);
        shader.bind();

        gl.activeTexture(gl.TEXTURE0);
        gl.bindTexture(gl.TEXTURE_2D, sources[0]._webGLTexture);

        shader.setUniformi(ShaderCompiler.TEXTURE_SAMPLER, 0);
        shader.setUniform2f(ShaderCompiler.DELTA, 1 / (sources[0].width) * this.direction[0], 1 / (sources[0].height) * this.direction[1]);

        this.applyIris(shader);

        PostProcess.bindData(this.canvasOrContext, this.manager.getDataBufferCfg(PostProcessDataManager.baseDataBufferCfgKey), shader);

        return shader;
    }

    protected static getVertexString() {
        return `
        #define KERNEL_OFFSET0 -15.36704337
        #define KERNEL_WEIGHT0 0.00226838
        #define KERNEL_OFFSET1 15.36704337
        #define KERNEL_WEIGHT1 0.00226838
        #define KERNEL_OFFSET2 -13.38352585
        #define KERNEL_WEIGHT2 0.00623279
        #define KERNEL_OFFSET3 13.38352585
        #define KERNEL_WEIGHT3 0.00623279
        #define KERNEL_OFFSET4 -11.40028041
        #define KERNEL_WEIGHT4 0.01489653
        #define KERNEL_OFFSET5 11.40028041
        #define KERNEL_WEIGHT5 0.01489653
        #define KERNEL_OFFSET6 -9.41727148
        #define KERNEL_WEIGHT6 0.03096924
        #define KERNEL_OFFSET7 9.41727148
        #define KERNEL_WEIGHT7 0.03096924
        #define KERNEL_OFFSET8 -7.4344613
        #define KERNEL_WEIGHT8 0.05600478
        #define KERNEL_OFFSET9 7.4344613
        #define KERNEL_WEIGHT9 0.05600478
        #define KERNEL_OFFSET10 -5.45181021
        #define KERNEL_WEIGHT10 0.08809971
        #define KERNEL_OFFSET11 5.45181021
        #define KERNEL_WEIGHT11 0.08809971
        #define KERNEL_OFFSET12 -3.46927703
        #define KERNEL_WEIGHT12 0.12055435
        #define KERNEL_OFFSET13 3.46927703
        #define KERNEL_WEIGHT13 0.12055435
        #define KERNEL_OFFSET14 -1.48681946
        #define KERNEL_WEIGHT14 0.14350044
        #define KERNEL_OFFSET15 1.48681946
        #define KERNEL_WEIGHT15 0.14350044
        #define KERNEL_OFFSET16 0.
        #define KERNEL_WEIGHT16 0.07494756

        #define SHADER_NAME vertex:kernelBlur
        precision highp float;
        attribute vec3 position;
        uniform vec2 delta;
        varying vec2 sampleCenter;
        varying vec2 sampleCoord0;
        varying vec2 sampleCoord1;
        varying vec2 sampleCoord2;
        varying vec2 sampleCoord3;
        varying vec2 sampleCoord4;
        varying vec2 sampleCoord5;
        varying vec2 sampleCoord6;
        varying vec2 sampleCoord7;
        varying vec2 sampleCoord8;
        varying vec2 sampleCoord9;
        varying vec2 sampleCoord10;
        varying vec2 sampleCoord11;
        varying vec2 sampleCoord12;
        varying vec2 sampleCoord13;
        varying vec2 sampleCoord14;
        varying vec2 sampleCoord15;
        varying vec2 sampleCoord16;
        const vec2 madd = vec2(0.5, 0.5);
        void main(void) {
            sampleCenter = (position.xy*madd+madd);
            sampleCoord0 = sampleCenter+delta*KERNEL_OFFSET0;
            sampleCoord1 = sampleCenter+delta*KERNEL_OFFSET1;
            sampleCoord2 = sampleCenter+delta*KERNEL_OFFSET2;
            sampleCoord3 = sampleCenter+delta*KERNEL_OFFSET3;
            sampleCoord4 = sampleCenter+delta*KERNEL_OFFSET4;
            sampleCoord5 = sampleCenter+delta*KERNEL_OFFSET5;
            sampleCoord6 = sampleCenter+delta*KERNEL_OFFSET6;
            sampleCoord7 = sampleCenter+delta*KERNEL_OFFSET7;
            sampleCoord8 = sampleCenter+delta*KERNEL_OFFSET8;
            sampleCoord9 = sampleCenter+delta*KERNEL_OFFSET9;
            sampleCoord10 = sampleCenter+delta*KERNEL_OFFSET10;
            sampleCoord11 = sampleCenter+delta*KERNEL_OFFSET11;
            sampleCoord12 = sampleCenter+delta*KERNEL_OFFSET12;
            sampleCoord13 = sampleCenter+delta*KERNEL_OFFSET13;
            sampleCoord14 = sampleCenter+delta*KERNEL_OFFSET14;
            sampleCoord15 = sampleCenter+delta*KERNEL_OFFSET15;
            sampleCoord16 = sampleCenter+delta*KERNEL_OFFSET16;
            gl_Position = vec4(position.xy, 0.0, 1.0);
        }
        `;
    }

    protected static getFragmentString(irisAreaMode: EIrisAreaMode = EIrisAreaMode.Circle) {
        return `
        #define KERNEL_OFFSET0 -15.36704337
        #define KERNEL_WEIGHT0 0.00226838
        #define KERNEL_OFFSET1 15.36704337
        #define KERNEL_WEIGHT1 0.00226838
        #define KERNEL_OFFSET2 -13.38352585
        #define KERNEL_WEIGHT2 0.00623279
        #define KERNEL_OFFSET3 13.38352585
        #define KERNEL_WEIGHT3 0.00623279
        #define KERNEL_OFFSET4 -11.40028041
        #define KERNEL_WEIGHT4 0.01489653
        #define KERNEL_OFFSET5 11.40028041
        #define KERNEL_WEIGHT5 0.01489653
        #define KERNEL_OFFSET6 -9.41727148
        #define KERNEL_WEIGHT6 0.03096924
        #define KERNEL_OFFSET7 9.41727148
        #define KERNEL_WEIGHT7 0.03096924
        #define KERNEL_OFFSET8 -7.4344613
        #define KERNEL_WEIGHT8 0.05600478
        #define KERNEL_OFFSET9 7.4344613
        #define KERNEL_WEIGHT9 0.05600478
        #define KERNEL_OFFSET10 -5.45181021
        #define KERNEL_WEIGHT10 0.08809971
        #define KERNEL_OFFSET11 5.45181021
        #define KERNEL_WEIGHT11 0.08809971
        #define KERNEL_OFFSET12 -3.46927703
        #define KERNEL_WEIGHT12 0.12055435
        #define KERNEL_OFFSET13 3.46927703
        #define KERNEL_WEIGHT13 0.12055435
        #define KERNEL_OFFSET14 -1.48681946
        #define KERNEL_WEIGHT14 0.14350044
        #define KERNEL_OFFSET15 1.48681946
        #define KERNEL_WEIGHT15 0.14350044
        #define KERNEL_OFFSET16 0.
        #define KERNEL_WEIGHT16 0.07494756

        #define SHADER_NAME fragment:kernelBlur
        precision highp float;

        varying vec2 sampleCenter;
        varying vec2 sampleCoord0;
        varying vec2 sampleCoord1;
        varying vec2 sampleCoord2;
        varying vec2 sampleCoord3;
        varying vec2 sampleCoord4;
        varying vec2 sampleCoord5;
        varying vec2 sampleCoord6;
        varying vec2 sampleCoord7;
        varying vec2 sampleCoord8;
        varying vec2 sampleCoord9;
        varying vec2 sampleCoord10;
        varying vec2 sampleCoord11;
        varying vec2 sampleCoord12;
        varying vec2 sampleCoord13;
        varying vec2 sampleCoord14;
        varying vec2 sampleCoord15;
        varying vec2 sampleCoord16;

        uniform sampler2D textureSampler;
        uniform vec2 delta;

        ${Blur.IrisFragmentDefined()}

        ${Blur.GetIrisMaskFragmentDefined(irisAreaMode)}

        void main(void) {
            float iris = ${Blur.GetIrisMaskFragmentCall(irisAreaMode, 'sampleCenter')};
            float computedWeight = 0.0;
            vec4 blend = vec4(0.);
            computedWeight = KERNEL_WEIGHT0;
            blend += texture2D(textureSampler, sampleCoord0)*computedWeight * iris;
            computedWeight = KERNEL_WEIGHT1;
            blend += texture2D(textureSampler, sampleCoord1)*computedWeight * iris;
            computedWeight = KERNEL_WEIGHT2;
            blend += texture2D(textureSampler, sampleCoord2)*computedWeight * iris;
            computedWeight = KERNEL_WEIGHT3;
            blend += texture2D(textureSampler, sampleCoord3)*computedWeight * iris;
            computedWeight = KERNEL_WEIGHT4;
            blend += texture2D(textureSampler, sampleCoord4)*computedWeight * iris;
            computedWeight = KERNEL_WEIGHT5;
            blend += texture2D(textureSampler, sampleCoord5)*computedWeight * iris;
            computedWeight = KERNEL_WEIGHT6;
            blend += texture2D(textureSampler, sampleCoord6)*computedWeight * iris;
            computedWeight = KERNEL_WEIGHT7;
            blend += texture2D(textureSampler, sampleCoord7)*computedWeight * iris;
            computedWeight = KERNEL_WEIGHT8;
            blend += texture2D(textureSampler, sampleCoord8)*computedWeight * iris;
            computedWeight = KERNEL_WEIGHT9;
            blend += texture2D(textureSampler, sampleCoord9)*computedWeight * iris;
            computedWeight = KERNEL_WEIGHT10;
            blend += texture2D(textureSampler, sampleCoord10)*computedWeight * iris;
            computedWeight = KERNEL_WEIGHT11;
            blend += texture2D(textureSampler, sampleCoord11)*computedWeight * iris;
            computedWeight = KERNEL_WEIGHT12;
            blend += texture2D(textureSampler, sampleCoord12)*computedWeight * iris;
            computedWeight = KERNEL_WEIGHT13;
            blend += texture2D(textureSampler, sampleCoord13)*computedWeight * iris;
            computedWeight = KERNEL_WEIGHT14;
            blend += texture2D(textureSampler, sampleCoord14)*computedWeight * iris;
            computedWeight = KERNEL_WEIGHT15;
            blend += texture2D(textureSampler, sampleCoord15)*computedWeight * iris;
            computedWeight = KERNEL_WEIGHT16;
            blend += texture2D(textureSampler, sampleCoord16)*computedWeight * iris;
            gl_FragColor = blend;
        }
        `;
    }
}

export interface IBlurOption {
    kernel: 3 | 5 | 9 | 13 | 15;
    direction?: [number, number];
}