
import { PostProcessDataManager } from "./data_manager";
import { PostProcess } from "./postprocess";
import { RenderTarget, RenderTexture } from "./render_target";
import { ShaderCompiler } from "./shader_compiler";
import { ManagedWebGLRenderingContext } from "./webgl";

/**
 * 均匀图块错位
 */
export class GlitchBlockOffset extends PostProcess {
    public static shaderName = `GlitchBlockOffset`;
    public static getShader(context: ManagedWebGLRenderingContext) {
        let manager = PostProcessDataManager.get(context);
        let shaderKey = `${GlitchBlockOffset.shaderName}`;
        let result = manager.getShaderCompiler(shaderKey);
        if (!result) {
            result = new ShaderCompiler(context, PostProcess.getVertexString(), GlitchBlockOffset.getFragmentString());
            manager.setShaderCompiler(shaderKey, result);
        }

        return result;
    }

    private static Fractor = 'fractor';
    private static Time = 'time';
    private static Shake = 'shake';
    private static Cutoff = 'u_cutoff';
    private static Debug = 'u_debug';

    blockSizeX: number = 0;
    blockSizeY: number = 20;
    time: number = 0.1;
    colorOffset: number = 0.1;
    cutoff: number = 0.9;
    debug: boolean = true;

    public apply(sources: RenderTexture[], target: RenderTarget): ShaderCompiler {
        super.apply(sources, target);

        const gl = this.canvasOrContext.gl;
        let shader = GlitchBlockOffset.getShader(this.canvasOrContext);
        shader.bind();

        gl.activeTexture(gl.TEXTURE0);
        gl.bindTexture(gl.TEXTURE_2D, sources[0]._webGLTexture);

        shader.setUniformi(ShaderCompiler.TEXTURE_SAMPLER, 0);
        shader.setUniform4f(GlitchBlockOffset.Fractor, this.blockSizeX, this.blockSizeY, this.time, this.colorOffset);
        shader.setUniformf(GlitchBlockOffset.Cutoff, this.cutoff);
        shader.setUniformf(GlitchBlockOffset.Debug, this.debug ? 1.0 : 0.0);

        PostProcess.bindData(this.canvasOrContext, this.manager.getDataBufferCfg(PostProcessDataManager.baseDataBufferCfgKey), shader);

        return shader;
    }

    protected _render() {
        // const gl = this.canvasOrContext.gl;

        // this._clear();

        // gl.disable(gl.CULL_FACE);
        // // gl.disable(gl.COLOR_WRITEMASK);
        // gl.disable(gl.DEPTH_TEST);
        // gl.bindBuffer(gl.ARRAY_BUFFER, null);
        // gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER, null);
        // // gl.bindVertexArray(null);

        // gl.activeTexture(gl.TEXTURE0);
        // gl.bindTexture(gl.TEXTURE_2D, this.sourceTexture);

        // this.shader.setUniformi(ShaderCompiler.TEXTURE_SAMPLER, 0);

        // this.shader.setUniform4f(GlitchBlockOffset.Fractor, this.blockSizeX, this.blockSizeY, this.time, this.colorOffset);
        // this.shader.setUniformf(GlitchBlockOffset.Cutoff, this.cutoff);
        // this.shader.setUniformf(GlitchBlockOffset.Debug, this.debug ? 1.0 : 0.0);
    }

    protected static getFragmentString() {
        return `
        #define SHADER_NAME fragment:GlitchBlockOffset
        precision highp float;

        #define SAMPLES 10

        varying vec2 sampleCenter;

        uniform sampler2D textureSampler;
        uniform vec4 fractor;
        uniform float u_cutoff;
        uniform float u_debug;

        float randomNoise(vec2 seed, float time, float speed)
        {
            return fract(sin(dot(seed * floor(time * speed), vec2(17.13, 3.71))) * 43758.5453123);
        }

        float nrand(float x, float y)
        {
            return fract(sin(dot(vec2(x, y), vec2(12.9898, 78.233))) * 43758.5453123);
        }

        float random(vec2 st)
        {
            return fract(sin(dot(st, vec2(12.9898, 78.233))) * 43758.5453);
        }

        float randomNoise1(float seed)
        {
            return nrand(seed, 1.0);
        }

        vec4 Block(vec2 source, vec4 fractor) {

            vec2 count = fractor.xy;
            float u_time = fractor.z;
            float colorOffset = fractor.w;
            float cutoff = u_cutoff;
            float debug = u_debug;

            vec2 st = source;
            st.y -= u_time;
            st *= count;

            vec2 ipos = floor(st);  // integer

            ipos = floor(st);  // integer

            vec2 rand = vec2(random(vec2(ipos.y, 0.0)), random(vec2(0.0, ipos.x)));

            float result = 0.0;
            result = random(rand);
            result *= result;

            // result = smoothstep(cutoff, cutoff + 0.001, result);
            result = step(cutoff, result);

            vec4 color = vec4(1.0, 1.0, 1.0, 1.0);

            if (result > 0.01) {
                vec4 src1 = texture2D(textureSampler, fract(source + vec2(result * colorOffset * 0.5, 0.0) ));
                vec4 src2 = texture2D(textureSampler, fract(source + vec2(result * colorOffset, 0.0) ));

                color = vec4(src1.r, src2.g, src1.b, (src1.a + src2.a) * 0.5);

                color = color * (1.0 - debug) + vec4(result, result, result, 1.0) * debug;
            }
            else {
                color = texture2D(textureSampler, fract(source + vec2(0.0, 0.0) ));
            }

            return color;
        }

        void main(void)
        {
            vec2 uv = sampleCenter;

            gl_FragColor = Block(uv, fractor);
        }
        `;
    }
}