
import { PostProcessDataManager } from "./data_manager";
import { PostProcess } from "./postprocess";
import { RenderTarget, RenderTexture } from "./render_target";
import { ShaderCompiler } from "./shader_compiler";
import { ManagedWebGLRenderingContext } from "./webgl";

/**
 * 均匀图块错位
 */
export class GlitchColorBlock extends PostProcess {
    public static shaderName = `GlitchColorBlock`;
    public static getShader(context: ManagedWebGLRenderingContext) {
        let manager = PostProcessDataManager.get(context);
        let shaderKey = `${GlitchColorBlock.shaderName}`;
        let result = manager.getShaderCompiler(shaderKey);
        if (!result) {
            result = new ShaderCompiler(context, PostProcess.getVertexString(), GlitchColorBlock.getFragmentString());
            manager.setShaderCompiler(shaderKey, result);
        }

        return result;
    }

    private static Fractor = 'fractor';
    private static ReplaceColor = 'replaceColor';

    blockSizeX: number = 0.1;
    blockSizeY: number = 0.1;
    time: number = 0.1;
    cutoff: number = 0.1;
    replaceColor: [number, number, number] = [0.3, 0.3, 0.3];

    public apply(sources: RenderTexture[], target: RenderTarget): ShaderCompiler {
        const gl = this.canvasOrContext.gl;
        let shader = GlitchColorBlock.getShader(this.canvasOrContext);
        shader.bind();

        gl.activeTexture(gl.TEXTURE0);
        gl.bindTexture(gl.TEXTURE_2D, sources[0]._webGLTexture);

        shader.setUniformi(ShaderCompiler.TEXTURE_SAMPLER, 0);
        shader.setUniform4f(GlitchColorBlock.Fractor, this.blockSizeX, this.blockSizeY, this.time, this.cutoff);
        shader.setUniform3f(GlitchColorBlock.ReplaceColor, this.replaceColor[0], this.replaceColor[1], this.replaceColor[2]);

        PostProcess.bindData(this.canvasOrContext, this.manager.getDataBufferCfg(PostProcessDataManager.baseDataBufferCfgKey), shader);

        return shader;
    }

    protected static getFragmentString() {
        return `
        #define SHADER_NAME fragment:GlitchImageBlock
        precision highp float;

        #define GLODEN_COS -0.7373688782616119
        #define GLODEN_SIN 0.675490294061441

        #define SAMPLES 10

        varying vec2 sampleCenter;

        uniform sampler2D textureSampler;
        uniform vec4 fractor;
        uniform vec3 replaceColor;

        float randomNoise(vec2 seed, float time, float speed)
        {
            return fract(sin(dot(seed * floor(time * speed), vec2(17.13, 3.71))) * 43758.5453123);
        }

        float nrand(float x, float y)
        {
            return fract(sin(dot(vec2(x, y), vec2(12.9898, 78.233))) * 43758.5453123);
        }

        float nrand2(vec2 st)
        {
            return fract(sin(dot(st, vec2(12.9898, 78.233))) * 43758.5453);
        }

        float randomNoise1(float seed)
        {
            return nrand(seed, 1.0);
        }

        vec4 Block(vec2 source, vec4 fractor) {

            float colordirty = randomNoise1(7.0);

            vec2 blockSize = fractor.xy;

            vec2 temp2 = floor(source * blockSize);
            // vec2 temp2 = source * blockSize;

            blockSize.x *= (randomNoise1(temp2.y) - 0.5);

            vec2 temp = floor(source * blockSize);
            // vec2 temp = source * blockSize;

            float time = fractor.z;
            float cutoff = fractor.w;
            float block = nrand2(temp + vec2(time, .0));
		    float displaceNoise = abs(block); //pow(block, 8.0) * pow(block, 3.0);

            displaceNoise = displaceNoise > cutoff ? 1.0 : 0.0;

            vec4 src0 = texture2D(textureSampler, fract(source));

            return vec4(src0.rgb * (1.0 - displaceNoise) + replaceColor * displaceNoise, src0.a);
        }

        void main(void)
        {
            vec2 uv = sampleCenter;

            gl_FragColor = Block(uv, fractor);
        }
        `;
    }
}