
import { PostProcessDataManager } from "./data_manager";
import { PostProcess } from "./postprocess";
import { RenderTexture, RenderTarget } from "./render_target";
import { ShaderCompiler } from "./shader_compiler";
import { ManagedWebGLRenderingContext } from "./webgl";

/**
 * 均匀图块错位
 */
export class GlitchImageBlock extends PostProcess {
    public static shaderName = `GlitchImageBlock`;
    public static getShader(context: ManagedWebGLRenderingContext) {
        let manager = PostProcessDataManager.get(context);
        let shaderKey = `${GlitchImageBlock.shaderName}`;
        let result = manager.getShaderCompiler(shaderKey);
        if (!result) {
            result = new ShaderCompiler(context, PostProcess.getVertexString(), GlitchImageBlock.getFragmentString());
            manager.setShaderCompiler(shaderKey, result);
        }

        return result;
    }

    private static Fractor = 'fractor';
    private static Time = 'time';
    private static Shake = 'shake';
    private static ColorDrift = 'colorDrift';

    blockSizeX: number = 0.1;
    blockSizeY: number = 0.1;
    time: number = 0.1;
    cutoff: number = 0.1;

    public apply(sources: RenderTexture[], target: RenderTarget): ShaderCompiler {
        const gl = this.canvasOrContext.gl;
        let shader = GlitchImageBlock.getShader(this.canvasOrContext);
        shader.bind();

        gl.activeTexture(gl.TEXTURE0);
        gl.bindTexture(gl.TEXTURE_2D, sources[0]._webGLTexture);
        shader.setUniformi(ShaderCompiler.TEXTURE_SAMPLER, 0);
        shader.setUniform4f(GlitchImageBlock.Fractor, this.blockSizeX, this.blockSizeY, this.time, this.cutoff);

        PostProcess.bindData(this.canvasOrContext, this.manager.getDataBufferCfg(PostProcessDataManager.baseDataBufferCfgKey), shader);

        return shader;
    }

    protected static getFragmentString() {
        return `
        #define SHADER_NAME fragment:GlitchImageBlock
        precision highp float;

        #define SAMPLES 10

        varying vec2 sampleCenter;

        uniform sampler2D textureSampler;
        uniform vec4 fractor;

        float randomNoise(vec2 seed, float time, float speed)
        {
            return fract(sin(dot(seed * floor(time * speed), vec2(17.13, 3.71))) * 43758.5453123);
        }

        float nrand(float x, float y)
        {
            return fract(sin(dot(vec2(x, y), vec2(12.9898, 78.233))) * 43758.5453123);
        }

        float nrand2(vec2 st)
        {
            return fract(sin(dot(st, vec2(12.9898, 78.233))) * 43758.5453);
        }

        float randomNoise1(float seed)
        {
            return nrand(seed, 1.0);
        }

        vec4 Block(vec2 source, vec4 fractor) {

            float colordirty = randomNoise1(7.0);

            vec2 blockSize = fractor.xy;

            vec2 temp2 = floor(source * blockSize);
            // vec2 temp2 = source * blockSize;

            blockSize.x *= (randomNoise1(temp2.y) - 0.5);

            vec2 temp = floor(source * blockSize);
            // vec2 temp = source * blockSize;

            float time = fractor.z;
            float cutoff = fractor.w;
            float block = nrand2(temp + vec2(time, .0));
		    float displaceNoise = abs(block); //pow(block, 8.0) * pow(block, 3.0);

            displaceNoise = displaceNoise > cutoff ? displaceNoise : 0.0;

            vec4 src1 = texture2D(textureSampler, fract(source + vec2(displaceNoise, 0.0)));
            vec4 src2 = texture2D(textureSampler, fract(source + vec2(displaceNoise * time * 0.1 * colordirty, 0.0)));
            // vec4 src3 = texture2D(textureSampler, fract(source - vec2(displaceNoise * time * 0.1 * colordirty, 0.0)));

            return vec4(src1.r, src2.g, src1.b, (src1.a + src2.b) * 0.5);
        }

        void main(void)
        {
            vec2 uv = sampleCenter;

            gl_FragColor = Block(uv, fractor);
        }
        `;
    }
}