
import { PostProcessDataManager } from "./data_manager";
import { PostProcess } from "./postprocess";
import { RenderTexture, RenderTarget } from "./render_target";
import { ShaderCompiler } from "./shader_compiler";
import { ManagedWebGLRenderingContext } from "./webgl";

/**
 * 均匀图块错位
 */
export class GlitchImageBlockOptimize extends PostProcess {
    public static shaderName = `GlitchImageBlockOptimize`;
    public static getShader(context: ManagedWebGLRenderingContext) {
        let manager = PostProcessDataManager.get(context);
        let shaderKey = `${GlitchImageBlockOptimize.shaderName}`;
        let result = manager.getShaderCompiler(shaderKey);
        if (!result) {
            result = new ShaderCompiler(context, GlitchImageBlockOptimize.getVertexString(), GlitchImageBlockOptimize.getFragmentString());
            manager.setShaderCompiler(shaderKey, result);
        }

        return result;
    }

    private static Fractor = 'fractor';
    private static Time = 'time';
    private static Shake = 'shake';
    private static ColorDrift = 'colorDrift';

    /**
     * 一次故障最大条纹尺寸
     */
    blockMaxSize: number = 0.3;
    /**
     * 一次故障最小条尺寸
     */
    blockMinSize: number = 0.02;
    /**
     * 一次故障最多条纹数
     */
    blockMaxCount: number = 5;
    /**
     * 一次故障最少条纹数
     */
    blockMinCount: number = 0;
    /**
     * 一次故障持续时间 - ms
     */
    blockLifetime: number = 1000;
    /**
     * 故障生成几率
     */
    probability: number = 0.1;
    /**
     * 故障扭曲强度
     */
    strength: number = 0.1;
    /**
     * 故障边界过渡因子
     */
    fade: number = 0.5;

    private lastRecordTime = 0;
    private recordTime = 0;

    public apply(sources: RenderTexture[], target: RenderTarget): ShaderCompiler {
        throw new Error(`Use render!`);
    }

    /**
     * 条纹生成点
     */
    private blockValues: number[] = [];
    /**
     * 条纹尺寸
     */
    private blockSizes: number[] = [];

    private minValue: number = 0;
    private maxValue: number = 1;

    /**
     * 渲染接口
     * @param source 源纹理
     * @param target 目标
     * @param time 当前时间 - ms
     */
    public render(source: RenderTexture, target: RenderTarget, viewport: [number, number, number, number], time: number) {
        // 计算故障
        if (this.lastRecordTime == 0) {
            this.lastRecordTime = time;
        }

        // 每一轮的开始重新计算
        if (this.recordTime == 0) {
            this.minValue = 1;
            this.maxValue = 0;
            this.blockValues.length = 0;
            this.blockSizes.length = 0;
            let count = this.blockMinCount + (this.blockMaxCount - this.blockMinCount) * (Math.max(0., this.probability - Math.random()) / this.probability);
            for (let i = 0; i < count; i++) {
                let value = Math.random();
                let size = this.blockMinSize + (this.blockMaxSize - this.blockMinSize) * Math.random();
                this.blockValues.push(value);
                this.blockSizes.push(size);

                // 条纹区域为条纹生成点 的  size / 2 邻域
                this.minValue = Math.min(this.minValue, value - size / 2);
                this.maxValue = Math.max(this.maxValue, value + size / 2);
            }

        }

        this.minValue = Math.min(this.minValue, this.maxValue);
        this.maxValue = Math.max(this.minValue, this.maxValue);
        this.recordTime += Math.max(0., time - this.lastRecordTime);

        //
        let lifePercent = Math.min(1., Math.max(this.recordTime / this.blockLifetime, 0.));
        // 渲染距离为 头部进入到尾部离开的范围, 中间经过一整个屏幕
        let moveDistance = (this.maxValue - this.minValue) * 2 + 1.;
        let moveStart = -(this.maxValue - this.minValue);

        // ------------------------------------------------

        let renderCount = this.blockValues.length;
        if (renderCount > 0) {
            const gl = this.canvasOrContext.gl;
            let dataBuffer = this.manager.getDataBufferCfg(PostProcessDataManager.baseDataBufferCfgKey);
            let shader = GlitchImageBlockOptimize.getShader(this.canvasOrContext);
            shader.bind();

            gl.activeTexture(gl.TEXTURE0);
            gl.bindTexture(gl.TEXTURE_2D, source._webGLTexture);
            shader.setUniformi(ShaderCompiler.TEXTURE_SAMPLER, 0);
            PostProcess.bindData(this.canvasOrContext, dataBuffer, shader);

            for (let i = 0; i < renderCount; i++) {
                let value = this.blockValues[i];
                let size = this.blockSizes[i];
                let start = moveStart + (value - this.minValue - size / 2) + lifePercent * moveDistance;

                PostProcess.glViewport(this.canvasOrContext, [viewport[0], Math.round(viewport[1] + start * viewport[3]), viewport[2], Math.round(size * viewport[3])], null);
                shader.setUniform4f(`uRang`, start, start + size, this.strength * (i % 2 ? -1 : 1), this.fade);
                PostProcess.drawData(this.canvasOrContext, dataBuffer);
            }
        }

        // ------------------------------------------------

        if (this.recordTime > this.blockLifetime) {
            this.recordTime = 0;
        }
        this.lastRecordTime = time;
    }

    protected static getVertexString() {
        return `
        #define SHADER_NAME vertex:GlitchImageBlockOptimize
        precision highp float;

        attribute vec3 position;

        uniform vec4 uRang;

        varying vec2 sampleCenter;
        const vec2 madd = vec2(0.5, 0.5);
        void main(void) {
            vec2 uv = (position.xy*madd+madd);
            vec2 xy = position.xy;
            uv.y    = uRang.x + uv.y * (uRang.y - uRang.x);
            sampleCenter = uv;
            gl_Position = vec4(position.xy, 0.0, 1.0);
        }
        `;
    }

    protected static getFragmentString() {
        return `
        #define SHADER_NAME fragment:GlitchImageBlockOptimize
        precision highp float;

        #define SAMPLES 10

        varying vec2 sampleCenter;
        uniform sampler2D textureSampler;

        uniform vec4 uRang;

        vec4 Block(vec2 source, float colordirty) {

            vec4 src1 = texture2D(textureSampler, vec2(fract(source.x + colordirty * 0.5), source.y));
            vec4 src2 = texture2D(textureSampler, vec2(fract(source.x + colordirty), source.y));
            // vec4 src3 = texture2D(textureSampler, vec2(colordirty, 0.0));

            return vec4(src1.r, src2.g, src1.b, (src1.a + src2.b) * 0.5);
        }

        void main(void)
        {
            vec2 uv = sampleCenter;

            float diff = uRang.y - uRang.x;
            float colordirty = (1.0 - smoothstep(uRang.w, 1., abs(uv.y - (uRang.x + diff / 2.)) / diff)) * uRang.z;

            gl_FragColor = Block(uv, colordirty);
        }
        `;
    }
}