
import { PostProcessDataManager } from "./data_manager";
import { PostProcess } from "./postprocess";
import { RenderTexture, RenderTarget } from "./render_target";
import { ShaderCompiler } from "./shader_compiler";
import { ManagedWebGLRenderingContext } from "./webgl";

export class HSV extends PostProcess {
    public static shaderName = `HSV`;
    public static getShader(context: ManagedWebGLRenderingContext) {
        let manager = PostProcessDataManager.get(context);
        let shaderKey = `${HSV.shaderName}`;
        let result = manager.getShaderCompiler(shaderKey);
        if (!result) {
            result = new ShaderCompiler(context, PostProcess.getVertexString(), HSV.getFragmentString());
            manager.setShaderCompiler(shaderKey, result);
        }

        return result;
    }

    public static HSV = 'u_hsv';
    /**
     * 色相 - 对应Photoshop的[-180, 180]
     */
    hue: number = 0;
    /**
     * 饱和度 对应Photoshop的[-100, 100]
     */
    saturate: number = 0;
    /**
     * 明度 对应Photoshop的[-100, 100]
     */
    brightness: number = 0;

    public apply(sources: RenderTexture[], target: RenderTarget): ShaderCompiler {
        super.apply(sources, target);

        const gl = this.canvasOrContext.gl;
        let shader = HSV.getShader(this.canvasOrContext);
        shader.bind();

        gl.activeTexture(gl.TEXTURE0);
        gl.bindTexture(gl.TEXTURE_2D, sources[0]._webGLTexture);
        shader.setUniformi(ShaderCompiler.TEXTURE_SAMPLER, 0);

        shader.setUniform3f(HSV.HSV, this.hue / 360, this.saturate / 100, this.brightness / 100);

        PostProcess.bindData(this.canvasOrContext, this.manager.getDataBufferCfg(PostProcessDataManager.baseDataBufferCfgKey), shader);

        return shader;
    }

    protected static getFragmentString() {
        return `

        #define SHADER_NAME fragment:hsv
        precision highp float;

        varying vec2 sampleCenter;
        uniform sampler2D textureSampler;
        uniform vec3 ${HSV.HSV};

        vec3 rgb2hsv(vec3 c)
        {
            vec4 K = vec4(0.0, -1.0 / 3.0, 2.0 / 3.0, -1.0);
            vec4 p = mix(vec4(c.bg, K.wz), vec4(c.gb, K.xy), step(c.b, c.g));
            vec4 q = mix(vec4(p.xyw, c.r), vec4(c.r, p.yzx), step(p.x, c.r));

            float d = q.x - min(q.w, q.y);
            float e = 1.0e-10;
            return vec3(abs(q.z + (q.w - q.y) / (6.0 * d + e)), d / (q.x + e), q.x);
        }

        vec3 hsv2rgb(vec3 c)
        {
            vec4 K = vec4(1.0, 2.0 / 3.0, 1.0 / 3.0, 3.0);
            vec3 p = abs(fract(c.xxx + K.xyz) * 6.0 - K.www);
            return c.z * mix(K.xxx, clamp(p - K.xxx, vec3(0.0), vec3(1.0)), c.y);
        }

        void main(void) {
            vec4 c = texture2D(textureSampler, sampleCenter);

            vec3 hsv = rgb2hsv(c.rgb);
            hsv.r += ${HSV.HSV}.r;
            c.rgb = hsv2rgb(hsv);

            // Note: When saturate is greater than 0, the formula is different from PS
            float gray = max(c.r, max(c.g, c.b)) + min(c.r, min(c.g, c.b));
            c.rgb = mix(c.rgb, vec3(0.5 * gray), -${HSV.HSV}.g);

            if (${HSV.HSV}.b >= 0.0) {
                c.rgb = mix(c.rgb, vec3(1.0), ${HSV.HSV}.b);
            } else {
                c.rgb *= 1.0 + ${HSV.HSV}.b;
            }

            gl_FragColor = c;
        }
        `;
    }
}