
import { IIrisBlur, KeyIrisBlur } from "../interface/index";
import { Blur, EIrisAreaMode } from "./blur";
import { PostProcessDataManager } from "./data_manager";
import { PostProcess } from "./postprocess";
import { RenderTexture, RenderTarget } from "./render_target";
import { ShaderCompiler } from "./shader_compiler";
import { ManagedWebGLRenderingContext } from "./webgl";

export class IrisBlur extends Blur {
    public static shaderName = KeyIrisBlur;
    public static getShader(context: ManagedWebGLRenderingContext, iteration: number, irisAreaMode: EIrisAreaMode) {
        let manager = PostProcessDataManager.get(context);
        let shaderKey = `${IrisBlur.shaderName}_${iteration}_${irisAreaMode}`;
        let result = manager.getShaderCompiler(shaderKey);
        if (!result) {
            result = new ShaderCompiler(context, Blur.getVertexString(), IrisBlur.getFragmentString(iteration, irisAreaMode));
            manager.setShaderCompiler(shaderKey, result);
        }

        return result;
    }

    private static BlurRadius = 'blurRadius';

    public apply(param: IIrisBlur, sources: RenderTexture[], target: RenderTarget, ctx: ManagedWebGLRenderingContext): ShaderCompiler {
        super.apply(param, sources, target, ctx);

        const gl = this.canvasOrContext.gl;
        let shader = IrisBlur.getShader(this.canvasOrContext, param.iteration, EIrisAreaMode.None);
        shader.bind();

        gl.activeTexture(gl.TEXTURE0);
        gl.bindTexture(gl.TEXTURE_2D, sources[0]._webGLTexture);
        shader.setUniformi(ShaderCompiler.TEXTURE_SAMPLER, 0);

        shader.setUniformf(IrisBlur.BlurRadius, param.blurRadius);

        let dataBuffer = this.manager.getDataBufferCfg(PostProcessDataManager.baseDataBufferCfgKey);
        PostProcess.bindData(this.canvasOrContext, dataBuffer, shader);
        PostProcess.drawData(this.canvasOrContext, dataBuffer);
        this.afterDraw();
        this.applyIris(shader);

        PostProcess.bindData(this.canvasOrContext, this.manager.getDataBufferCfg(PostProcessDataManager.baseDataBufferCfgKey), shader);

        return shader;
    }

    // https://www.shadertoy.com/view/4stSRM
    protected static getFragmentString(iteration: number = 6, irisAreaMode: EIrisAreaMode = EIrisAreaMode.Circle) {
        return `
        #define SHADER_NAME fragment:bokehblur
        precision highp float;

        #define GLODEN_COS -0.7373688782616119
        #define GLODEN_SIN 0.675490294061441

        #define SAMPLES 10

        uniform sampler2D textureSampler;

        varying vec2 sampleCenter;

        uniform float blurRadius;
        ${Blur.IrisFragmentDefined()}

        ${Blur.GetIrisMaskFragmentDefined(irisAreaMode)}

        const int ITERATION = ${iteration};

        vec4 BokehBlur(sampler2D tex, vec2 uv) {
            float iris = ${Blur.GetIrisMaskFragmentCall(irisAreaMode, 'sampleCenter')};
            if (iris < 0.0001) return texture2D(tex, uv.xy);

            mat2 rot = mat2(GLODEN_COS, GLODEN_SIN, -GLODEN_SIN, GLODEN_COS);

            vec4 accumulator = vec4(0.0);
            vec4 divisor = vec4(0.0);

            float r = 1.0;
            vec2 angle = vec2(0.0, blurRadius * iris);

            for (int j = 0; j < ITERATION; j++)
            {
                r += 1.0 / r;
                angle = rot * angle;

                uv = sampleCenter + vec2(r - 1.0) * angle;
                vec4 bokeh = texture2D(tex, uv);

                accumulator += bokeh * bokeh;
                divisor += bokeh;
            }

            return accumulator / divisor;
        }

        void main(void)
        {
            gl_FragColor = BokehBlur(textureSampler, sampleCenter);
        }
        `;
    }
}