
import { PostProcessDataManager } from "./data_manager";
import { PostProcess } from "./postprocess";
import { RenderTexture, RenderTarget } from "./render_target";
import { ShaderCompiler } from "./shader_compiler";
import { ManagedWebGLRenderingContext } from "./webgl";

/**
 * 均匀图块错位
 */
export class LineFlow extends PostProcess {
    public static shaderName = `LineFlow`;
    public static getShader(context: ManagedWebGLRenderingContext) {
        let manager = PostProcessDataManager.get(context);
        let shaderKey = `${LineFlow.shaderName}`;
        let result = manager.getShaderCompiler(shaderKey);
        if (!result) {
            result = new ShaderCompiler(context, LineFlow.getVertexString(), LineFlow.getFragmentString());
            manager.setShaderCompiler(shaderKey, result);
        }

        return result;
    }

    private static Fractor = 'u_fractor';
    private static Time = 'u_time';
    private static Shake = 'u_shake';
    private static Direction = 'u_direction';

    lineCountX: number = 0.8;
    lineCountY: number = 100;
    time: number = 0.1;
    cutoff: number = 0.95;
    /**
     * 径向角度
     */
    private _angle: number = 45 / 180 * Math.PI;
    public get angle() {
        return this._angle;
    }
    public set angle(v: number) {
        if (v == undefined || v === null) {
            v = 0;
        }

        v = v / 180 * Math.PI;

        if (this._angle !== v) {
            this._angle = v;
            this.direction[0] = Math.sin(this._angle);
            this.direction[1] = Math.cos(this._angle);
        }
    }
    private direction: [number, number] = [Math.sin(this._angle), Math.cos(this._angle)];

    public apply(sources: RenderTexture[], target: RenderTarget): ShaderCompiler {
        super.apply(sources, target);

        const gl = this.canvasOrContext.gl;
        let shader = LineFlow.getShader(this.canvasOrContext);
        shader.bind();

        gl.activeTexture(gl.TEXTURE0);
        gl.bindTexture(gl.TEXTURE_2D, sources[0]._webGLTexture);
        shader.setUniformi(ShaderCompiler.TEXTURE_SAMPLER, 0);

        shader.setUniform4f(LineFlow.Fractor, this.lineCountX, this.lineCountY, this.time, this.cutoff);
        shader.setUniform2f(LineFlow.Direction, this.direction[0], this.direction[1]);

        PostProcess.bindData(this.canvasOrContext, this.manager.getDataBufferCfg(PostProcessDataManager.baseDataBufferCfgKey), shader);

        return shader;
    }

    protected static getFragmentString() {
        return `
        #define SHADER_NAME fragment:LineFlow
        precision highp float;

        #define SAMPLES 10

        varying vec2 sampleCenter;

        uniform sampler2D textureSampler;
        uniform vec4 u_fractor;
        uniform vec2 u_direction;

        float randomNoise(vec2 seed, float time, float speed)
        {
            return fract(sin(dot(seed * floor(time * speed), vec2(17.13, 3.71))) * 43758.5453123);
        }

        float nrand(float x, float y)
        {
            return fract(sin(dot(vec2(x, y), vec2(12.9898, 78.233))) * 43758.5453123);
        }

        float random(vec2 st)
        {
            return fract(sin(dot(st, vec2(12.9898, 78.233))) * 43758.5453);
        }

        float randomNoise1(float seed)
        {
            return nrand(seed, 1.0);
        }

        float circle(vec2 xy, vec2 center, float radius, float smooth_edge) {
            float dist = distance(xy,center);
            dist = smoothstep(radius, radius + smooth_edge, dist);
            return dist;
        }

        vec4 Block(vec2 source, vec4 fractor, vec2 rotate) {

            vec2 count = fractor.xy;
            float u_time = fractor.z;
            float cutoff = fractor.w;

            vec2 st = vec2(source.x * rotate.y + source.y * rotate.x, source.y * rotate.y - source.x * rotate.x);
            st *= count;

            vec2 ipos = floor(st);  // integer
            st.x -= u_time * 2.0 * random(vec2(ipos.y, 1.0));

            ipos = floor(st);  // integer

            vec2 rand = vec2(random(vec2(ipos.y, 0.0)), random(vec2(0.0, ipos.x)));

            float result = 0.0;
            result = random(rand);

            result = smoothstep(cutoff, cutoff + 0.01, result);
            result = (1.0 - circle(fract(st), vec2(0.5), 0.25, 0.1)) * result;

            vec4 f_color = vec4(vec3(result), 1.0);
            vec4 src1 = texture2D(textureSampler, source);

            f_color.rgb = src1.rgb * (1.0 - result) + f_color.rgb * result;
            f_color.a = src1.a;

            return f_color;
        }

        void main(void)
        {
            vec2 uv = sampleCenter;

            gl_FragColor = Block(uv, u_fractor, u_direction);
        }
        `;
    }
}