
import { PostProcessDataManager } from "./data_manager";
import { PostProcess } from "./postprocess";
import { RenderTexture, RenderTarget } from "./render_target";
import { ShaderCompiler } from "./shader_compiler";
import { ManagedWebGLRenderingContext } from "./webgl";

/**
 * 均匀图块错位
 */
export class Magnifier extends PostProcess {
    public static shaderName = `Magnifier`;
    public static getShader(context: ManagedWebGLRenderingContext) {
        let manager = PostProcessDataManager.get(context);
        let shaderKey = `${Magnifier.shaderName}`;
        let result = manager.getShaderCompiler(shaderKey);
        if (!result) {
            result = new ShaderCompiler(context, Magnifier.getVertexString(), Magnifier.getFragmentString());
            manager.setShaderCompiler(shaderKey, result);
        }

        return result;
    }

    private static RadialCenter = 'radialCenter';
    private static Direction = 'direction';
    private static WaveInfo = 'waveInfo';
    private static ScreenInfo = 'screenInfo';
    private static WaveWidth = 'waveWidth';
    private static WaveRadius = 'waveRadius';
    /**
     *  是否处理为圆形
     */
    asCircle: boolean = false;
    /**
     * 波纹径向起点 0-2
     */
    waveStartRadius: number = 0.0;
    /**
     * 波纹径向终点 0-2
     */
    waveEndRadius: number = 0.5;
    /**
     * 波纹数量
     */
    waveCounter: number = 1;
    /**
     * 缩放值域
     */
    fector: number = 0.01;
    /**
     * 径向角度
     */
    private _angle: number = 0;
    public get angle() {
        return this._angle;
    }
    public set angle(v: number) {
        if (v == undefined || v === null) {
            v = 0;
        }

        v = v / 180 * Math.PI;

        if (this._angle !== v) {
            this._angle = v;
        }
    }
    private direction: [number, number] = [0, 0];

    private iteration = 20;

    public apply(sources: RenderTexture[], target: RenderTarget, alpha?: number): ShaderCompiler {
        super.apply(sources, target);

        const gl = this.canvasOrContext.gl;
        let shader = Magnifier.getShader(this.canvasOrContext);
        shader.bind();

        gl.activeTexture(gl.TEXTURE0);
        gl.bindTexture(gl.TEXTURE_2D, sources[0]._webGLTexture);
        shader.setUniformi(ShaderCompiler.TEXTURE_SAMPLER, 0);

        // const radius = this.blurRadius / this.sourceWidth / this.iteration;
        // this.direction[0] = Math.sin(this._angle) * radius ;
        // this.direction[1] = Math.cos(this._angle) * radius ;
        // this.shader.setUniform2f(Magnifier.Direction, this.direction[0], this.direction[1]);

        shader.setUniformf("alpha", alpha);
        shader.setUniform2f(Magnifier.ScreenInfo, this.asCircle ? sources[0].width : 1.0, this.asCircle ? sources[0].height : 1.0);
        shader.setUniform4f(Magnifier.WaveInfo, Math.min(this.waveStartRadius, this.waveEndRadius), Math.max(this.waveStartRadius, this.waveEndRadius), this.waveCounter, this.fector);

        PostProcess.bindData(this.canvasOrContext, this.manager.getDataBufferCfg(PostProcessDataManager.baseDataBufferCfgKey), shader);

        return shader;
    }

    protected static getFragmentString() {
        return `
        #define SHADER_NAME fragment:Magnifier
        precision highp float;

        varying vec2 sampleCenter;
        uniform sampler2D textureSampler;

        // uniform vec2 direction;
        uniform vec2 ${Magnifier.ScreenInfo};
        uniform vec4 ${Magnifier.WaveInfo};
        uniform float alpha;

        vec4 RadialBlur(sampler2D tex, vec2 uv) {

            vec4 acumulateColor = vec4(0.0);

            vec2 center = uv * 2.0 - 1.0; // [0, 1] -> [-1, 1]

            uv = (center * ${Magnifier.WaveInfo}.z + vec2(1.0)) * 0.5;

            center.y *= ${Magnifier.ScreenInfo}.y / ${Magnifier.ScreenInfo}.x;

            float len = abs(length(center));

            float temp = (${Magnifier.WaveInfo}.y - ${Magnifier.WaveInfo}.x) / 2.0;
            float temp2 = min(temp, 0.1);

            float weight = ${Magnifier.WaveInfo}.w * smoothstep(${Magnifier.WaveInfo}.x, ${Magnifier.WaveInfo}.x + temp2, len) * (1.0 - smoothstep(${Magnifier.WaveInfo}.y - temp, ${Magnifier.WaveInfo}.y, len));

            // len = len < ${Magnifier.WaveInfo}.x ? 0.0 : (len > ${Magnifier.WaveInfo}.y ? 0.0 : len - ${Magnifier.WaveInfo}.x);
            // len = len / (${Magnifier.WaveInfo}.y - ${Magnifier.WaveInfo}.x);
            float ll = len - ${Magnifier.WaveInfo}.x;
            ll = ${Magnifier.WaveInfo}.x + 1.0 / (1.0 - ${Magnifier.WaveInfo}.x) * ll * ll;
            ll = min(len, ll);
            // center = normalize(center) * mix(len, len * len, step(${Magnifier.WaveInfo}.x, len) * step(len, ${Magnifier.WaveInfo}.y));
            // uv += normalize(center) * len * len * ${Magnifier.WaveInfo}.w;
            // center = normalize(center) * ll;
            // uv = (center * ${Magnifier.WaveInfo}.z + vec2(1.0)) * 0.5;

            float dl = 1.0 - smoothstep(${Magnifier.WaveInfo}.x,  ${Magnifier.WaveInfo}.y, len);
            center = center * (1. - dl * ${Magnifier.WaveInfo}.w);
            uv = (center + vec2(1.0)) * 0.5;

            return texture2D(tex, uv);
        }

        void main(void) {
            vec4 color = RadialBlur(textureSampler, sampleCenter);
            color.a *= alpha;
            gl_FragColor = color;
        }
        `;
    }
}