
import { Blur, EIrisAreaMode } from "./blur";
import { PostProcessDataManager } from "./data_manager";
import { PostProcess } from "./postprocess";
import { RenderTexture, RenderTarget } from "./render_target";
import { ShaderCompiler } from "./shader_compiler";
import { ManagedWebGLRenderingContext } from "./webgl";

export enum EBlurMode {
    None = 0,
    BokehBlur = 1,
    DirectionalBlur = 2,
    RadialBlur = 3
}

export class MultiBlendPP extends Blur {
    public static shaderName = `MultiBlendPP`;
    public static getShader(context: ManagedWebGLRenderingContext, iteration: number, irisAreaMode: EIrisAreaMode) {
        let manager = PostProcessDataManager.get(context);
        let shaderKey = `${MultiBlendPP.shaderName}_${iteration}_${irisAreaMode}`;
        let result = manager.getShaderCompiler(shaderKey);
        if (!result) {
            result = new ShaderCompiler(context, MultiBlendPP.getVertexString(), MultiBlendPP.getFragmentString(iteration, irisAreaMode));
            manager.setShaderCompiler(shaderKey, result);
        }

        return result;
    }

    private static BlurRadius = 'blurRadius';
    private static RadialCenter = 'radialCenter';
    private static Direction = 'direction';
    private static BlurMode = 'blurMode';
    // https://evanw.github.io/glfx.js/demo/ - Vignette
    private static Vignette = 'vignette';
    private static VignetteColor = 'vignetteColor';
    /**
     * 模糊半径 - 像素值
     */
    blurRadius: number = 10;
    /**
     * 径向角度
     */
    private _angle: number = 0;
    public get angle() {
        return this._angle;
    }
    public set angle(v: number) {
        if (v == undefined || v === null) {
            v = 0;
        }

        v = v / 180 * Math.PI;

        if (this._angle !== v) {
            this._angle = v;
        }
    }
    private direction: [number, number] = [0, 0];

    public iteration = 6;

    public blurMode: EBlurMode = EBlurMode.None;

    public vignette: [number, number, number, number] = [0.2, 0.5, 2, 0.5];
    public vignetteColor: [number, number, number] = [0.1, 0.0, 0.0];

    public apply(sources: RenderTexture[], target: RenderTarget): ShaderCompiler {
        super.apply(sources, target);

        const gl = this.canvasOrContext.gl;
        let shader = MultiBlendPP.getShader(this.canvasOrContext, this.iteration, this.irisAreaMode);
        shader.bind();

        gl.activeTexture(gl.TEXTURE0);
        gl.bindTexture(gl.TEXTURE_2D, sources[0]._webGLTexture);
        shader.setUniformi(ShaderCompiler.TEXTURE_SAMPLER, 0);

        const radius = this.blurRadius / sources[0].width / this.iteration;
        this.direction[0] = Math.sin(this._angle) * radius ;
        this.direction[1] = Math.cos(this._angle) * radius ;
        shader.setUniform2f(MultiBlendPP.Direction, this.direction[0], this.direction[1]);
        shader.setUniformf(MultiBlendPP.BlurRadius, this.blurRadius / sources[0].width);
        shader.setUniformi(MultiBlendPP.BlurMode, Math.floor(this.blurMode));
        shader.setUniform4f(MultiBlendPP.Vignette, this.vignette[0], this.vignette[1], this.vignette[2], this.vignette[3]);
        shader.setUniform3f(MultiBlendPP.VignetteColor, this.vignetteColor[0], this.vignetteColor[1], this.vignetteColor[2]);

        this.applyIris(shader);

        PostProcess.bindData(this.canvasOrContext, this.manager.getDataBufferCfg(PostProcessDataManager.baseDataBufferCfgKey), shader);

        return shader;
    }

    protected static getFragmentString(iteration: number = 6, irisAreaMode: EIrisAreaMode = EIrisAreaMode.Circle) {
        return `
        #define SHADER_NAME fragment:directinalblur
        precision highp float;

        #define GLODEN_COS -0.7373688782616119
        #define GLODEN_SIN 0.675490294061441

        varying vec2 sampleCenter;

        uniform sampler2D textureSampler;
        uniform vec2 direction;

        uniform float ${MultiBlendPP.BlurRadius};

        uniform int ${MultiBlendPP.BlurMode};

        uniform vec4 ${MultiBlendPP.Vignette};
        uniform vec3 ${MultiBlendPP.VignetteColor};

        uniform vec4 ${Blur.IrisInfo};
        uniform float ${Blur.IrisTransition};
        uniform int ${Blur.IrisAreaMode};

        float IrisMask(vec2 uv, int irisAreaMode, vec4 irisInfo, float irisTransition)
        {
            if (irisAreaMode >= 2) {
                vec2 center = uv * 2.0 - 1.0 - irisInfo.zw; // [0,1] -> [-1,1]
                float x = clamp(abs(center.x) - irisInfo.x, 0.0, irisTransition) / irisTransition;
                float y = clamp(abs(center.y) - irisInfo.y, 0.0, irisTransition) / irisTransition;
                return max(x, y);
            }
            else if (irisAreaMode >= 1) {
                vec2 center = uv * 2.0 - 1.0 - irisInfo.zw; // [0,1] -> [-1,1]
                return clamp(abs(length(center)) - irisInfo.x, 0.0, irisTransition) / irisTransition;
            }
            else {
                return 1.0;
            }
        }

        vec4 DirectionBlur(sampler2D tex, vec2 uv, vec2 blurCenter, float blurRadius, float iris) {
            if (iris < 0.0001) return texture2D(tex, uv.xy);

            vec4 acumulateColor = vec4(0.0);

            for (int j = -${iteration}; j < ${iteration}; j++)
            {
                acumulateColor += texture2D(tex, uv.xy - iris * direction * float(j));
            }

            return acumulateColor / (${iteration}.0 * 2.0);
        }

        vec4 RadialBlur(sampler2D tex, vec2 uv, vec2 blurCenter, float blurRadius, float iris) {
            if (iris < 0.0001) return texture2D(tex, uv.xy);

            vec2 blurVector = (blurCenter - uv.xy) * blurRadius * iris;

            vec4 acumulateColor = vec4(0.0);

            for (int j = 0; j < ${iteration}; j++)
            {
                acumulateColor += texture2D(tex, uv.xy);
                uv.xy += blurVector;
            }

            return acumulateColor / ${iteration}.0;
        }

        vec4 BokehBlur(sampler2D tex, vec2 uv, vec2 blurCenter, float blurRadius, float iris) {
            if (iris < 0.0001) return texture2D(tex, uv.xy);

            mat2 rot = mat2(GLODEN_COS, GLODEN_SIN, -GLODEN_SIN, GLODEN_COS);

            vec4 accumulator = vec4(0.0);
            vec4 divisor = vec4(0.0);

            float r = 1.0;
            vec2 angle = vec2(0.0, blurRadius) * vec2(iris);

            for (int j = 0; j < ${iteration}; j++)
            {
                r += 1.0 / r;
                angle = rot * angle;

                uv = sampleCenter + vec2(r - 1.0) * angle;
                vec4 bokeh = texture2D(tex, uv);

                accumulator += bokeh * bokeh;
                divisor += bokeh;
            }

            return accumulator / divisor;
        }

        vec4 Blur(sampler2D tex, vec2 uv, int mode, vec2 blurCenter, float blurRadius, float iris) {
            if (mode == 3) {
                return RadialBlur(tex, uv.xy, blurCenter, blurRadius, iris);
            }
            else if (mode == 2) {
                return DirectionBlur(tex, uv.xy, blurCenter, blurRadius, iris);
            }
            else if (mode == 1) {
                return BokehBlur(tex, uv.xy, blurCenter, blurRadius, iris);
            }
            else {
                vec4 color = texture2D(tex, uv);
                return color;
            }
        }

        float Vignette(vec4 vignette, vec2 uv) {
            float dist = distance(uv, vec2(0.5, 0.5));
            return smoothstep(vignette.x, vignette.x + vignette.y, dist * vignette.z);
        }

        void main(void) {
            float iris = IrisMask(sampleCenter, ${Blur.IrisAreaMode}, ${Blur.IrisInfo}, ${Blur.IrisTransition});

            vec4 blurResult = Blur(textureSampler, sampleCenter, ${MultiBlendPP.BlurMode}, ${Blur.IrisInfo}.xy, ${MultiBlendPP.BlurRadius}, iris);

            float vignetteResult = Vignette(${MultiBlendPP.Vignette}, sampleCenter);

            blurResult.rgb = ${MultiBlendPP.VignetteColor} * vignetteResult + blurResult.rgb * (1.0 - vignetteResult);

            gl_FragColor = blurResult;
        }
        `;
    }
}