
import { PostProcessDataManager } from "./data_manager";
import { PostProcess } from "./postprocess";
import { RenderTexture, RenderTarget } from "./render_target";
import { ShaderCompiler } from "./shader_compiler";
import { ManagedWebGLRenderingContext } from "./webgl";

export class PointCloud extends PostProcess {
    public static shaderName = `PointCloud`;
    public static getShader(context: ManagedWebGLRenderingContext) {
        let manager = PostProcessDataManager.get(context);
        let shaderKey = `${PointCloud.shaderName}`;
        let result = manager.getShaderCompiler(shaderKey);
        if (!result) {
            result = new ShaderCompiler(context, PointCloud.getVertexString(), PointCloud.getFragmentString());
            manager.setShaderCompiler(shaderKey, result);
        }

        return result;
    }

    private static TextureSize = 'textureSize';
    private static Time = 'time';
    /**
     * 模糊半径 - 像素值
     */
    blurRadius: number = 10;
    radialCenter: [number, number] = [0.5, 0.5];

    private iteration = 20;

    public apply(sources: RenderTexture[], target: RenderTarget): ShaderCompiler {
        super.apply(sources, target);

        const gl = this.canvasOrContext.gl;
        let shader = PointCloud.getShader(this.canvasOrContext);
        shader.bind();

        gl.activeTexture(gl.TEXTURE0);
        gl.bindTexture(gl.TEXTURE_2D, sources[0]._webGLTexture);
        shader.setUniformi(ShaderCompiler.TEXTURE_SAMPLER, 0);

        shader.setUniform2f(PointCloud.TextureSize, sources[0].width, sources[0].height);
        shader.setUniformf(PointCloud.Time, Date.now() % 5000 / 5000);

        PostProcess.bindData(this.canvasOrContext, this.manager.getDataBufferCfg(PostProcessDataManager.baseDataBufferCfgKey), shader);

        return shader;
    }

    protected static getFragmentString() {
        return `

        #define SHADER_NAME fragment:pointcloud
        precision highp float;

        varying vec2 sampleCenter;
        uniform vec2 textureSize;
        uniform sampler2D textureSampler;

        uniform float time;

        uniform vec2 radialCenter;

        vec4 RadialBlur(sampler2D tex, vec2 uv) {
            float size = sin(time * 3.14159);

            vec2 lips = 10.0 / textureSize;

            uv = (uv - vec2(0.5, 0.5)) / 10.0 + vec2(0.5, 0.5);

            vec2 count = floor((uv - vec2(0.5, 0.5)) / lips / 10.0);

            vec2 sizeLips = vec2(size, size);

            vec2 newUV = uv;

            float flag = abs(floor(newUV.x / lips.x) * lips.x - newUV.x) < (0.5 * size) ? 1.0 : 0.0;

            vec4 color = texture2D(textureSampler, newUV);

            color *= flag;

            return color;
        }

        void main(void) {
            gl_FragColor = RadialBlur(textureSampler, sampleCenter);
        }
        `;
    }
}