
import { BlendMode } from "./base";
import { DataBufferCfg } from "./data_buffer";
import { PostProcessDataManager } from "./data_manager";
import { IPostprocessParam, IPostprocessTarget, IPostprocessTexture, KeyPostprocess } from "../interface/postprocess";
import { RenderTarget, RenderTexture } from "./render_target";
import { ShaderCompiler } from "./shader_compiler";
import { ManagedWebGLRenderingContext } from "./webgl";

export class PostProcess {
    private static PostProcessMgr: Map<KeyPostprocess, PostProcess> = new Map();
    public static regist(key: KeyPostprocess, item: PostProcess) {
        PostProcess.PostProcessMgr.set(key, item);
    }
    public static get(key: KeyPostprocess) {
        return PostProcess.PostProcessMgr.get(key);
    }
    canvasOrContext: ManagedWebGLRenderingContext;
    public get manager(): PostProcessDataManager {
        return PostProcessDataManager.get(this.canvasOrContext);
    }
    private _disposed: boolean = false;

    constructor() {
    }

    public apply(param: IPostprocessParam, sources: IPostprocessTexture[], target: IPostprocessTarget, ctx: ManagedWebGLRenderingContext) {
        this.canvasOrContext = ctx;
        let _context = this.canvasOrContext;
        _context.gl.bindFramebuffer(_context.gl.FRAMEBUFFER, target.framebuffer || null);
        _context.gl.viewport(target.x, target.y, target.width, target.height);
        _context.gl.scissor(target.x, target.y, target.width, target.height);
    }

    public afterDraw() {
        //
    }

    public draw(dataBufferCfg: DataBufferCfg) {
        PostProcess.drawData(this.canvasOrContext, dataBufferCfg);
    }

    protected static getVertexString() {
        return `
        #define SHADER_NAME vertex:copy
        precision highp float;

        attribute vec3 position;

        varying vec2 sampleCenter;
        const vec2 madd = vec2(0.5, 0.5);
        void main(void) {
            sampleCenter = (position.xy*madd+madd);
            gl_Position = vec4(position.xy, 0.0, 1.0);
        }
        `;
    }

    protected static getFragmentString() {
        return ``;
    }

    public dispose() {
        this._disposed = true;
    }

    public static glBindFrame(context: ManagedWebGLRenderingContext, targetRT: RenderTarget | null) {
        let gl = context.gl;

        gl.bindFramebuffer(gl.FRAMEBUFFER, targetRT?.framebuffer || null);
    }

    public static glViewport(context: ManagedWebGLRenderingContext, viewport: [number, number, number, number], scissor: [number, number, number, number] | null) {
        let gl = context.gl;

        gl.viewport(...viewport);

        if (scissor) {
            gl.enable(gl.SCISSOR_TEST);
            gl.scissor(...scissor);
        } else {
            gl.disable(gl.SCISSOR_TEST);
        }
    }

    public static glClear(context: ManagedWebGLRenderingContext, clearColor: [number, number, number, number] | null, clearDepth: boolean, clearStencil: boolean) {
        let gl = context.gl;

        let flag = 0;
        if (clearColor) {
            gl.clearColor(...clearColor);
            flag = flag | gl.COLOR_BUFFER_BIT;
        }
        if (clearDepth) {
            flag = flag | gl.DEPTH_BUFFER_BIT;
        }
        if (clearStencil) {
            flag = flag | gl.STENCIL_BUFFER_BIT;
        }
        if (flag > 0) {
            gl.clear(flag);
        }
    }

    public static bindData(context: ManagedWebGLRenderingContext, dataBufferCfg: DataBufferCfg, shader: ShaderCompiler) {
        let gl = context.gl;

        gl.bindBuffer(gl.ARRAY_BUFFER, <WebGLBuffer>dataBufferCfg.float_buffer);
        let location = <number>shader.getAttributeLocation(ShaderCompiler.POSITION);
        gl.vertexAttribPointer(location,
                                    dataBufferCfg.VERTEX_SIZE,
                                    gl.FLOAT,
                                    false,
                                    dataBufferCfg.FLOAT_SIZE * 4,
                                    dataBufferCfg.vertex_offset * 4
                                );

        gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER, <WebGLBuffer>dataBufferCfg.int_buffer);
    }

    public static drawData(context: ManagedWebGLRenderingContext, dataBufferCfg: DataBufferCfg) {
        let gl = context.gl;
        gl.drawElements(gl.TRIANGLES,
                            dataBufferCfg.face_data.length * dataBufferCfg.FACE_SIZE,
                            gl.UNSIGNED_SHORT,
                            dataBufferCfg.face_offset * 2
                        );
    }

    /**
     * 通用的屏幕绘制前的操作
     * @param context 上下文
     */
    public static beforeScreenDraw(context: ManagedWebGLRenderingContext) {
        let gl = context.gl;

        gl.disable(gl.CULL_FACE);
        // gl.disable(gl.COLOR_WRITEMASK);
        gl.disable(gl.DEPTH_TEST);
        gl.bindBuffer(gl.ARRAY_BUFFER, null);
        gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER, null);
        // gl.bindVertexArray(null);
    }

    public static blend(context: ManagedWebGLRenderingContext, mode: number) {
        if (mode > 0) {
            context.gl.enable(context.gl.BLEND);
            context.alphaBlendState(mode);
        } else {
            context.gl.disable(context.gl.BLEND);
        }
    }
}