
import { PostProcessDataManager } from "./data_manager";
import { IColorEffect, KeyColorEffect } from "../interface/postprocess";
import { PostProcess } from "./postprocess";
import { RenderTexture, RenderTarget } from "./render_target";
import { ShaderCompiler } from "./shader_compiler";
import { ManagedWebGLRenderingContext } from "./webgl";

export class PS extends PostProcess {
    public static shaderName = KeyColorEffect;
    public static getShader(context: ManagedWebGLRenderingContext) {
        let manager = PostProcessDataManager.get(context);
        let shaderKey = KeyColorEffect;
        let result = manager.getShaderCompiler(shaderKey);
        if (!result) {
            result = new ShaderCompiler(context, PostProcess.getVertexString(), PS.getFragmentString());
            manager.setShaderCompiler(shaderKey, result);
        }

        return result;
    }

    public static HSV = 'u_hsv';
    public static ColorBalance = 'u_cb';
    public static ColorScaleIn = 'u_csi';
    public static ColorScaleOut = 'u_cso';

    public apply(param: IColorEffect, sources: RenderTexture[], target: RenderTarget, ctx: ManagedWebGLRenderingContext) {
        super.apply(param, sources, target, ctx);

        const gl = this.canvasOrContext.gl;
        let shader = PS.getShader(this.canvasOrContext);
        shader.bind();

        gl.activeTexture(gl.TEXTURE0);
        gl.bindTexture(gl.TEXTURE_2D, sources[0]._webGLTexture);
        shader.setUniformi(ShaderCompiler.TEXTURE_SAMPLER, 0);

        shader.setUniform3f(PS.HSV, param.hue / 360, param.saturate / 100, param.brightness / 100);

        let centerCB = Math.max(Math.min(param.colorBalanceR, param.colorBalanceG), Math.min(Math.max(param.colorBalanceR, param.colorBalanceG), param.colorBalanceB));
        let cbR = Math.pow(0.5, ((param.colorBalanceR - centerCB) / 100));
        let cbG = Math.pow(0.5, ((param.colorBalanceG - centerCB) / 100));
        let cbB = Math.pow(0.5, ((param.colorBalanceB - centerCB) / 100));
        shader.setUniform3f(PS.ColorBalance, cbR, cbG, cbB);

        let inShadow = Math.min(param.colorScaleRInShadow, param.colorScaleRInHighlight) / 255;
        let inHightlight = Math.max(param.colorScaleRInShadow, param.colorScaleRInHighlight) / 255;
        let mid = Math.max(0.01, Math.min(9.99, param.colorScaleRInMid));

        shader.setUniform3f(PS.ColorScaleIn, inShadow, 1.0 / mid, inHightlight);
        let outShadow = Math.min(param.colorScaleROutShadow, param.colorScaleROutHighlight) / 255;
        let outHightlight = Math.max(param.colorScaleROutShadow, param.colorScaleROutHighlight) / 255;
        shader.setUniform3f(PS.ColorScaleOut, outShadow, outHightlight, 0.0);

        let dataBuffer = this.manager.getDataBufferCfg(PostProcessDataManager.baseDataBufferCfgKey);
        PostProcess.bindData(this.canvasOrContext, dataBuffer, shader);
        PostProcess.drawData(this.canvasOrContext, dataBuffer);
        this.afterDraw();
    }

    protected static getFragmentString() {
        return `

        #define SHADER_NAME fragment:${PS.shaderName}
        precision highp float;

        varying vec2 sampleCenter;
        uniform sampler2D textureSampler;
        uniform vec3 ${PS.HSV};
        uniform vec3 ${PS.ColorBalance};
        uniform vec3 ${PS.ColorScaleIn};
        uniform vec3 ${PS.ColorScaleOut};

        vec3 rgb2hsv(vec3 c)
        {
            vec4 K = vec4(0.0, -1.0 / 3.0, 2.0 / 3.0, -1.0);
            vec4 p = mix(vec4(c.bg, K.wz), vec4(c.gb, K.xy), step(c.b, c.g));
            vec4 q = mix(vec4(p.xyw, c.r), vec4(c.r, p.yzx), step(p.x, c.r));

            float d = q.x - min(q.w, q.y);
            float e = 1.0e-10;
            return vec3(abs(q.z + (q.w - q.y) / (6.0 * d + e)), d / (q.x + e), q.x);
        }

        vec3 hsv2rgb(vec3 c)
        {
            vec4 K = vec4(1.0, 2.0 / 3.0, 1.0 / 3.0, 3.0);
            vec3 p = abs(fract(c.xxx + K.xyz) * 6.0 - K.www);
            return c.z * mix(K.xxx, clamp(p - K.xxx, vec3(0.0), vec3(1.0)), c.y);
        }

        vec3 colorBalance(vec3 rgb, vec3 cb) {
            return pow(rgb, cb);
        }

        vec3 applyHSV(vec3 c, vec3 pHSV) {
            vec3 hsv = rgb2hsv(c.rgb);
            hsv.r += pHSV.r;
            c.rgb = hsv2rgb(hsv);

            // Note: When saturate is greater than 0, the formula is different from PS
            float gray = max(c.r, max(c.g, c.b)) + min(c.r, min(c.g, c.b));
            c.rgb = mix(c.rgb, vec3(0.5 * gray), -pHSV.g);

            if (pHSV.b >= 0.0) {
                c.rgb = mix(c.rgb, vec3(1.0), pHSV.b);
            } else {
                c.rgb *= 1.0 + pHSV.b;
            }

            return c;
        }

        vec3 applyColorScale(vec3 rgb, vec3 csInS, vec3 csInM, vec3 csInH, vec3 csOutS, vec3 csOutH) {
            vec3 csD = csInH - csInS;
            rgb = clamp((rgb - csInS) / csD, 0.0, 1.0);
            rgb = pow(rgb, csInM);
            csD = csOutH - csOutS;
            rgb = clamp(rgb * csD + csOutS, 0.0, 1.0);
            return rgb;
        }

        void main(void) {
            vec4 c = texture2D(textureSampler, sampleCenter);

            c.rgb = colorBalance(c.rgb, ${PS.ColorBalance});

            c.rgb = applyHSV(c.rgb, ${PS.HSV});

            c.rgb = applyColorScale(c.rgb, vec3(${PS.ColorScaleIn}.x), vec3(${PS.ColorScaleIn}.y), vec3(${PS.ColorScaleIn}.z), vec3(${PS.ColorScaleOut}.x), vec3(${PS.ColorScaleOut}.y));

            gl_FragColor = c;
        }
        `;
    }
}