
import { Blur, EIrisAreaMode } from "./blur";
import { PostProcessDataManager } from "./data_manager";
import { IRadialBlur, KeyRadialBlur } from "../interface/postprocess";
import { PostProcess } from "./postprocess";
import { RenderTarget, RenderTexture } from "./render_target";
import { ShaderCompiler } from "./shader_compiler";
import { ManagedWebGLRenderingContext } from "./webgl";

export class RadialBlur extends Blur {
    public static shaderName = KeyRadialBlur;
    public static getShader(context: ManagedWebGLRenderingContext, iteration: number, irisAreaMode: EIrisAreaMode) {
        let manager = PostProcessDataManager.get(context);
        let shaderKey = `${RadialBlur.shaderName}_${iteration}_${irisAreaMode}`;
        let result = manager.getShaderCompiler(shaderKey);
        if (!result) {
            result = new ShaderCompiler(context, RadialBlur.getVertexString(), RadialBlur.getFragmentString(iteration, irisAreaMode));
            manager.setShaderCompiler(shaderKey, result);
        }

        return result;
    }

    private static BlurRadius = 'blurRadius';
    private static RadialCenter = 'radialCenter';

    public apply(param: IRadialBlur, sources: RenderTexture[], target: RenderTarget, ctx: ManagedWebGLRenderingContext) {
        super.apply(param, sources, target, ctx);

        const gl = this.canvasOrContext.gl;
        let shader = RadialBlur.getShader(this.canvasOrContext, param.iteration, EIrisAreaMode.Circle);
        shader.bind();

        gl.activeTexture(gl.TEXTURE0);
        gl.bindTexture(gl.TEXTURE_2D, sources[0]._webGLTexture);
        shader.setUniformi(ShaderCompiler.TEXTURE_SAMPLER, 0);

        shader.setUniformf(RadialBlur.BlurRadius, param.blurRadius / sources[0].width);
        shader.setUniform2f(RadialBlur.RadialCenter, param.centerX, param.centerY);
        this.applyIris(shader);

        let dataBuffer = this.manager.getDataBufferCfg(PostProcessDataManager.baseDataBufferCfgKey);
        PostProcess.bindData(this.canvasOrContext, dataBuffer, shader);
        PostProcess.drawData(this.canvasOrContext, dataBuffer);
        this.afterDraw();
    }

    protected static getFragmentString(iteration: number = 6, irisAreaMode: EIrisAreaMode = EIrisAreaMode.Circle) {
        return `

        #define SHADER_NAME fragment:radialblur
        precision highp float;

        varying vec2 sampleCenter;

        uniform sampler2D textureSampler;
        uniform float blurRadius;
        uniform vec2 radialCenter;
        ${Blur.IrisFragmentDefined()}

        ${Blur.GetIrisMaskFragmentDefined(irisAreaMode)}

        vec4 RadialBlur(sampler2D tex, vec2 uv) {
            float iris = ${Blur.GetIrisMaskFragmentCall(irisAreaMode, 'uv')};
            if (iris < 0.0001) return texture2D(tex, uv.xy);

            vec2 blurVector = (radialCenter - uv.xy) * blurRadius * iris;

            vec4 acumulateColor = vec4(0.0);

            for (int j = 0; j < ${iteration}; j++)
            {
                acumulateColor += texture2D(tex, uv.xy);
                uv.xy += blurVector;
            }

            return acumulateColor / ${iteration}.0;
        }

        void main(void) {
            gl_FragColor = RadialBlur(textureSampler, sampleCenter);
        }
        `;
    }
}