
import { PostProcessDataManager } from "./data_manager";
import { IRadiusWave } from "../interface/postprocess";
import { PostProcess } from "./postprocess";
import { RenderTexture, RenderTarget } from "./render_target";
import { ShaderCompiler } from "./shader_compiler";
import { ManagedWebGLRenderingContext } from "./webgl";

export class RadiusWave extends PostProcess {
    public static shaderName = `RadiusWave`;
    public static getShader(context: ManagedWebGLRenderingContext) {
        let manager = PostProcessDataManager.get(context);
        let shaderKey = `${RadiusWave.shaderName}`;
        let result = manager.getShaderCompiler(shaderKey);
        if (!result) {
            result = new ShaderCompiler(context, RadiusWave.getVertexString(), RadiusWave.getFragmentString());
            manager.setShaderCompiler(shaderKey, result);
        }

        return result;
    }

    private static WaveInfo = 'waveInfo';
    private static ScreenInfo = 'screenInfo';

    public apply(param: IRadiusWave, sources: RenderTexture[], target: RenderTarget, ctx: ManagedWebGLRenderingContext): ShaderCompiler {
        super.apply(param, sources, target, ctx);

        const gl = this.canvasOrContext.gl;
        let shader = RadiusWave.getShader(this.canvasOrContext);
        shader.bind();

        gl.activeTexture(gl.TEXTURE0);
        gl.bindTexture(gl.TEXTURE_2D, sources[0]._webGLTexture);
        shader.setUniformi(ShaderCompiler.TEXTURE_SAMPLER, 0);

        // const radius = this.blurRadius / this.sourceWidth / this.iteration;
        // this.direction[0] = Math.sin(this._angle) * radius ;
        // this.direction[1] = Math.cos(this._angle) * radius ;
        // this.shader.setUniform2f(RadiusWave.Direction, this.direction[0], this.direction[1]);

        shader.setUniform2f(RadiusWave.ScreenInfo, param.asCircle ? sources[0].width : 1.0, param.asCircle ? sources[0].height : 1.0);
        shader.setUniform4f(RadiusWave.WaveInfo, Math.min(param.waveStartRadius, param.waveEndRadius), Math.max(param.waveStartRadius, param.waveEndRadius), param.waveCounter, param.waveWeight);

        let dataBuffer = this.manager.getDataBufferCfg(PostProcessDataManager.baseDataBufferCfgKey);
        PostProcess.bindData(this.canvasOrContext, dataBuffer, shader);
        PostProcess.drawData(this.canvasOrContext, dataBuffer);
        this.afterDraw();
    
        return shader;
    }

    protected static getFragmentString() {
        return `
        #define SHADER_NAME fragment:radial_wave
        precision highp float;

        varying vec2 sampleCenter;
        uniform sampler2D textureSampler;

        // uniform vec2 direction;
        uniform vec2 ${RadiusWave.ScreenInfo};
        uniform vec4 ${RadiusWave.WaveInfo};

        vec4 RadialBlur(sampler2D tex, vec2 uv) {

            vec4 acumulateColor = vec4(0.0);

            vec2 center = uv * 2.0 - 1.0; // [0, 1] -> [-1, 1]

            center.y *= ${RadiusWave.ScreenInfo}.y / ${RadiusWave.ScreenInfo}.x;

            float len = abs(length(center));

            float temp = (${RadiusWave.WaveInfo}.y - ${RadiusWave.WaveInfo}.x) / 2.0;
            float temp2 = min(temp, 0.1);

            float weight = ${RadiusWave.WaveInfo}.w * smoothstep(${RadiusWave.WaveInfo}.x, ${RadiusWave.WaveInfo}.x + temp2, len) * (1.0 - smoothstep(${RadiusWave.WaveInfo}.y - temp, ${RadiusWave.WaveInfo}.y, len));
            // weight = weight < 0.0 ? 0.0 : weight;

            len = len < ${RadiusWave.WaveInfo}.x ? 0.0 : (len > ${RadiusWave.WaveInfo}.y ? 0.0 : len - ${RadiusWave.WaveInfo}.x);
            len = len / (${RadiusWave.WaveInfo}.y - ${RadiusWave.WaveInfo}.x) * ${RadiusWave.WaveInfo}.z;

            float diff = weight * sin(len * 3.141592653589793);

            uv += diff;

            return texture2D(tex, uv);
        }

        void main(void) {
            gl_FragColor = RadialBlur(textureSampler, sampleCenter);
        }
        `;
    }
}