
import { PostProcessDataManager } from "./data_manager";
import { PostProcess } from "./postprocess";
import { RenderTarget, RenderTexture } from "./render_target";
import { ShaderCompiler } from "./shader_compiler";
import { ManagedWebGLRenderingContext } from "./webgl";

export class SobelEdge extends PostProcess {
    public static shaderName = `SobelEdge`;
    public static getShaderEdge(context: ManagedWebGLRenderingContext) {
        let manager = PostProcessDataManager.get(context);
        let shaderKey = `${SobelEdge.shaderName}`;
        let result = manager.getShaderCompiler(shaderKey);
        if (!result) {
            result = new ShaderCompiler(context, SobelEdge.getVertexString(), SobelEdge.getFragmentEdgeString());
            manager.setShaderCompiler(shaderKey, result);
        }

        return result;
    }

    public static shaderNameCombine = `SobelCombine`;
    public static getShaderCombine(context: ManagedWebGLRenderingContext) {
        let manager = PostProcessDataManager.get(context);
        let shaderKey = `${SobelEdge.shaderNameCombine}`;
        let result = manager.getShaderCompiler(shaderKey);
        if (!result) {
            result = new ShaderCompiler(context, SobelEdge.getVertexString(), SobelEdge.getCombineFragmentString());
            manager.setShaderCompiler(shaderKey, result);
        }

        return result;
    }

    public static readonly maxBlurRenderIteration = 4;
    public checkSize: number = 1;
    public clip: number = 0.0;
    public edgeColorR: number = 1;
    public edgeColorG: number = 1;
    public edgeColorB: number = 1;

    public apply(sources: RenderTexture[], target: RenderTarget): ShaderCompiler {
        super.apply(sources, target);

        throw new Error(`Not Work, use applyEdge() and applyCombineEdge().`);
    }

    /**
     * 视口 - fbo 绑定 - fbo 清理 均由调用方处理, 接口只关注自身功能
     * @param sourceRT 源 RT
     * @param targetRT 面板RT - 接收边缘信息 白色为边缘
     */
    public applyEdge(sourceRT: RenderTexture): ShaderCompiler {
        const gl = this.canvasOrContext.gl;
        let shader = SobelEdge.getShaderEdge(this.canvasOrContext);
        shader.bind();

        gl.activeTexture(gl.TEXTURE0);
        gl.bindTexture(gl.TEXTURE_2D, sourceRT._webGLTexture);
        shader.setUniformi(ShaderCompiler.TEXTURE_SAMPLER, 0);

        shader.setUniform2f(`uDiffUV`, this.checkSize / sourceRT.width, this.checkSize / sourceRT.height);

        PostProcess.bindData(this.canvasOrContext, this.manager.getDataBufferCfg(PostProcessDataManager.baseDataBufferCfgKey), shader);

        return shader;
    }

    /**
     * 合成 边缘和源
     * @param sourceRT 源图像
     * @param edgeDataRT 边缘信息
     * @param targetRT 渲染结果
     */
    public applyCombineEdge(sourceRT: RenderTarget, edgeDataRT: RenderTarget): ShaderCompiler {
        const gl = this.canvasOrContext.gl;
        let shader = SobelEdge.getShaderCombine(this.canvasOrContext);
        shader.bind();

        gl.activeTexture(gl.TEXTURE0);
        gl.bindTexture(gl.TEXTURE_2D, sourceRT._webGLTexture);
        shader.setUniformi(ShaderCompiler.TEXTURE_SAMPLER, 0);

        gl.activeTexture(gl[`TEXTURE1`]);
        gl.bindTexture(gl.TEXTURE_2D, edgeDataRT._webGLTexture);
        shader.setUniformi(`textureSampler${0 + 1}`, 0 + 1);

        shader.setUniform3f(`uEdgeColor`, this.edgeColorR, this.edgeColorG, this.edgeColorB);
        shader.setUniformf(`uClip`, this.clip);

        PostProcess.bindData(this.canvasOrContext, this.manager.getDataBufferCfg(PostProcessDataManager.baseDataBufferCfgKey), shader);

        return shader;
    }

    protected static getCombineFragmentString() {
        return `
        #define SHADER_NAME fragment:CombineEdgeAndSource
        precision highp float;

        uniform sampler2D textureSampler;
        uniform sampler2D textureSampler1;

        uniform vec3 uEdgeColor;
        uniform float uClip;

        varying vec2 sampleCenter;

        void main(void)
        {
            vec4 color =  texture2D( textureSampler, sampleCenter );
            vec4 color1 =  texture2D( textureSampler1, sampleCenter );
            float edge = clamp(color1.r, 0., 1.);
            edge = step(uClip, edge) * edge;
            gl_FragColor = vec4(uEdgeColor * edge + color.rgb * (1.0 - edge), color.a);
        }
        `;
    }

    protected static getFragmentEdgeString() {
        return `
        #define SHADER_NAME fragment:RGB2Edge
        precision highp float;

        uniform sampler2D textureSampler;
        varying vec2 sampleCenter;

        uniform vec2 uDiffUV;

        const vec3 S0 = vec3(1., 2., 1.);
        const vec3 S1 = vec3(1., 0., -1.);

        float rgb2gray(vec3 rgb) {
            return 0.2126 * rgb.r + 0.7150 * rgb.g + 0.0722 * rgb.b;
        }

        void main(void)
        {
            float g00 =  rgb2gray(texture2D( textureSampler, sampleCenter + uDiffUV * vec2(-1., -1.)).rgb);
            float g01 =  rgb2gray(texture2D( textureSampler, sampleCenter + uDiffUV * vec2(-0., -1.)).rgb);
            float g02 =  rgb2gray(texture2D( textureSampler, sampleCenter + uDiffUV * vec2( 1., -1.)).rgb);
            float g10 =  rgb2gray(texture2D( textureSampler, sampleCenter + uDiffUV * vec2(-1., -0.)).rgb);
            float g11 =  0.0;
            float g12 =  rgb2gray(texture2D( textureSampler, sampleCenter + uDiffUV * vec2( 1., -0.)).rgb);
            float g20 =  rgb2gray(texture2D( textureSampler, sampleCenter + uDiffUV * vec2(-1.,  1.)).rgb);
            float g21 =  rgb2gray(texture2D( textureSampler, sampleCenter + uDiffUV * vec2( 0.,  1.)).rgb);
            float g22 =  rgb2gray(texture2D( textureSampler, sampleCenter + uDiffUV * vec2( 1.,  1.)).rgb);

            mat3 a = mat3(
                g00, g01, g02,
                g10, g11, g12,
                g21, g21, g22
            );
            float gx = dot(S0, (S1 * a));
            float gy = dot(S1, (S0 * a));

            float g = sqrt(gx * gx + gy * gy);

            gl_FragColor = vec4(g, g, g, 1.0);
        }
        `;
    }
}