
import { PostProcessDataManager } from "./data_manager";
import { ITransition, KeyTransition } from "../interface/postprocess";
import { PostProcess } from "./postprocess";
import { RenderTexture, RenderTarget } from "./render_target";
import { ShaderCompiler } from "./shader_compiler";
import { ManagedWebGLRenderingContext } from "./webgl";

export class Transition extends PostProcess {
    public static shaderName = KeyTransition;
    public static getShader(context: ManagedWebGLRenderingContext) {
        let manager = PostProcessDataManager.get(context);
        let shaderKey = `${Transition.shaderName}`;
        let result = manager.getShaderCompiler(shaderKey);
        if (!result) {
            result = new ShaderCompiler(context, Transition.getVertexString(), Transition.getFragmentString());
            manager.setShaderCompiler(shaderKey, result);
        }

        return result;
    }

    private static Transition = 'uTransition';

    public apply(param: ITransition, sources: RenderTexture[], target: RenderTarget, ctx: ManagedWebGLRenderingContext): ShaderCompiler {
        super.apply(param, sources, target, ctx);

        const gl = this.canvasOrContext.gl;
        let shader = Transition.getShader(this.canvasOrContext);
        shader.bind();

        gl.activeTexture(gl.TEXTURE0);
        gl.bindTexture(gl.TEXTURE_2D, sources[0]._webGLTexture);
        shader.setUniformi(ShaderCompiler.TEXTURE_SAMPLER, 0);

        shader.setUniformf(Transition.Transition, param.transition);

        let dataBuffer = this.manager.getDataBufferCfg(PostProcessDataManager.baseDataBufferCfgKey);
        PostProcess.bindData(this.canvasOrContext, dataBuffer, shader);
        PostProcess.drawData(this.canvasOrContext, dataBuffer);
        this.afterDraw();
    
        return shader;
    }

    protected static getFragmentString() {
        return `
        #define SHADER_NAME fragment:radial_wave
        precision highp float;

        varying vec2 sampleCenter;
        uniform sampler2D textureSampler;

        uniform float ${Transition.Transition};

        float easeOutBack(float x) {
            float c1 = 1.70158;
            float c3 = c1 + 1.;

            return 1. + c3 * pow(x - 1., 3.) + c1 * pow(x - 1., 2.);
        }
        vec4 Transition(sampler2D tex, vec2 uv) {

            vec4 acumulateColor = vec4(0.0);

            vec2 center = uv * 2.0 - 1.0; // [0, 1] -> [-1, 1]

            float len = abs(length(center));

            float t = ${Transition.Transition};
            float a = min(1., t * 2.);
            float b = min(1., max(0., (t * 2. - 1.)));
            b = easeOutBack(b);

            vec2 scl0 = (center / (1.0 + a*a*a * 10. * len) + 1.0 ) * 0.5;

            float c = len; // max(abs(center.x), abs(center.y));
            // c = (2. - c);
            c = pow(c - 1., 5.) + 1.;
            vec2 scl1 = mix(center, normalize(center) * c, 1.-b);
            scl1 = ( scl1 + 1. ) * 0.5;
            uv = mix(scl0, mix(scl1, uv, step(1., t)), step(0.5, t));
            return texture2D(tex, uv);
        }

        void main(void) {
            gl_FragColor = Transition(textureSampler, sampleCenter);
        }
        `;
    }
}