
import { PostProcessDataManager } from "./data_manager";
import { IVignette } from "../interface/postprocess";
import { PostProcess } from "./postprocess";
import { RenderTexture, RenderTarget } from "./render_target";
import { ShaderCompiler } from "./shader_compiler";
import { ManagedWebGLRenderingContext } from "./webgl";

export enum EBlurMode {
    None = 0,
    BokehBlur = 1,
    DirectionalBlur = 2,
    RadialBlur = 3
}

export class Vignette extends PostProcess {
    public static shaderName = `Vignette`;
    public static getShader(context: ManagedWebGLRenderingContext) {
        let manager = PostProcessDataManager.get(context);
        let shaderKey = `${Vignette.shaderName}`;
        let result = manager.getShaderCompiler(shaderKey);
        if (!result) {
            result = new ShaderCompiler(context, Vignette.getVertexString(), Vignette.getFragmentString());
            manager.setShaderCompiler(shaderKey, result);
        }

        return result;
    }

    // https://evanw.github.io/glfx.js/demo/ - Vignette
    private static Vignette = 'vignette';
    private static VignetteColor = 'vignetteColor';

    public apply(param: IVignette, sources: RenderTexture[], target: RenderTarget, ctx: ManagedWebGLRenderingContext): ShaderCompiler {
        super.apply(param, sources, target, ctx);

        const gl = this.canvasOrContext.gl;
        let shader = Vignette.getShader(this.canvasOrContext);
        shader.bind();

        gl.activeTexture(gl.TEXTURE0);
        gl.bindTexture(gl.TEXTURE_2D, sources[0]._webGLTexture);
        shader.setUniformi(ShaderCompiler.TEXTURE_SAMPLER, 0);

        shader.setUniform4f(Vignette.Vignette, param.begin, param.end, param.scale, 0.0);
        shader.setUniform3f(Vignette.VignetteColor, param.vignetteColor[0], param.vignetteColor[1], param.vignetteColor[2]);

        PostProcess.bindData(this.canvasOrContext, this.manager.getDataBufferCfg(PostProcessDataManager.baseDataBufferCfgKey), shader);

        return shader;
    }

    protected static getFragmentString() {
        return `
        #define SHADER_NAME fragment:directinalblur
        precision highp float;

        #define GLODEN_COS -0.7373688782616119
        #define GLODEN_SIN 0.675490294061441

        varying vec2 sampleCenter;

        uniform sampler2D textureSampler;

        uniform vec4 ${Vignette.Vignette};
        uniform vec3 ${Vignette.VignetteColor};

        float Vignette(vec4 vignette, vec2 uv) {
            float dist = distance(uv, vec2(0.5, 0.5));
            return smoothstep(vignette.x, vignette.y, dist * vignette.z);
        }

        void main(void) {
            vec4 blurResult = texture2D(textureSampler, sampleCenter);

            float vignetteResult = Vignette(${Vignette.Vignette}, sampleCenter);

            blurResult.rgb = ${Vignette.VignetteColor} * vignetteResult + blurResult.rgb * (1.0 - vignetteResult);

            gl_FragColor = blurResult;
        }
        `;
    }
}