import { IRenderData } from "./base";
import { PostProcessDataManager } from "./data_manager";
import { PostProcess } from "./postprocess";
import { RenderTexture, RenderTarget } from "./render_target";
import { ShaderCompiler } from "./shader_compiler";
import { ManagedWebGLRenderingContext } from "./webgl";

export class WaveJitter extends PostProcess {
    public static shaderName = `WaveJitter`;
    public static getShader(context: ManagedWebGLRenderingContext) {
        let manager = PostProcessDataManager.get(context);
        let shaderKey = `${WaveJitter.shaderName}`;
        let result = manager.getShaderCompiler(shaderKey);
        if (!result) {
            result = new ShaderCompiler(context, WaveJitter.getVertexString(), WaveJitter.getFragmentString());
            manager.setShaderCompiler(shaderKey, result);
        }

        return result;
    }

    private static Frequency = 'frequency';
    private static RGBSplit = 'rgbSplit';
    private static Speed = 'speed';
    private static Amount = 'amount';
    private static TextureSize = 'textureSize';
    private static Time = 'time';

    frequency: number;
    rgbSplit: number;
    speed: number;
    amount: number;
    time: number = 0.1;
    textureSize: [number, number];

    constructor(canvasOrContext: ManagedWebGLRenderingContext) {

        super(canvasOrContext);

        this.frequency = 2.0;
        this.rgbSplit = 30;
        this.speed = 0.3;
        this.amount = 1;
        this.textureSize = [100, 100];
    }

    public apply(sources: RenderTexture[], target: RenderTarget): ShaderCompiler {
        super.apply(sources, target);

        const gl = this.canvasOrContext.gl;
        let shader = WaveJitter.getShader(this.canvasOrContext);
        shader.bind();

        gl.activeTexture(gl.TEXTURE0);
        gl.bindTexture(gl.TEXTURE_2D, sources[0]._webGLTexture);
        shader.setUniformi(ShaderCompiler.TEXTURE_SAMPLER, 0);

        shader.setUniform2f(WaveJitter.TextureSize, sources[0].width, sources[0].height);
        shader.setUniformf(WaveJitter.Frequency, this.frequency);
        shader.setUniformf(WaveJitter.RGBSplit, this.rgbSplit);
        shader.setUniformf(WaveJitter.Speed, this.speed);
        shader.setUniformf(WaveJitter.Amount, this.amount);
        shader.setUniformf(WaveJitter.Time, this.time);

        PostProcess.bindData(this.canvasOrContext, this.manager.getDataBufferCfg(PostProcessDataManager.baseDataBufferCfgKey), shader);

        return shader;
    }

    // https://github.com/QianMo/X-PostProcessing-Library/blob/master/Assets/X-PostProcessing/Effects/GlitchWaveJitter/Shader/GlitchWaveJitter.shader
    protected static getFragmentString() {
        return `
        #define SHADER_NAME fragment:wavejitter
        precision highp float;

        uniform sampler2D textureSampler;

        varying vec2 sampleCenter;
        uniform vec2 textureSize;

        uniform float frequency;
        uniform float rgbSplit;
        uniform float speed;
        uniform float amount;
        uniform float time;

        // https://github.com/QianMo/X-PostProcessing-Library/blob/master/Assets/X-PostProcessing/Shaders/XNoiseLibrary.hlsl
        #define NOISE_SIMPLEX_1_DIV_289 0.00346020761245674

        float mod289(float x)
        {
            return x - floor(x * NOISE_SIMPLEX_1_DIV_289) * 289.0;
        }

        vec2 mod289(vec2 x)
        {
            return x - floor(x * NOISE_SIMPLEX_1_DIV_289) * 289.0;
        }

        vec3 mod289(vec3 x)
        {
            return x - floor(x * NOISE_SIMPLEX_1_DIV_289) * 289.0;
        }

        vec4 mod289(vec4 x)
        {
            return x - floor(x * NOISE_SIMPLEX_1_DIV_289) * 289.0;
        }
        vec3 taylorInvSqrt(vec3 r)
        {
            return 1.79284291400159 - 0.85373472095314 * r;
        }

        vec4 taylorInvSqrt(vec4 r)
        {
            return 1.79284291400159 - r * 0.85373472095314;
        }
        // ( x*34.0 + 1.0 )*x =x*x*34.0 + x
        float permute(float x)
        {
            return mod289(x * x * 34.0 + x);
        }

        vec3 permute(vec3 x)
        {
            return mod289(x * x * 34.0 + x);
        }

        vec4 permute(vec4 x)
        {
            return mod289(x * x * 34.0 + x);
        }
        float snoise(vec2 v)
        {
            const vec4 C = vec4(0.211324865405187, // (3.0-sqrt(3.0))/6.0
            0.366025403784439, // 0.5*(sqrt(3.0)-1.0)
            - 0.577350269189626, // -1.0 + 2.0 * C.x
            0.024390243902439); // 1.0 / 41.0
            // First corner
            vec2 i = floor(v + dot(v, C.yy));
            vec2 x0 = v - i + dot(i, C.xx);

            // Other corners
            vec2 i1;
            i1.x = step(x0.y, x0.x);
            i1.y = 1.0 - i1.x;

            // x1 = x0 - i1  + 1.0 * C.xx;
            // x2 = x0 - 1.0 + 2.0 * C.xx;
            vec2 x1 = x0 + C.xx - i1;
            vec2 x2 = x0 + C.zz;

            // Permutations
            i = mod289(i); // Avoid truncation effects in permutation
            vec3 p = permute(permute(i.y + vec3(0.0, i1.y, 1.0))
            + i.x + vec3(0.0, i1.x, 1.0));

            vec3 m = max(0.5 - vec3(dot(x0, x0), dot(x1, x1), dot(x2, x2)), 0.0);
            m = m * m;
            m = m * m;

            // Gradients: 41 points uniformly over a line, mapped onto a diamond.
            // The ring size 17*17 = 289 is close to a multiple of 41 (41*7 = 287)
            vec3 x = 2.0 * fract(p * C.www) - 1.0;
            vec3 h = abs(x) - 0.5;
            vec3 ox = floor(x + 0.5);
            vec3 a0 = x - ox;

            // Normalise gradients implicitly by scaling m
            m *= taylorInvSqrt(a0 * a0 + h * h);

            // Compute final noise value at P
            vec3 g;
            g.x = a0.x * x0.x + h.x * x0.y;
            g.y = a0.y * x1.x + h.y * x1.y;
            g.z = a0.z * x2.x + h.z * x2.y;
            return 130.0 * dot(m, g);
        }


        vec4 Frag_Horizontal() {
            float strength = 0.0;

            strength = 0.5 + 0.5 * cos(time * frequency);

            // Prepare UV
            float uv_y = sampleCenter.y * textureSize.y;
            float noise_wave_1 = snoise(vec2(uv_y * 0.01, time * speed * 20.0)) * (strength * amount * 32.0);
            float noise_wave_2 = snoise(vec2(uv_y * 0.02, time * speed * 10.0)) * (strength * amount * 4.0);
            float noise_wave_x = noise_wave_1 * noise_wave_2 / textureSize.x;
            float uv_x = sampleCenter.x + noise_wave_x;

            float rgbSplit_uv_x = (rgbSplit * 50.0 + (20.0 * strength + 1.0)) * noise_wave_x / textureSize.x;

            // Sample RGB Color
            vec4 colorG = texture2D(textureSampler, vec2(uv_x, sampleCenter.y));
            vec4 colorRB = texture2D(textureSampler, vec2(uv_x + rgbSplit_uv_x, sampleCenter.y));

            return vec4(colorRB.r, colorG.g, colorRB.b, colorRB.a + colorG.a);
        }

        // vec4 Frag_Vertical() {

        // }

        void main(void)
        {
            gl_FragColor = Frag_Horizontal();
        }
        `;
    }
}