import { ExtensionTools } from "../interface/index";
import { BABYLON } from "../render3d/babylon";

export interface ICommonBufferScene {
    __commonBuffer: CommonBufferManager;
}

export class CommonVertexBuffer extends BABYLON.VertexBuffer {
    private _isDisposed: boolean = false;
    public isDisposed() {
        return this._isDisposed;
    }
    constructor(
        engine: BABYLON.Engine,
        kind: string,
        private _commonBuffer: CommonFloat32Buffer,
        public commonData: Float32Array,
        public readonly _commonIdx: number
    ) {
        super(engine, _commonBuffer.buffer, kind, true, undefined, undefined, false, _commonBuffer.blockSize * _commonIdx);
    }
    commonDirty() {
        // this.updateDirectly(this.commonData, 0)
        this._commonBuffer.dirty();
    }
    dispose(): void {
        this._commonBuffer._recycle(this);
        super.dispose();
    }
}

export class CommonFloat32Buffer {
    private static Record: Map<number, CommonFloat32Buffer> = new Map();
    private static TotalSize: number = 0;
    private static Counter: number = Number.MIN_SAFE_INTEGER;
    private _buffer: BABYLON.Buffer;
    public get buffer() {
        return this._buffer;
    }
    private _data: Float32Array;
    public readonly blockSize: number;
    private idxs: number[] = [];
    private _dirty: boolean = false;
    public readonly uid: number;
    constructor(
        private engine: BABYLON.Engine,
        blockSize: number,
        blockCount: number,
        private uncache: boolean = false,
    ) {
        this.blockSize = blockSize;
        for (let i = blockCount - 1; i >= 0; i--) {
            this.idxs.push(i);
        }
        this._data = new Float32Array(blockSize * blockCount);
        this._buffer = new BABYLON.Buffer(engine, this._data, true);
        this.uid = CommonFloat32Buffer.Counter++;
        CommonFloat32Buffer.Record.set(this.uid, this);
        CommonFloat32Buffer.TotalSize += this._data.byteLength;
    }
    vertexBuffer(kind: string, size: number) {
        if (this.blockSize < size || this.idxs.length == 0) {
            return;
        } else {
            let idx = this.idxs.pop();
            let commonData = new Float32Array(this._data.buffer, idx * this.blockSize * 4, this.blockSize);
            let buffer = new CommonVertexBuffer(this.engine, kind, this, commonData, idx);
            return buffer;
        }
    }
    _recycle(buffer: CommonVertexBuffer) {
        if (buffer.isDisposed() == false) {
            this.idxs.push(buffer._commonIdx);
        }
        if (this.uncache) {
            this.dispose();
        }
    }
    dirty() {
        this._dirty = true;
        if (this.uncache) {
            this.update();
        }
    }
    update() {
        if (this._dirty) {
            this.buffer.updateDirectly(this._data, 0);
            this.buffer._data = this._data;
            // this.buffer._rebuild();
        }
        this._dirty = false;
    }
    dispose() {
        CommonFloat32Buffer.Record.delete(this.uid);
        CommonFloat32Buffer.TotalSize -= this._data.byteLength;
        this._buffer.dispose();
        this._data = undefined;
        this._buffer = undefined;
    }
}

export class CommonBufferManager {
    public static MAX_COMMON_SIZE = 1024 * 64;
    constructor(private scene: BABYLON.Scene) {
        scene.onDisposeObservable.addOnce(() => {
            this.dispose();
        })
    }
    // private pool: Map<String, Map<number, CommonFloat32Buffer[]>> = new Map();
    private pool: Map<number, CommonFloat32Buffer[]> = new Map();

    public createVertexBuffer(kind: string, size: number) {
        let blockSize = Math.ceil(size / 256) * 256;
        let MAX_COMMON_SIZE = ExtensionTools.Option.cacheCustomGeometryDynamicBufferSize || 1;
        let blockCount = Math.ceil(MAX_COMMON_SIZE / blockSize / 4);
        if (blockSize * 4 > MAX_COMMON_SIZE) {
            blockCount = 1;
            let newCommon = new CommonFloat32Buffer(this.scene.getEngine(), blockSize, blockCount, true);
            return newCommon.vertexBuffer(kind, blockSize);
        }

        // let pool = this.pool.get(kind);
        // if (pool == undefined) {
        //     pool = new Map();
        //     this.pool.set(kind, pool);
        // }
        
        let pool = this.pool;
        let list = pool.get(blockSize);
        if (list == undefined) {
            list = [];
            pool.set(blockSize, list);
        }
        let commonCount = list.length;
        let result: CommonVertexBuffer;
        for (let i = 0; i < commonCount; i++) {
            result = list[i].vertexBuffer(kind, blockSize);
            if (result) {
                break;
            }
        }
        if (result == undefined) {
            let newCommon = new CommonFloat32Buffer(this.scene.getEngine(), blockSize, blockCount);
            list.push(newCommon);
            result = newCommon.vertexBuffer(kind, blockSize);
        }

        return result;
    }
    _onFrameUpdate() {
        // this.pool.forEach((pool) => {
        //     pool.forEach((list) => {
        //         list.forEach((item) => {
        //             item.update();
        //         });
        //     });
        // });
        this.pool.forEach((list) => {
            list.forEach((item) => {
                item.update();
            });
        });
    }
    dispose() {
        // this.pool.forEach((pool) => {
        //     pool.forEach((list) => {
        //         list.forEach((item) => {
        //             item.dispose();
        //         });
        //     });
        // });
        this.pool.forEach((list) => {
            list.forEach((item) => {
                item.dispose();
            });
        });
        this.pool.clear();
    }
}
