import { BABYLON } from "../babylon";

export function optBoundingInfo() {
    // BABYLON.BoundingSphere.prototype._update = function(worldMatrix: BABYLON.DeepImmutable<BABYLON.Matrix>): void {
    //     if (!worldMatrix.isIdentity()) {
    //         BABYLON.Vector3.TransformCoordinatesToRef(this.center, worldMatrix, this.centerWorld);
    //         const tempVector = (<any>BABYLON.BoundingSphere).TmpVector3[0];
    //         BABYLON.Vector3.TransformNormalFromFloatsToRef(1.0, 1.0, 1.0, worldMatrix, tempVector);
    //         this.radiusWorld = Math.max(Math.abs(tempVector.x), Math.abs(tempVector.y), Math.abs(tempVector.z)) * this.radius;
    //     }
    //     else {
    //         this.centerWorld.copyFrom(this.center);
    //         this.radiusWorld = this.radius;
    //     }
    //     this._worldMatrix = worldMatrix;
    // }

    // BABYLON.BoundingInfo = <any>BoundingInfo;
}

type DeepImmutable<T> = BABYLON.DeepImmutable<T>;
/**
 * Class used to store bounding sphere information
 */
export class BoundingSphere {
    /**
     * Gets the center of the bounding sphere in local space
     */
    public readonly center = BABYLON.Vector3.Zero();
    /**
     * Radius of the bounding sphere in local space
     */
    public radius: number;
    /**
     * Gets the center of the bounding sphere in world space
     */
    public readonly centerWorld = BABYLON.Vector3.Zero();
    /**
     * Radius of the bounding sphere in world space
     */
    public radiusWorld: number;
    /**
     * Gets the minimum vector in local space
     */
    public readonly minimum = BABYLON.Vector3.Zero();
    /**
     * Gets the maximum vector in local space
     */
    public readonly maximum = BABYLON.Vector3.Zero();

    private _worldMatrix: DeepImmutable<BABYLON.Matrix>;
    private static readonly TmpVector3 = [BABYLON.Vector3.Zero(), BABYLON.Vector3.Zero(), BABYLON.Vector3.Zero()];

    /**
     * Creates a new bounding sphere
     * @param min defines the minimum vector (in local space)
     * @param max defines the maximum vector (in local space)
     * @param worldMatrix defines the new world matrix
     */
    constructor(min: DeepImmutable<BABYLON.Vector3>, max: BABYLON.DeepImmutable<BABYLON.Vector3>, worldMatrix?: DeepImmutable<BABYLON.Matrix>) {
        this.reConstruct(min, max, worldMatrix);
    }

    /**
     * Recreates the entire bounding sphere from scratch as if we call the constructor in place
     * @param min defines the new minimum vector (in local space)
     * @param max defines the new maximum vector (in local space)
     * @param worldMatrix defines the new world matrix
     */
    public reConstruct(min: DeepImmutable<BABYLON.Vector3>, max: DeepImmutable<BABYLON.Vector3>, worldMatrix?: DeepImmutable<BABYLON.Matrix>) {
        this.minimum.copyFrom(min);
        this.maximum.copyFrom(max);

        var distance = BABYLON.Vector3.Distance(min, max);

        max.addToRef(min, this.center).scaleInPlace(0.5);
        this.radius = distance * 0.5;

        this._update(worldMatrix || BABYLON.Matrix.IdentityReadOnly);
    }

    /**
     * Scale the current bounding sphere by applying a scale factor
     * @param factor defines the scale factor to apply
     * @returns the current bounding box
     */
    public scale(factor: number): BoundingSphere {
        const newRadius = this.radius * factor;
        const tmpVectors = BoundingSphere.TmpVector3;
        const tempRadiusVector = tmpVectors[0].setAll(newRadius);
        const min = this.center.subtractToRef(tempRadiusVector, tmpVectors[1]);
        const max = this.center.addToRef(tempRadiusVector, tmpVectors[2]);

        this.reConstruct(min, max, this._worldMatrix);

        return this;
    }

    /**
     * Gets the world matrix of the bounding box
     * @returns a matrix
     */
    public getWorldMatrix(): DeepImmutable<BABYLON.Matrix> {
        return this._worldMatrix;
    }

    // Methods
    /** @hidden */
    public _update(worldMatrix: DeepImmutable<BABYLON.Matrix>): void {
        if (!worldMatrix.isIdentity()) {
            BABYLON.Vector3.TransformCoordinatesToRef(this.center, worldMatrix, this.centerWorld);
            const tempVector = BoundingSphere.TmpVector3[0];
            BABYLON.Vector3.TransformNormalFromFloatsToRef(1.0, 1.0, 1.0, worldMatrix, tempVector);
            this.radiusWorld = Math.max(Math.abs(tempVector.x), Math.abs(tempVector.y), Math.abs(tempVector.z)) * this.radius;
        }
        else {
            this.centerWorld.copyFrom(this.center);
            this.radiusWorld = this.radius;
        }
        this._worldMatrix = worldMatrix;
    }

    /**
     * Tests if the bounding sphere is intersecting the frustum planes
     * @param frustumPlanes defines the frustum planes to test
     * @returns true if there is an intersection
     */
    public isInFrustum(frustumPlanes: Array<DeepImmutable<BABYLON.Plane>>): boolean {
        let center = this.centerWorld;
        let radius = this.radiusWorld;
        for (let i = 0; i < 6; i++) {
            if (frustumPlanes[i].dotCoordinate(center) <= -radius) {
                return false;
            }
        }
        return true;
    }

    /**
     * Tests if the bounding sphere center is in between the frustum planes.
     * Used for optimistic fast inclusion.
     * @param frustumPlanes defines the frustum planes to test
     * @returns true if the sphere center is in between the frustum planes
     */
    public isCenterInFrustum(frustumPlanes: Array<DeepImmutable<BABYLON.Plane>>): boolean {
        let center = this.centerWorld;
        for (let i = 0; i < 6; i++) {
            if (frustumPlanes[i].dotCoordinate(center) < 0) {
                return false;
            }
        }
        return true;
    }

    /**
     * Tests if a point is inside the bounding sphere
     * @param point defines the point to test
     * @returns true if the point is inside the bounding sphere
     */
    public intersectsPoint(point: DeepImmutable<BABYLON.Vector3>): boolean {
        const squareDistance = BABYLON.Vector3.DistanceSquared(this.centerWorld, point);
        if (this.radiusWorld * this.radiusWorld < squareDistance) {
            return false;
        }

        return true;
    }

    // Statics
    /**
     * Checks if two sphere intersct
     * @param sphere0 sphere 0
     * @param sphere1 sphere 1
     * @returns true if the speres intersect
     */
    public static Intersects(sphere0: DeepImmutable<BoundingSphere>, sphere1: DeepImmutable<BoundingSphere>): boolean {
        const squareDistance = BABYLON.Vector3.DistanceSquared(sphere0.centerWorld, sphere1.centerWorld);
        const radiusSum = sphere0.radiusWorld + sphere1.radiusWorld;

        if (radiusSum * radiusSum < squareDistance) {
            return false;
        }

        return true;
    }
}


export class BoundingInfo {
    private static TmpVector3 = [BABYLON.Vector3.Zero(), BABYLON.Vector3.Zero()];
    private _isLocked: boolean;
    private _isImport: boolean;
    public boundingSphere: BoundingSphere;
    // PI_END
    /**
     * Constructs bounding info
     * @param minimum min vector of the bounding box/sphere
     * @param maximum max vector of the bounding box/sphere
     * @param worldMatrix defines the new world matrix
     */
    constructor(minimum, maximum, worldMatrix) {
        this._isLocked = false;
        // PI_BEGIN - 20200317 - 模型包围盒处理-可能已经设置了导出的包围盒
        /**
         * 可能已经设置了导出的包围盒
         */
        this._isImport = false;
        this.boundingSphere = new BoundingSphere(minimum, maximum, worldMatrix);
    }
    get isImport () {
        return this._isImport;
    }
    set isImport (value) {
        this._isImport = value;
    }
    /**
     * Recreates the entire bounding info from scratch as if we call the constructor in place
     * @param min defines the new minimum vector (in local space)
     * @param max defines the new maximum vector (in local space)
     * @param worldMatrix defines the new world matrix
     */
    public reConstruct(min, max, worldMatrix) {
        this.boundingSphere.reConstruct(min, max, worldMatrix);
    }
    get minimum() {
        return this.boundingSphere.minimum;
    }
    get maximum () {
        return this.boundingSphere.maximum;
    }
    get isLocked () {
        return this._isLocked;
    }
    set isLocked (value) {
        this._isLocked = value;
    }
    // Methods
    /**
     * Updates the bounding sphere and box
     * @param world world matrix to be used to update
     */
    public update(world) {
        if (this._isLocked) {
            return;
        }
        // this.boundingBox._update(world);
        this.boundingSphere._update(world);
        (<any>this.boundingSphere)._worldMatrix = world;
    };
    /**
     * Recreate the bounding info to be centered around a specific point given a specific extend.
     * @param center New center of the bounding info
     * @param extend New extend of the bounding info
     * @returns the current bounding info
     */
    public centerOn(center, extend) {
        var minimum = BoundingInfo.TmpVector3[0].copyFrom(center).subtractInPlace(extend);
        var maximum = BoundingInfo.TmpVector3[1].copyFrom(center).addInPlace(extend);
        // this.boundingBox.reConstruct(minimum, maximum, this.boundingBox.getWorldMatrix());
        this.boundingSphere.reConstruct(minimum, maximum, this.boundingSphere.getWorldMatrix());
        return this;
    };
    /**
     * Scale the current bounding info by applying a scale factor
     * @param factor defines the scale factor to apply
     * @returns the current bounding info
     */
    public scale (factor) {
        // this.boundingBox.scale(factor);
        this.boundingSphere.scale(factor);
        return this;
    };
    /**
     * Returns `true` if the bounding info is within the frustum defined by the passed array of planes.
     * @param frustumPlanes defines the frustum to test
     * @param strategy defines the strategy to use for the culling (default is BABYLON.AbstractMesh.CULLINGSTRATEGY_STANDARD)
     * @returns true if the bounding info is in the frustum planes
     */
    public isInFrustum (frustumPlanes, strategy) {
        if (strategy === void 0) { strategy = 0; }
        var inclusionTest = (strategy === 2 || strategy === 3);
        if (inclusionTest) {
            if (this.boundingSphere.isCenterInFrustum(frustumPlanes)) {
                return true;
            }
        }
        if (!this.boundingSphere.isInFrustum(frustumPlanes)) {
            return false;
        }
        // var bSphereOnlyTest = (strategy === 1 || strategy === 3);
        // if (bSphereOnlyTest) {
            return true;
        // }
        // return this.boundingBox.isInFrustum(frustumPlanes);
    };
    get diagonalLength () {
        // var boundingBox = this.boundingBox;
        // var diag = boundingBox.maximumWorld.subtractToRef(boundingBox.minimumWorld, BoundingInfo.TmpVector3[0]);
        // return diag.length();
        return this.boundingSphere.radiusWorld * 2;
    }
    /**
     * Checks if a cullable object (mesh...) is in the camera frustum
     * Unlike isInFrustum this cheks the full bounding box
     * @param frustumPlanes Camera near/planes
     * @returns true if the object is in frustum otherwise false
     */
    public isCompletelyInFrustum = function (frustumPlanes) {
        // return this.boundingBox.isCompletelyInFrustum(frustumPlanes);
        return true;
        // return this.boundingBox.isCompletelyInFrustum(frustumPlanes);
    };
    /** @hidden */
    public _checkCollision = function (collider) {
        let sphereCenter = this.boundingSphere.centerWorld;
        let sphereRadius = this.boundingSphere.radiusWorld;
        var distance = BABYLON.Vector3.Distance(collider._basePointWorld, sphereCenter);

        var max = Math.max(collider._radius.x, collider._radius.y, collider._radius.z);

        if (distance > collider._velocityWorldLength + max + sphereRadius) {
            return false;
        }

        // if (!intersectBoxAASphere(vecMin, vecMax, this._basePointWorld, this._velocityWorldLength + max)) {
        //     return false;
        // }

        return true;
        // return collider._canDoCollision(this.boundingSphere.centerWorld, this.boundingSphere.radiusWorld, this.boundingBox.minimumWorld, this.boundingBox.maximumWorld);
    };
    /**
     * Checks if a point is inside the bounding box and bounding sphere or the mesh
     * @see https://doc.babylonjs.com/babylon101/intersect_collisions_-_mesh
     * @param point the point to check intersection with
     * @returns if the point intersects
     */
    public intersectsPoint = function (point) {
        if (!this.boundingSphere.centerWorld) {
            return false;
        }
        if (!this.boundingSphere.intersectsPoint(point)) {
            return false;
        }
        // if (!this.boundingBox.intersectsPoint(point)) {
        //     return false;
        // }
        return true;
    };
    /**
     * Checks if another bounding info intersects the bounding box and bounding sphere or the mesh
     * @see https://doc.babylonjs.com/babylon101/intersect_collisions_-_mesh
     * @param boundingInfo the bounding info to check intersection with
     * @param precise if the intersection should be done using OBB
     * @returns if the bounding info intersects
     */
    public intersects = function (boundingInfo, precise) {
        if (!BoundingSphere.Intersects(this.boundingSphere, boundingInfo.boundingSphere)) {
            return false;
        }
        // if (!BoundingBox.Intersects(this.boundingBox, boundingInfo.boundingBox)) {
        //     return false;
        // }
        if (!precise) {
            return true;
        }
        return true;
    };
}

/**
 * Class used to store bounding box information
 */
export class BoundingBox {
    /**
     * Gets the 8 vectors representing the bounding box in local space
     */
    public readonly vectors: BABYLON.Vector3[] = [BABYLON.Vector3.Zero(), BABYLON.Vector3.Zero(), BABYLON.Vector3.Zero(), BABYLON.Vector3.Zero(),
        BABYLON.Vector3.Zero(), BABYLON.Vector3.Zero(), BABYLON.Vector3.Zero(), BABYLON.Vector3.Zero()
    ];
    /**
     * Gets the center of the bounding box in local space
     */
    public readonly center: BABYLON.Vector3 = BABYLON.Vector3.Zero();
    /**
     * Gets the center of the bounding box in world space
     */
    public readonly centerWorld: BABYLON.Vector3 = BABYLON.Vector3.Zero();
    /**
     * Gets the extend size in local space
     */
    public readonly extendSize: BABYLON.Vector3 = BABYLON.Vector3.Zero();
    /**
     * Gets the extend size in world space
     */
    public readonly extendSizeWorld: BABYLON.Vector3 = BABYLON.Vector3.Zero();
    /**
     * Gets the OBB (object bounding box) directions
     */
    public readonly directions: BABYLON.Vector3[] = [BABYLON.Vector3.Zero(), BABYLON.Vector3.Zero(), BABYLON.Vector3.Zero()
    ];
    /**
     * Gets the 8 vectors representing the bounding box in world space
     */
    public readonly vectorsWorld: BABYLON.Vector3[] = [BABYLON.Vector3.Zero(), BABYLON.Vector3.Zero(), BABYLON.Vector3.Zero(), BABYLON.Vector3.Zero(),
        BABYLON.Vector3.Zero(), BABYLON.Vector3.Zero(), BABYLON.Vector3.Zero(), BABYLON.Vector3.Zero()
    ];
    /**
     * Gets the minimum vector in world space
     */
    public readonly minimumWorld: BABYLON.Vector3 = BABYLON.Vector3.Zero();
    /**
     * Gets the maximum vector in world space
     */
    public readonly maximumWorld: BABYLON.Vector3 = BABYLON.Vector3.Zero();
    /**
     * Gets the minimum vector in local space
     */
    public readonly minimum: BABYLON.Vector3 = BABYLON.Vector3.Zero();
    /**
     * Gets the maximum vector in local space
     */
    public readonly maximum: BABYLON.Vector3 = BABYLON.Vector3.Zero();

    private _worldMatrix: DeepImmutable<BABYLON.Matrix>;
    private static readonly TmpVector3 = [BABYLON.Vector3.Zero(), BABYLON.Vector3.Zero(), BABYLON.Vector3.Zero()
    ];

    /**
     * @hidden
     */
    public _tag: number;

    /**
     * Creates a new bounding box
     * @param min defines the minimum vector (in local space)
     * @param max defines the maximum vector (in local space)
     * @param worldMatrix defines the new world matrix
     */
    constructor(min: DeepImmutable<BABYLON.Vector3>, max: DeepImmutable<BABYLON.Vector3>, worldMatrix?: DeepImmutable<BABYLON.Matrix>) {
        this.reConstruct(min, max, worldMatrix);
    }

    // Methods

    /**
     * Recreates the entire bounding box from scratch as if we call the constructor in place
     * @param min defines the new minimum vector (in local space)
     * @param max defines the new maximum vector (in local space)
     * @param worldMatrix defines the new world matrix
     */
    public reConstruct(min: DeepImmutable<BABYLON.Vector3>, max: DeepImmutable<BABYLON.Vector3>, worldMatrix?: DeepImmutable<BABYLON.Matrix>) {
        const minX = min.x, minY = min.y, minZ = min.z, maxX = max.x, maxY = max.y, maxZ = max.z;
        const vectors = this.vectors;

        this.minimum.copyFromFloats(minX, minY, minZ);
        this.maximum.copyFromFloats(maxX, maxY, maxZ);
        vectors[0].copyFromFloats(minX, minY, minZ);
        vectors[1].copyFromFloats(maxX, maxY, maxZ);
        vectors[2].copyFromFloats(maxX, minY, minZ);
        vectors[3].copyFromFloats(minX, maxY, minZ);
        vectors[4].copyFromFloats(minX, minY, maxZ);
        vectors[5].copyFromFloats(maxX, maxY, minZ);
        vectors[6].copyFromFloats(minX, maxY, maxZ);
        vectors[7].copyFromFloats(maxX, minY, maxZ);

        // OBB
        max.addToRef(min, this.center).scaleInPlace(0.5);
        max.subtractToRef(min, this.extendSize).scaleInPlace(0.5);

        this._worldMatrix = worldMatrix || BABYLON.Matrix.IdentityReadOnly;

        this._update(this._worldMatrix);
    }

    /**
     * Scale the current bounding box by applying a scale factor
     * @param factor defines the scale factor to apply
     * @returns the current bounding box
     */
    public scale(factor: number): BoundingBox {
        const tmpVectors = BoundingBox.TmpVector3;
        const diff = this.maximum.subtractToRef(this.minimum, tmpVectors[0]);
        const len = diff.length();
        diff.normalizeFromLength(len);
        const distance = len * factor;
        const newRadius = diff.scaleInPlace(distance * 0.5);

        const min = this.center.subtractToRef(newRadius, tmpVectors[1]);
        const max = this.center.addToRef(newRadius, tmpVectors[2]);

        this.reConstruct(min, max, this._worldMatrix);

        return this;
    }

    /**
     * Gets the world matrix of the bounding box
     * @returns a matrix
     */
    public getWorldMatrix(): DeepImmutable<BABYLON.Matrix> {
        return this._worldMatrix;
    }

    /** @hidden */
    public _update(world: DeepImmutable<BABYLON.Matrix>): void {
        const minWorld = this.minimumWorld;
        const maxWorld = this.maximumWorld;
        const directions = this.directions;
        const vectorsWorld = this.vectorsWorld;
        const vectors = this.vectors;

        if (!world.isIdentity()) {
            minWorld.setAll(Number.MAX_VALUE);
            maxWorld.setAll(-Number.MAX_VALUE);

            for (let index = 0; index < 8; ++index) {
                const v = vectorsWorld[index];
                BABYLON.Vector3.TransformCoordinatesToRef(vectors[index], world, v);
                minWorld.minimizeInPlace(v);
                maxWorld.maximizeInPlace(v);
            }

            // Extend
            maxWorld.subtractToRef(minWorld, this.extendSizeWorld).scaleInPlace(0.5);
            maxWorld.addToRef(minWorld, this.centerWorld).scaleInPlace(0.5);
        }
        else {
            minWorld.copyFrom(this.minimum);
            maxWorld.copyFrom(this.maximum);
            for (let index = 0; index < 8; ++index) {
                vectorsWorld[index].copyFrom(vectors[index]);
            }

            // Extend
            this.extendSizeWorld.copyFrom(this.extendSize);
            this.centerWorld.copyFrom(this.center);
        }

        BABYLON.Vector3.FromArrayToRef(world.m, 0, directions[0]);
        BABYLON.Vector3.FromArrayToRef(world.m, 4, directions[1]);
        BABYLON.Vector3.FromArrayToRef(world.m, 8, directions[2]);

        this._worldMatrix = world;
    }
}
