/**
 * CustomMaterial
 * @tip 一种效果需求, 在一个文件中实现 对应 材质的创建接口
 * @tip 在实现 的地方注册 一个 SHADER 名称， exampl: CUSTOM_MATERIAL_SHADER_NAMES['shader1'] = 'shader1';
 * @tip 创建 CustomMaterial 后 修改 name 属性为,对应 shader 名称
 * @example 'pi_babylon_demo\src\app_left\grid_line_spread\grid_line_spread.ts' - gltf 加载的材质
 * @example 'pi_babylon_demo\src\app_left\extends\wave2d.ts' - 程序运行时创建的材质
 */

import { BABYLON } from '../babylon';

export const CUSTOM_MATERIAL_SHADER_NAMES: {[index: string]: string} = {};
const CUSTOM_MATERIAL_RECORD: {[index: string]: number} = {};

export const initCustomMaterial = () => {
    // 扩展动画数据获取方法
    BABYLON.CustomMaterial.prototype.Builder = function(this: BABYLON.CustomMaterial, shaderName: string, uniforms: string[], uniformBuffers: string[], samplers: string[], defines: string[] | BABYLON.MaterialDefines, attributes?: string[]): string {

        if (attributes && this._customAttributes && this._customAttributes.length > 0) {
            attributes.push(...this._customAttributes);
        }

        this.ReviewUniform("uniform", uniforms);
        this.ReviewUniform("sampler", samplers);

        if (this._isCreatedShader) {
            return this._createdShaderName;
        }
        this._isCreatedShader = false;

        let name = "";
        if (CUSTOM_MATERIAL_SHADER_NAMES[this.name]) {
            name = CUSTOM_MATERIAL_SHADER_NAMES[this.name];
            CUSTOM_MATERIAL_RECORD[this.name] = (CUSTOM_MATERIAL_RECORD[this.name] || 0) + 1;
        }
        else {
            BABYLON.CustomMaterial.ShaderIndexer++;
            name = "custom_" + BABYLON.CustomMaterial.ShaderIndexer;
        }

        var fn_afterBind = this._afterBind.bind(this);
        this._afterBind = (m, e) => {
            if (!e) {
                return;
            }
            this.AttachAfterBind(m, e);
            try { fn_afterBind(m, e); }
            catch (e) { }
        };

        if (!CUSTOM_MATERIAL_RECORD[this.name] || CUSTOM_MATERIAL_RECORD[this.name] == 1) {
            BABYLON.Effect.ShadersStore[name + "VertexShader"] = this.VertexShader
                .replace('#define CUSTOM_VERTEX_BEGIN', (this.CustomParts.Vertex_Begin ? this.CustomParts.Vertex_Begin : ""))
                .replace('#define CUSTOM_VERTEX_DEFINITIONS', (this._customUniform ? this._customUniform.join("\n") : "") + (this.CustomParts.Vertex_Definitions ? this.CustomParts.Vertex_Definitions : ""))
                .replace('#define CUSTOM_VERTEX_MAIN_BEGIN', (this.CustomParts.Vertex_MainBegin ? this.CustomParts.Vertex_MainBegin : ""))
                .replace('#define CUSTOM_VERTEX_UPDATE_POSITION', (this.CustomParts.Vertex_Before_PositionUpdated ? this.CustomParts.Vertex_Before_PositionUpdated : ""))
                .replace('#define CUSTOM_VERTEX_UPDATE_NORMAL', (this.CustomParts.Vertex_Before_NormalUpdated ? this.CustomParts.Vertex_Before_NormalUpdated : ""))
                .replace('#define CUSTOM_VERTEX_MAIN_END', (this.CustomParts.Vertex_MainEnd ? this.CustomParts.Vertex_MainEnd : ""));

            if (this.CustomParts.Vertex_After_WorldPosComputed) {
                BABYLON.Effect.ShadersStore[name + "VertexShader"] = BABYLON.Effect.ShadersStore[name + "VertexShader"].replace('#define CUSTOM_VERTEX_UPDATE_WORLDPOS', this.CustomParts.Vertex_After_WorldPosComputed);
            }

            BABYLON.Effect.ShadersStore[name + "PixelShader"] = this.FragmentShader
                .replace('#define CUSTOM_FRAGMENT_BEGIN', (this.CustomParts.Fragment_Begin ? this.CustomParts.Fragment_Begin : ""))
                .replace('#define CUSTOM_FRAGMENT_MAIN_BEGIN', (this.CustomParts.Fragment_MainBegin ? this.CustomParts.Fragment_MainBegin : ""))
                .replace('#define CUSTOM_FRAGMENT_DEFINITIONS', (this._customUniform ? this._customUniform.join("\n") : "") + (this.CustomParts.Fragment_Definitions ? this.CustomParts.Fragment_Definitions : ""))
                .replace('#define CUSTOM_FRAGMENT_UPDATE_DIFFUSE', (this.CustomParts.Fragment_Custom_Diffuse ? this.CustomParts.Fragment_Custom_Diffuse : ""))
                .replace('#define CUSTOM_FRAGMENT_UPDATE_ALPHA', (this.CustomParts.Fragment_Custom_Alpha ? this.CustomParts.Fragment_Custom_Alpha : ""))
                .replace('#define CUSTOM_FRAGMENT_BEFORE_LIGHTS', (this.CustomParts.Fragment_Before_Lights ? this.CustomParts.Fragment_Before_Lights : ""))
                .replace('#define CUSTOM_FRAGMENT_BEFORE_FRAGCOLOR', (this.CustomParts.Fragment_Before_FragColor ? this.CustomParts.Fragment_Before_FragColor : ""));

            if (this.CustomParts.Fragment_Before_Fog) {
                BABYLON.Effect.ShadersStore[name + "PixelShader"] = BABYLON.Effect.ShadersStore[name + "PixelShader"].replace('#define CUSTOM_FRAGMENT_BEFORE_FOG', this.CustomParts.Fragment_Before_Fog);
            }
        }

        this._isCreatedShader = true;
        this._createdShaderName = name;

        return name;
    }
};