/**
 * 
 */
import { BABYLON } from '../babylon';
import { OfflineProvider } from '../offline_provider';
import { Scene } from '../scene';
import { BABYLON_GEOMETRY_TYPE, sizeOfGeometry } from '../../util/res_mgr';
import { IBufferArg, VertexBuffer } from './buffer';
import { PiBABYLONGeometry } from './geometry';
import { GLTFLoader as GLTFLoaderNew } from '../../gltf/glTFLoader';
import { IAccessor, IAnimation as IAnimationNew } from '../../gltf/glTF2Interface';
import { BABYLON_MATERIAL_NEW, ITempNodeAnimationAbout } from '../../gltf/base';
import { ExtensionTools, Res, ResTab } from '../../interface/index';
import { StandardMaterialData } from '../../material/standardmaterial';

declare let module;
let logger;
ExtensionTools.RegistStartUp(() => { logger = ExtensionTools.Logger(module.name, "BABYLON"); });

export enum GLTFInterpolationType {
    "LINEAR" = 1,
    "STEP" = 2,
    "CUBICSPLINE" = 3,
    "PICUBICSPLINE" = 4,
}
export enum GLTFInterpolationTypeOld {
    "STEP" = 0,
    "LINEAR" = 1,
    "PICUBICSPLINE" = 2,
    "CUBICSPLINE" = 3,
}
export enum EAnimationPropertyID {
    Position = 0,
    RotationQuaternion = 1,
    "Scaling",
    "MainTexUScale",
    "MainTexVScale",
    "MainTexUOffset",
    "MainTexVOffset",
    "Alpha",
    "MainColor",
    "ort",
    "per",
    "IsActive",
    "Euler",
    "intensity",
    "lightdiffuse",
    "AlphaCutOff",
    "cellId",
    "OpacityTexUScale",
    "OpacityTexVScale",
    "OpacityTexUOffset",
    "OpacityTexVOffset",
    "MaskCutoff",
    "MaskTexUScale",
    "MaskTexVScale",
    "MaskTexUOffset",
    "MaskTexVOffset",
}

const enum AnimationChannelTargetPath {
    /**
     * Translation
     */
    TRANSLATION = "translation",
    /**
     * Rotation
     */
    ROTATION = "rotation",
    // PI_BEGIN
    /**
     * rotation_eular
     */
    ROTATION_EULAR = "rotation_eular",
    // PI_END
    /**
     * Scale
     */
    SCALE = "scale",
    // PI_BEGIN
    // /**
    //  * Weights
    //  */
    WEIGHTS = "weights",
    // PI_END

    // PI_BEGIN
    USCALE = "uScale",

    VSCALE = "vScale",

    UOFFSET = "uOffset",

    VOFFSET = "vOffset",

    ALPHA = "alpha",

    COLOR = "color",

    ORT = "ort",

    PER = "per"
    // PI_END
}

const enum AnimationSamplerInterpolation {
    /**
     * The animated values are linearly interpolated between keyframes
     */
    LINEAR = "LINEAR",
    /**
     * The animated values remain constant to the output of the first keyframe, until the next keyframe
     */
    STEP = "STEP",
    /**
     * The animation's interpolation is computed using a cubic spline with specified tangents
     */
    CUBICSPLINE = "CUBICSPLINE",
    
    // PI_BEGIN
    PICUBICSPLINE = "PICUBICSPLINE",
    // PI_END
}

export type IAnimationChannel = number[];

// export type GLTFLoader = GLTFLoader;

export type GLTFLoader = BABYLON.GLTF2.GLTFLoader;
export type IAnimation = BABYLON.GLTF2.Loader.IAnimation;
export type IMeshPrimitive = BABYLON.GLTF2.Loader.IMeshPrimitive;
export type IBufferView = BABYLON.GLTF2.Loader.IBufferView;
// export type AnimationSamplerInterpolation = AnimationSamplerInterpolation;

type TGLTFLoadAnimaitonChannelFunctionNew = (loader: GLTFLoaderNew, context: string, animationContext: string, animation: IAnimationNew, channel: number[], babylonAnimationGroup: BABYLON.AnimationGroup) => Promise<void>;
export const GLTF_LOAD_ANIMATION_CHANNEL_ASYNC_LIST_NEW: TGLTFLoadAnimaitonChannelFunctionNew[] = [];

type TGLTFLoadAnimaitonChannelFunction = (loader: GLTFLoader, context: string, animationContext: string, animation: IAnimation, channel: number[], babylonAnimationGroup: BABYLON.AnimationGroup) => Promise<void>;
export const GLTF_LOAD_ANIMATION_CHANNEL_ASYNC_LIST: TGLTFLoadAnimaitonChannelFunction[] = [];

// tslint:disable-next-line:max-func-body-length
export const initGLTFLoader = () => {
    GLTFLoaderNew.AnimTargetPaths = [
        "translation",  "rotation",     "scale",            "uScale",       "vScale",
        "uOffset",      "vOffset",      "alpha",            "color",        "ort",
        "per",          "piIsActive",   "rotation_eular",   "intensity",    "lightdiffuse",
        "alphaCutOff",  "cellId",       "uScaleOp",         "vScaleOp",     "uOffsetOp",
        "vOffsetOp",    "maskCutoff",   "uScaleM",      "vScaleM",           "uOffsetM",
        "vOffsetM"
    ];

    // 文件加载适配
    BABYLON.GLTFFileLoader.prototype.requestFile = undefined;

    //#region New Loader
    GLTFLoaderNew.prototype._loadVertexAccessorAsync = function(this: GLTFLoaderNew, context: string, indexAccessor: number, accessor: IAccessor, kind: string): Promise<BABYLON.VertexBuffer> {

        if (accessor.sparse || (accessor.byteOffset && accessor.byteOffset % BABYLON.VertexBuffer.GetTypeByteLength(accessor.componentType) !== 0)) {
            throw new Error('Accessor does not support s, and its cost should be integer multiple of its ftype.');
        }

        let key = `${this.fileName}#${indexAccessor}`;
        let buffer: Res<any, BABYLON.Buffer> = VertexBuffer.resTab.get(BABYLON_GEOMETRY_TYPE, key);
        if (buffer) {
            let vb = new VertexBuffer(this.babylonScene.getEngine(), buffer.link, kind, false);
            vb.res = buffer;
            return Promise.resolve(vb);
        } else {
            // return this._loadFloatAccessorAsync(indexAccessor, accessor).then((data) => {
            //     return VertexBuffer.resTab.load(BABYLON_GEOMETRY_TYPE, key, [<IBufferArg>{ kind, engine: this.babylonScene.getEngine(), data }]).then(buffer => {
            //         let vb = new VertexBuffer(this.babylonScene.getEngine(), buffer.link, kind, false);
            //         vb.res = buffer;
            //         return vb;
            //     });
            // });
            
            var _this = this;
            let _babylonVertexBuffer;
            if (accessor.sparse) {
                _babylonVertexBuffer = this._loadFloatAccessorAsync(indexAccessor, accessor).then(function (data) {
                    return VertexBuffer.resTab.load(BABYLON_GEOMETRY_TYPE, key, [<IBufferArg>{ kind, engine: _this.babylonScene.getEngine(), data }]).then(buffer => {
                        let vb = new VertexBuffer(_this.babylonScene.getEngine(), buffer.link, kind, false);
                        vb.res = buffer;
                        return vb;
                    });
                });
            }
            // HACK: If byte offset is not a multiple of component type byte length then load as a float array instead of using Babylon buffers.
            else if (accessor.byteOffset && accessor.byteOffset % VertexBuffer.GetTypeByteLength(accessor.componentType) !== 0) {
                // Logger.Warn("Accessor byte offset is not a multiple of component type byte length");
                _babylonVertexBuffer = this._loadFloatAccessorAsync(indexAccessor, accessor).then(function (data) {
                    return VertexBuffer.resTab.load(BABYLON_GEOMETRY_TYPE, key, [<IBufferArg>{ kind, engine: _this.babylonScene.getEngine(), data }]).then(buffer => {
                        let vb = new VertexBuffer(_this.babylonScene.getEngine(), buffer.link, kind, false);
                        vb.res = buffer;
                        return vb;
                    });
                });
            }
            // Load joint indices as a float array since the shaders expect float data but glTF uses unsigned byte/short.
            // This prevents certain platforms (e.g. D3D) from having to convert the data to float on the fly.
            else if (kind === VertexBuffer.MatricesIndicesKind || kind === VertexBuffer.MatricesIndicesExtraKind) {
                _babylonVertexBuffer = this._loadFloatAccessorAsync(indexAccessor, accessor).then(function (data) {
                    return VertexBuffer.resTab.load(BABYLON_GEOMETRY_TYPE, key, [<IBufferArg>{ kind, engine: _this.babylonScene.getEngine(), data }]).then(buffer => {
                        let vb = new VertexBuffer(_this.babylonScene.getEngine(), buffer.link, kind, false);
                        vb.res = buffer;
                        return vb;
                    });
                });
            }
            else {
                const bufferView_2 = this.gltf.bufferViews[accessor.bufferView];
                _babylonVertexBuffer = this._loadVertexBufferViewAsync(accessor.bufferView, bufferView_2, kind).then(function (babylonBuffer) {
                    var size = GLTFLoaderNew._GetNumComponents(context, accessor.type);
                    return new VertexBuffer(_this.babylonScene.getEngine(), babylonBuffer, kind, false, false, bufferView_2.byteStride, false, accessor.byteOffset, size, accessor.componentType, accessor.normalized, true, 1, true);
                });
            }
            return _babylonVertexBuffer;
        }
    };

    GLTFLoaderNew.prototype._loadAnimationChannelAsync = function(this: GLTFLoaderNew, context: string, animationContext: string, animation: IAnimationNew, channel: number[], babylonAnimationGroup: BABYLON.AnimationGroup): Promise<void> {
        let loadFunction = GLTF_LOAD_ANIMATION_CHANNEL_ASYNC_LIST_NEW[channel[0]];
        if (loadFunction) {
            return loadFunction(this, context, animationContext, animation, channel, babylonAnimationGroup);
        } else {
            throw new Error(`${context}/target/path: Invalid value (${GLTFLoaderNew.AnimTargetPaths[channel[0]]})`);
        }
    };

    GLTFLoaderNew.prototype._loadAnimationKeysAsync = function(this: GLTFLoaderNew, context: string, animationContext: string, animation: IAnimationNew, channel: number[], targetPath: string): Promise<Res<void, BABYLON.IAnimationKey[]>> {
        // const _key = key + '/' + animationContext + '/channel/' + animation.buffer + ',' + channel[0] + ',' + channel[1] + ',' + channel[2] + ',' + channel[3] + ',' + channel[4] + ',' + channel[5];
        const fileName = this.fileName;
        const _key = `${fileName},${channel[0]},${channel[1]},${channel[2]},${channel[3]},${channel[4]},${channel[5]}`;
        const r = (<OfflineProvider>this.babylonScene.offlineProvider).loadAnimationKeys(_key);
        if (r) {
            return Promise.resolve(r);
        }
        let targetNode: ITempNodeAnimationAbout;
        if (channel[1] >= 0) {
            targetNode = this.loaderRecord.nodeAnimationAbout.get(channel[1]); // BABYLON.GLTF2.ArrayItem.Get(`${context}/target/node`, this.gltf.nodes, channel[1]);
        }

        // tslint:disable-next-line:max-func-body-length
        return this._loadPiAnimationAsync(context, (<any>animation).buffer, channel).then((data) => {

            let outputBufferOffset = 0;
            let getNextOutputValue: () => BABYLON.Vector3 | BABYLON.Quaternion | number[] | number | BABYLON.Color3;
            switch (targetPath) {
                case 'position': {
                    getNextOutputValue = () => {
                        const value = BABYLON.Vector3.FromArray(data.output, outputBufferOffset);
                        outputBufferOffset += 3;
    
                        return value;
                    };
                    break;
                }
                case 'rotationQuaternion': {
                    getNextOutputValue = () => {
                        const value = BABYLON.Quaternion.FromArray(data.output, outputBufferOffset);
                        outputBufferOffset += 4;
    
                        return value;
                    };
                    break;
                }
                case 'color': {
                    getNextOutputValue = () => {
                        const value = BABYLON.Color3.FromArray(data.output, outputBufferOffset);
                        outputBufferOffset += 3;
    
                        return value;
                    };
                    break;
                }
                case 'rotation': {
                    getNextOutputValue = () => {
                        const value = BABYLON.Vector3.FromArray(data.output, outputBufferOffset);
                        outputBufferOffset += 3;
    
                        return value;
                    };
                    break;
                }
                case 'scaling': {
                    getNextOutputValue = () => {
                        const value = BABYLON.Vector3.FromArray(data.output, outputBufferOffset);
                        outputBufferOffset += 3;
    
                        return value;
                    };
                    break;
                }
                case 'diffuse': {
                    getNextOutputValue = () => {
                        const value = BABYLON.Color3.FromArray(data.output, outputBufferOffset);
                        outputBufferOffset += 3;
    
                        return value;
                    };
                    break;
                }
                // case 'influence': {
                //     getNextOutputValue = () => {
                //         // tslint:disable-next-line:prefer-array-literal
                //         const value = new Array<number>(targetNode._numMorphTargets!);
                //         // tslint:disable-next-line:no-non-null-assertion
                //         for (let i = 0; i < targetNode._numMorphTargets!; i++) {
                //             value[i] = data.output[outputBufferOffset++];
                //         }
    
                //         return value;
                //     };
                //     break;
                // }
                default: {
                    getNextOutputValue = () => {
                        return data.output[outputBufferOffset++];
                    };
                }
            }
    
            let getNextKey: (frameIndex: number) => BABYLON.IAnimationKey;
            // tslint:disable-next-line:switch-default
            switch (GLTFLoaderNew.AnimInterpolationType[data.interpolation]) {
                case 'STEP': {
                    getNextKey = (frameIndex) => (new AnimationKey(
                        data.input[frameIndex],
                        getNextOutputValue(),
                        BABYLON.AnimationKeyInterpolation.STEP
                    ));
                    break;
                }
                case 'LINEAR': {
                    getNextKey = (frameIndex) => (new AnimationKey(
                        data.input[frameIndex],
                        getNextOutputValue()
                    ));
                    break;
                }
                case 'CUBICSPLINE': {
                    getNextKey = (frameIndex) => {
                        const inTangent = getNextOutputValue();
                        const value = getNextOutputValue();
                        const outTangent = getNextOutputValue();
    
                        return new AnimationKey(
                            data.input[frameIndex],
                            value,
                            undefined,
                            inTangent,
                            outTangent
                        );
                    };
                    break;
                }
                case 'PICUBICSPLINE' /* CUBICSPLINE */: {
                    getNextKey = (frameIndex) => {
                        const inAndOut = getNextOutputValue();
    
                        return new AnimationKey(
                            data.input[frameIndex],
                            getNextOutputValue(),
                            undefined,
                            inAndOut,
                            inAndOut
                        );
                    };
                }
            }
    
            // tslint:disable-next-line:prefer-array-literal
            const keys = new Array(data.input.length);
            for (let frameIndex = 0; frameIndex < data.input.length; frameIndex++) {
                // tslint:disable-next-line:no-non-null-assertion
                keys[frameIndex] = getNextKey!(frameIndex);
            }

            return (<OfflineProvider>this.babylonScene.offlineProvider).createAnimationKeys(_key, keys)
        });
    };
    //#endregion

    //#region Older Loader
    {
        BABYLON.GLTF2.GLTFLoader.AnimTargetPaths = [
            "translation",  "rotation",     "scale",            "uScale",       "vScale",
            "uOffset",      "vOffset",      "alpha",            "color",        "ort",
            "per",          "piIsActive",   "rotation_eular",   "intensity",    "lightdiffuse",
            "alphaCutOff",  "cellId",       "uScaleOp",         "vScaleOp",     "uOffsetOp",
            "vOffsetOp",    "maskCutoff",   "uScaleM",      "vScaleM",           "uOffsetM",
            "vOffsetM"
        ];
    
        // 文件加载适配
        BABYLON.GLTFFileLoader.prototype.requestFile = undefined;
    
        BABYLON.GLTF2.GLTFLoader.PiAsyncCall = (func: Function, arg: any) => {
            ExtensionTools.AddTask(func, [arg], 99999, 0);
        };
    
        (<any>BABYLON.GLTF2.GLTFLoader.prototype).loadAnimationAsync = function(this: GLTFLoader, context: string, animation: IAnimation): Promise<BABYLON.AnimationGroup> {
            const promise = (<any>this)._extensionsLoadAnimationAsync(context, animation);
            if (promise) {
    
                (<any>this).RecordPromise(`loadAnimationAsync - _extensions - ${(<any>this)._fileName} - ${context}`, promise);
    
                return promise;
            }
    
            const babylonAnimationGroup = new BABYLON.AnimationGroup(animation.name || `animation${animation.index}`, this.babylonScene);
            animation._babylonAnimationGroup = babylonAnimationGroup;
    
            const promises = new Array<Promise<any>>();
            // //////////////////////////////////
            if (animation.PiChannel) {
                BABYLON.GLTF2.ArrayItem.Assign(<any>animation.PiChannel);
                let tempAnimIDs = [];
                for (const channel of animation.PiChannel) {
                    let [propertyID, targetID, inputOffset, inputCount, inputBufferOffset, inputBufferLen, outputOffset, outputCount, outputBufferOffset, outputBufferLen, interpolation] = channel;
                    let aID = `${propertyID}#${targetID}`;
                    if (tempAnimIDs.includes(aID) == false) {
                        tempAnimIDs.push(aID);
                        promises.push((<any>this)._loadAnimationChannelAsync(`${context}/channels/${(<any>channel).index}`, context, animation, channel, babylonAnimationGroup));
                    }
                }
                let temp = <any>BABYLON.GLTF2.ArrayItem.Get(context + "/target/node", this.gltf.nodes, animation.nodeIndex);
                if (animation.nodeIndex && temp._babylonTransformNode) {
                    let targetNode = temp;
                    if (targetNode._babylonTransformNode) {
                        if (!(<any>targetNode._babylonTransformNode).animationGroups) {
                            (<any>targetNode._babylonTransformNode).animationGroups = [];
                        }
                        (<any>targetNode._babylonTransformNode).animationGroups.push(babylonAnimationGroup);
                    }
                    else {
                        //
                    }
                }
            } else if (animation.channels) {
                (<any>animation).buffer = 0;
                let lenChannel = animation.channels.length;
                for (let indexChannel = 0; indexChannel < lenChannel; indexChannel++) {
                    let channel = animation.channels[indexChannel];
                    
                    let propertyID: EAnimationPropertyID;
                    if (channel.target.path) {
                        switch (<any>channel.target.path) {
                            case (AnimationChannelTargetPath.TRANSLATION): {
                                propertyID = EAnimationPropertyID.Position;
                                break;
                            }
                            case (AnimationChannelTargetPath.SCALE): {
                                propertyID = EAnimationPropertyID.Scaling;
                                break;
                            }
                            case (AnimationChannelTargetPath.ROTATION): {
                                propertyID = EAnimationPropertyID.RotationQuaternion;
                                break;
                            }
                        }
                    }
                    if (channel.extras.property != undefined) {
                        propertyID = channel.extras.property;
                    }
    
                    let sampler = animation.samplers[channel.sampler];
                    let interpolation = GLTFInterpolationTypeOld.LINEAR;
                    if (sampler.interpolation) {
                        switch (<any>sampler.interpolation) {
                            case (AnimationSamplerInterpolation.STEP): {
                                interpolation = GLTFInterpolationTypeOld.STEP;
                                break;
                            }
                            case (AnimationSamplerInterpolation.CUBICSPLINE): {
                                interpolation = GLTFInterpolationTypeOld.CUBICSPLINE;
                                break;
                            }
                        }
                    } else if (sampler.extras?.interpolation != undefined) {
                        switch (<any>sampler.extras?.interpolation) {
                            case (AnimationSamplerInterpolation.STEP): {
                                interpolation = GLTFInterpolationTypeOld.STEP;
                                break;
                            }
                            case (AnimationSamplerInterpolation.PICUBICSPLINE): {
                                interpolation = GLTFInterpolationTypeOld.PICUBICSPLINE;
                                break;
                            }
                            case (AnimationSamplerInterpolation.CUBICSPLINE): {
                                interpolation = GLTFInterpolationTypeOld.CUBICSPLINE;
                                break;
                            }
                        }
                    }
    
                    let targetID = channel.target.node;
                    let input = this.gltf.accessors[sampler.input];
                    let output = this.gltf.accessors[sampler.output];
                    let inputBufferview = this.gltf.bufferViews[input.bufferView];
                    let outputBufferview = this.gltf.bufferViews[output.bufferView];
    
                    // let [propertyID, targetID, inputOffset, inputCount, inputBufferOffset, inputBufferLen, outputOffset, outputCount, outputBufferOffset, outputBufferLen, interpolation]
                    let channelOld = [propertyID, targetID, input.byteOffset, input.count, inputBufferview.byteOffset, inputBufferview.byteLength, output.byteOffset, output.count, outputBufferview.byteOffset, outputBufferview.byteLength, interpolation];
                    
                    promises.push((<any>this)._loadAnimationChannelAsync(`${context}/channels/${indexChannel}`, context, animation, channelOld, babylonAnimationGroup));
                }
            }
            // PI_END
    
            // PI_BEGIN - 20200814 - 动画事件
            if (animation.events) {
                var babylonAnimation = new BABYLON.Animation(`event_${animation.name}`, `_pi_meta_event`, 1, BABYLON.Animation.ANIMATIONTYPE_FLOAT);
                const count = animation.events.length;
                const keys = [];
                for (let i = 0; i < count; i++) {
                    const event = animation.events[i];
                    keys.push({
                        frame: event[0],
                        value: 0
                    });
                    const animEvent = new BABYLON.AnimationEvent(event[0], <any>undefined, false);
                    animEvent._pi_eventKey = event[1];
                    animEvent.action = function(curframe: number) {
                        const listenList: (() => boolean)[] = (<any>this)._pi_listenList;
                        if (listenList) {
                            listenList.forEach((func: any, index: number) => {
                                if (func) {
                                    const onlyOnce = func(curframe);
                                    if (onlyOnce) {
                                        listenList[index] = <any>undefined;
                                    }
                                }
                            });
                        }
                    };
    
                    babylonAnimation.addEvent(animEvent);
                }
                babylonAnimation.setKeys(keys);
                babylonAnimationGroup.addTargetedAnimation(babylonAnimation, babylonAnimation); // 目标不限,
            }
            // PI_END
    
            // PI_BEGIN - 20200902 - 骨骼动画缓存
            if (animation.SkinAnim) {
                const [, , start, end, totalFrame, , , skinIndex, skinName, animName] = animation.SkinAnim;
                BABYLON.Animation.AddSkinAnimCfg(skinName, animName, animation.SkinAnim);
                var babylonAnimation = new BABYLON.Animation(`skin_cache_${animation.name}`, `pi_actionFrame`, 1, BABYLON.Animation.ANIMATIONTYPE_FLOAT_PRECISION);
    
                const keys = [
                    {
                        frame: 0,
                        value: 0
                    },
                    {
                        frame: (end - start),
                        value: totalFrame - 1
                    }
                ];
                babylonAnimation.setKeys(keys);
                babylonAnimation.pi_floatPrecision = 1;
    
                const skeletonId = `skeleton${skinIndex}`;
    
                let skin: BABYLON.Skeleton = (<any>this._skeletons).find((v: BABYLON.GLTF2.ISkin) => {
                    return v.name === skinName;
                });
                if (!skin) {
                    skin = new BABYLON.Skeleton(skinName, skeletonId, this.babylonScene);
                    skin.metadata = { gltf: {} };
                    this._skeletons.push(skin);
                }
                skin.pi_cacheMatrixInfoMap.set(animName, animation.SkinAnim);
                babylonAnimationGroup.addTargetedAnimation(babylonAnimation, skin);
                babylonAnimationGroup.onAnimationGroupPlayObservable.add((e, s) => {
                    skin.pi_actionName = animName;
                });
            }
            // PI_END
    
            return Promise.all(promises).then(() => {
                babylonAnimationGroup.normalize(0);
                return babylonAnimationGroup;
            }).catch((e) => {
                console.error(e);
                return babylonAnimationGroup;
            });
        };
        // PI_END
    
        (<any>BABYLON.GLTF2.GLTFLoader.prototype)._loadVertexBufferViewAsync = function(this: GLTFLoader, bufferView: IBufferView, _kind: string): Promise<ArrayBufferView> {
            if (bufferView._babylonBuffer) {
                return <any>bufferView._babylonBuffer;
            }
    
            bufferView._babylonBuffer = this.loadBufferViewAsync(`#/bufferViews/${bufferView.index}`, bufferView).then((data) => {
                return <any>data;
            });
    
            return <any>bufferView._babylonBuffer;
        };
    
        (<any>BABYLON.GLTF2.GLTFLoader.prototype)._loadVertexAccessorAsync = function(this: GLTFLoader, context: string, accessor: BABYLON.GLTF2.Loader.IAccessor, kind: string): Promise<BABYLON.VertexBuffer> {
            if (accessor._babylonVertexBuffer) {
                return accessor._babylonVertexBuffer;
            }
    
            if (accessor.sparse || (accessor.byteOffset && accessor.byteOffset % BABYLON.VertexBuffer.GetTypeByteLength(accessor.componentType) !== 0)) {
                throw new Error('Accessor does not support s, and its cost should be integer multiple of its ftype.');
            }
    
            // const bufferView = BABYLON.GLTF2.ArrayItem.Get(`${context}/bufferView`, this.gltf.bufferViews, accessor.bufferView);
            // accessor._babylonVertexBuffer = this._loadVertexBufferViewAsync(<any>bufferView, kind).then((babylonBuffer) => {
            //     const size = (<any>BABYLON.GLTF2.GLTFLoader)._GetNumComponents(context, accessor.type);
            //     return new BABYLON.VertexBuffer(this.babylonScene.getEngine(), data, kind, false, false, null,
            //         false, accessor.byteOffset, size, accessor.componentType, accessor.normalized, true);
            // });
    
            let key = `${(<any>this)._rootUrl}${(<any>this)._fileName}#${accessor.index}`;
            let buffer: Res<any, BABYLON.Buffer> = VertexBuffer.resTab.get(BABYLON_GEOMETRY_TYPE, key);
            if (buffer) {
                let vb = new VertexBuffer(this.babylonScene.getEngine(), buffer.link, kind, false);
                vb.res = buffer;
                accessor._babylonVertexBuffer = Promise.resolve(vb);
            } else {

                // accessor._babylonVertexBuffer = this._loadFloatAccessorAsync("/accessors/" + accessor.index, accessor).then((data) => {
                //     return VertexBuffer.resTab.load(BABYLON_GEOMETRY_TYPE, key, [<IBufferArg>{ kind, engine: this.babylonScene.getEngine(), data }]).then(buffer => {
                //         let vb = new VertexBuffer(this.babylonScene.getEngine(), buffer.link, kind, false);
                //         vb.res = buffer;
                //         return vb;
                //     });
                // });
                
                var _this = this;
                if (accessor.sparse) {
                    accessor._babylonVertexBuffer = this._loadFloatAccessorAsync("/accessors/" + accessor.index, accessor).then(function (data) {
                        return VertexBuffer.resTab.load(BABYLON_GEOMETRY_TYPE, key, [<IBufferArg>{ kind, engine: _this.babylonScene.getEngine(), data }]).then(buffer => {
                            let vb = new VertexBuffer(_this.babylonScene.getEngine(), buffer.link, kind, false);
                            vb.res = buffer;
                            return vb;
                        });
                    });
                }
                // HACK: If byte offset is not a multiple of component type byte length then load as a float array instead of using Babylon buffers.
                else if (accessor.byteOffset && accessor.byteOffset % VertexBuffer.GetTypeByteLength(accessor.componentType) !== 0) {
                    // Logger.Warn("Accessor byte offset is not a multiple of component type byte length");
                    accessor._babylonVertexBuffer = this._loadFloatAccessorAsync("/accessors/" + accessor.index, accessor).then(function (data) {
                        return VertexBuffer.resTab.load(BABYLON_GEOMETRY_TYPE, key, [<IBufferArg>{ kind, engine: _this.babylonScene.getEngine(), data }]).then(buffer => {
                            let vb = new VertexBuffer(_this.babylonScene.getEngine(), buffer.link, kind, false);
                            vb.res = buffer;
                            return vb;
                        });
                    });
                }
                // Load joint indices as a float array since the shaders expect float data but glTF uses unsigned byte/short.
                // This prevents certain platforms (e.g. D3D) from having to convert the data to float on the fly.
                else if (kind === VertexBuffer.MatricesIndicesKind || kind === VertexBuffer.MatricesIndicesExtraKind) {
                    accessor._babylonVertexBuffer = this._loadFloatAccessorAsync("/accessors/" + accessor.index, accessor).then(function (data) {
                        return VertexBuffer.resTab.load(BABYLON_GEOMETRY_TYPE, key, [<IBufferArg>{ kind, engine: _this.babylonScene.getEngine(), data }]).then(buffer => {
                            let vb = new VertexBuffer(_this.babylonScene.getEngine(), buffer.link, kind, false);
                            vb.res = buffer;
                            return vb;
                        });
                    });
                }
                else {
                    var bufferView_2 = BABYLON.GLTF2.ArrayItem.Get(context + "/bufferView", this._gltf.bufferViews, accessor.bufferView);
                    accessor._babylonVertexBuffer = this._loadVertexBufferViewAsync(bufferView_2, kind).then(function (babylonBuffer) {
                        var size = BABYLON.GLTF2.GLTFLoader._GetNumComponents(context, accessor.type);
                        return new VertexBuffer(_this._babylonScene.getEngine(), babylonBuffer, kind, false, false, bufferView_2.byteStride, false, accessor.byteOffset, size, accessor.componentType, accessor.normalized, true, 1, true);
                    });
                }
            }
    
            return accessor._babylonVertexBuffer;
        };
    
        (<any>BABYLON.GLTF2.GLTFLoader.prototype)._loadVertexDataAsync = function(this: GLTFLoader, context: string, primitive: IMeshPrimitive, babylonMesh: BABYLON.Mesh): Promise<BABYLON.Geometry> {
            const indices_accessor = BABYLON.GLTF2.ArrayItem.Get(`${context}/indices`, this.gltf.accessors, primitive.indices);
            const bufferView = BABYLON.GLTF2.ArrayItem.Get(`#/accessors/${(<any>indices_accessor).index}/bufferView`, this.gltf.bufferViews, indices_accessor.bufferView);
            const buffer = BABYLON.GLTF2.ArrayItem.Get(`#/bufferViews/${(<any>bufferView).index}/buffer`, this.gltf.buffers, bufferView.buffer);
            // const url = (<any>this)._rootUrl + buffer.uri;
    
            const extensionPromise = (<any>this)._extensionsLoadVertexDataAsync(context, primitive, babylonMesh);
            if (extensionPromise) {
                return extensionPromise;
            }
    
            const m1 = context.slice(0, context.lastIndexOf('/'));
            const m2 = m1.slice(0, m1.lastIndexOf('/'));
            const index = parseInt(m2.slice(m2.lastIndexOf('/') + 1, m2.length), 10);
            const key = (<Scene>this.babylonScene).piID + "#" + (<any>BABYLON.GLTF2.ArrayItem.Get(m2, this.gltf.meshes, index)).geometry + primitive.indices + ((<any>this).geometryForMeshParticle ? Date.now() : "");
            (<any>this).geometryForMeshParticle = false;
        
            const attributes = primitive.attributes;
            if (!attributes) {
                throw new Error(`${context}: Attributes are missing`);
            }
        
            // tslint:disable-next-line:prefer-array-literal
            const promises = new Array<Promise<any>>();
        
            const babylonGeometry: PiBABYLONGeometry = <PiBABYLONGeometry>new BABYLON.Geometry(key, this.babylonScene);
            (<any>babylonGeometry).pi_indices = primitive.indices;
        
            if (primitive.indices === undefined) {
                // babylonMesh.isUnIndexed = true;
            } else {
                promises.push((<any>this)._loadIndicesAccessorAsync(`#/accessors/${(<any>indices_accessor).index}`, indices_accessor).then((data) => {
                    babylonGeometry.setIndices(data);
                }));
            }
        
            const loadAttribute = (attribute: string, kind: string, callback?: (accessor: BABYLON.GLTF2.Loader.IAccessor) => void) => {
                if (attributes[attribute] === undefined) {
                    return;
                }
        
                const accessor = BABYLON.GLTF2.ArrayItem.Get(`${context}/attributes/${attribute}`, this.gltf.accessors, attributes[attribute]);
                promises.push((<any>this)._loadVertexAccessorAsync(`#/accessors/${(<any>accessor).index}`, (<any>accessor), kind).then((babylonVertexBuffer) => {
                    babylonGeometry.setVerticesBuffer(babylonVertexBuffer, accessor.count);
                }));
        
                if (callback) {
                    callback(<any>accessor);
                }
            };
        
            loadAttribute('POSITION', BABYLON.VertexBuffer.PositionKind);
            loadAttribute('NORMAL', BABYLON.VertexBuffer.NormalKind);
            loadAttribute('TANGENT', BABYLON.VertexBuffer.TangentKind);
            loadAttribute('TEXCOORD_0', BABYLON.VertexBuffer.UVKind);
            loadAttribute('TEXCOORD_1', BABYLON.VertexBuffer.UV2Kind);
            loadAttribute('JOINTS_0', BABYLON.VertexBuffer.MatricesIndicesKind);
            loadAttribute('WEIGHTS_0', BABYLON.VertexBuffer.MatricesWeightsKind);
            loadAttribute('COLOR_0', BABYLON.VertexBuffer.ColorKind, (accessor) => {
                // if (accessor.type === 'VEC4' /* VEC4 */) {
                //     babylonMesh.hasVertexAlpha = true;
                // }
            });
            
            (<any>babylonGeometry)._bakeframes = primitive.extras?.bakeframes;
        
            return Promise.all(promises).then(() => {
        
                if (OfflineProvider.onoffCollectInfo) {
                    OfflineProvider.piBABYLONGeometryRecord.set(key, sizeOfGeometry(babylonGeometry));
                }
        
                babylonGeometry.attrsOKCall();
                return babylonGeometry;
            });
        };
    
        (<any>BABYLON.GLTF2.GLTFLoader.prototype)._loadAnimationChannelAsync = function(this: GLTFLoader, context: string, animationContext: string, animation: IAnimation, channel: IAnimationChannel, babylonAnimationGroup: BABYLON.AnimationGroup): Promise<void> {
            const targetNode: any = BABYLON.GLTF2.ArrayItem.Get(`${context}/target/node`, this.gltf.nodes, channel[1]);
    
            let loadFunction = GLTF_LOAD_ANIMATION_CHANNEL_ASYNC_LIST[channel[0]];
            if (loadFunction) {
                return loadFunction(this, context, animationContext, animation, channel, babylonAnimationGroup);
            } else {
                throw new Error(`${context}/target/path: Invalid value (${this.targetPathType[channel[0]]})`);
            }
        };
    
        (<any>BABYLON.GLTF2.GLTFLoader.prototype)._loadAnimationKeysAsync = function(this: GLTFLoader, context: string, animationContext: string, animation: IAnimation, channel: IAnimationChannel, key: string, targetPath: string): Promise<Res<void, BABYLON.IAnimationKey[]>> {
            // const _key = key + '/' + animationContext + '/channel/' + animation.buffer + ',' + channel[0] + ',' + channel[1] + ',' + channel[2] + ',' + channel[3] + ',' + channel[4] + ',' + channel[5];
            const fileName = ExtensionTools.FileNameWithoutDirectory(key);
            const _key = `${fileName}/${animationContext}/channel/${(<any>animation).buffer},${channel[0]},${channel[1]},${channel[2]},${channel[3]},${channel[4]},${channel[5]}`;
            const r = (<OfflineProvider>this.babylonScene.offlineProvider).loadAnimationKeys(_key);
            if (r) {
                return Promise.resolve(r);
            }
            let targetNode: any;
            if (channel[1] >= 0) {
                targetNode = BABYLON.GLTF2.ArrayItem.Get(`${context}/target/node`, this.gltf.nodes, channel[1]);
            }
    
            // tslint:disable-next-line:max-func-body-length
            return (<any>this)._loadPiAnimationAsync(context, (<any>animation).buffer, channel).then((data) => {
    
                const r = (<OfflineProvider>this.babylonScene.offlineProvider).loadAnimationKeys(_key);
                if (r) {
                    return r;
                }
    
                let outputBufferOffset = 0;
                let getNextOutputValue: () => BABYLON.Vector3 | BABYLON.Quaternion | number[] | number | BABYLON.Color3;
                switch (targetPath) {
                    case 'position': {
                        getNextOutputValue = () => {
                            const value = BABYLON.Vector3.FromArray(data.output, outputBufferOffset);
                            outputBufferOffset += 3;
    
                            return value;
                        };
                        break;
                    }
                    case 'rotationQuaternion': {
                        getNextOutputValue = () => {
                            const value = BABYLON.Quaternion.FromArray(data.output, outputBufferOffset);
                            outputBufferOffset += 4;
    
                            return value;
                        };
                        break;
                    }
                    case 'color': {
                        getNextOutputValue = () => {
                            const value = BABYLON.Color3.FromArray(data.output, outputBufferOffset);
                            outputBufferOffset += 3;
    
                            return value;
                        };
                        break;
                    }
                    case 'rotation': {
                        getNextOutputValue = () => {
                            const value = BABYLON.Vector3.FromArray(data.output, outputBufferOffset);
                            outputBufferOffset += 3;
    
                            return value;
                        };
                        break;
                    }
                    case 'scaling': {
                        getNextOutputValue = () => {
                            const value = BABYLON.Vector3.FromArray(data.output, outputBufferOffset);
                            outputBufferOffset += 3;
    
                            return value;
                        };
                        break;
                    }
                    case 'diffuse': {
                        getNextOutputValue = () => {
                            const value = BABYLON.Color3.FromArray(data.output, outputBufferOffset);
                            outputBufferOffset += 3;
    
                            return value;
                        };
                        break;
                    }
                    case 'influence': {
                        getNextOutputValue = () => {
                            // tslint:disable-next-line:prefer-array-literal
                            const value = new Array<number>(targetNode._numMorphTargets!);
                            // tslint:disable-next-line:no-non-null-assertion
                            for (let i = 0; i < targetNode._numMorphTargets!; i++) {
                                value[i] = data.output[outputBufferOffset++];
                            }
    
                            return value;
                        };
                        break;
                    }
                    default: {
                        getNextOutputValue = () => {
                            return data.output[outputBufferOffset++];
                        };
                    }
                }
    
                let getNextKey: (frameIndex: number) => BABYLON.IAnimationKey;
                // tslint:disable-next-line:switch-default
                switch (this.interpolationType[data.interpolation]) {
                    case 'STEP': {
                        getNextKey = (frameIndex) => (new AnimationKey(
                            data.input[frameIndex],
                            getNextOutputValue(),
                            BABYLON.AnimationKeyInterpolation.STEP
                        ));
                        break;
                    }
                    case 'LINEAR': {
                        getNextKey = (frameIndex) => (new AnimationKey(
                            data.input[frameIndex],
                            getNextOutputValue()
                        ));
                        break;
                    }
                    case 'CUBICSPLINE': {
                        getNextKey = (frameIndex) => {
                            const inTangent = getNextOutputValue();
                            const value = getNextOutputValue();
                            const outTangent = getNextOutputValue();
    
                            return new AnimationKey(
                                data.input[frameIndex],
                                value,
                                undefined,
                                inTangent,
                                outTangent
                            );
                        };
                        break;
                    }
                    case 'PICUBICSPLINE' /* CUBICSPLINE */: {
                        getNextKey = (frameIndex) => {
                            const inAndOut = getNextOutputValue();
    
                            return new AnimationKey(
                                data.input[frameIndex],
                                getNextOutputValue(),
                                undefined,
                                inAndOut,
                                inAndOut
                            );
                        };
                    }
                }
    
                // tslint:disable-next-line:prefer-array-literal
                const keys = new Array(data.input.length);
                for (let frameIndex = 0; frameIndex < data.input.length; frameIndex++) {
                    // tslint:disable-next-line:no-non-null-assertion
                    keys[frameIndex] = getNextKey!(frameIndex);
                }
    
                return (<OfflineProvider>this.babylonScene.offlineProvider).createAnimationKeys(_key, keys);
            });
        };
    
        (<any>BABYLON.GLTF2.GLTFLoader.prototype).requestFile = undefined;
    
        /** @hidden */
        interface _IAnimationSamplerData {
            input: Float32Array;
            interpolation: AnimationSamplerInterpolation;
            output: Float32Array;
        };
    
        // PI_BEGIN - ANIMATION
        (<any>BABYLON.GLTF2.GLTFLoader.prototype)._loadPiAnimationAsync = function(this: GLTFLoader, context: string, bufferIndex: number, channel: any): Promise<_IAnimationSamplerData> {
            let _data = Promise.all([
                (<any>this)._loadPiFloatAccessorAsync(channel[2], channel[3], channel[4], channel[5], bufferIndex, -1),
                (<any>this)._loadPiFloatAccessorAsync(channel[6], channel[7], channel[8], channel[9], bufferIndex, channel[0])
            ]).then(([inputData, outputData]) => {
                return {
                    input: inputData,
                    interpolation: channel[10],
                    output: outputData,
                };
            });
            // let a =
            //     this._loadPiFloatAccessorAsync(channel[2],channel[3],channel[4],channel[5], bufferIndex,-1),
            //    b= this._loadPiFloatAccessorAsync(channel[6],channel[7],channel[8],channel[9], bufferIndex,channel[0]);
    
            //     cb( {
            //         input: a,
            //         interpolation: channel[10],
            //         output: b,
            //     });
    
            return _data;
        };
    
        (<any>BABYLON.GLTF2.GLTFLoader.prototype)._loadPiFloatAccessorAsync = function(this: GLTFLoader, byteOffset: number, count: number, bufferByteOffset: number, bufferLength: number, index: number, type: number): Promise<Float32Array> {
            let type1 = "SCALAR";
            /**
             * translation
             * scale
             * color
             * rotation_eular
             * lightdiffuse
             */
            if (type == 0 || type == 2 || type == 8 || type == 12 || type == 14) {
                type1 = "VEC3";
            }else if (type == 1) {
                type1 = "VEC4";
            }
            // TODO: support normalized and stride
            const numComponents = _GetPiComponents('', type1);
            const length = numComponents * count;
    
            let _data = (<any>this).loadPiBufferAsync(bufferByteOffset, bufferLength, index).then((data) => {
                return (<any>BABYLON.GLTF2.GLTFLoader)._GetTypedArray('', 5126, data, byteOffset, length);
            });
    
            // let _data = this.loadPiBufferAsync(bufferByteOffset,bufferLength,index);
            //     return GLTFLoader._GetTypedArray('', 5126, _data, byteOffset, length);
    
            return _data as Promise<Float32Array>;
        };
    
        (<any>BABYLON.GLTF2.GLTFLoader.prototype).loadPiBufferAsync = function(this: GLTFLoader, bufferByteOffset: number, bufferLength: number, index: number): Promise<ArrayBufferView> {
    
            const buffer = BABYLON.GLTF2.ArrayItem.Get(`buffer`, this.gltf.buffers, index);
            let _data = (<any>this)._loadBufferAsync(`#/buffers/${index}`, buffer, bufferByteOffset, bufferLength).then((data) => {
                try {
                    return new Uint8Array(data.buffer, (bufferByteOffset || 0), bufferLength);
                }
                catch (e) {
                    throw new Error(`loadPiBufferAsync: ${e.message}`);
                }
            });
    
            // return new Uint8Array(this.buffer.buffer, this.buffer.byteOffset + (bufferByteOffset || 0), bufferLength);
    
            return _data;
        }
    
        function _GetPiComponents(context: string, type: string): number {
            switch (type) {
                case "SCALAR": return 1;
                case "VEC2": return 2;
                case "VEC3": return 3;
                case "VEC4": return 4;
                case "MAT2": return 4;
                case "MAT3": return 9;
                case "MAT4": return 16;
            }
    
            throw new Error(`${context}: Invalid type (${type})`);
        }
        // PI_END
        
        /** @hidden */
        (<any>BABYLON.GLTF2.GLTFLoader.prototype).dispose = function() {
            if (this._disposed) {
                return;
            }

            this._disposed = true;

            for (const name in this._extensions) {
                const extension = this._extensions[name];
                extension.dispose && extension.dispose();
                delete this._extensions[name];
            }

            // PI_BEGIN
            this.cameraPromiseMap.clear();
            this.meshPromiseMap.clear();
            // PI_END

            (this._gltf as any) = null;
            (this._babylonScene as any) = null;
            (this._rootBabylonMesh as any) = null;

            this._completePromises = undefined;
            this._particles = undefined;
            this._localLights = undefined;
            this._babylonLights = undefined;
            this._skeletons = undefined;
            (<any>this)._gltf = undefined;

            (<any>this)._particleCloneInfo = undefined;
            (<any>this)._skeletonCloneInfo = undefined;

            this._parent.dispose();
        };
    
    (<any>BABYLON.GLTF2.GLTFLoader.prototype)._loadMaterialAsync = function (context, material, babylonMesh, babylonDrawMode, assign) {
        let resTab = this._metadata.resTab;
        let key = `${this._babylonScene.uid}#${this._rootUrl}${this._fileName}${context}`;
        if (!babylonMesh.geometryForMeshParticle && ExtensionTools.Option.optimizeMaterialWithResTab) {
            let pmat = resTab.get(BABYLON_MATERIAL_NEW, key);
            let promise;
            if (pmat == undefined) {
                if (assign === void 0) { assign = function () { }; }
                var extensionPromise = this._extensionsLoadMaterialAsync(context, material, babylonMesh, babylonDrawMode, assign);
                if (extensionPromise) {
                    return extensionPromise;
                }
                material._data = material._data || {};
                var babylonData = material._data[babylonDrawMode];
                if (!babylonData) {
                    this.logOpen(context + " " + (material.name || ""));
                    var babylonMaterial = this.createMaterial(context, material, babylonDrawMode);
                    babylonData = {
                        babylonMaterial: babylonMaterial,
                        babylonMeshes: [],
                        promise: this.loadMaterialPropertiesAsync(context, material, babylonMaterial)
                    };
                    material._data[babylonDrawMode] = babylonData;
                    BABYLON.GLTF2.GLTFLoader.AddPointerMetadata(babylonMaterial, context);
                    this._parent.onMaterialLoadedObservable.notifyObservers(babylonMaterial);
                    this.logClose();
                }
                pmat = resTab.createRes(key, BABYLON_MATERIAL_NEW, babylonData.babylonMaterial, []);
                promise = babylonData.promise;
            } else {
                promise = Promise.resolve();
            }

            if (babylonMesh) {
                // babylonData.babylonMeshes.push(babylonMesh);
                // babylonMesh.onDisposeObservable.addOnce(function () {
                //     var index = babylonData.babylonMeshes.indexOf(babylonMesh);
                //     if (index !== -1) {
                //         babylonData.babylonMeshes.splice(index, 1);
                //     }
                // });
            }
            assign(pmat.link);
            return promise.then(function () {
                return pmat.link;
            });
        } else {
            if (assign === void 0) { assign = function () { }; }
            var extensionPromise = this._extensionsLoadMaterialAsync(context, material, babylonMesh, babylonDrawMode, assign);
            if (extensionPromise) {
                return extensionPromise;
            }
            material._data = material._data || {};
            var babylonData = material._data[babylonDrawMode];
            if (!babylonData) {
                this.logOpen(context + " " + (material.name || ""));
                var babylonMaterial = this.createMaterial(context, material, babylonDrawMode);
                babylonData = {
                    babylonMaterial: babylonMaterial,
                    babylonMeshes: [],
                    promise: this.loadMaterialPropertiesAsync(context, material, babylonMaterial)
                };
                material._data[babylonDrawMode] = babylonData;
                BABYLON.GLTF2.GLTFLoader.AddPointerMetadata(babylonMaterial, context);
                this._parent.onMaterialLoadedObservable.notifyObservers(babylonMaterial);
                this.logClose();
            }
            if (babylonMesh) {
                babylonData.babylonMeshes.push(babylonMesh);
                babylonMesh.onDisposeObservable.addOnce(function () {
                    var index = babylonData.babylonMeshes.indexOf(babylonMesh);
                    if (index !== -1) {
                        babylonData.babylonMeshes.splice(index, 1);
                    }
                });
            }
            assign(babylonData.babylonMaterial);
            return babylonData.promise.then(function () {
                return babylonData.babylonMaterial;
            });
        }
    };
    }
    //#endregion

    /** @hidden */
    interface _IAnimationSamplerData {
        input: Float32Array;
        interpolation: AnimationSamplerInterpolation;
        output: Float32Array;
    };

    // PI_END
};

type AnimationValue = number | BABYLON.Vector2 | BABYLON.Vector3 | BABYLON.Vector4 | BABYLON.Matrix  | BABYLON.Quaternion | number[] | BABYLON.Color3 ;
export class AnimationKey<T extends AnimationValue > {
    public constructor(
        public frame:     number,
        public value: T,
        public interpolation?: BABYLON.AnimationKeyInterpolation,
        public inTangent?: T,
        public outTangent?: T,
    ) {
        
    }
}
