import { BABYLON } from "../babylon";

export class MathDataPool {
    private static _Color3: BABYLON.Color3[] = [];
    private static _Color4: BABYLON.Color4[] = [];
    private static _Vector3: BABYLON.Vector3[] = [];
    private static _Vector4: BABYLON.Vector4[] = [];
    private static _Quaternion: BABYLON.Quaternion[] = [];
    private static _Matrix: BABYLON.Matrix[] = [];
    public static MAX_CACHE: number = 10000;
    public static Init() {
        BABYLON.Vector3.Zero = function() {
            return MathDataPool.Vector3(0, 0, 0);
        }
        BABYLON.Vector3.One = function() {
            return MathDataPool.Vector3(1, 1, 1);
        }
        BABYLON.Vector3.FromArray = function(array: BABYLON.DeepImmutable<ArrayLike<number>>, offset: number = 0) {
            return MathDataPool.Vector3(array[offset + 0], array[offset + 1], array[offset + 2]);
        }
        BABYLON.Color3.White = function() {
            return MathDataPool.Color3(1, 1, 1);
        }
        BABYLON.Color3.Black = function() {
            return MathDataPool.Color3(0, 0, 0);
        }
        BABYLON.Color3.FromArray = function(array: BABYLON.DeepImmutable<ArrayLike<number>>, offset: number = 0) {
            return MathDataPool.Color3(array[offset + 0], array[offset + 1], array[offset + 2]);
        }
        BABYLON.Color4.FromArray = function(array: BABYLON.DeepImmutable<ArrayLike<number>>, offset: number = 0) {
            return MathDataPool.Color4(array[offset + 0], array[offset + 1], array[offset + 2], array[offset + 3]);
        }
        BABYLON.Vector4.Zero = function() {
            return MathDataPool.Vector4(0, 0, 0, 0);
        }
        BABYLON.Vector4.One = function() {
            return MathDataPool.Vector4(1, 1, 1, 1);
        }
        BABYLON.Quaternion.Zero = function() {
            return MathDataPool.Quaternion(0, 0, 0, 0);
        }
        BABYLON.Quaternion.Identity = function() {
            return MathDataPool.Quaternion(0, 0, 0, 1);
        }
        BABYLON.Matrix.Identity = function() {
            return MathDataPool.Matrix();
        }
        BABYLON.Matrix.Zero = function() {
            let result = MathDataPool.Matrix();
            BABYLON.Matrix.FromValuesToRef(0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, result);
            return result;
        }
        BABYLON.Matrix.FromArray = function(array: BABYLON.DeepImmutable<ArrayLike<number>>, offset: number = 0) {
            let result = MathDataPool.Matrix();
            BABYLON.Matrix.FromArrayToRef(array, offset, result);
            return result;
        }
        BABYLON.Matrix.FromValues = function(initialM11: number, initialM12: number, initialM13: number, initialM14: number, initialM21: number, initialM22: number, initialM23: number, initialM24: number, initialM31: number, initialM32: number, initialM33: number, initialM34: number, initialM41: number, initialM42: number, initialM43: number, initialM44: number) {
            let result = MathDataPool.Matrix();
            BABYLON.Matrix.FromValuesToRef(initialM11, initialM12, initialM13, initialM14, initialM21, initialM22, initialM23, initialM24, initialM31, initialM32, initialM33, initialM34, initialM41, initialM42, initialM43, initialM44, result);
            return result;
        }

        /**
         * Remove a transform node for the list of scene's transform nodes
         * @param toRemove defines the transform node to remove
         * @returns the index where the transform node was in the transform node list
         */
        BABYLON.Node.prototype.dispose = function(doNotRecurse?: boolean, disposeMaterialAndTextures = false): number {
            this._isDisposed = true;
    
            if (!doNotRecurse) {
                const nodes = this.getDescendants(true);
                for (const node of nodes) {
                    node.dispose(doNotRecurse, disposeMaterialAndTextures);
                }
            }
    
            if (!this.parent) {
                this._removeFromSceneRootNodes();
            } else {
                this.parent = null;
            }
    
            // Callback
            this.onDisposeObservable.notifyObservers(this);
            this.onDisposeObservable.clear();
    
            // Behaviors
            for (var behavior of this._behaviors) {
                behavior.detach();
            }
    
            this._behaviors = [];

            // MathDataPool.Recycle(this._position);
            // MathDataPool.Recycle(this._rotation);
            // MathDataPool.Recycle(this._rotationQuaternion);
            // MathDataPool.Recycle(this._scaling);
            MathDataPool.Recycle(this._forward);
            MathDataPool.Recycle(this._up);
            MathDataPool.Recycle(this._localMatrix);
            MathDataPool.Recycle(this._worldMatrix);
            MathDataPool.Recycle(this._absolutePosition);
            MathDataPool.Recycle(this._absoluteScaling);
            MathDataPool.Recycle(this._absoluteRotationQuaternion);
            MathDataPool.Recycle(this._pivotMatrix);
            MathDataPool.Recycle(this._poseMatrix);
            // this._position = undefined;
            // this._rotation = undefined;
            // this._rotationQuaternion = undefined;
            // this._scaling = undefined;
            this._forward = undefined;
            this._up = undefined;
            this._worldMatrix = undefined;
            this._localMatrix = undefined;
            this._absolutePosition = undefined;
            this._absoluteScaling = undefined;
            this._absoluteRotationQuaternion = undefined;
            this._pivotMatrix = undefined;
            this._poseMatrix = undefined;
        }
    }
    public static Color3(r: number = 0, g: number = 0, b: number = 0) {
        let result = MathDataPool._Color3.pop();
        if (result == undefined) {
            result = new BABYLON.Color3();
        }

        result.copyFromFloats(r, g, b);

        return result;
    }
    public static Color4(r: number = 0, g: number = 0, b: number = 0, a: number = 0) {
        let result = MathDataPool._Color4.pop();
        if (result == undefined) {
            result = new BABYLON.Color4();
        }

        result.copyFromFloats(r, g, b, a);

        return result;
    }
    public static Vector3(x: number = 0, y: number = 0, z: number = 0) {
        let result = MathDataPool._Vector3.pop();
        if (result == undefined) {
            result = new BABYLON.Vector3();
        }

        result.copyFromFloats(x, y, z);

        return result;
    }
    public static Quaternion(x: number = 0, y: number = 0, z: number = 0, w: number = 1) {
        let result = MathDataPool._Quaternion.pop();
        if (result == undefined) {
            result = new BABYLON.Quaternion();
        }

        result.copyFromFloats(x, y, z, w);

        return result;
    }
    public static Matrix() {
        let result = MathDataPool._Matrix.pop();
        if (result == undefined) {
            result = new BABYLON.Matrix();
        }

        BABYLON.Matrix.IdentityToRef(result);

        return result;
    }
    public static Vector4(x: number = 0, y: number = 0, z: number = 0, w: number = 0) {
        let result = MathDataPool._Vector4.pop();
        if (result == undefined) {
            result = new BABYLON.Vector4(x, y, z, w);
        }

        result.copyFromFloats(x, y, z, w);

        return result;
    }
    public static Recycle(item: BABYLON.Color3 | BABYLON.Color4 | BABYLON.Vector3 | BABYLON.Vector4 | BABYLON.Matrix | BABYLON.Quaternion) {
        if (!item) return;
        if (item instanceof BABYLON.Color3) {
            if (MathDataPool._Color3.length > MathDataPool.MAX_CACHE) return;
            MathDataPool._Color3.push(item);
        } else if (item instanceof BABYLON.Color4) {
            if (MathDataPool._Color4.length > MathDataPool.MAX_CACHE) return;
            MathDataPool._Color4.push(item);
        } else if (item instanceof BABYLON.Vector3) {
            if (MathDataPool._Vector3.length > MathDataPool.MAX_CACHE) return;
            MathDataPool._Vector3.push(item);
        } else if (item instanceof BABYLON.Vector4) {
            if (MathDataPool._Vector4.length > MathDataPool.MAX_CACHE) return;
            MathDataPool._Vector4.push(item);
        } else if (item instanceof BABYLON.Quaternion) {
            if (MathDataPool._Quaternion.length > MathDataPool.MAX_CACHE) return;
            MathDataPool._Quaternion.push(item);
        } else if (item instanceof BABYLON.Matrix) {
            if (MathDataPool._Matrix.length > MathDataPool.MAX_CACHE) return;
            MathDataPool._Matrix.push(item);
        }
    }
}