
/**
 *
 */

import { BABYLON } from './babylon';
import { Engine } from './engine';
import { ResLoader } from './loader';
import { IMeshParticleSystemScene } from '../interface/particlesystem';
import { MeshParticleSystemBatch } from '../mesh_particle/mesh_particle_system_batch';
import { ISpriteAtlas, ISpriteManager, ISpriteScene } from '../interface/sprite';
import { SpriteManager } from '../sprite/sprite_manager';
import { Constants, ExtensionTools, ResTab } from '../interface/index';
import { GlobalAtlasManager } from '../sprite/sprite_atlas';
import { CommonBufferManager } from './commonbuffer';
import { IPickingInfo } from '../render/types';
import { PickingInfo } from '../render/collisions/pickingInfo';

/**
 * 封装项目层使用的 场景命中信息
 * * 底层处理过命中检测 - 只通过包围盒检测
 */
// tslint:disable-next-line:interface-name
export interface IRayCastInfo extends BABYLON.PickingInfo {
    hit: boolean;
    pickedPoint: BABYLON.Nullable<BABYLON.Vector3>;
    /**
     * 命中的目标mesh
     */
    target: BABYLON.AbstractMesh;
    rayID: number;
}

// 重载BABYLON.Scene， 释放资源
export class Scene extends BABYLON.Scene implements IMeshParticleSystemScene, ISpriteScene {
    public static IDCounter: number = 0;
    public loader: ResLoader;
    public piID: number;
    /**
     * 场景在window中的显示范围, 如果attachControl传入gui的document,且场景渲染范围与gui渲染范围对应,则传入偏移为0,宽高为gui的document的宽高
     */
    clientRect: { left: number, top: number, width: number, height: number } = { left: 0, top: 0, width: 100, height: 100 };
    constructor(engine: Engine, opt?: BABYLON.SceneOptions) {
        super(engine, opt);
        this.offlineProvider = Engine.ofl;
        this.loader = new ResLoader(ExtensionTools.CreateResTab());
        this.piID = Scene.IDCounter++;

        this.__commonBuffer = new CommonBufferManager(this);

        this.onBeforeRenderObservable.add(() => {
            this._updateParticleSystemBatch();
            this._updateSprites();

            this._updateCommonBuffer();
        });
        this.onAfterRenderObservable.add(() => {
            if (this.autoClearSpite) {
                this.cleanSpiteManagerNoSprite();
            }
        });
    }
    //#region MeshParticleSystemBatch
    public particleSystemTime = 0;
    private particleSystemBatches: Map<string, MeshParticleSystemBatch> = new Map();
    public autoClearSpite: boolean = true;
    getParticleSystemBatche(key: string): MeshParticleSystemBatch {
        let result = this.particleSystemBatches.get(key);
        if (result == undefined) {
            result = new MeshParticleSystemBatch(this, key);
            this.particleSystemBatches.set(key, result);
        }
        return result;
    }
    private _updateParticleSystemBatch = () => {
        let time = (window.performance || Date).now();
        this.particleSystemBatches.forEach((item) => {
            item._update();
        });
        this.particleSystemTime = (window.performance || Date).now() - time;
    }
    //#endregion

    //#region Sprite
    public spriteUpdateTime = 0;
    private _spriteManagers: Map<string, ISpriteManager> = new Map();
    registSpritAtlas(cfg: ISpriteAtlas, maxSprite?: number, material?: BABYLON.Material, pass?: Constants.PassTag): void {
        if (this._spriteManagers.get(cfg.image) == undefined) {
            let atlas = GlobalAtlasManager.getAtlas(cfg);
            maxSprite = GlobalAtlasManager.atlasSpriteLimit.get(cfg.image).spriteLimit || maxSprite || SpriteManager.MAX_SPRITE;
            let item = new SpriteManager(this, atlas, maxSprite, material, pass);
            this._spriteManagers.set(cfg.image, item);
        }
    }
    getSpritManager(atlas: string): ISpriteManager {
        atlas = GlobalAtlasManager.queryRealAtlasName(atlas);
        let result = this._spriteManagers.get(atlas);
        if (result == undefined) {
            let cfg = GlobalAtlasManager.getAtlas(atlas);
            if (cfg) {
                this.registSpritAtlas(cfg.atlas);
            }
            result = this._spriteManagers.get(atlas);
        }
        return result;
    }
    delSpritManager(managr: ISpriteManager): void {
        this._spriteManagers.delete(managr.atlas.image);
    }
    /**
     * 清除没有Sprite的SpriteManager - 会清除内部 Mesh,Texture
     */
    cleanSpiteManagerNoSprite() {
        this._spriteManagers.forEach((item) => {
            if ((<SpriteManager>item).spriteCount() == 0) {
                item.dispose();
            }
        });
    }
    private _updateSprites = () => {
        let time = (window.performance || Date).now();
        this._spriteManagers.forEach((item) => {
            (<SpriteManager>item)._onFrameUpdate();
        });
        this.spriteUpdateTime = (window.performance || Date).now() - time;
    }
    //#endregion

    //#region CommonBuffer
    private __commonBuffer: CommonBufferManager;
    private _updateCommonBuffer() {
        this.__commonBuffer._onFrameUpdate();
    }
    //#endregion

    /**
     * 3D 场景点击命中结果
     * @return
     * * .pickedMesh.link: 命中的mesh
     */
    public pick3D() {
        const pickRes: BABYLON.PickingInfo = this.pickWithBoundingInfo(
            this.pointerX,
            this.pointerY,
            (mesh) => {
                return mesh.isPickable && mesh.isEnabled();
            },
            true
        );
        (<IRayCastInfo>pickRes).target = pickRes.pickedMesh;
        (<IRayCastInfo>pickRes).rayID = (<IRayCastInfo>pickRes).target ? (<any>(<IRayCastInfo>pickRes).target)?.rayID : null;

        return (<IRayCastInfo>pickRes);
    }

    pickWithColliderInfo(x: number, y: number, camera?: BABYLON.Camera): IPickingInfo {
        // Dummy info if picking as not been imported
        const pi = new PickingInfo();

        if (!camera) {
            if (!this.activeCamera) {
                return undefined;
            }
            camera = this.activeCamera;
        }

        pi.hit = false;
        var pickingInfo = new BABYLON.PickingInfo();
        pickingInfo.ray.update(x, y, 1, 1, BABYLON.Matrix.IdentityReadOnly, BABYLON.Matrix.IdentityReadOnly, camera.getTransformationMatrix());
        let rayFunction = (world: BABYLON.Matrix) => {
            return pickingInfo.ray;
        }
        let trianglePredicate = undefined;
        for (var meshIndex = 0; meshIndex < this.meshes.length; meshIndex++) {
            var mesh = this.meshes[meshIndex];
            if (!mesh.isEnabled() || !mesh.isVisible || !mesh.isPickable) {
                continue;
            }
            var world = mesh.skeleton && mesh.skeleton.overrideMesh ? mesh.skeleton.overrideMesh.getWorldMatrix() : mesh.getWorldMatrix();
            if (mesh.hasThinInstances && (<BABYLON.Mesh>mesh).thinInstanceEnablePicking) {
                // first check if the ray intersects the whole bounding box/sphere of the mesh
                var result = this._internalPickForMesh(pickingInfo, rayFunction, mesh, world, true, true, trianglePredicate);
                break;
            }
            else {
                var result = this._internalPickForMesh(pickingInfo, rayFunction, mesh, world, true, true, trianglePredicate);
                if (result) {
                    pickingInfo = result;
                    break;
                }
            }
        }
        if (pickingInfo.hit) {
            pi.hit = true;
            pi.pickedMesh = pickingInfo.pickedMesh;
            pi.pickedPoint = pickingInfo.pickedPoint;
            pi._pickingUnavailable = true;
        }

        return pi;
    }

    dispose(): void {
        this.loader.dispose();
        super.dispose();
    }
}