import { BABYLON } from "../render3d/babylon";
import { RibbonBuilder } from "./ribbon";
import { SceneStruct } from "../scene_struct";
import { IPath, IRibbonRenderer } from "./interface";

export class RibbonRender implements IRibbonRenderer {
    /**
     * 最大点数
     */
    private readonly maxCount: number;
    /**
     * 两点之间最小有效距离
     */
    private readonly minimumDistance = 0.02;
    /**
     * 水波扩散宽度
     */
    private size = 0.35;
    /**
     * 脏更新标志
     */
    private dirty = false;
    /**
     * 水波流动速度
     */
    private readonly textureUOffSpeed = 0.10;
    private readonly textureVOffSpeed = 0.10;

    private readonly uvs: Float32Array;
    private readonly colors: Float32Array;
    private readonly indices: Uint16Array;
    private readonly positions: Float32Array;

    private _image: string;
    public get image() {
        return this._image;
    }
    public set image(path: string) {
        if (this._image != path) {
            this._image = path;
            if (this.material.diffuseTexture) {
                (<BABYLON.Texture>this.material.diffuseTexture).updateURL(this._image);
            } else {
                let texture = new BABYLON.Texture(null, this.scene.impl, true, false);
                texture.updateURL(this._image);
                this.material.diffuseTexture = texture;
            }
        }
    }

    public get alphaIndex() {
        return this.lineSource.alphaIndex;
    }
    public set alphaIndex(val: number) {
        this.lineSource.alphaIndex = val;
    }

    constructor(private readonly scene: SceneStruct, maxCount = 128) {
        this.maxCount = maxCount;
        this.uvs = new Float32Array(this.maxCount * 3 * 2);
        this.colors = new Float32Array(this.maxCount * 3 * 4);
        this.indices = new Uint16Array((this.maxCount - 1) * 12);
        this.positions = new Float32Array(this.maxCount * 3 * 3);

        this.createLineSourceMesh();
        this.scene.registerBeforeRenderCall(this.render);
    }

    /**
     * 添加一个点
     */
    public addPoint(newPoint: BABYLON.Vector3) {
        this.basePath.push(newPoint);
        this.dirty = true;
    }

    public resetPoint(newPoint: BABYLON.Vector3[]) {
        this.basePath = newPoint;
        this.dirty = true;
    }

    public updateScale(uScale: number, vScale: number) {
        if (this.isLine) {
            let texture = <BABYLON.Texture>(<BABYLON.StandardMaterial>this.lineSource?.material)?.diffuseTexture;
            if (texture) {
                texture.uScale = uScale;
                texture.vScale = vScale;
            }
        }

        if (this.isQuad) {
            let texture = <BABYLON.Texture>(<BABYLON.StandardMaterial>this.quadSource?.material)?.diffuseTexture;
            if (texture) {
                texture.uScale = uScale;
                texture.vScale = vScale;
            }
        }
    }

    //#region 调试数据
    private isDebug: boolean = false;
    private debugBoxSource: BABYLON.Mesh;
    private readonly debugBoxs: BABYLON.InstancedMesh[] = [];
    private createDebugBox() {
        let debugBoxMaterial = new BABYLON.StandardMaterial("debugMaterial", this.scene.impl);
        debugBoxMaterial.backFaceCulling = false;
        debugBoxMaterial.emissiveColor.copyFromFloats(1, 1, 1);
        debugBoxMaterial.disableLighting = true;
        this.debugBoxSource = BABYLON.BoxBuilder.CreateBox("Box", { size: this.size }, this.scene.impl);
        this.debugBoxSource.material = debugBoxMaterial;
        let ins = this.debugBoxSource.createInstance(`${-1}`);
        ins.parent = this.debugBoxSource;
        this.isDebug = true;
    }
    //#endregion

    //#region 点模式数据
    private isQuad: boolean = false;
    private quadSource: BABYLON.Mesh;
    private readonly quads: BABYLON.InstancedMesh[] = [];
    private quadPoints: [BABYLON.Vector3, number][] = [];
    public createQuadSourceMesh() {
        this.quadSource = BABYLON.PlaneBuilder.CreatePlane("quadSource", { size: 1 }, this.scene.impl);
        let material = new BABYLON.StandardMaterial("quadSourceMaterial", this.scene.impl);
        material.backFaceCulling = false;
        material.emissiveColor.copyFromFloats(1, 1, 1);
        material.disableLighting = true;
        let texture = new BABYLON.Texture(null, this.scene.impl, true, false);
        texture.updateURL("app_b/new_city/images/scene/eff_wsg_xulie_lz_002.png");
        texture.uScale = 1;
        texture.vScale = 1;
        texture.hasAlpha = true;
        material.diffuseTexture = texture;
        material.useAlphaFromDiffuseTexture = true;
        this.quadSource.material = material;
        let ins = this.quadSource.createInstance("a");
        ins.setEnabled(false);
        this.isQuad = true;
    }
    //#endregion

    /**
     * 外部逻辑添加的点
     */
    private basePath: BABYLON.Vector3[] = [];
    /**
     * 构建出来的水波点
     */
    private readonly resultPath: IPath = { points: [] };
    private isLine: boolean = false;
    private lineSource: BABYLON.Mesh;
    private material: BABYLON.StandardMaterial;
    private createLineSourceMesh() {
        let material = new BABYLON.StandardMaterial("meshSourceMaterial", this.scene.impl);
        material.backFaceCulling = false;
        material.emissiveColor.copyFromFloats(1, 1, 1);
        material.disableLighting = true;
        material.useAlphaFromDiffuseTexture = true;
        this.lineSource = new BABYLON.Mesh("meshSource", this.scene.impl);
        this.lineSource.setVerticesData(BABYLON.VertexBuffer.PositionKind, this.positions, true);
        this.lineSource.setVerticesData(BABYLON.VertexBuffer.ColorKind, this.colors, true);
        this.lineSource.setVerticesData(BABYLON.VertexBuffer.UVKind, this.uvs, true);
        this.lineSource.setIndices(this.indices, this.maxCount * 3, true);
        this.lineSource.material = material;
        this.lineSource.useVertexColors = true;
        this.lineSource.hasVertexAlpha = true;
        this.lineSource.alwaysSelectAsActiveMesh = true;
        this.isLine = true;
    }

    private destroyLineSourceMesh() {
        this.resultPath.points = [];
        if (this.lineSource) {
            this.lineSource.dispose();
            this.lineSource = null;
        }
    }

    public init() {
        // this.createDebugBox();
        // this.createQuadSourceMesh();
    }

    public destroy() {
        this.destroyLineSourceMesh();
        this.scene.unregisterBeforeRenderCall(this.render);
    }

    /**
     * 渲染
     */
    private render = () => {
        if (this.isLine) {
            let texture = <BABYLON.Texture>(<BABYLON.StandardMaterial>this.lineSource?.material)?.diffuseTexture;
            if (texture) {
                texture.uOffset = Date.now() % (1000 / this.textureUOffSpeed) / (1000 / this.textureUOffSpeed);
                texture.vOffset = Date.now() % (1000 / this.textureVOffSpeed) / (1000 / this.textureVOffSpeed);
            }
        }

        if (this.isQuad) {
            let texture = <BABYLON.Texture>(<BABYLON.StandardMaterial>this.quadSource?.material)?.diffuseTexture;
            if (texture) {
                texture.uOffset = Date.now() % (1000 / this.textureUOffSpeed) / (1000 / this.textureUOffSpeed);
                texture.vOffset = Date.now() % (1000 / this.textureVOffSpeed) / (1000 / this.textureVOffSpeed);
            }
        }

        if (this.dirty && this.basePath.length >= 2) {
            RibbonBuilder.FormatPathForceNormal(this.basePath, this.size, this.resultPath, false, this.minimumDistance, new BABYLON.Vector3(0, 0, 1));
            this.dirty = false;
            this.updateLines();
            this.updateDebug();
        }

        this.updateQuads();
    }

    /**
     * 更新线模式
     */
    private updateLines() {
        let count = this.resultPath.points.length;
        RibbonBuilder.Build(this.resultPath, 0, this.positions, this.colors, this.uvs, this.indices);
        this.lineSource.updateVerticesData(BABYLON.VertexBuffer.PositionKind, this.positions);
        this.lineSource.updateVerticesData(BABYLON.VertexBuffer.ColorKind, this.colors);
        this.lineSource.updateVerticesData(BABYLON.VertexBuffer.UVKind, this.uvs);
        this.lineSource.updateIndices(this.indices.slice(0, Math.max(count - 1, 0) * 12));
    }

    /**
     * 更新调试数据
     */
    private updateDebug() {
        if (!this.isDebug) return;
        let count = this.resultPath.points.length;
        this.debugBoxs.forEach((v) => {
            v.dispose();
        });
        for (let i = 0; i < count; i++) {
            for (let j = 0; j < 3; j++) {
                let ins = this.debugBoxSource.createInstance(`${i}`);
                ins.position.copyFromFloats(this.positions[(i * 3 + j) * 3 + 0], this.positions[(i * 3 + j) * 3 + 1], this.positions[(i * 3 + j) * 3 + 2]);
                this.debugBoxs.push(ins);
                ins.parent = this.debugBoxSource;
            }
        }
    }

    /**
     * 更新点模式数据
     */
    private updateQuads() {
        if (!this.isQuad) return;
        if (!this.dirty) return;
        this.quads.forEach((v) => {
            v.dispose();
        });
        this.quads.length = 0;
        this.quadPoints.forEach(([point, size], idx) => {
            let ins = this.quadSource.createInstance(`${idx}`);
            ins.position.copyFrom(point);
            ins.scaling.copyFromFloats(size, size, size);
            this.quads.push(ins);
        });
    }
}