import { BoundingBox, BoundingSphere } from "./render3d/extends/bounding_info";
import { BABYLON } from "./render3d/babylon";
import { Ray } from "./render3d/extends/ray";
import { IColliderItem, IPiCollider } from "./interface/extends";

export class PiCollider implements IPiCollider {
    public static OutSideScale = Math.sqrt(3);
    public static CutSideScale = Math.sqrt(2) / PiCollider.OutSideScale;
    public static IncutSideScale = 1 / PiCollider.OutSideScale;
    boundingSphere: BoundingSphere;
    boundingBox: BoundingBox;
    /**
     * Constructs bounding info
     * @param minimum min vector of the bounding box/sphere
     * @param maximum max vector of the bounding box/sphere
     * @param worldMatrix defines the new world matrix
     */
    constructor(minimum: BABYLON.DeepImmutable<BABYLON.Vector3>, maximum: BABYLON.DeepImmutable<BABYLON.Vector3>, worldMatrix?: BABYLON.DeepImmutable<BABYLON.Matrix>, public intersectionTreshold: number = 0) {
        this.boundingSphere = new BoundingSphere(minimum, maximum, worldMatrix);
        this.boundingBox = new BoundingBox(minimum, maximum, worldMatrix);
    }
    /**
     * Recreates the entire bounding info from scratch as if we call the constructor in place
     * @param min defines the new minimum vector (in local space)
     * @param max defines the new maximum vector (in local space)
     * @param worldMatrix defines the new world matrix
     */
    public reConstruct(min: BABYLON.DeepImmutable<BABYLON.Vector3>, max: BABYLON.DeepImmutable<BABYLON.Vector3>, worldMatrix?: BABYLON.DeepImmutable<BABYLON.Matrix>) {
        this.boundingSphere.reConstruct(min, max, worldMatrix);
        this.boundingBox?.reConstruct(min, max, worldMatrix);
    }

    /**
     * Updates the bounding sphere and box
     * @param world world matrix to be used to update
     */
    public update(world: BABYLON.DeepImmutable<BABYLON.Matrix>) {
        this.boundingSphere._update(world);
        this.boundingBox?._update(world);
    }
    
    public intersects(ray: Ray) {
        return ray.intersectsSphere(this.boundingSphere, this.boundingSphere.radiusWorld * this.intersectionTreshold) && ray.intersectsBoxMinMax(this.boundingBox.minimumWorld, this.boundingBox.maximumWorld);
    }
}

export class ScenePickManager {
    public static Initial() {
        (<any>BABYLON.TransformNode.prototype)._afterComputeWorldMatrix = function() {
            let _colliderBoundle = <PiCollider>this._colliderBoundle;
            if (_colliderBoundle) {
                _colliderBoundle.update(this.worldMatrixFromCache);
            }
        }
    }
    public listeners: Set<IColliderItem> = new Set();
    public pointUplisteners: Map<IColliderItem, Function> = new Map();
    public pointDownlisteners: Map<IColliderItem, Function> = new Map();
    public picked: BABYLON.Mesh;
    protected colliderItems: IColliderItem[] = [];
    protected ray = new Ray(new BABYLON.Vector3(), new BABYLON.Vector3(0, 0, 1));
    constructor(public readonly scene: BABYLON.Scene) {

    }
    public addCollider(target: IColliderItem, minimum?: BABYLON.DeepImmutable<BABYLON.Vector3>, maximum?: BABYLON.DeepImmutable<BABYLON.Vector3>) {
        if (target._colliderBoundle == undefined && target._boundingInfo == undefined) {
            if (!maximum) {
                maximum = new BABYLON.Vector3( 0.5,  0.5,  0.5);
            }
            if (!maximum) {
                minimum = new BABYLON.Vector3(-0.5, -0.5, -0.5);
            }
            target._colliderBoundle = new PiCollider(minimum, maximum, target.worldMatrixFromCache);
        } else {
            if (minimum && maximum) {
                if (target._colliderBoundle) {
                    target._colliderBoundle.reConstruct(minimum, maximum, target.worldMatrixFromCache);
                } else if (target._boundingInfo) {
                    target._boundingInfo.reConstruct(minimum, maximum, target.worldMatrixFromCache);
                }
            }
        }
        if (this.listeners.has(target) == false) {
            this.listeners.add(target);
            target.onDisposeObservable.addOnce(() => {
                this.listeners.delete(target);
            });
        }
    }
    public addEventListener(target: IColliderItem, etype: "pointerup" | "pointerdown", func: Function) {
        if (this.listeners.has(target) == false) {
            if (!target._colliderBoundle && !target._boundingInfo) {
                let minimum = new BABYLON.Vector3(-0.5, -0.5, -0.5);
                let maximum = new BABYLON.Vector3( 0.5,  0.5,  0.5);
                target._colliderBoundle = new PiCollider(minimum, maximum, target.worldMatrixFromCache);
            }
            this.listeners.add(target);
            target.onDisposeObservable.addOnce(() => {
                this.listeners.delete(target);
            });
        }

        if (etype == "pointerup") {
            this.pointUplisteners.set(target, func);
        } else {
            this.pointDownlisteners.set(target, func);
        }
    }
    public onPointerUp = (pointerX: number, pointerY: number) => {
        this.colliderItems.length = 0;
        this.pointUplisteners.forEach((_func, key) => {
            this.colliderItems.push(key);
        });
        this.colliderItems.sort((a, b) => b.alphaIndex - a.alphaIndex);
        let target = this.pick(pointerX, pointerY, this.colliderItems);
        if (target) {
            let func = this.pointUplisteners.get(target);
            try {
                func();
            } catch (error) {
                
            }
        }
    }
    public onPointerDown = (pointerX: number, pointerY: number) => {
        this.colliderItems.length = 0;
        this.pointDownlisteners.forEach((_func, key) => {
            this.colliderItems.push(key);
        });
        this.colliderItems.sort((a, b) => b.alphaIndex - a.alphaIndex);
        let target = this.pick(pointerX, pointerY, this.colliderItems);
        if (target) {
            let func = this.pointDownlisteners.get(target);
            try {
                func();
            } catch (error) {
                
            }
        }
    }
    public pick(pointerX: number, pointerY: number, items?: IColliderItem[]): IColliderItem | undefined {
        if (items == undefined) {
            items = this.colliderItems;
            items.length = 0;
            this.listeners.forEach((key) => {
                items.push(key);
            });
        }
        let camera = this.scene.activeCamera;
        this.ray.update(pointerX, pointerY, 1, 1, BABYLON.Matrix.IdentityReadOnly, BABYLON.Matrix.IdentityReadOnly, camera.getTransformationMatrix());

        items.sort((a, b) => b.alphaIndex - a.alphaIndex);
        let lastIndex = 0;
        let isHit = false;
        let hitTarget: IColliderItem;
        let distance: number = Number.MAX_SAFE_INTEGER;
        for (let target of items) {
            if (lastIndex != target.alphaIndex && isHit) {
                break;
            }
            if (target.isEnabled() && target._colliderBoundle) {
                if (target._colliderBoundle.intersects(this.ray)) {
                    let tmpdistance = target.getDistanceToCamera(camera);
                    if (tmpdistance < distance) {
                        distance = tmpdistance;
                        hitTarget = target;
                    }
                    isHit = true;
                }
            }
            lastIndex = target.alphaIndex;
        }

        return hitTarget;
    }
}

