import { SceneStructBase } from "./scene_struct_base";
import { Scene, IRayCastInfo } from "./render3d/scene";
import { BABYLON } from "./render3d/babylon";
import { Scene3DEventInfo, SceneRenderFlags, IModelOpt, IModelObj } from "./interface/modelobj";
import { ModelObj } from "./model_obj";
import { factoryScene3D } from "./base_factory";
import { InputManager } from "./render/inputs/inputManager";
import { Nullable } from "./interface/types";
import { EventListener } from "./interface/event";

/**
 * 项目应用层的场景数据结构
 */
export class SceneStruct extends SceneStructBase {
    public static SceneMap: Map<string, SceneStruct> = new Map();
    /**
     * 底层场景对象
     */
    declare public impl: Scene;
    public viewportX: number;
    /**
     * 场景内渲染光表
     */
    public readonly lightMap: Map<string, BABYLON.Light> = new Map();
    // /**
    //  * 场景内 场景环境级别节点表
    //  */
    // public readonly appendMeshMap: Map<string, AppendModelObj> = new Map();
    public readonly pointerDownListenMap: Map<Function, (info: Scene3DEventInfo) => any> = new Map();
    public readonly pointerMoveListenMap: Map<Function, (info: Scene3DEventInfo) => any> = new Map();
    public readonly pointerUpListenMap: Map<Function, (info: Scene3DEventInfo) => any> = new Map();
    public readonly pointerClickListenMap: Map<Function, (info: Scene3DEventInfo) => any> = new Map();
    public readonly onBeforeDisposeObserver: BABYLON.Observable<any> = new BABYLON.Observable();
    public onBeforeDispose(func: () => void) {
        this.onBeforeDisposeObserver.addOnce(func);
    }
    private _prefabID: number = 0;
    public prefabID() {
        return this._prefabID++;
    }
    /**
     *
     * @param sceneName 场景命名
     * @param engine 引擎
     */
    constructor(sceneName: string, engine: BABYLON.Engine) {
        super(sceneName, engine);
        SceneStruct.SceneMap.set(sceneName, this);
        this.impl.defaultMaterial = new BABYLON.StandardMaterial("Default", this.impl);
        this.onPointerDown = (e: PointerEvent) => {
            (<any>this.impl)._inputManager._onPointerDown(e);
        };
        this.onPointerMove = (e: PointerEvent) => {
            (<any>this.impl)._inputManager._onPointerMove(e);
        };
        this.onPointerUp = (e: PointerEvent) => {
            (<any>this.impl)._inputManager._onPointerUp(e);
        };
        this.onPointerClick = (e: any) => {
            const rayInfo = this.impl.pick3D();
            this.pointerClickListenMap.forEach((func) => {
                func({ rayInfo, e, s: null });
            });
        };
        this.inputManager = new InputManager(this.impl);
    }

    public applyEnvironmentBRDFTexture(path: string) {
        if (!this.impl.environmentBRDFTexture) {
            let texture = new BABYLON.Texture(null, this.impl, true, false, BABYLON.Texture.BILINEAR_SAMPLINGMODE);
            this.impl.environmentBRDFTexture = texture;
            (<any>texture).updateURL(path, undefined, undefined, () => {
                if (texture._texture && !texture._texture.isReady) {
                    texture.isRGBD = true;
                    texture.wrapU = BABYLON.Texture.CLAMP_ADDRESSMODE;
                    texture.wrapV = BABYLON.Texture.CLAMP_ADDRESSMODE;
                    BABYLON.RGBDTextureTools.ExpandRGBDTexture(texture);
                }
            });
        }
    }

    public create(engine: BABYLON.Engine) {
        this.impl = factoryScene3D(<any>engine);
        this.impl.renderTargetsEnabled = true;
    }
    /**
     * 导入场景模型
     * @param objName 模型命名
     * * 用于项目层管理
     * @param opt 模型加载配置
     */
    public importScene(objName: string, opt: IModelOpt = {}) {
        // const model = new AppendModelObj(objName, this, opt);

        // this.appendMeshMap.set(objName, model);
    }
    /**
     * 插入模型
     * @param objName 模型命名
     * * 用于项目层管理
     * @param opt 模型加载配置
     */
    public insertMesh(objName: string, opt: IModelOpt = {}): IModelObj {
        // tslint:disable-next-line:no-unnecessary-local-variable
        const model = new ModelObj(objName, <any>this, opt);

        return model;
    }
    /**
     * 销毁
     */
    public dispose() {
        if (this.impl) {
            SceneStruct.SceneMap.delete(this.name);
            this.disposeObserver(this.camera);
            this.onBeforeDisposeObserver.notifyObservers(undefined);
            // this.appendMeshMap.clear();
            this.impl.dispose();
            this.camera = undefined;
            this.renderFlag = SceneRenderFlags.dispose;
        }
    }
    /**
     * 停止所有动画
     */
    public stopAnim() {
        this.impl.stopAllAnimations();
    }

    /**
     * 获取操作点在场景的命中信息
     * @deprecated 使用 modelprefab.onPointerObserver 
     * @param listener 
     */
    public pickCall = (e: BABYLON.PointerInfo, s: BABYLON.EventState) => {
        let rayInfo: IRayCastInfo;
        switch (e.type) {
            case BABYLON.PointerEventTypes.POINTERDOWN: {
                if (this.pointerDownListenMap.size > 0) {
                    // rayInfo = this.impl.pick3D();
                    rayInfo = <any>e.pickInfo;
                    rayInfo.target = rayInfo.pickedMesh;
                    if ((<any>rayInfo.pickedMesh)) {
                        rayInfo.rayID = (<any>rayInfo.pickedMesh).rayID;
                    }

                    this.pointerDownListenMap.forEach((func) => {
                        func({ rayInfo, e, s });
                    });
                }
                break;
            }
            case BABYLON.PointerEventTypes.POINTERMOVE: {
                rayInfo = this.impl.pick3D();
                this.pointerMoveListenMap.forEach((func) => {
                    func({ rayInfo, e, s });
                });
                break;
            }
            case BABYLON.PointerEventTypes.POINTERUP: {
                // rayInfo = this.impl.pick3D();

                rayInfo = <any>e.pickInfo;
                rayInfo.target = rayInfo.pickedMesh;
                if ((<any>rayInfo.pickedMesh)) {
                    rayInfo.rayID = (<any>rayInfo.pickedMesh).rayID;
                }

                this.pointerUpListenMap.forEach((func) => {
                    func({ rayInfo, e, s });
                });
                break;
            }
            // tslint:disable-next-line:no-empty
            default: { }
        }
    }
    /**
     * @deprecated 使用 modelprefab.onPointerObserver
     * @param listener 
     */
    public addDownListen = (listener: (info: Scene3DEventInfo) => any) => {
        this.pointerDownListenMap.set(listener, listener);
    }
    /**
     * @deprecated 使用 modelprefab.onPointerObserver
     * @param listener 
     */
    public removeDownListen = (listener: (info: Scene3DEventInfo) => any) => {
        this.pointerDownListenMap.delete(listener);
    }
    /**
     * @deprecated 使用 modelprefab.onPointerObserver
     * @param listener 
     */
    public addUpListen = (listener: (info: Scene3DEventInfo) => any) => {
        this.pointerUpListenMap.set(listener, listener);
    }
    /**
     * @deprecated 使用 modelprefab.onPointerObserver
     * @param listener 
     */
    public removeUpListen = (listener: (info: Scene3DEventInfo) => any) => {
        this.pointerUpListenMap.delete(listener);
    }
    /**
     * @deprecated 使用 modelprefab.onPointerObserver
     * @param listener 
     */
    public addMoveListen = (listener: (info: Scene3DEventInfo) => any) => {
        this.pointerMoveListenMap.set(listener, listener);
    }
    /**
     * @deprecated 使用 modelprefab.onPointerObserver
     * @param listener 
     */
    public removeMoveListen = (listener: (info: Scene3DEventInfo) => any) => {
        this.pointerMoveListenMap.delete(listener);
    }
    /**
     * @deprecated 使用 modelprefab.onPointerObserver
     * @param listener 
     */
    public addClickListen = (listener: (info: Scene3DEventInfo) => any) => {
        this.pointerClickListenMap.set(listener, listener);
    }
    /**
     * @deprecated 使用 modelprefab.onPointerObserver
     * @param listener 
     */
    public removeClickListen = (listener: (info: Scene3DEventInfo) => any) => {
        this.pointerClickListenMap.delete(listener);
    }

    protected inputManager: InputManager;
    attachControl(attachUp = true, attachDown = true, attachMove = true, elementToAttachTo: Nullable<{
        addEventListener(type: string, listener: EventListener, options?: boolean | AddEventListenerOptions): void;
        removeEventListener(type: string, listener: EventListener, options?: boolean | AddEventListenerOptions): void;
    }> = null) {
        this.inputManager.attachControl(attachUp, attachDown, attachMove, elementToAttachTo);
    }
}
