import { BABYLON } from "./render3d/babylon";
import { Engine } from "./render3d/engine";
import { SceneRenderFlags } from "./interface/modelobj";
import { factorySceneGUI } from "./base_factory";
import { ModelObj } from "./model_obj";

export class SceneStructBase {
    public isDirty: boolean = true;
    public engine: Engine;
    /**
     * 是否启用自动优化
     * * 需要确保该场景是渲染到 FBO 的
     * * 需要确保场景渲染的目标FBO不会被其他逻辑 clear
     */
    public autoOptimize: boolean = false;
    protected _frameStep: number = 1;
    /**
     * 场景脏渲染的帧间隔
     */
    public get frameStep() {
        return this._frameStep;
    }
    /**
     * 场景脏渲染的帧间隔
     */
    public set frameStep(v: number) {
        this._frameStep = Math.max(0, Math.floor(v));
    }
    protected _frameCount = 1;
    /** @hidden */
    public _modelMap: Map<ModelObj, ModelObj> = new Map();
    /**
     * 场景名称
     */
    public name: string;
    /**
     * 底层场景对象
     */
    public impl: BABYLON.Scene;
    /**
     * 场景渲染状态
     */
    public renderFlag: string = SceneRenderFlags.pause;
    /**
     * 场景当前活动相机
     */
    public camera: BABYLON.Camera;
    /**
     * 场景内相机表
     */
    public cameraMap: Map<string, BABYLON.Camera> = new Map();
    public beforeRenderCallList: ((scene: SceneStructBase) => void)[] = [];
    public onBeforeDrawPhaseCallList: ((scene: SceneStructBase) => void)[] = [];
    public afterRenderCallList: ((scene: SceneStructBase) => void)[] = [];
    public onPointerUp: (e: PointerEvent) => void;
    public onPointerClick: (e: PointerEvent) => void;
    public onPointerMove: (e: PointerEvent) => void;
    public onPointerDown: (e: PointerEvent) => void;
    /**
     *
     * @param sceneName 场景命名
     * @param engine 引擎
     */
    constructor(sceneName: string, engine: BABYLON.Engine) {
        this.name = sceneName;
        this.engine = <Engine>engine;
        this.create(engine);
    }
    public create(engine: BABYLON.Engine) {
        this.impl = factorySceneGUI(engine);
    }
    /**
     * @deprecated
     */
    public render() {
        if (this.renderFlag === SceneRenderFlags.active) {
            // const gl: WebGLRenderingContext = <any>this.impl.getEngine()._gl;
            // gl.disable(gl.SCISSOR_TEST);
            this.beforeRenderCallList.forEach((f) => f(this));

            if (this.impl.animatables.length > 0 || this.impl.particleSystems.length > 0) {
                this.isDirty = true;
            }
            this._frameCount++;

            if (!this.autoOptimize || (this.isDirty && (this._frameCount % this.frameStep == 0))) {
                this.impl.onBeforeDrawPhaseObservable.addOnce(this.onBeforeDrawPhaseOnceCall);
                (<any>this.impl.getEngine())._setProgram(null);
                // this.impl.getEngine().setAlphaMode(BABYLON.Engine.ALPHA_DISABLE);
                this.impl.render();
                this.impl.getEngine().wipeCaches(true);
                this.isDirty = false;
            } else {
                this.onBeforeDrawPhaseOnceCall();
            }

            this.afterRenderCallList.forEach((f) => f(this));
        }
        // console.warn(`Models: ${this._modelMap.size}`);
    }
    public activeObserver = (camera: BABYLON.Camera) => {
        console.warn(`Scene ${this.name} active!`);
    }
    public disposeObserver = (camera: BABYLON.Camera) => {
        console.warn(`Scene ${this.name} dispose!`);
    }
    public changeCameraObserver = (camera: BABYLON.Camera) => {
        console.log(`Scene ${this.name} camera change!`);
    }
    public removeCameraObserver = (camera: BABYLON.Camera) => {
        console.log(`Scene ${this.name} camera change!`);
    }
    /**
     * 设置当前活动相机
     * @param cameraName 目标相机名称
     * * 在场景内部相机表中查找
     */
    public setCurrCamera(cameraName: string) {
        const camera = this.cameraMap.get(cameraName);
        if (!!camera) {
            this.camera = camera;
            this.impl.activeCamera = camera;
            this.changeCameraObserver(this.camera);
        } else {
            console.warn(`SceneStruct.setCurrCamera：Not Found Camera which name ${cameraName}.`);
        }
    }
    /**
     * 销毁相机
     * @param cameraName 目标相机名称
     */
    public removeCamera(cameraName: string) {
        const camera = this.cameraMap.get(cameraName);
        if (camera !== undefined) {
            if (this.camera === camera) {
                this.removeCameraObserver(this.camera);
                this.camera = undefined;
                this.renderFlag = SceneRenderFlags.pause;
            }
            this.cameraMap.delete(cameraName);
            camera.dispose();
        }
    }
    /**
     * 添加相机
     * @param camera 目标相机对象
     */
    public addCamera(camera: BABYLON.Camera) {
        this.cameraMap.set(camera.name, camera);
    }

    public pickCall = (e: BABYLON.PointerInfo, s: BABYLON.EventState) => {
        // 由子类实现
    }
    /**
     * 销毁
     */
    public dispose() {
        this.disposeObserver(this.camera);
        this.impl.dispose();
        this.camera = undefined;
        this.renderFlag = SceneRenderFlags.dispose;
    }
    /**
     * 暂停渲染
     */
    public pause = () => {
        this.renderFlag = SceneRenderFlags.pause;
        this.impl.onPointerObservable.removeCallback(this.pickCall);
    }
    /**
     * 激活渲染
     */
    public active = () => {
        if (this.camera !== undefined) {
            if (this.renderFlag !== SceneRenderFlags.dispose) {
                if (!this.impl.onPointerObservable.hasObservers()) {
                    this.impl.onPointerObservable.add(this.pickCall);
                }
                this.renderFlag = SceneRenderFlags.active;
                this.activeObserver(this.camera);
            }
        }
    }

    public registerBeforeRenderCall(f: (scene: SceneStructBase) => void) {
        if (this.beforeRenderCallList.indexOf(f) < 0) {
            this.beforeRenderCallList.push(f);
        }
    }
    public unregisterBeforeRenderCall(f: (scene: SceneStructBase) => void) {
        const index = this.beforeRenderCallList.indexOf(f);
        if (index >= 0) {
            this.beforeRenderCallList.splice(index, 1);
        }
    }

    public registerAfterRenderCall(f: (scene: SceneStructBase) => void) {
        if (this.afterRenderCallList.indexOf(f) < 0) {
            this.afterRenderCallList.push(f);
        }
    }
    public unregisterAfterRenderCall(f: (scene: SceneStructBase) => void) {
        const index = this.afterRenderCallList.indexOf(f);
        if (index >= 0) {
            this.afterRenderCallList.splice(index, 1);
        }
    }

    public registerBeforeDrawPhaseCall(f: (scene: SceneStructBase) => void) {
        if (this.onBeforeDrawPhaseCallList.indexOf(f) < 0) {
            this.onBeforeDrawPhaseCallList.push(f);
        }
    }
    public unregisterBeforeDrawPhaseCall(f: (scene: SceneStructBase) => void) {
        const index = this.onBeforeDrawPhaseCallList.indexOf(f);
        if (index >= 0) {
            this.onBeforeDrawPhaseCallList.splice(index, 1);
        }
    }

    protected onBeforeDrawPhaseOnceCall = () => {
        this.onBeforeDrawPhaseCallList.forEach((f) => {
            if (f) {
                f(this);
            }
        });
    }
}