import { BABYLON } from "../render3d/babylon";
import { CustomScene, IModelOpt, IOfflineAnimationCfg, ModelPrefab, PrefabModelObjPool, Sprite } from "../index";
import { ModelObj } from "../model_obj";
import { AnimationKey } from "../render3d/extends/gltf_loader";
import { TexturePackAtlas } from "./sprite_atlas";
import { MathDataPool } from "../render3d/extends/math";
import { ExtensionTools, IOfflineAnimationCfgBin, Res, ResTab } from "../interface/index";

type NodeName = string;
type Queue = number;
type SpritNode = 0 | 1;
type PosX = number;
type PosY = number;
type PosZ = number;
type RotX = number;
type RotY = number;
type RotZ = number;
type SclX = number;
type SclY = number;
type SclZ = number;
export type IAnimeSpritTree = [NodeName, SpritNode, Queue, PosX, PosY, PosZ, RotX, RotY, RotZ, SclX, SclY, SclZ, IAnimeSpritTree[]];
export interface IAnimeSpriteDesc {
    atlas: string,
    frame: string,
    atlasAnim?: { duration: number, animFrom: number, animTo: number },
}
export interface IAnimeSpriteOpt extends IModelOpt {
    key: string;
    tree?: IAnimeSpritTree;
    appendAnimation?: IOfflineAnimationCfg;
    defaultFrames: { [index: string]: IAnimeSpriteDesc };
}
export interface IAnimeSpriteBinPathOpt extends IAnimeSpriteOpt {
    spriteBinPath: string;
}

export const RES_ANIM_SPRITE_CFG = "RES_ANIM_SPRITE_CFG";
export class AnimeSpriteRes {
    constructor(
        public tree: IAnimeSpritTree,
        public appendAnimation: IOfflineAnimationCfg,
        public size: number,
    ) {

    }
}

export class AnimeSprite extends ModelObj {
    private static registResLoader() {
        ExtensionTools.ResLoadRegist(
            RES_ANIM_SPRITE_CFG,
            function(resTab: ResTab, type: typeof RES_ANIM_SPRITE_CFG, filePath: string, ...arg: any[]): Promise<AnimeSpriteRes> {
                return ExtensionTools.LoadFile(resTab, filePath).then((buffer: ArrayBuffer) => {
                    if (ArrayBuffer.isView(buffer)) {
                        buffer = buffer.buffer.slice(buffer.byteOffset, buffer.byteOffset + buffer.byteLength);
                    }
                    let jsonLen = (new Int32Array(<ArrayBuffer>buffer, 0, 1))[0];
                    let jsonStr = ExtensionTools.UTF8Decode(buffer.slice(4, 4 + jsonLen));
                    let json = JSON.parse(jsonStr);
                    let [tree, anim] = <[IAnimeSpritTree, IOfflineAnimationCfgBin]>json;
                    (<any>anim)[2] = new Uint8Array(buffer.slice(4 + jsonLen, 4 + jsonLen + anim[2]));

                    return new AnimeSpriteRes(tree, <any>anim, buffer.byteLength + 200);
                });
            },
            function(link: AnimeSpriteRes) {
                link.appendAnimation = undefined;
                link.tree = undefined;
            },
            function(res: AnimeSpriteRes) {
                return res.size
            }
        )
    }
    public static registBySpriteBinPath(scene: CustomScene, key: string, spriteBinPath: string, defaultFrames: { [index: string]: IAnimeSpriteDesc }, single: boolean = false, cacheCount: number = 8) {
        if (AnimeSprite.registResLoader) {
            AnimeSprite.registResLoader();
            AnimeSprite.registResLoader = undefined;
        }

        // if (scene.prefabPool.cfgmap.has(key)) return;
        scene.prefabPool.registPrefabCfg(key, { path: "", fileName: "", single, cacheCount, defaultFrames, key, spriteBinPath});
        PrefabModelObjPool.registPrefabFactory(key, ModelPrefab.Creation);
        PrefabModelObjPool.registPrefabResFactory(key, AnimeSprite.CreationWithSpriteBinPath);
    }
    public static CreationWithSpriteBinPath(meshName: string, scene: CustomScene, opt?: IAnimeSpriteBinPathOpt): AnimeSprite {
        let model = new AnimeSprite(meshName, scene, opt);
        let loadRes: Promise<Res<any, AnimeSpriteRes>>;
        let res: Res<any, AnimeSpriteRes> = model.resTab.get(RES_ANIM_SPRITE_CFG, opt.spriteBinPath);
        if (res == undefined) {
            loadRes = model.resTab.load(RES_ANIM_SPRITE_CFG, opt.spriteBinPath, []);
        } else {
            loadRes = Promise.resolve(res);
        }
        (model)._loadPromise = loadRes.then((res) => {
            opt.tree = res.link.tree;
            opt.appendAnimation = res.link.appendAnimation;
            let root = AnimeSprite.CreateNode(scene, opt.tree, model, opt.defaultFrames);
            (model).loaded(root, [root], [], [], [], [], []);
        });
        return model;
    }
    public static regist(scene: CustomScene, key: string, defaultFrames: { [index: string]: IAnimeSpriteDesc }, tree: IAnimeSpritTree, appendAnimation: IOfflineAnimationCfg, single: boolean = false, cacheCount: number = 8) {
        // if (scene.prefabPool.cfgmap.has(key)) return;
        scene.prefabPool.registPrefabCfg(key, { path: "", fileName: "", single, cacheCount, tree, appendAnimation, defaultFrames , key});
        PrefabModelObjPool.registPrefabFactory(key, ModelPrefab.Creation);
        PrefabModelObjPool.registPrefabResFactory(key, AnimeSprite.Creation);
    }
    public static Creation(meshName: string, scene: CustomScene, opt?: IAnimeSpriteOpt): AnimeSprite {
        let model = new AnimeSprite(meshName, scene, opt);
        let root = AnimeSprite.CreateNode(scene, opt.tree, model, opt.defaultFrames);
        (model)._loadPromise = Promise.resolve().then(() => {
            (model).loaded(root, [root], [], [], [], [], []);
        });
        return model;
    }
    protected appendAnimationEnd(animationGroup: BABYLON.AnimationGroup): void {
    }
    declare protected option: IAnimeSpriteOpt;
    protected sprites: Map<string, Sprite> = new Map();
    public static CreateNode(scene: CustomScene, node: IAnimeSpritTree, model: AnimeSprite, defaultFrames: { [index: string]: IAnimeSpriteDesc }): BABYLON.TransformNode | Sprite {
        let result: BABYLON.TransformNode;
        if (node[1] == 0) {
            result = new BABYLON.TransformNode(node[0], scene.impl);
        } else {
            let param = defaultFrames[node[0]];
            result = new Sprite(param.atlas, param.frame, scene.impl, 0, node[2]);
            result.name = node[0];

            if (model._colliderCustom == undefined) {
                model._colliderBoundle = (<Sprite>result)._colliderBoundle;
            }

            model.sprites.set(node[0], <Sprite>result);
        }

        result.position.copyFromFloats(node[3], node[4], node[5]);
        result.rotation.copyFromFloats(node[6], node[7], node[8]);
        result.scaling .copyFromFloats(node[9], node[10], node[11]);

        node[12].forEach(function(node) {
            let item = AnimeSprite.CreateNode(scene, node, model, defaultFrames);
            item.parent = result;
        });

        return result;
    }
    set extendData(val: [string, string][]) {
        super.extendData = val;
        if (val) {

            this.sprites.forEach((sprite) => {
                sprite.setEnabled(false);
            });
    
            val.forEach((item) => {
                let [key, frame] = item;
                let sprite = this.sprites.get(key);
                if (sprite) {
                    sprite.spriteFrame = frame;
                    sprite.setEnabled(true);
                }
            })
        }
    }
}