import { ExtensionTools, IPrefabCreation } from "../interface/index";
import { BABYLON } from "../render3d/babylon";
import { ModelObj, SceneStruct } from "../scene_base";
import { ModelPrefab, PrefabModelObjPool } from "../model_prefab";
import { CustomScene } from "../index";

export interface ICutResult {
    vertexs: number[];
    faces: number[];
    uvs: number[];
    image: string;
}

export interface IBigMapInfo {
    width: number;
    height: number;
    mapInfos: ICutResult[];
}

export namespace BigMapLoader {
    export let BasePath = () => {
        return '';
    };
    export function load(mapName: string, scene: SceneStruct, waitForOK: boolean = false) {
        const jsonFileName = `${BasePath()}${mapName}/${mapName}.jmap`;
        const root = new BABYLON.TransformNode(`${mapName}`, scene.impl);
        
        return ExtensionTools.LoadFile(undefined, jsonFileName).then((res: ArrayBuffer) => {
            if (ArrayBuffer.isView(res)) {
                res = (<Uint8Array>res).slice().buffer;
            }
            const str = ExtensionTools.UTF8Decode(res);
            const json: IBigMapInfo = JSON.parse(str);

            return parseJson(mapName, scene, json.mapInfos, waitForOK).then((meshes) => {
                const count = meshes.length;

                for (let i = 0; i < count; i++) {
                    const mesh = meshes[i];
                    mesh.parent = root;
                }

                return { root, width: json.width, height: json.height, meshes };
            });
        });
    }
    export function parseJson(mapName: string, scene: SceneStruct, resultArray: ICutResult[], waitForOK: boolean = false) {
        const count = resultArray.length;

        const promiseArray: Promise<BABYLON.Mesh>[] = [];

        for (let i = 0; i < count; i++) {
            const result = resultArray[i];

            const texName = result.image.replace('./', '');

            const customMesh = new BABYLON.Mesh(`${mapName}_${i}`, scene.impl);
            const vertexData = new BABYLON.VertexData();
            vertexData.positions = result.vertexs;
            vertexData.indices = result.faces;   
            vertexData.uvs = result.uvs;
            vertexData.applyToMesh(customMesh);

            const mat = new BABYLON.StandardMaterial(`${mapName}_${i}`, scene.impl);
            mat.specularColor = BABYLON.Color3.Black();
            mat.emissiveColor = BABYLON.Color3.White();
            mat.disableLighting = true;

            customMesh.material = mat;

            let tex: BABYLON.Texture;

            if (waitForOK) {
                const promise = new Promise((resolve: (res: BABYLON.Mesh) => void, reject) => {
                    tex = new BABYLON.Texture(`${BasePath()}${mapName}/${texName}`, scene.impl, true, undefined, undefined, () => {

                        tex.hasAlpha = true;
                        tex.wrapU = BABYLON.Texture.CLAMP_ADDRESSMODE;
                        tex.wrapV = BABYLON.Texture.CLAMP_ADDRESSMODE;
                        mat.useAlphaFromDiffuseTexture = true;
                        mat.diffuseTexture = tex;

                        resolve(customMesh);
                    }, reject);
                });

                promiseArray.push(promise);
            }
            else {
                tex = new BABYLON.Texture(`${BasePath()}${mapName}/${texName}`, scene.impl, true);

                tex.hasAlpha = true;
                tex.wrapU = BABYLON.Texture.CLAMP_ADDRESSMODE;
                tex.wrapV = BABYLON.Texture.CLAMP_ADDRESSMODE;
                mat.useAlphaFromDiffuseTexture = true;
                mat.diffuseTexture = tex;

                const promise = Promise.resolve(customMesh);
                promiseArray.push(promise);
            }
        }

        return Promise.all(promiseArray);
    }
}

export interface IBigMapModelOpt extends IPrefabCreation {
    json: IBigMapInfo;
    imageDir: string;
}
export class BigMapModel extends ModelObj {
    public static parseJson(mapName: string, imageDir: string, scene: SceneStruct, resultArray: ICutResult[], waitForOK: boolean = false) {
        const count = resultArray.length;

        const promiseArray: Promise<BABYLON.Mesh>[] = [];

        for (let i = 0; i < count; i++) {
            const result = resultArray[i];

            const texName = result.image.replace('./', '');

            const customMesh = new BABYLON.Mesh(`${mapName}_${i}`, scene.impl);
            const vertexData = new BABYLON.VertexData();
            vertexData.positions = result.vertexs;
            vertexData.indices = result.faces;   
            vertexData.uvs = result.uvs;
            vertexData.applyToMesh(customMesh);

            const mat = new BABYLON.StandardMaterial(`${mapName}_${i}`, scene.impl);
            mat.specularColor = BABYLON.Color3.Black();
            mat.emissiveColor = BABYLON.Color3.White();
            mat.disableLighting = true;
            mat.backFaceCulling = false;
            mat.transparencyMode = 2;

            customMesh.material = mat;

            let tex: BABYLON.Texture;

            if (waitForOK) {
                const promise = new Promise((resolve: (res: BABYLON.Mesh) => void, reject) => {
                    tex = new BABYLON.Texture(null, scene.impl, true, false, undefined, () => {

                        tex.hasAlpha = true;
                        tex.wrapU = BABYLON.Texture.CLAMP_ADDRESSMODE;
                        tex.wrapV = BABYLON.Texture.CLAMP_ADDRESSMODE;

                        resolve(customMesh);
                    }, reject);
                    tex.updateURL(`${imageDir}${texName}`);
                    mat.diffuseTexture = tex;
                    mat.useAlphaFromDiffuseTexture = true;
                });

                promiseArray.push(promise);
            }
            else {
                tex = new BABYLON.Texture(null, scene.impl, true, false);
                tex.updateURL(`${imageDir}${texName}`);

                tex.hasAlpha = true;
                tex.wrapU = BABYLON.Texture.CLAMP_ADDRESSMODE;
                tex.wrapV = BABYLON.Texture.CLAMP_ADDRESSMODE;
                mat.useAlphaFromDiffuseTexture = true;
                mat.diffuseTexture = tex;

                const promise = Promise.resolve(customMesh);
                promiseArray.push(promise);
            }
        }

        return Promise.all(promiseArray);
    }
    public static regist(prefabPool: PrefabModelObjPool, key: string, json: IBigMapInfo, imageDir: string) {
        prefabPool.registPrefabCfg(key, { path: "", fileName: "", single: false, cacheCount: 0, json, imageDir });
        PrefabModelObjPool.registPrefabFactory(key, ModelPrefab.Creation);
        PrefabModelObjPool.registPrefabResFactory(key, BigMapModel.Create);
    }
    public static Create(meshName: string, scene: CustomScene, opt?: IBigMapModelOpt): BigMapModel {
        let result = new BigMapModel(meshName, scene, opt);

        const root = new BABYLON.TransformNode(`${meshName}`, scene.impl);
        result.rootImpl = root;
        result._loadPromise = BigMapModel.parseJson(meshName, opt.imageDir, scene, opt.json.mapInfos, true).then((meshes) => {
            const count = meshes.length;

            for (let i = 0; i < count; i++) {
                const mesh = meshes[i];
                mesh.parent = root;
            }

            result.loaded(root, [root, ...meshes], meshes, [], [], [], []);
        });

        return result;
    }
}