import { IMeshMaterialProperty, IMeshRenderQueue } from "../interface/mesh";
import { BABYLON } from "../render3d/babylon";
import { IIsometryNode, INode } from "../interface/node";
import { SpriteManager } from "./sprite_manager";
import { EventState, IScene, ITransformNode, Observable } from "../interface/index";
import { Scene } from "../render3d/scene";
import { ISprite, ISpriteAtlas, ISpriteAtlasCells, ISpriteAtlasRects, ISpriteManager, ITexturePackAtlas, ITexturePackFrame } from "../interface/sprite";
import { TextureTillOffData, TextureTillOffDataScene } from "../material/standardmaterial";
import { PointerInfo } from "../render/events/pointerEvents";
import { GlobalAtlasManager } from "./sprite_atlas";
import { PiCollider } from "../scene_pick";

class Record<K, V> {
    private _dirty: boolean = false;
    private map: Map<K, V> = new Map();
    set(k: K, v: V) {
        this._dirty = true;
        this.map.set(k, v);
    }
    get(k: K) {
        return this.map.get(k);
    }
    delete(k: K) {
        this._dirty = this.map.delete(k);
    }
    iter() {
        return this.map.entries();
    }
    undirty() {
        this._dirty = false;
    }
    get dirty() {
        return this._dirty;
    }
}

export class Sprite extends BABYLON.TransformNode implements IMeshRenderQueue, ITransformNode, IMeshMaterialProperty, ISprite {
    private static Counter: number = Number.MIN_SAFE_INTEGER;
    private static Vx: BABYLON.Vector3 = new BABYLON.Vector3(1, 0, 0);
    private static Vy: BABYLON.Vector3 = new BABYLON.Vector3(0, 1, 0);
    /**
     * 
     * @param hue 色相 - 对应Photoshop的[-180, 180]
     * @param saturate 饱和度 对应Photoshop的[-100, 100]
     * @param brightness 明度 对应Photoshop的[-100, 100]
     */
    public static HSB(
        hue: number = 0,
        saturate: number = 0,
        brightness: number = 0
    ) {
        return new BABYLON.Color3(hue / 360, saturate / 100, brightness / 100);
    }
    protected _color: BABYLON.Color3 = BABYLON.Color3.White();
    get color() {
        return this._color;
    }
    set color(val: BABYLON.Color3) {
        this._color = val;
        this.mgr?.dirty();
    }
    setColor(r: number, g: number, b: number) {
        this._color.copyFromFloats(r, g, b);
        this.mgr?.dirty();
        return this;
    }
    protected _alpha: number = 1;
    get alpha() {
        return this._alpha;
    }
    set alpha(val: number) {
        this._alpha = val;
        this.mgr?.dirty();
    }
    duration: number = 1.0;

    protected mgr: ISpriteManager;
    layerMask: number;
    protected _renderingGroupId: number;
    set renderingGroupId(val: number) {
        if (this._renderingGroupId != val) {
            this._renderingGroupId = val;
            this.mgr.dirty();
        }
    }
    get renderingGroupId() {
        return this._renderingGroupId;
    }
    protected _alphaIndex: number;
    set alphaIndex(val: number) {
        if (this._alphaIndex != val) {
            this._alphaIndex = val;
            this.mgr.dirty();
        }
    }
    get alphaIndex() {
        return this._alphaIndex;
    }
    
    public onPointerObservable = new Observable<PointerInfo>();

    isSynchronized(): boolean {
        let result = super.isSynchronized();
        if (!result) {
            this.mgr.dirty();
        }
        return result;
    }

    sortParam() {
        this.computeWorldMatrix();
        return this.getAbsolutePosition().y;
    }

    setEnabled(val: boolean): void {
        super.setEnabled(val);
        this.mgr.dirty();
    }
    isDisposed(): boolean {
        return this._isDisposed;
    }
    dispose(doNotRecurse?: boolean, disposeMaterialAndTextures?: boolean): void {
        this.onPointerObservable.clear();
        this.tilloff?.dispose();
        this.mgr.delSprite(this);
        super.dispose(doNotRecurse, disposeMaterialAndTextures);
    }
    private _sx: number = 1;
    private _sy: number = 1;
    private _dx: number = 0;
    private _dy: number = 0;
    vertexPosition(data: Float32Array, offset: number): number {
        let point = 4;
        let matrix = this.computeWorldMatrix();

        let sx = 1;
        let sy = 1;
        let dx = 0;
        let dy = 0;

        if (this._texturePackFrame) {
            let s = this._texturePackFrame.sourceSize;
            let sw = s.w;
            let sh = s.h;
            let ss = this._texturePackFrame.spriteSourceSize;
            let ssw = ss.w;
            let ssh = ss.h;
            sx = ssw / sw * 0.5;
            sy = ssh / sh * 0.5;
            dx = (ss.x + ssw * 0.5) - sw * 0.5;
            dx /= sw;
            dy = (ss.y + ssh * 0.5) - sh * 0.5;
            dy /= sh;
        } else {
            sx = 0.5;
            sy = 0.5;
        }
        
        if (this.mgr.atlas.scale) {
            sx *= this.mgr.atlas.scale[0];
            sy *= this.mgr.atlas.scale[1];
            dx *= this.mgr.atlas.scale[0];
            dy *= this.mgr.atlas.scale[1];
        }

        if (this.mgr.atlas.isInvertY) {
            sy *= -1;
        }

        Sprite.modify(dx, dy, sx, sy, this, undefined);

        BABYLON.Vector3.TransformCoordinatesFromFloatsToRef(dx - sx + Sprite.DiffData[0], dy - sy + Sprite.DiffData[1], 0, matrix, Sprite.Vx);
        data[offset + 0 * 3 + 0] = Sprite.Vx.x; data[offset + 0 * 3 + 1] = Sprite.Vx.y; data[offset + 0 * 3 + 2] = Sprite.Vx.z;
        BABYLON.Vector3.TransformCoordinatesFromFloatsToRef(dx + sx + Sprite.DiffData[2], dy - sy + Sprite.DiffData[3], 0, matrix, Sprite.Vx);
        data[offset + 1 * 3 + 0] = Sprite.Vx.x; data[offset + 1 * 3 + 1] = Sprite.Vx.y; data[offset + 1 * 3 + 2] = Sprite.Vx.z;
        BABYLON.Vector3.TransformCoordinatesFromFloatsToRef(dx - sx + Sprite.DiffData[4], dy + sy + Sprite.DiffData[5], 0, matrix, Sprite.Vx);
        data[offset + 2 * 3 + 0] = Sprite.Vx.x; data[offset + 2 * 3 + 1] = Sprite.Vx.y; data[offset + 2 * 3 + 2] = Sprite.Vx.z;
        BABYLON.Vector3.TransformCoordinatesFromFloatsToRef(dx + sx + Sprite.DiffData[6], dy + sy + Sprite.DiffData[7], 0, matrix, Sprite.Vx);
        data[offset + 3 * 3 + 0] = Sprite.Vx.x; data[offset + 3 * 3 + 1] = Sprite.Vx.y; data[offset + 3 * 3 + 2] = Sprite.Vx.z;

        this._sx = sx;
        this._sy = sy;
        this._dx = dx;
        this._dy = dy;
        if (this.___colliderBoundle) {
            this.___colliderBoundle.reConstruct(new BABYLON.Vector3(dx - sx, dy - sy, -0.1), new BABYLON.Vector3(dx + sx, dy + sy, 0.1), matrix);
        }

        return offset + point * 3;
    }

    vertexColor(data: Float32Array, offset: number): number {
        let point = 4;
        for (let i = 0; i < point; i++) {
            data[offset + i * 4 + 0] = this._color.r;
            data[offset + i * 4 + 1] = this._color.g;
            data[offset + i * 4 + 2] = this._color.b;
            data[offset + i * 4 + 3] = this.alpha;
        }
        return offset + point * 4;
    }

    vertexUV(data: Float32Array, offset: number): number {
        let point = 4;
        if (this.tilloff) {
            let frame = GlobalAtlasManager.getFrame(this._spriteFrame, this.atlas);
            if (frame) {
                let u0 = this.tilloff.uOffset; let u1 = this.tilloff.uScale + this.tilloff.uOffset;
                let v0 = this.tilloff.vOffset; let v1 = this.tilloff.vScale + this.tilloff.vOffset;
    
                data[offset + 0 * 2 + 0] = u0; data[offset + 0 * 2 + 1] = v0;
                data[offset + 1 * 2 + 0] = u1; data[offset + 1 * 2 + 1] = v0;
                data[offset + 2 * 2 + 0] = u0; data[offset + 2 * 2 + 1] = v1;
                data[offset + 3 * 2 + 0] = u1; data[offset + 3 * 2 + 1] = v1;
            } else {
                for (let i = point * 2; i > 0; i--) {
                    data[offset + i] = 0;
                }
            }
        } else {
            let frame = GlobalAtlasManager.getFrame(this._spriteFrame, this.atlas);
            if (frame) {
                let u0 = frame[2]; let u1 = frame[0] + frame[2];
                let v0 = frame[3]; let v1 = frame[1] + frame[3];
                // data[offset + 0 * 2 + 0] = u0; data[offset + 0 * 2 + 1] = v0;
                // data[offset + 1 * 2 + 0] = u1; data[offset + 1 * 2 + 1] = v0;
                // data[offset + 2 * 2 + 0] = u0; data[offset + 2 * 2 + 1] = v1;
                // data[offset + 3 * 2 + 0] = u0; data[offset + 3 * 2 + 1] = v1;
                // data[offset + 4 * 2 + 0] = u1; data[offset + 4 * 2 + 1] = v0;
                // data[offset + 5 * 2 + 0] = u1; data[offset + 5 * 2 + 1] = v1;
    
                if (this._texturePackFrame?.rotated) {
                    data[offset + 2 * 2 + 0] = u0; data[offset + 2 * 2 + 1] = v0;
                    data[offset + 0 * 2 + 0] = u1; data[offset + 0 * 2 + 1] = v0;
                    data[offset + 3 * 2 + 0] = u0; data[offset + 3 * 2 + 1] = v1;
                    data[offset + 1 * 2 + 0] = u1; data[offset + 1 * 2 + 1] = v1;
                } else {
                    data[offset + 0 * 2 + 0] = u0; data[offset + 0 * 2 + 1] = v0;
                    data[offset + 1 * 2 + 0] = u1; data[offset + 1 * 2 + 1] = v0;
                    data[offset + 2 * 2 + 0] = u0; data[offset + 2 * 2 + 1] = v1;
                    data[offset + 3 * 2 + 0] = u1; data[offset + 3 * 2 + 1] = v1;
                }
            } else {
                for (let i = point * 2; i > 0; i--) {
                    data[offset + i] = 0;
                }
            }
        }

        return offset + point * 2;
    }

    protected tilloff: TextureTillOffData;
    public get uScale() {
        if (this.tilloff == undefined) this.tilloff = TextureTillOffData.Create(<TextureTillOffDataScene>this._scene);
        return this.tilloff.uScale;
    }
    public set uScale(val: number) {
        if (this.tilloff == undefined) this.tilloff = TextureTillOffData.Create(<TextureTillOffDataScene>this._scene);
        this.tilloff.uScale = val;
        this.mgr?.dirty();
    }
    public get vScale() {
        if (this.tilloff == undefined) this.tilloff = TextureTillOffData.Create(<TextureTillOffDataScene>this._scene);
        return this.tilloff.vScale;
    }
    public set vScale(val: number) {
        if (this.tilloff == undefined) this.tilloff = TextureTillOffData.Create(<TextureTillOffDataScene>this._scene);
        this.tilloff.vScale = val;
        this.mgr?.dirty();
    }
    public get uOffset() {
        if (this.tilloff == undefined) this.tilloff = TextureTillOffData.Create(<TextureTillOffDataScene>this._scene);
        return this.tilloff.uOffset;
    }
    public set uOffset(val: number) {
        if (this.tilloff == undefined) this.tilloff = TextureTillOffData.Create(<TextureTillOffDataScene>this._scene);
        this.tilloff.uOffset = val;
        this.mgr?.dirty();
    }
    public get vOffset() {
        if (this.tilloff == undefined) this.tilloff = TextureTillOffData.Create(<TextureTillOffDataScene>this._scene);
        return this.tilloff.vOffset;
    }
    public set vOffset(val: number) {
        if (this.tilloff == undefined) this.tilloff = TextureTillOffData.Create(<TextureTillOffDataScene>this._scene);
        this.tilloff.vOffset = val;
        this.mgr?.dirty();
    }
    public tillOffset(uScale: number, vScale: number, uOffset: number, vOffset: number) {
        if (this.tilloff == undefined) this.tilloff = TextureTillOffData.Create(<TextureTillOffDataScene>this._scene);
        this.tilloff.uScale = uScale;
        this.tilloff.vScale = vScale;
        this.tilloff.uOffset = uOffset;
        this.tilloff.vOffset = vOffset;
        this.mgr?.dirty();
        return this;
    }

    set position(val: BABYLON.Vector3) {
        (<any>this)._position = val;
        this.mgr.dirty();
    }
    get position() {
        return (<any>this)._position;
    }
    setPosition(x: number, y: number, z: number) {
        (<any>this)._position.copyFromFloats(x, y, z);
        this.mgr.dirty();
        return this;
    }
    set scaling(val: BABYLON.Vector3) {
        this._scaling = val;
        this.mgr.dirty();
    }
    get scaling() {
        return this._scaling;
    }
    setScaling(x: number, y: number, z: number) {
        (<any>this)._scaling.copyFromFloats(x, y, z);
        this.mgr.dirty();
        return this;
    }
    set rotation(val: BABYLON.Vector3) {
        (<any>this)._rotation = val;
        this.mgr.dirty();
    }
    get rotation() {
        return (<any>this)._rotation;
    }
    setRotation(x: number, y: number, z: number) {
        (<any>this)._rotation.copyFromFloats(x, y, z);
        this.mgr.dirty();
        return this;
    }

    public readonly spriteId: number;
    protected atlasData: ISpriteAtlas;
    constructor(protected atlas: string, protected _spriteFrame: string | number, scene: Scene, renderingGroupId: number = 0, alphaIndex: number = 999999) {
        let spriteId: number = Sprite.Counter++;
        let mgr = scene.getSpritManager(atlas);
        super(`${spriteId}`, scene);

        this.atlasData = GlobalAtlasManager.atlasMap.get(atlas)?.atlas;
        this.mgr = mgr;
        this.spriteId = spriteId;
        this._renderingGroupId = renderingGroupId;
        this._alphaIndex = alphaIndex;
        this.mgr?.addSprite(this);
        this._texturePackFrame = GlobalAtlasManager.getTexturePackFrame(this._spriteFrame, this.atlas);
        let renderMode = (<any>mgr)._atlasFrames.renderMode;
        if (renderMode == 1 || renderMode == 2 || renderMode == 3) {
            this._color.copyFromFloats(0, 0, 0);
        }
    }
    _onFrameUpdate(): boolean {
        return this._animIsLoop !== undefined || this.tilloff?.dirty;
    }
    protected _texturePackFrame: ITexturePackFrame;
    get spriteFrame() {
        if (this._animIsLoop !== undefined) {
            if ((<ITexturePackAtlas>this.atlasData)?.meta) {
                let animations = (<ITexturePackAtlas>this.atlasData)?.animations;
                if (animations) {
                    if (this._animeName === undefined) {
                        for (let animeName in animations) {
                            if (animations[animeName].indexOf(<string>this._spriteFrame) >= 0) {
                                this._animeName = animeName;
                                break;
                            }
                        }
                    }
                    if (this._animeName !== undefined) {
                        this._animedTime += this._scene.deltaTime;
                        let frames = animations[this._animeName];
                        let runnedFrame = this._animedTime / (1000 * this.duration) * frames.length;
                        if (runnedFrame >= frames.length && this._animIsLoop === false) {
                            let frame = frames.length - 1;
                            this._spriteFrame = frames[frame];
                            this._animIsLoop = undefined;
                        } else {
                            let frame = Math.floor(runnedFrame) % frames.length;
                            this._spriteFrame = frames[frame];
                        }
                    }
                }
            } else if ((<ISpriteAtlasRects>this.atlasData)?.rects) {
                // this._animedTime += this._scene.deltaTime;
            } else {
                let atlas = <ISpriteAtlasCells>this.atlasData;
                let frames = atlas.hCount * atlas.vCount;
                this._animedTime += this._scene.deltaTime;
                let runnedFrame = this._animedTime / (1000 * this.duration) * frames;
                if (runnedFrame >= frames && this._animIsLoop === false) {
                    let frame = frames - 1;
                    this._spriteFrame = frame;
                    this._animIsLoop = undefined;
                } else {
                    let frame = Math.floor(runnedFrame) % frames;
                    this._spriteFrame = frame;
                }
            }
        }
        return this._spriteFrame;
    }
    set spriteFrame(val: string | number) {
        if (this._spriteFrame != val) {
            this._spriteFrame = val;
            this._texturePackFrame = GlobalAtlasManager.getTexturePackFrame(this._spriteFrame, this.atlas);
            this.mgr.dirty();
        }
    }

    private _animedTime: number = 0;
    private _animeName: string | number;
    private _animIsLoop: boolean;
    setAnim(isLoop: boolean): void {
        this._animedTime = 0;
        if ((<ITexturePackAtlas>this.atlasData)?.meta) {
            if ((<ITexturePackAtlas>this.atlasData)?.animations) {
                this._animIsLoop = isLoop;
            }
        } else {
            this._animIsLoop = isLoop;
        }
    }

    ___colliderBoundle: PiCollider;
    get _colliderBoundle(): PiCollider {
        if (this.___colliderBoundle == undefined) {
            this.___colliderBoundle = new PiCollider(new BABYLON.Vector3(this._dx - this._sx, this._dy - this._sy, -0.001), new BABYLON.Vector3(this._dx + this._sx, this._dy + this._sy, 0.001), this.worldMatrixFromCache);
        }
        return this.___colliderBoundle;
    }

    public static DiffData: [number, number, number, number, number, number, number, number, ] = [0, 0, 0, 0, 0, 0, 0, 0];
    public static modify(dx: number, dy: number, sx: number, sy: number, sprite: Sprite, events?: [number, number, number, number][]) {

        let ddx1 = 0;
        let ddy1 = 0;
        let ddx2 = 0;
        let ddy2 = 0;
        let ddx3 = 0;
        let ddy3 = 0;
        let ddx4 = 0;
        let ddy4 = 0;
        if (events) {
            let distance = 0;
            let dx1 = 0;
            let dy1 = 0;
            let dx2 = 0;
            let dy2 = 0;
            let dx3 = 0;
            let dy3 = 0;
            let dx4 = 0;
            let dy4 = 0;
            let x = dx;
            let y = dy;
            let tileWidth = sx;
            let tileHeight = sy;
            let halfW = sx / 2;
            let halfH = sy / 2;

            let x1 = (x - halfW) * 1; // Tile.X_POSITION_SCALE;
            let x2 = (x + halfW) * 1; // Tile.X_POSITION_SCALE;
            let x3 = (x + halfW) * 1; // Tile.X_POSITION_SCALE;
            let x4 = (x - halfW) * 1; // Tile.X_POSITION_SCALE;
            let y1 = (y - halfH) * 1; // Tile.Y_POSITION_SCALE;
            let y2 = (y - halfH) * 1; // Tile.Y_POSITION_SCALE;
            let y3 = (y + halfH) * 1; // Tile.Y_POSITION_SCALE;
            let y4 = (y + halfH) * 1; // Tile.Y_POSITION_SCALE;
            events?.forEach((item) => {
                let cx = Math.abs(x - item[0]);
                let cy = Math.abs(y - item[1]);
                if (cx > 0.55 || cy > 0.55) {
                    if (cx > (cy + 0.2)) { cx = 1; cy = 0; } else if ((cx + 0.2) < cy) { cx = 0; cy = 1; } else { cx = 0.5; cy = 0.5; }
                    let diff = item[3] * 0.5;
                    dx1 = x1 - item[0]; dy1 = y1 - item[1]; dx1 *= cx; dy1 *= cy;
                    distance = Math.sqrt(dx1 * dx1 + dy1 * dy1);
                    ddx1 += distance > 0 ? dx1 / distance * (1.0 - Math.min(1., distance / item[2])) * diff : 0;
                    ddy1 += distance > 0 ? dy1 / distance * (1.0 - Math.min(1., distance / item[2])) * diff : 0;
                    dx2 = x2 - item[0]; dy2 = y2 - item[1]; dx2 *= cx; dy2 *= cy;
                    distance = Math.sqrt(dx2 * dx2 + dy2 * dy2);
                    ddx2 += distance > 0 ? dx2 / distance * (1.0 - Math.min(1., distance / item[2])) * diff : 0;
                    ddy2 += distance > 0 ? dy2 / distance * (1.0 - Math.min(1., distance / item[2])) * diff : 0;
                    dx3 = x3 - item[0]; dy3 = y3 - item[1]; dx3 *= cx; dy3 *= cy;
                    distance = Math.sqrt(dx3 * dx3 + dy3 * dy3);
                    ddx3 += distance > 0 ? dx3 / distance * (1.0 - Math.min(1., distance / item[2])) * diff : 0;
                    ddy3 += distance > 0 ? dy3 / distance * (1.0 - Math.min(1., distance / item[2])) * diff : 0;
                    dx4 = x4 - item[0]; dy4 = y4 - item[1]; dx4 *= cx; dy4 *= cy;
                    distance = Math.sqrt(dx4 * dx4 + dy4 * dy4);
                    ddx4 += distance > 0 ? dx4 / distance * (1.0 - Math.min(1., distance / item[2])) * diff : 0;
                    ddy4 += distance > 0 ? dy4 / distance * (1.0 - Math.min(1., distance / item[2])) * diff : 0;
                }
            });
        }

        Sprite.DiffData[0] = ddx1;
        Sprite.DiffData[1] = ddy1;
        Sprite.DiffData[2] = ddx2;
        Sprite.DiffData[3] = ddy2;
        Sprite.DiffData[4] = ddx3;
        Sprite.DiffData[5] = ddy3;
        Sprite.DiffData[6] = ddx4;
        Sprite.DiffData[7] = ddy4;
    }
}
