import { ISprite, ISpriteAtlas, EAtlasRenderBase, EAtlasRenderHS, EAtlasRenderHSB, EAtlasRenderOverlay, ISpriteManager, ITexturePackAtlas, ITexturePackFrame } from "../interface/index";
import { CustomGeometry } from "../customgeometry/custom_geometry";
import { BABYLON } from "../render3d/babylon";
import { Constants } from "../interface/constants";
import { Scene } from "../render3d/scene";
import { Observable } from "../interface/observable";
import { Atlas, GlobalAtlasManager } from "./sprite_atlas";
import { TextureHSVMaterial } from "../material/hsv";
import { TextureHueSaturateMaterial } from "../material/hue_saturate";
import { OverlayColorMaterial } from "../material/overlay_color";
import { EAtlasRenderMode } from "../interface/texturepacker";
import { TextureProgressMaterial } from "../material/texture_progress";

export class SpriteManager implements ISpriteManager {
    public static MAX_SPRITE = 128;
    get material() {
        return this._material;
    }
    _scene: Scene;
    _isDisposed: boolean = false;
    setEnabled(val: boolean): void {
        // throw new Error("Method not implemented.");
    }
    isDisposed(): boolean {
        return this._isDisposed;
    }
    dispose(): void {
        if (this._isDisposed) return;

        this._scene.delSpritManager(this);

        this._sprites.forEach((item) => {
            item.dispose(false, true);
        });

        this._material.dispose(false, true);
        this._meshes.forEach((item) => {
            item.geo?.uninstallMesh();
            item.geo?.dispose();
            item.mesh.dispose(false, true);
        });
        this._meshes.clear();
    }

    protected _geometry: CustomGeometry;
    protected _material: BABYLON.Material;
    protected _sprites: Map<number, ISprite> = new Map();
    protected _meshes: Map<string, { mesh: BABYLON.Mesh, geo: CustomGeometry }> = new Map();
    protected _atlasFrames: Atlas;
    protected _dirty: boolean = false;
    public readonly atlasImage: string;
    public readonly atlas: ISpriteAtlas;
    public static MaterialFactory: Map<EAtlasRenderMode | number, (scene: Scene, atlasImage: string, noMipmap: boolean, samplerMode: number, alphaMode: number, pass?: Constants.PassTag) => BABYLON.Material> = new Map();
    public static CreateMaterial(scene: Scene, atlasImage: string, noMipmap: boolean, samplerMode: number, alphaMode: number, pass?: Constants.PassTag, renderMode: EAtlasRenderMode | number = 0) {
        let func = SpriteManager.MaterialFactory.get(renderMode) || SpriteManager.MaterialFactory.get(0);
        return func(scene, atlasImage, noMipmap, samplerMode, alphaMode, pass);
    }
    constructor(scene: Scene, atlas: Atlas, public readonly maxSprite: number = SpriteManager.MAX_SPRITE, material?: BABYLON.Material, private pass?: Constants.PassTag) {
        this._scene = scene;
        this._atlasFrames = atlas;
        this.atlas = atlas.atlas;
        this.atlasImage = this._atlasFrames.atlasImage;

        if (material) {
            this._material = <BABYLON.StandardMaterial>material;
        } else {
            this._material = SpriteManager.CreateMaterial(this._scene, atlas.atlasImage, atlas.noMipmap, atlas.samplerMode, atlas.alphaMode, pass, atlas.renderMode);
        }
    }

    extendAtlasFrame(key: string | number, rect: [number, number, number, number]): void {
        this._atlasFrames.extendAtlasFrame(key, rect);
    }
    getFrame(item: ISprite): [number, number, number, number] {
        return undefined;
        // return this._atlasFrames.getFrame(item.spriteFrame);
    }
    getTexturePackFrame(item: ISprite): ITexturePackFrame {
        return undefined;
        // return GlobalAtlasManager.getTexturePackFrame(item.spriteFrame, this.atlasImage);
    }
    addSprite(item: ISprite): boolean {
        if (this._sprites.size < this.maxSprite) {
            this._sprites.set(item.spriteId, item);
            this._dirty = true;
            return true;
        } else {
            return false;
        }
    }
    delSprite(item: ISprite): void {
        this._sprites.delete(item.spriteId);
        this._dirty = true;
    }
    dirty(): void {
        this._dirty = true;
    }
    spriteCount() {
        return this._sprites.size;
    }
    onDisposeObservable: Observable<number>;
    name?: string;
    _onFrameUpdate(): void {
        if (!this._dirty) {
            for (let sprite of this._sprites.values()) {
                this._dirty = this._dirty || (sprite.isEnabled() && (!(<any>sprite).isSynchronized() || sprite._onFrameUpdate()));
                if (this._dirty) {
                    break;
                }
            }
            if (!this._dirty) {
                return;
            }
        }
        this._dirty = false;
        let queues: Map<number, Map<number, { id: number, param: any }[]>> = new Map();
        this._sprites.forEach((sprite, uniqueId) => {
            if (sprite.isEnabled()) {
                let renderingGroupId = sprite.renderingGroupId;
                let alphaIndex = sprite.alphaIndex;
                let sameGroup = queues.get(renderingGroupId);
                if (sameGroup == undefined) {
                    sameGroup = new Map();
                    queues.set(renderingGroupId, sameGroup);
                }
                let sameAlphaindex = sameGroup.get(alphaIndex);
                if (sameAlphaindex == undefined) {
                    sameAlphaindex = [];
                    sameGroup.set(alphaIndex, sameAlphaindex);
                }
                sameAlphaindex.push({ id: uniqueId, param: sprite.sortParam() });
            }
        });

        this._meshes.forEach((mesh) => {
            mesh.mesh.setEnabled(false);
        });

        let useIndex = 0;
        queues.forEach((sameGroup, groupId) => {
            sameGroup.forEach((sameAlphaindex, alphaIndex) => {
                useIndex += 1;
                this.sort(sameAlphaindex);
                let key = `${groupId}#${useIndex}`;
                let info = this._meshes.get(key);
                let spriteCount = sameAlphaindex.length;
                if (info == undefined) {
                    let mesh = new BABYLON.Mesh(`${this.atlasImage}#${key}`, this._scene);
                    mesh.renderingGroupId = groupId;
                    mesh.alphaIndex = alphaIndex;
                    mesh.alwaysSelectAsActiveMesh = true;
                    mesh.doNotSyncBoundingInfo = true;
                    mesh.material = this._material;
                    let geo = new CustomGeometry(mesh, `${this.atlasImage}#${key}`, [Constants.VERTEX_ATTR_POSITION, Constants.VERTEX_ATTR_COLOR4, Constants.VERTEX_ATTR_UV], spriteCount * 4, spriteCount * 2, true);
                    info = { mesh, geo };
                    this._meshes.set(key, info);

                    let indices = geo.indices();
                    for (let i = 0; i < this.maxSprite; i++) {
                        indices[i * 6 + 0] = i * 4 + 0;
                        indices[i * 6 + 1] = i * 4 + 1;
                        indices[i * 6 + 2] = i * 4 + 2;
                        indices[i * 6 + 3] = i * 4 + 2;
                        indices[i * 6 + 4] = i * 4 + 1;
                        indices[i * 6 + 5] = i * 4 + 3;
                    }
                } else if (info.geo.maxVertex < spriteCount * 4) {
                    info.geo?.uninstallMesh();
                    info.geo?.dispose();
                    info.geo = undefined;
                    let geo = new CustomGeometry(info.mesh, `${this.atlasImage}#${key}`, [Constants.VERTEX_ATTR_POSITION, Constants.VERTEX_ATTR_COLOR4, Constants.VERTEX_ATTR_UV], spriteCount * 4, spriteCount * 2, true);
                    let indices = geo.indices();
                    for (let i = 0; i < spriteCount; i++) {
                        indices[i * 6 + 0] = i * 4 + 0;
                        indices[i * 6 + 1] = i * 4 + 1;
                        indices[i * 6 + 2] = i * 4 + 2;
                        indices[i * 6 + 3] = i * 4 + 2;
                        indices[i * 6 + 4] = i * 4 + 1;
                        indices[i * 6 + 5] = i * 4 + 3;
                    }
                    info.geo = geo;
                    this._meshes.set(key, info);
                }

                let positions = info.geo.vertices(Constants.VERTEX_ATTR_POSITION)[0];
                let colors = info.geo.vertices(Constants.VERTEX_ATTR_COLOR4)[0];
                let uvs = info.geo.vertices(Constants.VERTEX_ATTR_UV)[0];
                let count = 0;
                let offsetPosition = 0;
                let offsetColor = 0;
                let offsetUV = 0;
                sameAlphaindex.forEach((item) => {
                    let sprite = this._sprites.get(item.id);
                    if (sprite && sprite.isEnabled() && count < spriteCount) {
                        offsetPosition = sprite.vertexPosition(positions, offsetPosition);
                        offsetColor = sprite.vertexColor(colors, offsetColor);
                        offsetUV = sprite.vertexUV(uvs, offsetUV);

                        count += 1;
                    }
                });

                info.geo.update(count * 6);
                info.mesh.setEnabled(count > 0);
                info.mesh.renderingGroupId = groupId;
                info.mesh.alphaIndex = alphaIndex;
            });
        })
    }
    /**
     * 默认param 为 sprite 的全局坐标的 y 值
     * @param list 
     */
    sort(list: { param: any; }[]): void {
        list.sort((a, b) => b.param - a.param);
    }
}

SpriteManager.MaterialFactory.set(0, function(scene: Scene, atlasImage: string, noMipmap: boolean, samplerMode: number, alphaMode: number, pass?: Constants.PassTag) {
    let _material = new BABYLON.StandardMaterial(`${atlasImage}`, scene);
    let texture = GlobalAtlasManager.getTexture(atlasImage, scene);
    if (texture == undefined) {
        texture = new BABYLON.Texture(null, scene, noMipmap, false, samplerMode || BABYLON.Texture.LINEAR_LINEAR);
        texture.updateURL(atlasImage);
        texture.hasAlpha = true;
        texture.wrapU = BABYLON.Texture.CLAMP_ADDRESSMODE;
        texture.wrapV = BABYLON.Texture.CLAMP_ADDRESSMODE;
    }
    _material.diffuseTexture = texture;
    _material.disableLighting = true;
    _material.emissiveColor = new BABYLON.Color3(1, 1, 1);
    _material.disableDepthWrite = true;
    _material.backFaceCulling = false;
    _material.alphaMode = alphaMode;
    _material.transparencyMode = BABYLON.Material.MATERIAL_ALPHATESTANDBLEND;
    _material.useAlphaFromDiffuseTexture = true;
    _material.cutoffEdgeAlphaDiff = 0.0;
    _material.alphaCutOff = 0.0;
    return _material;
});
SpriteManager.MaterialFactory.set(1, function(scene: Scene, atlasImage: string, noMipmap: boolean, samplerMode: number, alphaMode: number, pass?: Constants.PassTag) {
    let _material = new TextureHSVMaterial(`${atlasImage}`, scene);
    let texture = GlobalAtlasManager.getTexture(atlasImage, scene);
    if (texture == undefined) {
        texture = new BABYLON.Texture(null, scene, noMipmap, false, samplerMode || BABYLON.Texture.LINEAR_LINEAR);
        texture.updateURL(atlasImage);
        texture.hasAlpha = true;
        texture.wrapU = BABYLON.Texture.CLAMP_ADDRESSMODE;
        texture.wrapV = BABYLON.Texture.CLAMP_ADDRESSMODE;
    }
    _material.diffuseTexture = texture;
    _material.emissiveColor = new BABYLON.Color3(1, 1, 1);
    _material.disableDepthWrite = true;
    _material.backFaceCulling = false;
    _material.alphaMode = alphaMode;
    _material.transparencyMode = BABYLON.Material.MATERIAL_ALPHATESTANDBLEND;
    return _material;
});

SpriteManager.MaterialFactory.set(2, function(scene: Scene, atlasImage: string, noMipmap: boolean, samplerMode: number, alphaMode: number, pass?: Constants.PassTag) {
    let _material = new TextureHueSaturateMaterial(`${atlasImage}`, scene);
    let texture = GlobalAtlasManager.getTexture(atlasImage, scene);
    if (texture == undefined) {
        texture = new BABYLON.Texture(null, scene, noMipmap, false, samplerMode || BABYLON.Texture.LINEAR_LINEAR);
        texture.updateURL(atlasImage);
        texture.hasAlpha = true;
        texture.wrapU = BABYLON.Texture.CLAMP_ADDRESSMODE;
        texture.wrapV = BABYLON.Texture.CLAMP_ADDRESSMODE;
    }
    _material.diffuseTexture = texture;
    _material.emissiveColor = new BABYLON.Color3(1, 1, 1);
    _material.disableDepthWrite = true;
    _material.backFaceCulling = false;
    _material.alphaMode = alphaMode;
    _material.transparencyMode = BABYLON.Material.MATERIAL_ALPHATESTANDBLEND;
    return _material;
});

SpriteManager.MaterialFactory.set(3, function(scene: Scene, atlasImage: string, noMipmap: boolean, samplerMode: number, alphaMode: number, pass?: Constants.PassTag) {
    let _material = new OverlayColorMaterial(`${atlasImage}`, scene);
    let texture = GlobalAtlasManager.getTexture(atlasImage, scene);
    if (texture == undefined) {
        texture = new BABYLON.Texture(null, scene, noMipmap, false, samplerMode || BABYLON.Texture.LINEAR_LINEAR);
        texture.updateURL(atlasImage);
        texture.hasAlpha = true;
        texture.wrapU = BABYLON.Texture.CLAMP_ADDRESSMODE;
        texture.wrapV = BABYLON.Texture.CLAMP_ADDRESSMODE;
    }
    _material.diffuseTexture = texture;
    _material.emissiveColor = new BABYLON.Color3(1, 1, 1);
    _material.disableDepthWrite = true;
    _material.backFaceCulling = false;
    _material.alphaMode = alphaMode;
    _material.transparencyMode = BABYLON.Material.MATERIAL_ALPHATESTANDBLEND;
    return _material;
});