
import { SceneStruct } from "../scene_struct";
import { BABYLON } from "../render3d/babylon";
import { MainOpacityMaterial } from "../index";
import { ITileObjectKey, ITileState } from "./interface";
import { TileMapAsset } from "./assets";
import { TileObjectMesh } from "./tile_mesh";
import { TileObject } from "./tile_object";

export class TileObjectRender {
    protected map: Map<ITileObjectKey, TileObjectMesh> = new Map();
    protected objectmap: Map<any, TileObject> = new Map();
    protected root: BABYLON.TransformNode;
    protected meshBase: BABYLON.TransformNode;
    constructor(
        protected scene: SceneStruct,
    ) {
        this.root = new BABYLON.TransformNode(`TileObjects`, scene.impl);
        this.meshBase = new BABYLON.TransformNode(`TileObjectsMeshes`, scene.impl);
        this.meshBase.parent = this.root;
    }
    /**
     * 创建一个对象
     * @param id ID
     * @param state 渲染状态
     * @param layer 层级
     * @param assetMap 资源
     * @param materials 材质
     */
    public create(
        id: any,
        state: ITileState,
        layer: number,
    ): void {
        let obj = new TileObject(id, state, layer);

        let mesh = this.get_mesh(layer, state.assetSource);
        let old = mesh.get(id);
        if (old) {
            console.warn(`Add TileObj ${id} Again !!!`);
        }
        mesh.add(obj);
        this.objectmap.set(obj.id, obj);
    }
    protected get_mesh(
        layer: number,
        assetSource: string,
    ) {
        let key = TileObject.key(layer, assetSource);
        let mesh = this.map.get(key);
        if (!mesh) {
            mesh = new TileObjectMesh(this.scene, layer, assetSource);
            this.map.set(key, mesh);
        }
        return mesh;

    }

    /**
     * 每帧自动更新
     * @param tileMapAsset 
     * @param materials 
     */
    update = (
        tileMapAsset: TileMapAsset,
        materials: MainOpacityMaterial,
        events?: [number, number, number, number][],
    ) => {
        this.map.forEach((mesh, key) => {
            mesh.build(tileMapAsset, materials, this.root, events);
        });
    }

    /**
     * 修改对象尺寸
     * @param id ID
     * @param value 数据
     */
    width(id: any, value: number) {
        let old: TileObject;
        old = this.objectmap.get(id);
        if (old && this.map) {
            let mesh = this.map.get(old.key);
            if (mesh) {
                mesh.dirtyPosition = true;
            }
            old.width = value;
        }
    }
    /**
     * 修改对象尺寸
     * @param id ID
     * @param value 数据
     */
    height(id: any, value: number) {
        let old: TileObject;
        old = this.objectmap.get(id);
        if (old && this.map) {
            let mesh = this.map.get(old.key);
            if (mesh) {
                mesh.dirtyPosition = true;
            }
            old.height = value;
        }
    }

    /**
     * 修改对象位置
     * @param id ID
     * @param value 数据
     */
    x(id: any, value: number) {
        let old: TileObject;
        old = this.objectmap.get(id);
        if (old && this.map) {
            let mesh = this.map.get(old.key);
            if (mesh) {
                mesh.dirtyPosition = true;
            }
            old.x = value;
        }
    }
    /**
     * 修改对象位置
     * @param id ID
     * @param value 数据
     */
    y(id: any, value: number) {
        let old: TileObject;
        old = this.objectmap.get(id);
        if (old && this.map) {
            let mesh = this.map.get(old.key);
            if (mesh) {
                mesh.dirtyPosition = true;
            }
            old.y = value;
        }
    }
    /**
     * 修改对象可见性
     * @param id ID
     * @param value 数据
     */
    alpha(id: any, a: number) {
        let old: TileObject;
        old = this.objectmap.get(id);
        if (old && this.map) {
            let mesh = this.map.get(old.key);
            if (mesh) {
                mesh.dirtyColor = true;
            }
            old.a = a;
        }
    }
    /**
     * 获取对象可见性
     * @param id ID
     * @param value 数据
     */
    isEnabled(id: any) {
        return this.objectmap.get(id).visible
    }
    /**
     * 获取位置
     * @param id ID
     * @param value 数据
     */
    getPosition(id: any) {
        const old = this.objectmap.get(id);
        if (!old) return [];
        return [old.x, 0, old.y]
    }
    /**
     * 删除对象
     * @param id ID
     */
    delete(id: any) {
        let old: TileObject;
        old = this.objectmap.get(id);
        if (old) {
            if (this.map) {
                let mesh = this.map.get(old.key);
                if (mesh) {
                    old = mesh.remove(id);
                    mesh.dirty = true;
                }
            }
            old && old.dispose();
        }
        this.objectmap.delete(id);
    }
    dispose() {
        this.map.forEach((mesh) => {
            mesh.dispose();
        });

        this.map = undefined;
    }
}
