/**
 * BABYLON 资源加载
 */
import { PiBABYLONGeometry } from '../render3d/extends/geometry';
import { OfflineProvider } from '../render3d/offline_provider';
import { BABYLON } from '../render3d/babylon';
import { ExtensionTools, KTXInfo, Res, ResTab } from '../interface/index';

declare let module;
let logV, logD, logI, logW, logE;
let logger;
ExtensionTools.RegistStartUp(() => {
    logger = ExtensionTools.Logger(module.name, "BABYLON");
    let temp = ExtensionTools.GetLogger(module.name, "BABYLONResMgr");
    logV = temp.logV, logD = temp.logD, logI = temp.logI, logW = temp.logW, logE = temp.logE;
});

export const BABYLON_TEXTURE_TYPE   = 'BABYLON_TEXTURE_TYPE';
export const BABYLON_GEOMETRY_TYPE  = 'BABYLON_GEOMETRY_TYPE';
export const BABYLON_IMAGE_TYPE     = 'BABYLON_IMAGE_TYPE';
export const BABYLON_ANIMATION_TYPE = 'BABYLON_ANIMATION_TYPE';
export const BABYLON_GLTF_TYPE = 'BABYLON_GLTF_TYPE';

// tslint:disable-next-line:no-reserved-keywords
type LoadCall = (tab: ResTab, type: string, name: string, ...args: any[]) => Promise<any>;

type IImage = HTMLImageElement;

type IAnimation = BABYLON.IAnimationKey[];

export interface ITextureRectModifyInfo {
    /**
     * 更改的目标矩形区域
     * @tip [使用的图片的矩形区域 x,y,w,h , 目标纹理尺寸w,h]
     */
    r: [number, number, number, number];
    /**
     * 更改使用的颜色
     */
    c?: [number, number, number];
    /**
     * 更改使用的 图片链接
     */
    i?: string;
    /**
     * 更改使用的 图片的矩形区域
     * @tip [使用的图片的矩形区域 x,y,w,h , 图片尺寸w,h] - 当原图尺寸与构建出的图片尺寸不同时,设置原图尺寸为 作为图片尺寸w,h
     */
    ir?: [number, number, number, number, number?, number?];
    /**
     * 更改使用图片时, 图片混合方式, 未设置则为无混合替换
     */
    ib?: number;
    /**
     * 更改使用图片时, 图片是否Y翻转
     */
    iy?: number;
    /**
     * 更改使用的 HSV 变换参数
     */
    t?: [number, number, number];
}

export interface ITextureModifyInfo {
    /**
     * 纹理类型 1 : RGBA , 否则 jpg
     */
    f?: number;
    /**
     * 使用的原图的路径
     */
    s?: string;
    /**
     * 不使用原图,则设定大小
     */
    is?: [number, number];
    m: ITextureRectModifyInfo[];
}

export interface ITextureArgCfg {
    engine: BABYLON.Engine;
    invertY: boolean;
    noMipmap: boolean;
    scene: BABYLON.Scene;
    samplingMode: number;
    internalTexture: BABYLON.InternalTexture;
    buffer?: HTMLImageElement | string;
    isCompressed?: boolean;
}

/**
 * 纹理资源创建参数
 */
export interface IParamPiTexture {
    /**
     * 是否转换为 2 的幂尺寸内容
     */
    useForPowerOfTwo?: boolean;
    /**
     * 是否激活 mipMap
     */
    mipMap?: boolean;
    /**
     * 平铺方式
     */
    wrap?: [EWrap, EWrap];
    /**
     * 采样方式
     */
    filter?: [EFilter, EFilter];
}

export enum EWrap {
    REPEAT = 0,
    CLAMP_TO_EDGE = 1,
    MIRRORED_REPEAT = 2
}

export enum EFilter {
    LINEAR = 0,
    NEAREST = 1,
    NEAREST_MIPMAP_NEAREST = 2,
    LINEAR_MIPMAP_NEAREST = 3,
    NEAREST_MIPMAP_LINEAR = 4,
    LINEAR_MIPMAP_LINEAR = 5
}

/**
 * 统一的纹理资源
 */
export class PiTexture {
    /**
     * gl webgl 上下文
     * @mark 必须在应用初始化时设置 gl
     */
    public static gl: WebGLRenderingContext;
    /**
     * 纹理宽度 - 像素
     * @hidden
     */
    protected _width: number = 0;
    /**
     * 纹理高度 - 像素
     * @hidden
     */
    protected _height: number = 0;
    /**
     * gl纹理对象
     * @hidden
     */
    protected _texture: WebGLTexture = null;

    /**
     * 纹理路径
     * @hidden
     */
    protected _url: string = '';
    /**
     * 销毁标识
     * @hidden
     */
    protected _destroyed: boolean = false;
    /**
     * 所属资源表
     * @hidden
     */
    protected resTab: ResTab = ExtensionTools.CreateResTab();
    /**
     * 资源文件路径
     */
    public set src(v: string) {
        this._url = v;
        this.load();
    }
    /**
     * 资源文件路径
     */
    public get src() {
        return this._url;
    }
    /**
     * 纹理宽度 - 像素
     */
    public get width() {
        return this._width;
    }
    /**
     * 纹理高度 - 像素
     */
    public get height() {
        return this._height;
    }
    /**
     * gl纹理对象
     */
    public get webglTex() {
        return this._texture;
    }
    /**
     * gl纹理对象
     */
    protected _size: number = 0;
    public get size() {
        return this._size;
    }
    constructor() {
    }
    /**
     * 创建时参数
     */
    public args: [IParamPiTexture?];
    /**
     * 纹理格式
     */
    public format: number;
    /**
     * 是否 noMipmap
     */
    public noMipmap: boolean;
    /**
     * 是否反转Y
     */
    public invertY: boolean;
    /**
     * 采样模式
     */
    public samplingMode: number;
    /**
     * 加载成功回调
     */
    public onloaded: (tex: PiTexture) => void;
    /**
     * 加载错误回调
     */
    public onerror: (e) => void;

    /**
     * 加载
     * @hidden
     */
    protected load() {
    }

    /**
     * 加载后处理
     * @tip IParamPiTexture 提供参考，决定目标纹理是否使用为 非2的幂尺寸内容
     * @tip 作为2的幂尺寸的纹理时，是否启用 mipMap
     * @tip 作为2的幂尺寸的纹理时 其他设置由使用者自己再处理
     * @param image 加载到的图片
     * @hidden
     */
    private _loaded(image: any) {
        const [param] = this.args;

        const gl = PiTexture.gl;

        gl.pixelStorei(gl.UNPACK_FLIP_Y_WEBGL, 0);
        if (!param || !param.mipMap || !param.useForPowerOfTwo) {
            this._width = image.width;
            this._height = image.height;
            this._texture = gl.createTexture();

            gl.activeTexture(gl.TEXTURE0);
            gl.bindTexture(gl.TEXTURE_2D, this._texture);
            if ((<KTXInfo>image).mipmaps) {
                this._size = 0;
                for (let i = 0; i < (<KTXInfo>image).mipmapCount; i++) {
                    const data = (<KTXInfo>image).mipmaps[i];
                    gl.compressedTexImage2D(gl.TEXTURE_2D, i, (<KTXInfo>image).format, data.width, data.height, 0, data.data);
                    this._size += data.data.byteLength;
                }
            }
            else {
                gl.texImage2D(gl.TEXTURE_2D, 0, this.format, this.format, gl.UNSIGNED_BYTE, <HTMLImageElement>image);
                this._size = this._width * this._height * (this.format == PiTexture.gl.RGB ? 3 : 4);
            }

            // No, it's not a power of 2. Turn of mips and set wrapping to clamp to edge
            wrap(gl, EWrap.CLAMP_TO_EDGE, gl.TEXTURE_WRAP_S);
            wrap(gl, EWrap.CLAMP_TO_EDGE, gl.TEXTURE_WRAP_T);
            filter(gl, EFilter.NEAREST, gl.TEXTURE_MIN_FILTER);
            filter(gl, EFilter.NEAREST, gl.TEXTURE_MAG_FILTER);

            gl.bindTexture(gl.TEXTURE_2D, null);
        }
        else {
            const is2 = isPowerOfTwo(image.width) && isPowerOfTwo(image.height);

            this._width = image.width;
            this._height = image.height;
            this._texture = PiTexture.gl.createTexture();

            gl.activeTexture(gl.TEXTURE0);
            gl.bindTexture(gl.TEXTURE_2D, this._texture);

            if (!is2) {
                logE(`Image size needs to be a power of 2 - ${this._url} - ${image.src}`);

                // Fill the texture with a 1x1 blue pixel. - 洋红色
                gl.texImage2D(gl.TEXTURE_2D, 0, gl.RGBA, 1, 1, 0, gl.RGBA, gl.UNSIGNED_BYTE,
                    new Uint8Array([207, 0, 112, 255]));
                    
                this._size = 4;
            }
            else {
                if ((<KTXInfo>image).mipmaps) {
                    this._size = 0;
                    for (let i = 0; i < (<KTXInfo>image).mipmapCount; i++) {
                        const data = (<KTXInfo>image).mipmaps[i];
                        gl.compressedTexImage2D(gl.TEXTURE_2D, i, (<KTXInfo>image).format, data.width, data.height, 0, data.data);
                        this._size += data.data.byteLength;
                    }
                } else {
                    gl.texImage2D(gl.TEXTURE_2D, 0, this.format, this.format, gl.UNSIGNED_BYTE, image);
                    this._size = this._width * this._height * (this.format == PiTexture.gl.RGB ? 3 : 4);
                }
            }

            if (param.wrap) {
                wrap(gl, param.wrap[0], gl.TEXTURE_WRAP_S);
                wrap(gl, param.wrap[1], gl.TEXTURE_WRAP_T);
            } else {
                wrap(gl, null, gl.TEXTURE_WRAP_S);
                wrap(gl, null, gl.TEXTURE_WRAP_T);
            }
            if (param.filter) {
                filter(gl, param.filter[0], gl.TEXTURE_MIN_FILTER);
                filter(gl, param.filter[1], gl.TEXTURE_MAG_FILTER);
            } else {
                filter(gl, null, gl.TEXTURE_MIN_FILTER);
                filter(gl, null, gl.TEXTURE_MAG_FILTER);
            }

            if (param.mipMap) {
                // Yes, it's a power of 2. Generate mips.
                gl.generateMipmap(gl.TEXTURE_2D);
            }

            gl.bindTexture(gl.TEXTURE_2D, null);
        }

    }
    /**
     * 销毁
     */
    public destroy() {
        this._destroyed = true;
        if (this._texture) {
            PiTexture.gl.deleteTexture(this._texture);
            this._texture = null;
        }
        this.resTab.clear();
    }
}

export function wrap(gl: WebGLRenderingContext, mode: EWrap, wrapFlag: number) {
    switch (mode) {
        case EWrap.MIRRORED_REPEAT: {
            gl.texParameteri(gl.TEXTURE_2D, wrapFlag, gl.MIRRORED_REPEAT);
            break;
        }
        case EWrap.REPEAT: {
            gl.texParameteri(gl.TEXTURE_2D, wrapFlag, gl.REPEAT);
            break;
        }
        default: {
            gl.texParameteri(gl.TEXTURE_2D, wrapFlag, gl.CLAMP_TO_EDGE);
            break;
        }
    }
}
export function filter(gl: WebGLRenderingContext, mode: EFilter, filterFlag: number) {
    switch (mode) {
        case EFilter.LINEAR: {
            gl.texParameteri(gl.TEXTURE_2D, filterFlag, gl.LINEAR);
            break;
        }
        case EFilter.LINEAR_MIPMAP_LINEAR: {
            gl.texParameteri(gl.TEXTURE_2D, filterFlag, gl.LINEAR_MIPMAP_LINEAR);
            break;
        }
        case EFilter.LINEAR_MIPMAP_NEAREST: {
            gl.texParameteri(gl.TEXTURE_2D, filterFlag, gl.LINEAR_MIPMAP_NEAREST);
            break;
        }
        case EFilter.NEAREST_MIPMAP_LINEAR: {
            gl.texParameteri(gl.TEXTURE_2D, filterFlag, gl.NEAREST_MIPMAP_LINEAR);
            break;
        }
        case EFilter.NEAREST_MIPMAP_NEAREST: {
            gl.texParameteri(gl.TEXTURE_2D, filterFlag, gl.NEAREST_MIPMAP_NEAREST);
            break;
        }
        default: {
            gl.texParameteri(gl.TEXTURE_2D, filterFlag, gl.NEAREST);
            break;
        }
    }
}

function isPowerOfTwo(value: number) {
	return (value & (value - 1)) === 0 && value !== 0;
}

export class PiBABYLONTexture extends PiTexture {
    public static isCompressed(url: string) : boolean {
        return (/\.ktx$/).test(url);
    }
    private static modifyInfosMap: Map<string, ITextureModifyInfo> = new Map();
    public static delTextureModifiyInfos(url: string) {
        PiBABYLONTexture.modifyInfosMap.delete(url);
    }
    public static getTextureModifiyInfos(url: string): ITextureModifyInfo {
        return PiBABYLONTexture.modifyInfosMap.get(url);
    }
    public static setTextureModifiyInfos(url: string, info: ITextureModifyInfo) {
        PiBABYLONTexture.modifyInfosMap.set(url, info);
    }
    public internalTexture: BABYLON.InternalTexture;
    public p_args: [ITextureArgCfg];
    protected load() {
        const argCfg                    = <ITextureArgCfg>this.p_args[0];
        const engine: BABYLON.Engine    = argCfg.engine;
        const scene: BABYLON.Scene      = argCfg.scene;

        this.noMipmap       = argCfg.noMipmap;
        this.invertY        = argCfg.invertY;
        this.samplingMode   = argCfg.samplingMode;

        let tex: BABYLON.InternalTexture = argCfg.internalTexture;
        if (tex) {
            if (tex.isReady) {
                Promise.resolve(() => {
                    this._size = tex.width * tex.height * 4;
                    this._width     = tex.width;
                    this._height    = tex.height;
                    if (argCfg.isCompressed) {
                        this._size  = ExtensionTools.getFileSize(tex.url) || this._size;
                    }
                    this.onloaded && this.onloaded(this);
                });
            }
            else {
                tex.onLoadedObservable.add(() => {
                    this._size = tex.width * tex.height * 4;
                    this._width     = tex.width;
                    this._height    = tex.height;
                    if (argCfg.isCompressed) {
                        this._size  = ExtensionTools.getFileSize(tex.url) || this._size;
                    }
                    this.onloaded && this.onloaded(this);
                });
            }
        } else {
            // alert(that.src);
            // const purePath = PISYS.Depend.getPureResPath(that.src);
            // const forceExtension = that.src.replace(purePath, '');
            let modifyinfo = PiBABYLONTexture.getTextureModifiyInfos(this.src);
            if (modifyinfo) {
                tex = PiBABYLONTexture.loadWithModify(argCfg.engine, argCfg.scene, this.src, this, argCfg, modifyinfo,
                    () => {
                        this._texture   = tex._webGLTexture;
                        this._size      = tex.width * tex.height * 4;
                        this._width     = tex.width;
                        this._height    = tex.height;
                        if (argCfg.isCompressed) {
                            this._size  = ExtensionTools.getFileSize(tex.url) || this._size;
                        }
                        this.onloaded && this.onloaded(this);
                    },
                    (msg) => {
                        this.onerror && this.onerror(msg);
                    }
                );
            } else {
                tex = engine.createTexture(
                    this.src,
                    this.noMipmap,
                    this.invertY,
                    scene,
                    this.samplingMode,
                    () => {
                        this._texture   = tex._webGLTexture;
                        this._size      = tex.width * tex.height * 4;
                        this._width     = tex.width;
                        this._height    = tex.height;
                        if (argCfg.isCompressed) {
                            this._size  = ExtensionTools.getFileSize(tex.url) || this._size;
                        }
                        this.onloaded && this.onloaded(this);
                        if (typeof argCfg.buffer == 'string') {
                            argCfg.buffer = "";
                        } else if (argCfg.buffer) {
                            argCfg.buffer.src = "";
                        }
                    },
                    (msg, e) => {
                        // alert(msg);
                        this.onerror && this.onerror(msg);
                        if (typeof argCfg.buffer == 'string') {
                            argCfg.buffer = "";
                        } else if (argCfg.buffer) {
                            argCfg.buffer.src = "";
                        }
                    },
                    argCfg.buffer
                    // undefined,
                    // undefined,
                    // forceExtension
                );
            }
        }

        this.internalTexture = tex;
    }
    public destroy() {

        this._destroyed = true;

        if (this.internalTexture) {
            this.internalTexture.dispose();
            this.internalTexture = null;
        }
    }
    private static loadWithModify(engine: BABYLON.Engine, scene: BABYLON.Scene, uuid: string, target: PiBABYLONTexture, arg: ITextureArgCfg, modifyinfo: ITextureModifyInfo, success: () => void, fail: (err: any) => void): BABYLON.InternalTexture {

        let gl = <WebGLRenderingContext>(<any>engine._gl);
        let selfTexture = new BABYLON.InternalTexture(engine, 1);
        selfTexture.url = uuid;
        selfTexture.generateMipMaps = !arg.noMipmap;
        selfTexture.samplingMode = arg.samplingMode;
        selfTexture.invertY = arg.invertY;

        gl.bindTexture(gl.TEXTURE_2D, null);

        let resTab = ExtensionTools.CreateResTab();
        let promiseArray: Promise<void>[] = [];

        let needTexturesUrls = [];
        let needTexturesInvert = [];

        if (modifyinfo.s) {
            needTexturesUrls.push(modifyinfo.s);
            needTexturesInvert.push(!!arg.invertY);
        }

        let rectInfos = modifyinfo.m;
        let count = rectInfos.length;
        for (let i = 0; i < count; i++) {
            let url = rectInfos[i].i;
            if (url && !needTexturesUrls.includes(url)) {
                needTexturesUrls.push(url);
                needTexturesInvert.push(!!rectInfos[i].iy);
            }
        }

        let textureResMap: Map<string, PiBABYLONTexture> = new Map();
        let texCount = needTexturesUrls.length;
        if (texCount > 0) {
            for (let i = 0; i < texCount; i++) {
                let key = needTexturesUrls[i];
                let loadP = resTab.load(
                    BABYLON_TEXTURE_TYPE,
                    key,
                    [{
                        engine: scene.getEngine(),
                        invertY: needTexturesInvert[i],
                        scene,
                        noMipmap: true,
                        samplingMode: BABYLON.Texture.NEAREST_SAMPLINGMODE,
                        internalTexture: undefined
                    }]
                ).then((res) => {
                    textureResMap.set(key, res.link);
                });

                promiseArray.push(loadP);
            }
        }

        Promise.all(promiseArray).then(() => {
            let width = modifyinfo.is[0];
            let height = modifyinfo.is[1];
            if (modifyinfo.s) {
                let sourceTexture = textureResMap.get(modifyinfo.s);
                width = sourceTexture.width;
                height = sourceTexture.height;
            }

            gl.bindTexture(gl.TEXTURE_2D, selfTexture._webGLTexture);
            (<any>engine)._prepareWebGLTexture(selfTexture, scene, width, height, selfTexture.invertY, arg.noMipmap, arg.isCompressed, (potWidth, potHeight, continuationCallback) => {
                var isPot = (width === potWidth && height === potHeight);
                let internalFormat = modifyinfo.f ? gl.RGBA : gl.RGB;

                if (isPot) {
                    gl.texImage2D(gl.TEXTURE_2D, 0, internalFormat, potWidth, potHeight, 0, internalFormat, gl.UNSIGNED_BYTE, null);
                    return false;
                }

                let maxTextureSize = engine._caps.maxTextureSize;

                if (width > maxTextureSize || height > maxTextureSize || !(<any>engine)._supportsHardwareTextureRescaling) {
                    engine._prepareWorkingCanvas();
                    gl.texImage2D(gl.TEXTURE_2D, 0, internalFormat, potWidth, potHeight, 0, internalFormat, gl.UNSIGNED_BYTE, null);

                    selfTexture.width = potWidth;
                    selfTexture.height = potHeight;

                    return false;
                } else {
                    // Using shaders when possible to rescale because canvas.drawImage is lossy
                    let source = new BABYLON.InternalTexture(engine, 2);
                        engine._bindTextureDirectly(gl.TEXTURE_2D, source, true);
                        gl.texImage2D(gl.TEXTURE_2D, 0, internalFormat, potWidth, potHeight, 0, internalFormat, gl.UNSIGNED_BYTE, null);

                    engine._rescaleTexture(source, selfTexture, scene, internalFormat, () => {
                        engine._releaseTexture(source);
                        engine._bindTextureDirectly(gl.TEXTURE_2D, selfTexture, true);

                        continuationCallback();
                    });
                }

                return true;
            }, arg.samplingMode);
            gl.bindTexture(gl.TEXTURE_2D, null);

            PiBABYLONTexture.textureModifyRenderCall(<any>scene.getEngine().getRenderingCanvas(), target, modifyinfo, textureResMap, () => {
                Promise.resolve().then(() => { resTab.clear() });
                success();
            }, fail);
        });

        return selfTexture;
    }
    /**
     * 外部设置 更改纹理内容的渲染调用
     * @param source 源纹理
     * @param infos 更改配置
     * @param texResMap 纹理map
     * @param success 成功回调
     * @param fail 失败回调
     */
    public static textureModifyRenderCall = (canvas: HTMLCanvasElement, source: PiBABYLONTexture, infos: ITextureModifyInfo, texResMap: Map<string, PiBABYLONTexture>, success: () => void, fail: (err: any) => void) => {
        console.warn(`Not Implements textureModifyRenderCall!`);

        success();
    }
}

// ======================= 立即执行


const loadAnimationRes: LoadCall = (tab: ResTab, _type: 'BABYLON_ANIMATION_TYPE', _name: string, ...args: any[]): Promise<IAnimation> => {
    // tslint:disable-next-line:no-unnecessary-local-variable
    const p = new Promise((resolve: (res: IAnimation) => void, rejects) => {
        resolve(createAnimationRes(_type, _name, args[0], undefined));
    });

    return p;
};

const createAnimationRes = (_type: string, _name: string, keys: BABYLON.IAnimationKey[], _: any): IAnimation => {
    return keys;
};

const destroyAnimationRes = (r: IAnimation) => {
    // r.link.src = '';
};

// ====================================================================
export function sizeOfGeometry(geo: BABYLON.Geometry) {
    let vertexSize = 0;
    let VertexBuffers = geo.getVertexBuffers();
    for (let key in VertexBuffers) {
        vertexSize += VertexBuffers[key].getSize();
    }

    let size = (geo.getTotalVertices() || (<any>geo)._totalVertices) * vertexSize + geo.getIndices()?.length * 2;

    return  size || (1024 * 100);  // 10240; 
}
// register(BABYLON_TEXTURE_TYPE, loadTextureRes, destroyTextureRes, (tex: BABYLON.InternalTexture) => { return tex._size; });
// register(BABYLON_GEOMETRY_TYPE, loadGeometryRes, destroyGeometryRes, sizeOfGeometry);

// register(BABYLON_IMAGE_TYPE, loadImageRes, destroyImageRes);
ExtensionTools.RegistStartUp(() => {
    ExtensionTools.ResLoadRegist(BABYLON_ANIMATION_TYPE, loadAnimationRes, destroyAnimationRes, (ani: any[]) => { return (ani.length || 1) * 96; });
    ExtensionTools.ResLoadRegist<typeof BABYLON_TEXTURE_TYPE, ITextureArgCfg, PiBABYLONTexture>(
        BABYLON_TEXTURE_TYPE,
        (resTab: ResTab, type: typeof BABYLON_TEXTURE_TYPE, filePath: string, ...args: [ITextureArgCfg]): Promise<PiBABYLONTexture> => {
            const piTex = new PiBABYLONTexture();
            piTex.p_args = args;
    
            return new Promise((resolve, reject) => {
                piTex.onloaded = () => {
                    resolve(piTex);
                    if (OfflineProvider.onoffCollectInfo) {
                        if (OfflineProvider.piBABYLONResCollect[BABYLON_TEXTURE_TYPE] == undefined) {
                            OfflineProvider.piBABYLONResCollect[BABYLON_TEXTURE_TYPE] = 0;
                        }
                        
                        OfflineProvider.piBABYLONResCollect[BABYLON_TEXTURE_TYPE] += piTex.size;
                        OfflineProvider.piBABYLONTextureRecord.set(filePath, [piTex.width, piTex.height]);
                    }
                };
                piTex.onerror = (e) => {
                    reject(e);
                };
                piTex.src = filePath;
            });
        },
        (link: PiBABYLONTexture) => {
            if (OfflineProvider.onoffCollectInfo) {
                OfflineProvider.piBABYLONResCollect[BABYLON_TEXTURE_TYPE] -= link.size;
            }
            link.destroy();
        },
        function(link: PiBABYLONTexture) {
            return link.size;
        }
    );
});


// export interface IGLTFArg {
//     engine: BABYLON.Engine,
// }
// function loadGLTF(
//     tab: ResTab, _type: 'RES_BABYLON_GLTF', _name: string, args: IGLTFArg
// ): Promise<GLTFRes> {
//     return UriLoader._loadFile(_name, tab).then((res) => {
//         let rootDir = _name;
//         let rootUrl = _name;

//         const str = PISYS.UTF8.utf8Decode(res);
//         let json = JSON.parse(str);
//         json = OfflineProvider.GLTFFormat(_name, json);

//         let result = new GLTFRes();
//         result.path = _name;
//         result.buffers = [];
//         result.accessorbuffers = [];
//         result.json = json;
//         result.size += res.byteLength;

//         let bufferPromises = [];
//         result.json.buffers?.forEach((info, index) => {
//             let uri = info.uri;
//             if (UriLoader.IsBase64(uri)) {
//                 const data = new Uint8Array(UriLoader.DecodeBase64(uri));
//                 LoaderLog.log(`Decoded ${uri.substr(0, 64)}... (${data.length} bytes)`);
//                 result.buffers[index] = data;
//                 bufferPromises.push( Promise.resolve(undefined) );
//             } else {
//                 let url = UriLoader.formatUrl(rootDir, uri);
//                 bufferPromises.push( UriLoader._loadFile(url, tab).then((data) => {
//                     result.buffers[index] = data;
//                     result.size += data.byteLength;
//                 }) );
//             }
//         });

//         return Promise.all(bufferPromises).then(() => {
//             result.json.accessors?.forEach((accessor, index) => {
//                 let bufferview = result.json.bufferViews[accessor.bufferView];
//                 let data = result.buffers[bufferview.buffer];
//                 switch (accessor.componentType) {
//                     case (GLTF2.AccessorComponentType.UNSIGNED_SHORT):  {
//                         let temp = new Uint16Array(data.buffer, data.byteOffset + (accessor.byteOffset || 0), accessor.count * AccessorLoader._GetNumComponents(accessor.type));
//                         this.accessorbuffers[index] = temp;
//                         break;
//                     }
//                     case (GLTF2.AccessorComponentType.UNSIGNED_INT): {
//                         let temp = new Uint32Array(data.buffer, data.byteOffset + (accessor.byteOffset || 0), accessor.count * AccessorLoader._GetNumComponents(accessor.type));
//                         this.accessorbuffers[index] = temp;
//                         break;
//                     }
//                     case (GLTF2.AccessorComponentType.FLOAT): {
//                         let temp = new Float32Array(data.buffer, data.byteOffset + (accessor.byteOffset || 0), accessor.count * AccessorLoader._GetNumComponents(accessor.type));
//                         this.accessorbuffers[index] = temp;
//                         break;
//                     }
//                     default: {
//                         throw new Error(``);
//                     }
//                 }
//             });

//             result.json.meshes?.forEach((mesh, indexmesh) => {
//                 mesh.primitives.forEach((primitive, indexprimitive) => {
//                     if (primitive.indices != undefined) {
//                         let index_accessor = primitive.indices;
//                         result.indices[index_accessor] = this.accessorbuffers[index_accessor];
//                     }
//                     for (let key in primitive.attributes) {
//                         let index_accessor = primitive.attributes[key];
//                         result.renderbuffers[index_accessor] = new BABYLON.Buffer(args.engine, this.accessorbuffers[index_accessor], false);
//                     }
//                 });
//             });

//             result.json.animations?.forEach((anime, indexGroup) => {
//                 let samplers = anime.samplers;
//                 anime.channels.forEach((channel, indexChannel) => {
//                     let sampler = samplers[channel.sampler];

//                     let propertyID: EAnimationPropertyID;
//                     if (channel.target.path) {
//                         switch (channel.target.path) {
//                             case (AnimationChannelTargetPath.TRANSLATION): {
//                                 propertyID = EAnimationPropertyID.Position;
//                                 break;
//                             }
//                             case (AnimationChannelTargetPath.SCALE): {
//                                 propertyID = EAnimationPropertyID.Scaling;
//                                 break;
//                             }
//                             case (AnimationChannelTargetPath.ROTATION): {
//                                 propertyID = EAnimationPropertyID.RotationQuaternion;
//                                 break;
//                             }
//                         }
//                     }
//                     if (channel.extras.property != undefined) {
//                         propertyID = channel.extras.property;
//                     }

//                     let times = result.accessorbuffers[sampler.input];
//                     let values = result.accessorbuffers[sampler.output];
//                     if (times && values) {
//                         indexGroup += 1;
//                         indexChannel += 1;
//                         let key = `${rootUrl}#${indexGroup}#${indexChannel}`;
//                         let interpolation = GLTFInterpolationType.LINEAR;
//                         if (sampler.interpolation) {
//                             switch (sampler.interpolation) {
//                                 case (AnimationSamplerInterpolation.STEP): {
//                                     interpolation = GLTFInterpolationType.STEP;
//                                     break;
//                                 }
//                                 case (AnimationSamplerInterpolation.CUBICSPLINE): {
//                                     interpolation = GLTFInterpolationType.CUBICSPLINE;
//                                     break;
//                                 }
//                             }
//                         } else if (sampler.extras?.interpolation != undefined) {
//                             interpolation = sampler.extras.interpolation;
//                         }
                        
//                         let outputBufferOffset = 0;
//                         let getNextOutputValue: () => BABYLON.Vector3 | BABYLON.Quaternion | number[] | number | BABYLON.Color3;
//                         switch (propertyID) {
//                             case EAnimationPropertyID.Position: {
//                                 getNextOutputValue = () => {
//                                     const value = BABYLON.Vector3.FromArray(values, outputBufferOffset);
//                                     outputBufferOffset += 3;

//                                     return value;
//                                 };
//                                 break;
//                             }
//                             case EAnimationPropertyID.RotationQuaternion: {
//                                 getNextOutputValue = () => {
//                                     const value = BABYLON.Quaternion.FromArray(values, outputBufferOffset);
//                                     outputBufferOffset += 4;

//                                     return value;
//                                 };
//                                 break;
//                             }
//                             case EAnimationPropertyID.MainColor: {
//                                 getNextOutputValue = () => {
//                                     const value = BABYLON.Color3.FromArray(values, outputBufferOffset);
//                                     outputBufferOffset += 3;

//                                     return value;
//                                 };
//                                 break;
//                             }
//                             case EAnimationPropertyID.Euler: {
//                                 getNextOutputValue = () => {
//                                     const value = BABYLON.Vector3.FromArray(values, outputBufferOffset);
//                                     outputBufferOffset += 3;

//                                     return value;
//                                 };
//                                 break;
//                             }
//                             case EAnimationPropertyID.Scaling: {
//                                 getNextOutputValue = () => {
//                                     const value = BABYLON.Vector3.FromArray(values, outputBufferOffset);
//                                     outputBufferOffset += 3;

//                                     return value;
//                                 };
//                                 break;
//                             }
//                             case EAnimationPropertyID.lightdiffuse: {
//                                 getNextOutputValue = () => {
//                                     const value = BABYLON.Color3.FromArray(values, outputBufferOffset);
//                                     outputBufferOffset += 3;

//                                     return value;
//                                 };
//                                 break;
//                             }
//                             // case EAnimationPropertyID.intensity: {
//                             //     getNextOutputValue = () => {
//                             //         // tslint:disable-next-line:prefer-array-literal
//                             //         const value = new Array<number>(targetNode._numMorphTargets!);
//                             //         // tslint:disable-next-line:no-non-null-assertion
//                             //         for (let i = 0; i < targetNode._numMorphTargets!; i++) {
//                             //             value[i] = values[outputBufferOffset++];
//                             //         }

//                             //         return value;
//                             //     };
//                             //     break;
//                             // }
//                             default: {
//                                 getNextOutputValue = () => {
//                                     return values[outputBufferOffset++];
//                                 };
//                             }
//                         }

//                         let getNextKey: (frameIndex: number) => BABYLON.IAnimationKey;
//                         // tslint:disable-next-line:switch-default
//                         switch (interpolation) {
//                             case GLTFInterpolationType.STEP: {
//                                 getNextKey = (frameIndex) => ({
//                                     frame: times[frameIndex],
//                                     value: getNextOutputValue(),
//                                     interpolation: BABYLON.AnimationKeyInterpolation.STEP
//                                 });
//                                 break;
//                             }
//                             case GLTFInterpolationType.LINEAR: {
//                                 getNextKey = (frameIndex) => ({
//                                     frame: times[frameIndex],
//                                     value: getNextOutputValue()
//                                 });
//                                 break;
//                             }
//                             case GLTFInterpolationType.CUBICSPLINE: {
//                                 getNextKey = (frameIndex) => {
//                                     const inTangent = getNextOutputValue();
//                                     const value = getNextOutputValue();
//                                     const outTangent = getNextOutputValue();

//                                     return {
//                                         frame: times[frameIndex],
//                                         inTangent: inTangent,
//                                         value: value,
//                                         outTangent: outTangent
//                                     };
//                                 };
//                                 break;
//                             }
//                             case GLTFInterpolationType.PICUBICSPLINE /* CUBICSPLINE */: {
//                                 getNextKey = (frameIndex) => {
//                                     const inAndOut = getNextOutputValue();

//                                     return ({
//                                         frame: times[frameIndex],
//                                         inTangent: inAndOut,
//                                         value: getNextOutputValue(),
//                                         outTangent: inAndOut
//                                     });
//                                 };
//                             }
//                         }

//                         // tslint:disable-next-line:prefer-array-literal
//                         const keys: BABYLON.IAnimationKey[] = new Array(times.length);
//                         for (let frameIndex = 0; frameIndex < times.length; frameIndex++) {
//                             // tslint:disable-next-line:no-non-null-assertion
//                             keys[frameIndex] = getNextKey!(frameIndex);
//                         }

//                         result.animetionKeys.set(key, keys);
//                     }
//                 });
//             });

//             result.buffers.length = 0;
//             result.accessorbuffers.length = 0;

//             return result;
//         });
//     });
// }
// register<typeof BABYLON_GLTF_TYPE, IGLTFArg, GLTFRes>(
//     BABYLON_GLTF_TYPE,
//     loadGLTF,
//     (link: GLTFRes) => link.dispose(),
//     (link: GLTFRes) => {
//         return link.size;
//     }
// );
